/* Configuration directives.
 *
 * PegSoft configuration parser library (c) 2004 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

/** 
 * \file conf/directives.h
 * \brief Configuration directives header file.
 */

#ifndef CONF_DIRECTIVES_H
#define CONF_DIRECTIVES_H

/*************************************************************************/

#if HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <map>
#include <string>
#include <vector>

#include <conf/parser.h>
#include <misc/ratio.h>
#include <misc/string.h>
#include <misc/uri.h>

/*************************************************************************/

/// Namespace for all configuration parser declarations.

namespace conf 
{

/*************************************************************************/

class Conf;

/*************************************************************************/

/// Configuration directive base class.

/** ConfDir is the configuration directive base class; all directive
  * types ultimately inherit from it. 
  * 
  * Each directive must have an unique name in the list it is added to.
  *
  * This class is essentially an interface to allow the parser to set
  * directives in a standard way. It does nothing and must not be used
  * directly. Other generic classes are defined for specific directive
  * types, and you may also create your own by inheriting this class
  * or any derived class and overriding the begin(), set() and end()
  * functions. The label(), label_count() and sub() functions may
  * also be overriden. */

class ConfDir
{
	public:
		/// Main constructor.
		/** This constructor is used to construct a ConfDir object
		  * from scratch.
		  * \param[in] name The name of the directive, that must be unique
		  * within a directives list (case-insensitive).
		  * \param[in] reloadable Whether the directives' begin(), set()
		  * and end() functions should have an effect when reloading
		  * configuration.
		  * \param[in] min The minimum number of parameters required. The
		  * set() function will yield false if there aren't enough parameters.
		  * \param[in] max The maximum number of parameters. The set() function
		  * will yield false if there are too many parameters.
		  */
		explicit ConfDir(std::string const &name, bool reloadable,
				unsigned short min = 0, unsigned short max = 0) : m_name(name),
				m_reloadable(reloadable), m_ready(false), m_minparams(min),
				m_maxparams(max) { }
		/// Copy constructor.
		ConfDir(ConfDir const &cd) : m_name(cd.m_name),
				m_reloadable(cd.m_reloadable), m_ready(cd.m_ready),
				m_minparams(cd.m_minparams), m_maxparams(cd.m_maxparams) { }
		/// Destructor.
		virtual ~ConfDir() { }
		/// Assignment operator
		virtual ConfDir &operator=(ConfDir const &cd);
		
		/// Notifies the directive that parsing has begun.
		/** Notifies the directive that parsing has begun. This
		  * is usually called from the Conf::begin() function of the
		  * object to which the directive has been added. It must
		  * take care to initialize the directive so that's it's
		  * ready to be set.
		  *
		  * The default implementation just sets m_ready to true.
		  * Overriders will need to ensure that this variable is
		  * set to true ONLY if the begin() function is successful,
		  * and to not do any initialization if m_ready is already
		  * true, because begin() may be called multiple times.
		  *
		  * \param[in] conf The Conf object making the call. 
		  * \return true if the function is successful, false
		  * otherwise (conf.error() should have be called with
		  * an appropriate error message in that case).
		  *
		  * \sa Conf::begin(), Conf::error() */
		virtual bool begin(Conf &conf) { return (m_ready = true); }
		/// Sets the directive.
		/** Passes an array of parameters to be used as the
		  * value of the directive. The directive may do whatever
		  * it wants with it (the most common thing being to
		  * set a value). This is generally called automatically
		  * from the Conf::parse() or Conf::parse_file() functions.
		  * The begin() function of the directive is guaranteed to
		  * have been called at this point of time.
		  *
		  * \param[in] conf The Conf object making the call.
		  * \param[in] count Number of parameters in \a params.
		  * \param[in] params Parameters passed to the directive
		  * (NULL if count is 0).
		  * 
		  * \return true if the function is successful, false
		  * otherwise (conf.error() should have be called with
		  * an appropriate error message in that case).
		  *
		  * \sa Conf::parse(), Conf::parse_file(), begin(),
		  * Conf::error() */
		virtual bool set(Conf &conf, unsigned short count,
				std::string const params[]);
		/// Notifies the directive that parsing is finished.
		/** Notifies the directive that parsing is finished. This
		  * is usually called from the end() function of the Conf
		  * object to which the directive has been added. It must
		  * do any action necessary to finish the configuration
		  * of the directive (if that's needed, of course).
		  *
		  * The default implementation just sets m_ready to false.
		  * Overriders will need to ensure that this variable is
		  * set to false ONLY if the end() function is successful,
		  * and to not do any finalization if m_ready is not
		  * true, because begin() may be called multiple times.
		  *
		  * \param[in] conf The Conf object making the call.
		  * \return true if the function is successful, false
		  * otherwise (conf.error() should have be called with
		  * an appropriate error message in that case).
		  *
		  * \sa Conf::end(), Conf::error() */
		virtual bool end(Conf &conf) { m_ready = false; return true; }
		
		/// Returns the name of the directive.
		std::string const &name() const { return m_name; }
		/// Returns whether directive is reloadable.
		bool reloadable() const { return m_reloadable; }
		/// Returns whether begin() has been called, but not yet end().
		bool ready() const { return m_ready; }
		
		/// Returns a label describing a value of the directive.
		/** Returns a label describing a value of the directive.
		  * A label is a string that has been generated based on
		  * one of the directive's value.
		  *
		  * The default implementation always returns 0. Derived
		  * classes may choose to override this function to provide
		  * an appropriate label, but that's not required.
		  *
		  * \param[in] index The index of the value that needs to
		  * be retrieved, which must be between 0 and
		  * label_count() - 1.
		  * \return A pointer to a label, or 0 if there is no
		  * label for this value.
		  * \sa label_count() */
		virtual char const *label(size_t index) const { return 0; }
		/// Returns the number of labels available.
		/** Returns the number of labels available for this
		  * directive.
		  *
		  * The default implementation always returns zero.
		  * Derived classes must override this function if they
		  * override the label() function to provide labels.
		  *
		  * \return The number of labels available. */
		virtual size_t label_count() const { return 0; }
		/// Returns the sub-directives.
		/** Returns the sub-directives as a configuration parser
		  * filled with the sub-directives of this directive.
		  *
		  * The default implementation always returns zero. Derived
		  * classes may override this function if they have
		  * sub-directives, but that's not required.
		  *
		  * \return A pointer to the configuration parser with
		  * a list of configuration directives added. */
		virtual Conf const *sub() const { return 0; }
	protected:
		/// Directive name.
		std::string m_name;
		/// Is directive reloadable?
		bool m_reloadable;
		
		/// Has begin() been called, but not yet end() ?
		bool m_ready;
		
		/// Minimum number of parameters required.
		unsigned short m_minparams;
		/// Maximum number of parameters.
		unsigned short m_maxparams;
};

/*************************************************************************/
/*********************** Generic directives classes **********************/
/*************************************************************************/

/// Template class implementing basics of value-setting directives.

/** This template class is used as a base class for value-setting directive
  * types. The template parameter is the type of the value, default value
  * and old value that will be aggregated in the class. It must have a
  * default constructor and a compatible operator==() function.
  *
  * Each of these directives have a value and an optional default value to
  * which the value will be set if the set() function is not called before
  * the end() function is (i.e. when the parser did not encounter the
  * directive). If there is no default value, then setting the directive
  * is REQUIRED and end() will yield false if it isn't the case.
  *
  * On reload, the old value will saved so that you can do changes that
  * would require you to know the old value easily. */

template <class T>
class ConfValue : public ConfDir
{
	public:
		/// The value type.
		typedef T value_type;
	
		/// Constructor setting directive name, reloadability and min./max.
		explicit ConfValue(std::string const &name, bool reloadable,
				unsigned short min_params = 1, unsigned short max_params = 1) :
				ConfDir(name, reloadable, min_params, max_params), m_value(),
				m_value_set(false), m_value_changed(false), m_defvalue(),
				m_has_default(false), m_oldvalue(), m_label() { }
		/// Constructor also setting the default value...
		ConfValue(std::string const &name, T const &defval, bool reloadable,
				unsigned short min_params = 1, unsigned short max_params = 1) :
				ConfDir(name, reloadable, min_params, max_params), m_value(),
				m_value_set(false), m_value_changed(false), m_defvalue(defval),
				m_has_default(true), m_oldvalue(), m_label() { }
		/// Copy constructor.
		ConfValue(ConfValue const &cd) : ConfDir(cd), m_value(cd.m_value),
				m_value_set(cd.m_value_set),
				m_value_changed(cd.m_value_changed), m_defvalue(cd.m_defvalue),
				m_has_default(cd.m_has_default), m_oldvalue(cd.m_oldvalue),
				m_label(cd.m_label) { }
		/// Assignment operator.
		virtual ConfValue<T> &operator=(ConfValue<T> const &right);
		
		/// Begin parsing.
		/** Sets the old value.
		  * \return Always true. */
		virtual bool begin(Conf &conf);
		/// Sets the value.
		/** Just calls assign to set the value... What a lazy function. ;) */
		virtual bool set(Conf &conf, unsigned short count,
				std::string const params[]);
		/// End parsing.
		/** Ensures a value has been set by the set() function. If so,
		  * calls apply_change() if the conf.reloading() is true and value
		  * changed; else sets it to the default value if available or
		  * sends an error message.
		  * \return true if successful, false otherwise. */
		virtual bool end(Conf &conf);
		
		/// Gets the value.
		/** Gets the current value the directive is set to. Beware that
		  * it is not meaningful unless begin() and end() were
		  * called successfully.
		  * \return The value the directive is set to.
		  */
		T const &value() const { return m_value; }
		/// Returns whether the value was changed by a configuration reload.
		bool has_value_changed() const { return m_value_changed; }
		
		/// Returns whether a default value is available.
		bool has_default() const { return m_has_default; }
		
		/// Returns the old value.
		/** Returns the previous value the directive was set to. Beware
		  * that it is not meaningful unless begin() and end() were called
		  * and the configuration was \b reloaded.
		  * \return The previous value. */
		T const &old_value() const { return m_oldvalue; }
		
		/// Implements ConfDir::label().
		virtual char const *label(size_t index) const
				{ return m_label.c_str(); }
		/// Implements ConfDir::label_count().
		virtual size_t label_count() const { return 1; }
	protected:
		/// The value.
		T m_value;
		/// Has m_value been set? (Set to false again by end())
		bool m_value_set;
		/// Is m_value different from m_oldvalue ?
		bool m_value_changed;
		
		/// The default value (don't use if m_has_default is false!)
		T m_defvalue;
		/// Is m_default a valid default value?
		bool m_has_default;
		
		/// The old value (valid only on reload, set in begin())
		T m_oldvalue;
		
		/// The label of the value.
		mutable std::string m_label;
		
		/// Sets the value.
		/** Transforms the string array passed by the parser into the
		  * directive's value. This function must ensure that the value is
		  * valid through a call to check() before setting it (you don't
		  * need to set m_value_set to true, this will be handled by set()
		  * when the function returns successfully).
		  *
		  * \param[in] conf The Conf object that called set().
		  * \param[in] count Number of parameters in \a params.
		  * \param[in] params Parameters passed to the directive
		  * (NULL if count is 0).
		  * 
		  * \return true if the function is successful, false
		  * otherwise (conf.error() should have be called with
		  * an appropriate error message in that case). */
		virtual bool assign(Conf &conf, unsigned short count,
				std::string const params[]) = 0;
		/// Checks that a value is valid.
		/** Checks that the given value is valid.
		  * 
		  * \param[in] conf Conf object that called set().
		  * \param[in] value The value to check.
		  *
		  * \return true if the value is valid, false otherwise
		  * (conf.error() should have be called with an appropriate
		  * error message in that case). */
		virtual bool check(Conf &conf, T const &value) { return true; }
		
		/// Does stuff based on the directive new value.
		/** Does arbitrary stuff based on the directive new value. This
		  * is guaranteed to be called by the end() function when the
		  * directive is successfully loaded or reloaded.
		  *
		  * \return true or false; the latter value will make the end()
		  * function return false too (conf.error() should have been called
		  * with an appropriate error message in that case). */
		virtual bool apply(Conf &conf) { return true; }
		
		/// Lets the directive apply change.
		/** Like apply(), but will only be called after the directive
		  * has been reloaded, if and only if the new value is different
		  * than the old. This function is called before apply().
		  *
		  * \return true or false; the latter value will make the end()
		  * function return false too (conf.error() should have been called
		  * with an appropriate error message in that case). */
		virtual bool apply_change(Conf &conf) { return true; }
};

/*************************************************************************/

/// Class implementing boolean directive type.

/** This class is used to store bool values. If param is "yes"
  * the value is set to true, and if it is "no", the value is
  * set to false.
  *
  * A conversion operator to bool is also provided for easy value access. */

class ConfBool : public ConfValue<bool>
{
	public:
		/// Constructor setting the directive name.
		explicit ConfBool(std::string const &name, bool reloadable) :
				ConfValue<bool>(name, reloadable) { }
		/// Constructor setting the directive name and default value.
		ConfBool(std::string const &name, bool const &defval,
				bool reloadable) :
				ConfValue<bool>(name, defval, reloadable) { }
		/// Conversion operator.
		operator bool() const { return m_value; }
		
		/// Implements ConfDir::label().
		virtual char const *label(size_t index) const
				{ return (m_value ? "yes" : "no"); }
	protected:
		virtual bool assign(Conf &conf, unsigned short count,
				std::string const params[]);
};

/*************************************************************************/

/// Class implementing a duration directive type.

/** This class is used to store duration values.
  *
  * It converts string expressions of a duration to a number of seconds.
  * The conversion function used to do the job is duration_scan() from
  * the misc library. See its help for more information about valid
  * expressions.
  *
  * A conversion operator to time_t is provided for easy value access. */

class ConfDuration : public ConfValue<time_t>
{
	public:
		/// Constructor without default value.
		/** This constructor sets the directive name and whether it
		  * can be reloaded, along with the minimum and maximum duration
		  * that will be accepted when the directive is set.
		  * \param[in] name The name of the directive
		  * \param[in] reloadable true if the directive can be reloaded,
		  * false otherwise.
		  * \param[in] mindur The minimum duration to accept as a value.
		  * \param[in] maxdur The maximum duration to accept as a value. */
		explicit ConfDuration(std::string const &name, bool reloadable,
				time_t mindur, time_t maxdur) :
				ConfValue<time_t>(name, reloadable), m_mindur(mindur),
				m_maxdur(maxdur) { }
		/// Constructor with default value.
		/** This constructor sets the directive name, default value and whether
		  * it can be reloaded, along with the minimum and maximum duration
		  * that will be accepted when the directive is set.
		  * \param[in] name The name of the directive
		  * \param[in] defval The default value.
		  * \param[in] reloadable true if the directive can be reloaded,
		  * false otherwise.
		  * \param[in] mindur The minimum duration to accept as a value.
		  * \param[in] maxdur The maximum duration to accept as a value. */
		ConfDuration(std::string const &name, time_t defval,
				bool reloadable, time_t mindur,
				time_t maxdur) :
				ConfValue<time_t>(name, defval, reloadable), m_mindur(mindur),
				m_maxdur(maxdur) { }
		/// ConfArray-specific constructor
		ConfDuration(std::string const &name, bool reloadable) :
				ConfValue<time_t>(name, reloadable), m_mindur(0),
				m_maxdur(2147483647) { }
		/// Copy constructor.
		ConfDuration(ConfDuration const &cd) : ConfValue<time_t>(cd),
				m_mindur(cd.m_mindur), m_maxdur(cd.m_maxdur) { }
		/// Assignment operator.
		virtual ConfDuration &operator=(ConfDuration const &right);
		/// Conversion operator.
		operator time_t() const { return m_value; }
		
		/// Implements ConfDir::label().
		virtual char const *label(size_t index) const;
	protected:
		/// The minimum duration.
		time_t m_mindur;
		/// The maximum duration.
		time_t m_maxdur;
	
		virtual bool assign(Conf &conf, unsigned short count,
				std::string const params[]);
		
		/// Checks if the duration is in the range m_mindur..m_maxdur.
		virtual bool check(Conf &conf, time_t value) const;
};

/*************************************************************************/

/// Class implementing an enum directive type.

/** This class is used to store integer values that are set using predefined
  * string representations. 
  *
  * This class must not be used directly. Instead, it must be inherited,
  * and the add() function should be overriden to add all possible values
  * to m_values.
  *
  * The template parameter is the type in which values will be stored.
  * A conversion operator to this type is provided for easy access to the
  * value. */

template <class T = u_int32_t>
class ConfEnum : public ConfValue<T>
{
	public:
	    typedef std::map<std::string, T, misc::insensitive_less> ValueList;

		/// Constructor setting the directive name.
		explicit ConfEnum(std::string const &name, bool reloadable) :
				ConfValue<T>(name, reloadable), m_values() { }
		/// Constructor setting the directive name and default value.
		ConfEnum(std::string const &name, T const &defval, bool reloadable) :
				ConfValue<T>(name, defval, reloadable), m_values() { }
		/// Copy constructor.
		ConfEnum(ConfEnum const &cd) : ConfValue<T>(cd),
				m_values(cd.m_values) { }
		/// Assignment operator.
		virtual ConfEnum<T> &operator=(ConfEnum<T> const &right);
		/// Conversion operator.
		operator T const &() const { return ConfValue<T>::m_value; }
		
		/// Implements ConfDir::label().
		virtual char const *label(size_t index) const;
	protected:
		/// The list of values.
		ValueList m_values;
		
		/// Adds valid values to the list.
		/** This function adds valid values to the m_values list. It must
		  * be defined in an inherited class.
		  * Example of adding a value to the list.
		  * m_values["SOMETHING"] = 3;
		  * \param[in] conf The Conf object that called the set() function.
		  * \return true if successful, false otherwise. You must
		  * log an error message through conf.error() in that case. */
		virtual bool add(Conf &conf) = 0;
		
		/// Sets the value using its string representation in params[0].
		virtual bool assign(Conf &conf, unsigned short count,
				std::string const params[]);
};

/*************************************************************************/

/// Operator== for in_addr classes.

inline bool operator==(struct in_addr const &in1, struct in_addr const &in2)
{
	return (in1.s_addr == in2.s_addr);
}

/// Class implementing an hostname directive type.

/** This class is used to store IP addresses in a in_addr class.
  *
  * It takes a hostname (example: localhost) or a string IP address (example:
  * 127.0.0.1) and resolves/converts it to an integer IP address stored
  * in the s_addr field of struct in_addr. If the parameter is not a valid
  * hostname/IP address, then an error is generated.
  *
  * A conversion operator to struct in_addr is also provided for easier
  * value access. */

class ConfHostname : public ConfValue<struct in_addr>
{
	public:
		/// Constructor setting the directive name.
		explicit ConfHostname(std::string const &name,
				bool reloadable) :
				ConfValue<struct in_addr>(name, reloadable) { }
		/// Constructor setting the directive name and default value.
		ConfHostname(std::string const &name, struct in_addr const &defval,
				bool reloadable) :
				ConfValue<struct in_addr>(name, defval, reloadable) { }
		/// Conversion operator.
		operator struct in_addr const &() const { return m_value; }
		
		/// Implements ConfDir::label().
		virtual char const *label(size_t index) const;
	protected:
		virtual bool assign(Conf &conf, unsigned short count,
				std::string const params[]);
};

/*************************************************************************/

/// Class implementing an integer directive type.

/** This class is used to store integer values (the exact integer type
  * is specified by the template parameter).
  *
  * Conversions are done with strtol() (therefore the range of possible
  * values is LONG_MIN..LONG_MAX). This can be changed by overriding
  * the convert() and/or my_strtol() functions in derived classes. The
  * class also overrides check() to refine range checking to the caller's need.
  *
  * A conversion operator to the integer type T is provided as well for
  * easy access to the value. */

template <class T>
class ConfInt : public ConfValue<T>
{
	public:
		/// Constructor without default value.
		/** This constructor sets the directive name and whether it
		  * can be reloaded, along with the minimum and maximum values
		  * that will be accepted when the directive is set.
		  * \param[in] name The name of the directive
		  * \param[in] reloadable true if the directive can be reloaded,
		  * false otherwise.
		  * \param[in] minvalue The minimum to accept as a value.
		  * \param[in] maxvalue The maximum to accept as a value. */
		explicit ConfInt(std::string const &name, bool reloadable,
				T minvalue, T maxvalue) :
				ConfValue<T>(name, reloadable), m_minvalue(minvalue),
				m_maxvalue(maxvalue) { }
		/// Constructor with default value.
		/** This constructor sets the directive name, default value and
		  * whether it can be reloaded, along with the minimum and maximum
		  * values that will be accepted when the directive is set.
		  * \param[in] name The name of the directive.
		  * \param[in] defval The default value.
		  * \param[in] reloadable true if the directive can be reloaded,
		  * false otherwise.
		  * \param[in] minvalue The minimum to accept as a value.
		  * \param[in] maxvalue The maximum to accept as a value. */
		ConfInt(std::string const &name, T const &defval, bool reloadable,
				T minvalue, T maxvalue) :
				ConfValue<T>(name, defval, reloadable), m_minvalue(minvalue),
				m_maxvalue(maxvalue) { }
		/// ConfArray-specific constructor.
		ConfInt(std::string const &name, bool reloadable) :
				ConfValue<T>(name, reloadable), m_minvalue(LONG_MIN),
				m_maxvalue(LONG_MAX) { }
		/// Copy constructor.
		ConfInt(ConfInt const &cd) : ConfValue<T>(cd),
				m_minvalue(cd.m_minvalue), m_maxvalue(cd.m_maxvalue) { }
		/// Assignment operator.
		virtual ConfInt<T> &operator=(ConfInt<T> const &right);
		/// Conversion operator.
		operator T() const { return ConfValue<T>::m_value; }
		
		/// Implements ConfDir::label().
		virtual char const *label(size_t index) const;
	protected:
		T m_minvalue;
		T m_maxvalue;
	
		virtual bool assign(Conf &conf, unsigned short count,
				std::string const params[]);
		
		/// Checks if the value is in the m_minvalue..m_maxvalue range.
		virtual bool check(Conf &conf, T const &value) const;
		
		/// Converts the string to an integer.
		/** Converts the given string to an integer, if possible.
		  * The my_strtol() function of the class is used to do the
		  * conversion.
		  * \param[in]  conf The Conf object that called the set() function.
		  * \param[in]  expr The string to convert to an integer.
		  * \param[out] result The function will set this variable before
		  * returning if successful.
		  * \return true if the conversion went well, false otherwise. */
		virtual bool convert(Conf &conf, std::string const &expr,
				T &result) const;
		
		/// Calls the strtol() function.
		/** This function just calls the strtol() function. You may
		  * override it to call other strtol-like functions... Parameters
		  * are the same than those of the strtol() function, but the return
		  * value is different. */
		virtual T my_strtol(char const *nptr, char **endptr, int base) const
				{ return strtol(nptr, endptr, base); }
};

/*************************************************************************/

/// Class implementing a non-zero integer directive type.

/** This class is used to store integer values (the exact
  * integer type is specified by the template parameter). It is exactly
  * the same than ConfInt but it adds an additional check to ensure
  * the value is non-zero. */

template <class T>
class ConfIntNZ : public ConfInt<T>
{
	public:
		/// Constructor without default value.
		explicit ConfIntNZ(std::string const &name, bool reloadable,
				T minvalue, T maxvalue) :
				ConfInt<T>(name, reloadable, minvalue, maxvalue) { }
		/// Constructor with default value.
		ConfIntNZ(std::string const &name, T const &defval, bool reloadable,
				T minvalue, T maxvalue) :
				ConfInt<T>(name, defval, reloadable, minvalue, maxvalue) { }
		/// ConfArray-specific constructor.
		ConfIntNZ(std::string const &name, bool reloadable) :
				ConfInt<T>(name, reloadable) { }
	protected:
		/// Checks if the value != 0 then calls ConfInt<T>::check().
		virtual bool check(Conf &conf, T const &value) const;
};

/*************************************************************************/

/// Like ConfInt, but with defaults suitable for port numbers.

class ConfPort : public ConfInt<uint16_t>
{
	public:
		/// Constructor without default value.
		explicit ConfPort(std::string const &name, bool reloadable) :
				ConfInt<uint16_t>(name, reloadable, 0, 65535) { }
		/// Constructor with default value.
		ConfPort(std::string const &name, uint16_t const &defval,
				bool reloadable) : ConfInt<uint16_t>(name, defval,
				reloadable, 0, 65535) { }
};

/*************************************************************************/

/// Class implementing a ratio directive type.

/** This class is used to store ratio values.
  *
  * It converts string expressions of a ratio in the form x:y where
  * x and y are integers, to a value of the Ratio class. */

class ConfRatio : public ConfValue<misc::Ratio>
{
	public:
		/// Constructor without default value.
		/** This constructor sets the directive name and whether it
		  * can be reloaded, along with the minimum and maximum X and Y
		  * values that will be accepted when the directive is set.
		  * \param[in] name The name of the directive
		  * \param[in] reloadable true if the directive can be reloaded,
		  * false otherwise.
		  * \param[in] minx The minimum X to accept as a value.
		  * \param[in] maxx The maximum X to accept as a value.
		  * \param[in] miny The minimum Y to accept as a value.
		  * \param[in] maxy The maximum Y to accept as a value. */
		explicit ConfRatio(std::string const &name, bool reloadable,
				int32_t minx, int32_t maxx, int32_t miny, int32_t maxy) :
				ConfValue<misc::Ratio>(name, reloadable), m_minx(minx),
				m_maxx(maxx), m_miny(miny), m_maxy(maxy) { }
		/// Constructor with default value.
		/** This constructor sets the directive name, default value and whether
		  * it can be reloaded, along with the minimum and maximum X and Y
		  * values that will be accepted when the directive is set.
		  * \param[in] name The name of the directive
		  * \param[in] defval The default value.
		  * \param[in] reloadable true if the directive can be reloaded,
		  * false otherwise.
		  * \param[in] minx The minimum X to accept as a value.
		  * \param[in] maxx The maximum X to accept as a value.
		  * \param[in] miny The minimum Y to accept as a value.
		  * \param[in] maxy The maximum Y to accept as a value. */
		ConfRatio(std::string const &name, misc::Ratio const &defval,
				bool reloadable, int32_t minx, int32_t maxx, int32_t miny,
				int32_t maxy) : ConfValue<misc::Ratio>(name, defval, reloadable),
				m_minx(minx), m_maxx(maxx), m_miny(miny), m_maxy(maxy) { }
		/// ConfArray-specific constructor
		ConfRatio(std::string const &name, bool reloadable) :
				ConfValue<misc::Ratio>(name, reloadable), m_minx(-2147483647),
				m_maxx(2147483647), m_miny(-2147483647), m_maxy(2147483647) { }
		/// Copy constructor.
		ConfRatio(ConfRatio const &cd) : ConfValue<misc::Ratio>(cd),
				m_minx(cd.m_minx), m_maxx(cd.m_maxx),
				m_miny(cd.m_miny), m_maxy(cd.m_maxy) { }
		/// Assignment operator.
		virtual ConfRatio &operator=(ConfRatio const &right);
		
		/// Implements ConfDir::label().
		virtual char const *label(size_t index) const;
	protected:
		/// The minimum X value.
		time_t m_minx;
		/// The maximum X value.
		time_t m_maxx;
		/// The minimum Y value.
		time_t m_miny;
		/// The maximum Y value.
		time_t m_maxy;
	
		virtual bool assign(Conf &conf, unsigned short count,
				std::string const params[]);
		
		/// Checks for ranges of value.x() and value.y().
		virtual bool check(Conf &conf, misc::Ratio value) const;
};

/*************************************************************************/

/// Class implementing a set directive type.

/** This class is used to store "bits" in 32-bits int values. Each bit
  * has a name that will be used in configuration to set or clear
  * the bit to and from the value. 
  * 
  * For example, for a bit named "SOMETHING", if a parameter "+SOMETHING"
  * is set to a directive of this type, then the SOMETHING bit will be set to
  * the value, and if a parameter "-SOMETHING" is set to a directive of this
  * type, then the SOMETHING bit will be cleared. Of course, directives of this
  * type may be set multiple times to set or clear bits.
  *
  * This class must not be used directly. Instead, it must be inherited,
  * and the add() function should be overriden to add all possible bits
  * to m_bits.
  *
  * The template parameter is the type in which bits will be stored.
  * A conversion operator to this type is provided for easy access to the
  * value. */

template <class T = u_int32_t>
class ConfSet : public ConfValue<T>
{
	public:
	    typedef std::map<std::string, T, misc::insensitive_less> BitList;

		/// Constructor setting the directive name.
		explicit ConfSet(std::string const &name, bool reloadable) :
				ConfValue<T>(name, reloadable), m_bits() { }
		/// Constructor setting the directive name and default value.
		ConfSet(std::string const &name, T const &defval, bool reloadable) :
				ConfValue<T>(name, defval, reloadable), m_bits() { }
		/// Copy constructor.
		ConfSet(ConfSet const &cd) : ConfValue<T>(cd), m_bits(cd.m_bits) { }
		/// Assignment operator.
		virtual ConfSet<T> &operator=(ConfSet<T> const &right);
		/// Conversion operator.
		operator T const &() const { return ConfValue<T>::m_value; }
			
		/// Implements ConfDir::label().
		virtual char const *label(size_t index) const;
	protected:
		/// The list of bits.
		BitList m_bits;
		
		/// Adds valid bits to the list.
		/** This function adds valid bits to the m_bits list. It must
		  * be defined in an inherited class.
		  * Example of adding a bit to the list.
		  * m_bits["SOMETHING"] = 0x00080000;
		  * \param[in] conf The Conf object that called the set() function.
		  * \return true if successful, false otherwise. You must
		  * log an error message through conf.error() in that case. */
		virtual bool add(Conf &conf) = 0;
		
		virtual bool assign(Conf &conf, unsigned short count,
				std::string const params[]);
};

/*************************************************************************/

/// Class implementing a string directive type.

/** This class is used to store string values. So as you guess, there's
  * no conversion going on... it's just a copy. */

class ConfStr : public ConfValue<std::string>
{
	public:
		/// Constructor without default value.
		/** This constructor sets the directive name and whether it
		  * can be reloaded, along with the minimum and maximum string
		  * length that will be accepted when the directive is set.
		  * \param[in] name The name of the directive
		  * \param[in] reloadable true if the directive can be reloaded,
		  * false otherwise.
		  * \param[in] minlen The minimum length to accept as a value.
		  * \param[in] maxlen The maximum length to accept as a value.
		  * \param[in] validchars All characters that may be in the
		  * value (an error is generated if a character which is not in
		  * this string is used). Can be an empty string to accept all
		  * characters. */
		explicit ConfStr(std::string const &name, bool reloadable,
				std::string::size_type minlen,
				std::string::size_type maxlen, std::string const &validchars) :
				ConfValue<std::string>(name, reloadable), m_minlen(minlen),
				m_maxlen(maxlen), m_validchars(validchars) { }
		/// Constructor with default value.
		/** This constructor sets the directive name, default value and whether
		  * it can be reloaded, along with the minimum and maximum string
		  * length that will be accepted when the directive is set.
		  * \param[in] name The name of the directive
		  * \param[in] defval The default value.
		  * \param[in] reloadable true if the directive can be reloaded,
		  * false otherwise.
		  * \param[in] minlen The minimum length to accept as a value.
		  * \param[in] maxlen The maximum length to accept as a value.
		  * \param[in] validchars All characters that may be in the
		  * value (an error is generated if a character which is not in
		  * this string is used). Can be an empty string to accept all
		  * characters. */
		ConfStr(std::string const &name, std::string const &defval,
				bool reloadable, std::string::size_type minlen,
				std::string::size_type maxlen, std::string const &validchars) :
				ConfValue<std::string>(name, defval, reloadable),
				m_minlen(minlen), m_maxlen(maxlen), m_validchars(validchars)
				{ }
		/// ConfArray-specific constructor.
		ConfStr(std::string const &name, bool reloadable) :
				ConfValue<std::string>(name, reloadable), m_minlen(0),
				m_maxlen(2147483647), m_validchars("") { }
		/// Copy constructor.
		ConfStr(ConfStr const &cd) : ConfValue<std::string>(cd),
				m_minlen(cd.m_minlen), m_maxlen(cd.m_maxlen),
				m_validchars(cd.m_validchars) { }
		/// Assignment operator.
		virtual ConfStr &operator=(ConfStr const &right);
		/// == operator
		bool operator==(ConfStr const &right) const
				{ return (m_value == right.m_value); }
		/// == operator for std::string r-values
		bool operator==(std::string const &right) const
				{ return (m_value == right); }
		/// != operator
		bool operator!=(ConfStr const &right) const
				{ return (m_value != right.m_value); }
		/// != operator for std::string r-values
		bool operator!=(std::string const &right) const
				{ return (m_value != right); }
		/// < operator
		bool operator<(ConfStr const &right) const
				{ return (m_value < right.m_value); }
		/// < operator for std::string r-values
		bool operator<(std::string const &right) const
				{ return (m_value < right); }
		/// <= operator
		bool operator<=(ConfStr const &right) const
				{ return (m_value <= right.m_value); }
		/// <= operator for std::string r-values
		bool operator<=(std::string const &right) const
				{ return (m_value <= right); }
		/// > operator
		bool operator>(ConfStr const &right) const
				{ return (m_value > right.m_value); }
		/// > operator for std::string r-values
		bool operator>(std::string const &right) const
				{ return (m_value > right); }
		/// >= operator
		bool operator>=(ConfStr const &right) const
				{ return (m_value >= right.m_value); }
		/// >= operator for std::string r-values
		bool operator>=(std::string const &right) const
				{ return (m_value >= right); }
		
		/// Implements ConfDir::label().
		virtual char const *label(size_t index) const
				{ return m_value.c_str(); }
	protected:
		std::string::size_type m_minlen;
		std::string::size_type m_maxlen;
		std::string m_validchars;
	
		virtual bool assign(Conf &conf, unsigned short count,
				std::string const params[]);
		
		/// Checks if the string length is in the range m_minlen..m_maxlen.
		virtual bool check(Conf &conf, std::string const &value) const;
};

/*************************************************************************/

/// Like ConfStr but adds a conversion operator to char *.

class ConfStrChar : public ConfStr
{
	public:
		/// Constructor without default value.
		explicit ConfStrChar(std::string const &name, bool reloadable,
				std::string::size_type minlen, std::string::size_type maxlen,
				std::string const &validchars) :
				ConfStr(name, reloadable, minlen, maxlen, validchars) { }
		/// Constructor with default value.
		ConfStrChar(std::string const &name, std::string const &defval,
				bool reloadable, std::string::size_type minlen,
				std::string::size_type maxlen, std::string const &validchars) :
				ConfStr(name, defval, reloadable, minlen, maxlen,
				validchars) { }
		/// ConfArray-specific constructor.
		ConfStrChar(std::string const &name, bool reloadable) :
				ConfStr(name, reloadable) { }
		/// Conversion operator.
		operator char const *() const
				{ return (m_value.size() ? m_value.c_str() : 0); }
};

/*************************************************************************/

/// Like ConfStr but adds a conversion operator to std::string.

class ConfString : public ConfStr
{
	public:
		/// Constructor without default value.
		explicit ConfString(std::string const &name, bool reloadable,
				std::string::size_type minlen, std::string::size_type maxlen,
				std::string const &validchars) :
				ConfStr(name, reloadable, minlen, maxlen, validchars) { }
		/// Constructor with default value.
		ConfString(std::string const &name, std::string const &defval,
				bool reloadable, std::string::size_type minlen,
				std::string::size_type maxlen, std::string const &validchars) :
				ConfStr(name, defval, reloadable, minlen, maxlen,
				validchars) { }
		/// ConfArray-specific constructor.
		ConfString(std::string const &name, bool reloadable) :
				ConfStr(name, reloadable) { }
		/// [] operator
		char operator[](std::string::size_type idx) const
				{ return m_value[idx]; }
		/// Conversion operator.
		operator std::string() const { return m_value; }
};

/*************************************************************************/

/// Class implementing an unsigned integer directive type.

/** This class is used to store unsigned integer values (the exact
  * integer type is specified by the template parameter). It inherits
  * from ConfInt.
  *
  * Conversions are done with strtoul() instead of strtol() (therefore the
  * range of possible values is 0..ULONG_MAX). */

template <class T>
class ConfUint : public ConfInt<T>
{
	public:
		/// Constructor without default value.
		/** This constructor sets the directive name and whether it
		  * can be reloaded, along with the minimum and maximum values
		  * that will be accepted when the directive is set.
		  * \param[in] name The name of the directive.
		  * \param[in] reloadable true if the directive can be reloaded,
		  * false otherwise.
		  * \param[in] minvalue The minimum to accept as a value.
		  * \param[in] maxvalue The maximum to accept as a value. */
		explicit ConfUint(std::string const &name, bool reloadable,
				T minvalue, T maxvalue) :
				ConfInt<T>(name, reloadable, minvalue, maxvalue) { }
		/// Constructor with default value.
		/** This constructor sets the directive name, default value and
		  * whether it can be reloaded, along with the minimum and maximum
		  *  values that will be accepted when the directive is set.
		  * \param[in] name The name of the directive.
		  * \param[in] defval The default value.
		  * \param[in] reloadable true if the directive can be reloaded,
		  * false otherwise.
		  * \param[in] minvalue The minimum to accept as a value.
		  * \param[in] maxvalue The maximum to accept as a value. */
		ConfUint(std::string const &name, T const &defval, bool reloadable,
				T minvalue, T maxvalue) :
				ConfInt<T>(name, defval, reloadable, minvalue, maxvalue) { }
		/// ConfArray-specific constructor.
		ConfUint(std::string const &name, bool reloadable) :
				ConfInt<T>(name, reloadable) { }
		
		/// Implements ConfDir::label().
		virtual char const *label(size_t index) const;
	protected:
		/// Checks if the value is in the m_minvalue..m_maxvalue range.
		virtual bool check(Conf &conf, T const &value) const;
		
		/// Ensures expr doesn't start with - then calls ConfInt<T>::convert().
		virtual bool convert(Conf &conf, std::string const &expr,
				T &result) const;
		/// Calls the strtoul() function.
		virtual T my_strtol(char const *nptr, char **endptr, int base) const
				{ return strtoul(nptr, endptr, base); }
};

/*************************************************************************/

/// Class implementing a URI directive type.

/** This class is used to store URI values.
  *
  * It converts string expressions of a URI using the misc::URI class. */

class ConfURI : public ConfValue<misc::URI>
{
	public:
		/// Constructor without default value.
		explicit ConfURI(std::string const &name, bool reloadable) :
				ConfValue<misc::URI>(name, reloadable) { }
		/// Constructor with default value.
		ConfURI(std::string const &name, misc::URI const &defval,
					bool reloadable) : 
					ConfValue<misc::URI>(name, defval, reloadable) { }
		
		/// Implements ConfDir::label().
		virtual char const *label(size_t index) const;
	protected:
		virtual bool assign(Conf &conf, unsigned short count,
				std::string const params[]);
};

/*************************************************************************/

/// Class implementing an array of values directive type.

/** This class is a metaclass implementing an array of values. To achieve
  * that, it creates new ConfDir (or derived) objects on the fly, giving
  * them a name in the form name[n] where name is the directive name and n
  * is their index, and calls begin(), set() and end() to set them accordingly.
  *
  * The template also lets you choose the container type you want to
  * use to implement the array (it will work as long as the container has
  * insert(iterator pos, T const &value), clear() and size() functions --
  * which is the case of all containers of the C++ STL.
  *
  * Operators * and -> are overloaded so you can easily access the internal
  * container object. A conversion operator to the container type is also
  * provided.
  *
  * Example of use: 
  *
  * ConfArray<ConfBool, std::vector<ConfBool> > array;
  *
  * Caveat: 
  *
  * - when using ConfArray for ConfBlock values, apply_change() will be
  * called at each reload, regardless of whether there really was a change
  * or not (ConfBlock::operator== always yields false). You should override
  * ConfBlock::operator== in derived classes if you want to make sure it
  * is called only when there are changes. */

template <class ValueT, class ContainerT = std::vector<ValueT> >
class ConfArray : public ConfValue <ContainerT>
{
	public:
		typedef ValueT value_type;
		typedef ContainerT container_type;
	
		/// Constructor.
		/** This constructor sets the directive name and the
		  * maximum number of values that can be added to the
		  * container.
		  * \param[in] name The directive name.
		  * \param[in] reloadable Whether the directive can be reloaded or
		  *            not.
		  * \param[in] max The maximum number of values. */
		explicit ConfArray(std::string const &name, bool reloadable = true,
				typename ContainerT::size_type max = INT_MAX) :
				ConfValue <ContainerT>(name, reloadable), m_max_values(max) { }
		/// Constructor setting a default value.
		/** This constructor sets the directive name, default value
		  * and the maximum number of values that can be added to the
		  * container.
		  * \param[in] name The directive name.
		  * \param[in] defval The default value.
		  * \param[in] reloadable Whether the directive can be reloaded or
		  *            not.
		  * \param[in] max The maximum number of values. */
		ConfArray(std::string const &name, ContainerT const &defval,
				bool reloadable = true,
				typename ContainerT::size_type max = INT_MAX) :
				ConfValue <ContainerT>(name, defval, reloadable),
						m_max_values(max) { }
		/// Copy constructor.
		ConfArray(ConfArray const &cd) : ConfValue<ContainerT>(cd),
				m_max_values(cd.m_max_values) { }
		/// Assignment operator.
		virtual ConfArray<ValueT, ContainerT>
				&operator=(ConfArray<ValueT, ContainerT> const &right);
		/// Conversion operator.
		operator ContainerT const &() const { return ConfValue <ContainerT>::m_value; }
		/// Provides access to the underlying container.
		ContainerT const &operator*() const { return ConfValue <ContainerT>::m_value; }
		/// Provides access to the underlying container.
		ContainerT const *operator->() const { return &(this->ConfValue<ContainerT>::m_value); }
		
		/// Implements ConfDir::label().
		virtual char const *label(size_t index) const;
		/// Implements ConfDir::label_count().
		virtual size_t label_count() const
				{ return ConfValue <ContainerT>::m_value.size(); }
	protected:
		typename ContainerT::size_type m_max_values;
		
		virtual bool assign(Conf &conf, unsigned short count,
				std::string const params[]);
		
		/// Inserts a new ValueT object in the container.
		/** Inserts a new ValueT directive (that the function is responsible
		  * to create) in the container, through a call to the
		  * ContainerT::insert(iterator, ValueT) function. It must name the
		  * new directive as the name[n] where name is this directive name
		  * and n is ContainerT::size().
		  *
		  * \return An iterator pointing to the newly-added directive. */
		virtual typename ContainerT::iterator insert();
};

/*************************************************************************/

typedef ConfArray<ConfBool> ConfBoolA;
typedef ConfArray<ConfDuration> ConfDurationA;
typedef ConfArray<ConfHostname> ConfHostnameA;
typedef ConfArray<ConfPort> ConfPortA;
typedef ConfArray<ConfStr> ConfStrA;
typedef ConfArray<ConfStrChar> ConfStrCharA;
typedef ConfArray<ConfString> ConfStringA;

/*************************************************************************/
/************************ Block directives classes ***********************/
/*************************************************************************/

/// Class implementing a block directive type.

/** This class implements a type of directive taking a block of directives
  * as its value.
  *
  * It can't be used directly; instead, you must create a derived class
  * that will override the add() function to add new directives to the
  * Conf object that is created in the begin() function. 
  * 
  * In your derived classes, you'll generally want to aggregate the directives
  * objects too. That'll allow you to use the ConfBlock class with the
  * ConfArray template class. */

class ConfBlock : public ConfDir
{
	public:
		/// Constructor setting directive name and reloadability.
		explicit ConfBlock(std::string const &name, bool reloadable = true) :
				ConfDir(name, reloadable, 1, 1), m_conf(0) { }
		/// Copy constructor.
		ConfBlock(ConfBlock const &cd);
		/// Destructor.
		virtual ~ConfBlock();
		/// Assignment operator.
		virtual ConfBlock &operator=(ConfBlock const &right);
		/// == operator
		/** This operator is provided in order to use the ConfArray
		  * template class. It always return false though (how the heck
		  * would we check reliably that two directives have the same
		  * value?! We don't even know if they _have_ a value! ;) */
		virtual bool operator==(ConfBlock const &) const { return false; }
		
		/// Begin parsing.
		/** Creates the conf parser, deleting the old parser if necessary,
		  * then calls add() to initialize the list of directives, then
		  * Conf::begin() to initialize it and sets the m_conf object to
		  * the newly-created object. */
		virtual bool begin(Conf &conf);
		/// Parses the block.
		/** Parses the block of directives in params[0]. */
		virtual bool set(Conf &conf, unsigned short count,
				std::string const params[]);
		/// End parsing.
		/** Calls apply_changes() and m_conf->end(). */
		virtual bool end(Conf &conf);
		
		/// Implements ConfDir::sub().
		virtual Conf const *sub() const { return m_conf; }
	protected:
		/// Created parser -- 0 until first successful call to begin().
		Conf *m_conf;
	
		/// Adds directives to a Conf object.
		/** This function must be overriden in derived classes to add 
		  * directives to the given Conf object.
		  *
		  * \param conf The Conf object that called the ConfBlock::begin()
		  * function. DO NOT add directives to this object! ;)
		  * \param new_conf The Conf object to which you must add the
		  * directives.
		  *
		  * \return true if initialization went well. Else it returns false and
		  * an error message should have been logged through conf.error(). */
		virtual bool add(Conf &conf, Conf &new_conf) = 0;
		
		/// Applies changes.
		/** This function is called automatically by end() whenever the
		  * configuration is reloaded. Use it to apply changes to the
		  * program if necessary.
		  * \return true if successful, false otherwise (which will
		  * make end() return false too.) */
		virtual bool apply_changes(Conf &conf) { return true; }
};

/*************************************************************************/

/// Class implementing a boolean block directive type.

/** This class implements a boolean block directive type, derived from
  * ConfBlock. In this class, m_conf will only be initialized from the set()
  * function; this means that, if the directive is not encountered by the
  * parser, the directives of the block will be totally ignored. They won't be
  * set to their default value, or anything like that, because their begin()
  * and end() functions will NEVER BE CALLED !
  *
  * This is useful when implementing a directive that should be able
  * to turn off a feature altogether when it is not set explicitely. If you
  * were using a ConfBlock class, if there was a directive with no default
  * value in the block, then the end() function would return false, even
  * if the directive was not set explicitely, which is what this class
  * will prevent. 
  
  * Of course, if the directive is set, and you don't set a value for a
  * required directive in the block, end() will still cause an error. To sum
  * up, this class implements a block directive that is optional, but that
  * must have values for all of its directives whenever it is set explicitely!
  * I hope it is clear. :P
  *
  * There is a conversion operator to bool that is useful to know whether the
  * directive was set or not (hence the name of the class ;). */

class ConfBoolBlock : public ConfBlock
{
	public:
		/// Constructor setting directive name and reloadability.
		explicit ConfBoolBlock(std::string const &name,
				bool reloadable = true) : ConfBlock(name, reloadable),
				m_set(false) { }
		/// Copy constructor.
		ConfBoolBlock(ConfBoolBlock const &cd) : ConfBlock(cd),
				m_set(cd.m_set) { }
		/// Assignment operator.
		virtual ConfBoolBlock &operator=(ConfBoolBlock const &right);
		/// Conversion operator.
		operator bool() const { return m_set; }
		
		/// Just sets m_set to false.
		virtual bool begin(Conf &) { m_set = false; return true; }
		/// Calls ConfBlock::begin(), then ConfBlock::set() and sets m_set.
	 	virtual bool set(Conf &conf, unsigned short count,
				std::string const params[]);
		
		/// Implements ConfDir::label().
		virtual char const *label(size_t index) const
				{ return (m_set ? "yes" : "no"); }
		/// Implements ConfDir::label_count().
		virtual size_t label_count() const { return 1; }
	protected:
		bool m_set;
};

/*************************************************************************/

typedef ConfArray<ConfBlock> ConfBlockA;
typedef ConfArray<ConfBoolBlock> ConfBoolBlockA;

/*************************************************************************/
/********************* Specialized directives classes ********************/
/*************************************************************************/

/// Class implementing a host/port block directive type.

/** This class implements a block directive type that have two directives
  * "hostname" and "port" of types ConfHostname and ConfPort. */

class ConfHostPort : public ConfBlock
{
	public:
		/* Data */
		/// hostname directive object.
		ConfHostname hostname;
		/// port directive object.
		ConfPort port;
		
		/* Functions */
		/// Constructor setting directive name and reloadability.
		explicit ConfHostPort(std::string const &name,
				bool reloadable = true) : ConfBlock(name, reloadable),
				hostname("hostname", reloadable), port("port", reloadable) { }
		/// Constructor setting the directive name and default values.
		ConfHostPort(std::string const &name, struct in_addr const &defhost, 
				uint16_t defport, bool reloadable = true) : 
				ConfBlock(name, reloadable), 
				hostname("hostname", defhost, reloadable),
				port("port", defport, reloadable) { }
		/// Constructor setting the directive name and port default value.
		ConfHostPort(std::string const &name, uint16_t defport,
				bool reloadable = true) : ConfBlock(name, reloadable), 
				hostname("hostname", reloadable),
				port("port", defport, reloadable) { }
		/// Copy constructor.
		ConfHostPort(ConfHostPort const &cd) : ConfBlock(cd),
				hostname(cd.hostname), port(cd.port) { }
		/// Assignment operator.
		virtual ConfHostPort &operator=(ConfHostPort const &right);
	protected:
		virtual bool add(Conf &conf, Conf &new_conf);
};

/*************************************************************************/

/// Class implementing a list of rules directive type.

/* This class implements a list of rules block directive type. There
 * is no directive in the block by default; you should create them using
 * the ConfRule class in derived classes. */

class ConfRules : public ConfBlock
{
	public:
		/* Classes */
		
		/// Class representing a rule.
		/** This class is the representation of a rule, which has a
		  * type and a value (both caller-defined). */
		class Rule
		{
			public:
				/// Constructor.
				/** Initializes a Rule object.
				  * \param[in] type The type of the rule (caller-defined).
				  * \param[in] value The value of the rule (caller-defined). */
				Rule(int type, std::string const &value) : m_type(type),
						m_value(value) { }
				/// Copy constructor.
				Rule(Rule const &r) : m_type(r.m_type),
						m_value(r.m_value) { }
				/// Assignment operator.
				Rule &operator=(Rule const &right);
				
				/// Returns the rule type.
				int type() const { return m_type; }
				/// Returns the rule value.
				std::string const &value() const { return m_value; }
			protected:
				/// The rule type.
				int m_type;
				/// The rule value.
				std::string m_value;
		};
		
		/* Typedefs */
		typedef std::vector<Rule> container_type;
		
		/* Classes */
		
		/// Rule configuration directive.
		/** This class implements a rule configuration directive,
		  * that may be used to the ConfRules block directive.
		  * The check() function may be overriden to ensure
		  * the value is correct. */ 
		class ConfRule : public ConfDir
		{
			public:
				/// Constructor.
				/** Constructor.
				  * \param[in] name The directive name.
				  * \param[in] reloadable Determines whether the
				  * directive is reloadable or not.
				  * \param[in] container The container to which
				  * new rules are added by this directive (must
				  * be valid until the directive is destroyed).
				  * \param[in] type The type of rule this directive
				  * creates. */
				ConfRule(std::string const &name, bool reloadable,
						ConfRules::container_type &container, int type) :
						ConfDir(name, reloadable, 1, 1),
						m_container(&container), m_type(type) { }
				/// Copy constructor.
				ConfRule(ConfRule const &cd) : ConfDir(cd),
						m_container(cd.m_container), m_type(cd.m_type) { }
				/// Assignment operator.
				virtual ConfRule &operator=(ConfRule const &right);
				
				/// Creates a new rule and adds it to the list.
				virtual bool set(Conf &conf, unsigned short count,
						std::string const params[]);
				
				/// Returns the container.
				ConfRules::container_type *container() const
						{ return m_container; }
				/// Sets the container.
				ConfRules::container_type *
						container(ConfRules::container_type &c)
						{ return (m_container = &c); }
			protected:
				/// Checks whether a value is correct.
				/** Checks whether a value is correct. This is
				  * called automatically by the set() function.
				  *
				  * \param[in] value The value to check.
				  * \return true if the value is correct, false
				  * otherwise (which will make set() return false
				  * too).
				  * \sa set() */
				virtual bool check(std::string const &value) const
						{ return true; }
			protected:
				/// The container to which new rules are added.
				ConfRules::container_type *m_container;
				/// The type of created rules.
				int m_type;
		};
		
		/* Functions */
		
		/// Constructor setting directive name and reloadability.
		ConfRules(std::string const &name, bool reloadable) :
				ConfBlock(name, reloadable), m_rules() { }
		/// Copy constructor.
		ConfRules(ConfRules const &cd) : ConfBlock(cd), m_rules(cd.m_rules) { }
		/// Assignment operator
		virtual ConfRules &operator=(ConfRules const &right);
		
		/// Clears the rules list on reload.
		virtual bool begin(Conf &conf);
		
		/// Returns the list of rules.
		container_type const &rules() const { return m_rules; }
	protected:
		/// List of rules.
		container_type m_rules;
};

/*************************************************************************/
/********************* Informative directives classes ********************/
/*************************************************************************/

/// Class implementing a warning directive type.

/** This class will log a programmer-defined warning message through
  * Conf::error() when set. */

class ConfWarning : public ConfDir
{
	public:
		/// Constructor setting directive name and default message.
		explicit ConfWarning(std::string const &name,
				std::string const &msg) : ConfDir(name, true, 0, 65535),
				m_msg(msg) { }
		/// Copy constructor.
		ConfWarning(ConfWarning const &cd) : ConfDir(cd),
				m_msg(cd.m_msg) { }
		/// Assignment operator.
		virtual ConfWarning &operator=(ConfWarning const &right);
		
		/// Logs a message through Conf::error().
		virtual bool set(Conf &conf, unsigned short count,
				std::string const params[]);
		
		/// Returns the message what will be logged when directive is set.
		std::string const &msg() const { return m_msg; }
	protected:
		std::string m_msg;
};

/*************************************************************************/

/// Class implementing an error directive type.

/** Same as ConfWarning, but set will return false, thus stopping
  * the configuration parsing. */

class ConfError : public ConfWarning
{
	public:
		/// Constructor setting directive name.
		explicit ConfError(std::string const &name, std::string const &msg) :
				ConfWarning(name, msg) { }
		
		/// Calls ConfWarning::set and return false.
		virtual bool set(Conf &conf, unsigned short count,
				std::string const params[]);
};

/*************************************************************************/
/*********************** Global directives classes ***********************/
/*************************************************************************/

/// Class implementing else global configuration directive.

/** This class implements the else global configuration directive.
  *
  * ConfIf directives (and derived) must call the can_process() member
  *  function of the ConfElse object (obtained through 
  * conf.global().else_directive() */

class ConfElse : public ConfDir
{
	public:
		/// Constructor setting the name of the directive.
		explicit ConfElse(std::string const &name) : ConfDir(name, true, 1, 1),
				m_process(false) { }
		/// Copy constructor.
		ConfElse(ConfElse const &cd) : ConfDir(cd), m_process(cd.m_process) { }
		/// Assignment operator
		virtual ConfElse &operator=(ConfElse const &cd);
		
		/// Determines whether to process params[0] next time.
		/** Determines whether to process params[0] next time the directive
		  * is used.
		  * \return true if it must be processed, false otherwise.
		  */
		bool can_process() const { return m_process; }
		/// Sets whether to process params[0] next time.
		/** Sets whether to process params[0] next time the directive
		  * is used.
		  * \param process Set this to true to process params[0]
		  * next time, false otherwise.
		  * \return true if params[0] will be processed next time, 
		  * false otherwise. */
		bool can_process(bool process) { return (m_process = process); }
		
		/// Executes params[0] if (can_process() == true).
		virtual bool set(Conf &conf, unsigned short count, std::string const
				params[]);
	private:
		bool m_process;
};

/*************************************************************************/

/// Base class for all conditional directives.

/** This class implements the basics of conditional directives.
  *
  * Conditional directives can inherit this class and override
  * the can_process() member function to have a brand new and fully
  * functional conditional directive. */

class ConfIf : public ConfDir
{
	public:
		/// Constructor setting the name of the directive.
		explicit ConfIf(std::string const &name, unsigned short min = 2,
				unsigned short max = 2) : ConfDir(name, true, min, max) { }
		
		/// Calls can_process(params[0]) and processes params[1] if true.
		virtual bool set(Conf &conf, unsigned short count,
				std::string const params[]);
	
	protected:
		/// Determines whether to process the block.
		/** Determines whether to process the block of directives in params[1],
		  * based on the evaluation of the given expression (which is
		  * implementer-defined).
		  * \param[in] conf The Conf object that called set().
		  * \param[in] expr The expression to evaluate.
		  * \return true to process the block, false otherwise. */
		virtual bool can_process(Conf &conf, std::string const &expr) const = 0;
};

/*************************************************************************/

/// Class implementing ifnset global configuration directive.

/** This class implements the ifnset global configuration directive. */

class ConfIfnset : public ConfIf
{
	public:
		/// Constructor setting the name of the directive.
		explicit ConfIfnset(std::string const &name) : ConfIf(name) { }
		
	protected:
		/// Checks whether the variable expr doesn't exist.
		virtual bool can_process(Conf &conf, std::string const &expr) const;
};

/*************************************************************************/

/// Class implementing ifset global configuration directive.

/** This class implements the ifset global configuration directive. */

class ConfIfset : public ConfIf
{
	public:
		/// Constructor setting the name of the directive.
		explicit ConfIfset(std::string const &name) : ConfIf(name) { }
		
	protected:
		/// Checks whether the variable expr exists.
		virtual bool can_process(Conf &conf, std::string const &expr) const;
};

/*************************************************************************/

/// Class implementing include global configuration directive.

/** This class implements the include global configuration directive. */

class ConfInclude : public ConfDir
{
	public:
		/// Constructor setting the name of the directive.
		explicit ConfInclude(std::string const &name) :
				ConfDir(name, true, 1, 1) { }
		
		/// Includes file params[0].
		virtual bool set(Conf &conf, unsigned short count,
				std::string const params[]);
};

/*************************************************************************/

/// Class implementing path global configuration directive.

/** This class implements the path global configuration directive. */

class ConfPath : public ConfDir
{
	public:
		/// Constructor setting the name of the directive.
		explicit ConfPath(std::string const &name) :
				ConfDir(name, true, 1, 1) { }
		
		/// Adds directory params[0] to the include path.
		virtual bool set(Conf &conf, unsigned short count,
				std::string const params[]);
};

/*************************************************************************/

/// Class implementing set global configuration directive.

/** This class implements the set global configuration directive. */

class ConfVarSet : public ConfDir
{
	public:
		/// Constructor setting the name of the directive.
		explicit ConfVarSet(std::string const &name) :
				ConfDir(name, true, 1, 2) { }
		
		/// Sets a configuration variable params[0] to params[1].
		virtual bool set(Conf &conf, unsigned short count,
				std::string const params[]);
};

/*************************************************************************/

/// Class implementing unset global configuration directive.

/** This class implements the unset global configuration directive. */

class ConfUnset : public ConfDir
{
	public:
		/// Constructor setting the name of the directive.
		explicit ConfUnset(std::string const &name) :
				ConfDir(name, true, 1, 1) { }
		
		/// Unsets the configuration variable params[0].
		virtual bool set(Conf &conf, unsigned short count,
				std::string const params[]);
};

/*************************************************************************/
/*************************************************************************/

template <class T>
ConfValue<T> &ConfValue<T>::operator=(ConfValue<T> const &right)
{
	if (&right != this) 
	{
		ConfDir::operator=(right);
		
		m_value = right.m_value;
		m_value_set = right.m_value_set;
		m_value_changed = right.m_value_changed;
		
		m_has_default = right.m_has_default;
		m_defvalue = right.m_defvalue;
		
		m_oldvalue = right.m_oldvalue;
		
		m_label = right.m_label;
	}
	
	return *this;
}

/*************************************************************************/

template <class T>
bool ConfValue<T>::begin(Conf &conf)
{
	if (m_ready)
		return true;
	
	/* m_value_set set to false by the constructor and end() */
	
	if (conf.reloading())
	{
		m_oldvalue = m_value;
		m_label = "";
	}
	
	return ConfDir::begin(conf);
}

/*************************************************************************/

template <class T>
bool ConfValue<T>::end(Conf &conf)
{
	if (!m_ready)
		return true;
	
	if (!m_value_set)
	{
		if (m_has_default) {
#if DEBUG
			conf.debugf(Conf::dbgProcess, 
					"Setting directive %s to its default value",
					m_name.c_str());
#endif
			m_value = m_defvalue;
		} else {
			conf.errorf("Directive %s has not been set and there is "
					"no default value", m_name.c_str());
			return false;
		}
	}
	
	if (conf.reloading() && !(m_value == m_oldvalue))
	{
		m_value_changed = true;
		if (!apply_change(conf))
			return false;
	} else {
		m_value_changed = false;
	}
	
	if (!apply(conf))
		return false;
	
	m_value_set = false;
	return ConfDir::end(conf);
}

/*************************************************************************/

template <class T>
bool ConfValue<T>::set(Conf &conf, unsigned short count,
		std::string const params[])
{
	if (!ConfDir::set(conf, count, params) || !assign(conf, count, params))
		return false;
	
	m_value_set = true;
	return true;
}

/*************************************************************************/

template <class T>
ConfEnum<T> &ConfEnum<T>::operator=(ConfEnum<T> const &right)
{
	if (&right != this)
	{
		ConfValue<T>::operator=(right);
		
		m_values = right.m_values;
	}
	
	return *this;
}

/*************************************************************************/

template <class T>
bool ConfEnum<T>::assign(Conf &conf, unsigned short count,
		std::string const params[])
{
	if (!ConfValue<T>::m_value_set && !add(conf))
		return false;
	
	typename ValueList::const_iterator pos;
	
	if ((pos = m_values.find(params[0])) == m_values.end())
	{
		conf.errorf("Invalid option name %s to %s", params[0].c_str(),
				ConfValue<T>::m_name.c_str());
		return false;
	}
	
	ConfValue<T>::m_value = pos->second;
	
#if DEBUG
	conf.debugf(Conf::dbgProcess,
			"Set %s to %s (%ld)",
			ConfValue<T>::m_name.c_str(), params[0].c_str(),
			(long) pos->second);
#endif
	return true;
}

/*************************************************************************/

template <class T>
char const* ConfEnum<T>::label(size_t index) const
{
	if (ConfValue<T>::m_label.empty())
	{
		typename ValueList::const_iterator it, end;
	
		for (it = m_values.begin(), end = m_values.end(); it != end; ++it)
		{
			if (it->second == ConfValue<T>::m_value)
			{
				ConfValue<T>::m_label = it->first;
				break;
			}
		}
	}
	
	return ConfValue<T>::m_label.c_str();
}

/*************************************************************************/

template <class T>
ConfInt<T> &ConfInt<T>::operator=(ConfInt<T> const &right)
{
	if (&right != this)
	{
		ConfValue<T>::operator=(right);
		
		m_minvalue = right.m_minvalue;
		m_maxvalue = right.m_maxvalue;
	}
	
	return *this;
}

/*************************************************************************/

template <class T>
bool ConfInt<T>::assign(Conf &conf, unsigned short count,
		std::string const params[])
{
	if (params[0].empty())
	{
		conf.errorf("Can't set directive %s: empty string is not a valid value",
				ConfValue<T>::m_name.c_str());
		return false;
	}
	
	T value;
	
	if (!convert(conf, params[0], value) || !check(conf, value))
		return false;

	ConfValue<T>::m_value = value;

#if DEBUG
	conf.debugf(Conf::dbgProcess, "Set %s to %s", ConfValue<T>::m_name.c_str(),
			params[0].c_str());
#endif
	return true;
}

/*************************************************************************/

template <class T>
bool ConfInt<T>::check(Conf &conf, T const &value) const
{	
	if ((value < m_minvalue) || (value > m_maxvalue))
	{
		conf.errorf("Can't set directive %s to %ld: value must be "
				"between %ld and %ld", ConfValue<T>::m_name.c_str(),
				(long) value, (long) m_minvalue, (long) m_maxvalue);
		return false;
	}
	
	return true;
}

/*************************************************************************/

template <class T>
bool ConfInt<T>::convert(Conf &conf, std::string const &expr, T &result) const
{
	char *end;
	T value = my_strtol(expr.c_str(), &end, 0);
	
	if (*end != 0)
	{
		conf.errorf("Can't set directive %s: value %s is not a valid number",
				ConfValue<T>::m_name.c_str(), expr.c_str());
		return false;
	}
	
	if (errno == ERANGE)
	{
		conf.errorf("Can't set directive %s: value %s is out-of-range.",
				ConfValue<T>::m_name.c_str(), expr.c_str());
		return false;
	}
				
	result = value;
	return true;
}

/*************************************************************************/

template <class T>
char const* ConfInt<T>::label(size_t index) const
{
	if (ConfValue<T>::m_label.empty())
	{
		char buf[64];
		snprintf(buf, sizeof(buf), "%ld", (long) ConfValue<T>::m_value);
		
		ConfValue<T>::m_label = buf;
	}
	
	return ConfValue<T>::m_label.c_str();
}

/*************************************************************************/

template <class T>
bool ConfIntNZ<T>::check(Conf &conf, T const &value) const
{
	if (value == 0)
	{
		conf.errorf("Can't set directive %s: value must be non-zero ",
				ConfInt<T>::m_name.c_str());
		return false;
	}
	
	return ConfInt<T>::check(conf, value);
}

/*************************************************************************/

template <class T>
ConfSet<T> &ConfSet<T>::operator=(ConfSet<T> const &right)
{
	if (&right != this)
	{
		ConfValue<T>::operator=(right);
		
		m_bits = right.m_bits;
	}
	
	return *this;
}

/*************************************************************************/

template <class T>
bool ConfSet<T>::assign(Conf &conf, unsigned short count,
		std::string const params[])
{
	if (!ConfValue<T>::m_value_set && !add(conf))
		return false;
	
	if (params[0].size() > 1)
	{
		bool add = true;
		
		switch (params[0][0])
		{
			case '-':
				add = false;
				break;
			case '+':
				add = true;
				break;
			default:
				conf.errorf("Invalid option modifier %c for directive %s",
						params[0][0], ConfValue<T>::m_name.c_str());
				return false;
		}
		
		std::string name(params[0], 1);
		
		typename BitList::const_iterator pos;
		
		if ((pos = m_bits.find(name)) == m_bits.end())
		{
			conf.errorf("Invalid option name %s to %s", name.c_str(),
					ConfValue<T>::m_name.c_str());
			return false;
		} else {
			if (add)
				ConfValue<T>::m_value |= pos->second;
			else
				ConfValue<T>::m_value &= ~pos->second;

#if DEBUG
			conf.debugf(Conf::dbgProcess,
					"%s option %s (0x%08lx) to %s",
					(add ? "Added" : "Removed"), name.c_str(),
					(long) pos->second, ConfValue<T>::m_name.c_str());
#endif
			return true;
		}
	}
	
	conf.errorf("Invalid option %s for directive %s", 
			(params[0].empty() ? "\"\"" : params[0].c_str()), ConfValue<T>::m_name.c_str());
	return false;
}

/*************************************************************************/

template <class T>
char const *ConfSet<T>::label(size_t index) const
{
	ConfValue<T>::m_label.clear();
	
	typename BitList::const_iterator it, end;
	
	for (it = m_bits.begin(), end = m_bits.end(); it != end; ++it)
	{
		if (ConfValue<T>::m_value & it->second)
		{
			ConfValue<T>::m_label += "\"+";
			ConfValue<T>::m_label += it->first;
			ConfValue<T>::m_label += "\" ";
		}
	}
	
	return ConfValue<T>::label(index);
}

/*************************************************************************/

template <class T>
bool ConfUint<T>::check(Conf &conf, T const &value) const
{
	if ((value < ConfInt<T>::m_minvalue) || (value > ConfInt<T>::m_maxvalue))
	{
		conf.errorf("Can't set directive %s to %lu: value must be "
				"between %lu and %lu", ConfInt<T>::m_name.c_str(),
				(unsigned long) value, (unsigned long) ConfInt<T>::m_minvalue,
				(unsigned long) ConfInt<T>::m_maxvalue);
		return false;
	}
	
	return true;
}

/*************************************************************************/

template <class T>
bool ConfUint<T>::convert(Conf &conf, std::string const &expr, T &result) const
{
	if (expr[0] == '-')
	{
		conf.errorf("Can't set directive %s to %s: value must be a positive "
				"number", ConfInt<T>::m_name.c_str(), expr.c_str());
		return false;
	}
	
	return ConfInt<T>::convert(conf, expr, result);
}

/*************************************************************************/

template <class T>
char const* ConfUint<T>::label(size_t index) const
{
	if (ConfValue<T>::m_label.empty())
	{
		char buf[64];
		snprintf(buf, sizeof(buf), "%lu",
				(unsigned long) ConfValue<T>::m_value);
		
		ConfValue<T>::m_label = buf;
	}
	
	return ConfValue<T>::m_label.c_str();
}

/*************************************************************************/

template <class ValueT, class ContainerT>
ConfArray<ValueT, ContainerT> &ConfArray<ValueT, ContainerT>::operator=
		(ConfArray<ValueT, ContainerT> const &right)
{
	if (&right != this) 
	{
		ConfValue<ContainerT>::operator=(right);
		
		m_max_values = right.m_max_values;
	}
	
	return *this;
}

/*************************************************************************/

template <class ValueT, class ContainerT>
bool ConfArray<ValueT, ContainerT>::assign(Conf &conf, unsigned short count,
				std::string const params[])
{
	/* Clear the array on reload, if no value has been added yet. */
	if (conf.reloading() && !ConfValue<ContainerT>::m_value_set)
		ConfValue<ContainerT>::m_value.clear();
	
	if (m_max_values == ConfValue<ContainerT>::m_value.size())
	{
		conf.errorf("Can't accept more than %ld values for directive %s",
				(long) m_max_values, ConfValue<ContainerT>::m_name.c_str());
		return false;
	}
	
	typename container_type::iterator pos = insert();
	
	if (!pos->begin(conf))
		return false;
	
	if (!pos->set(conf, 1, params))
		return false;
		
	if (!pos->end(conf))
		return false;
	
	return true;
}

/*************************************************************************/

template <class ValueT, class ContainerT>
typename ContainerT::iterator ConfArray<ValueT, ContainerT>::insert()
{
	using namespace std;

	char buf[512];
	snprintf(buf, sizeof(buf), "%s[%ld]",
			ConfValue<ContainerT>::m_name.c_str(), 
			(long) ConfValue<ContainerT>::m_value.size());
	
	return ConfValue<ContainerT>::m_value.insert(ConfValue<ContainerT>::m_value.end(), ValueT(buf, ConfValue<ContainerT>::m_reloadable));
}

/*************************************************************************/

template <class ValueT, class ContainerT>
char const *ConfArray<ValueT, ContainerT>::label(size_t index) const
{
	if (ConfValue<ContainerT>::m_value.size() <= index)
		return 0;
	
	ValueT const &dir = ConfValue<ContainerT>::m_value[index];
	
	if (dir.label_count() != 1)
		return 0;
	
	return dir.label(0);
}

/*************************************************************************/
/*************************************************************************/

inline ConfElse &ConfElse::operator=(ConfElse const &right)
{
	if (&right != this) 
	{
		ConfDir::operator=(right);
		
		m_process = right.m_process;
	}
	
	return *this;
}

/*************************************************************************/

} /* namespace conf */

/*************************************************************************/

#endif /* CONF_DIRECTIVES_H */
