/* Parser global settings.
 *
 * PegSoft configuration parser library (c) 2004 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

#include <system/headers.h>

#if HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif

#include <misc/misc.h>

#include "conf/parser.h"
#include "conf/directives.h"
#include "conf/global.h"

using namespace std;
using misc::absolute_path;

namespace conf
{

/*************************************************************************/

ConfGlobal::~ConfGlobal()
{
	delete m_else;
	delete m_ifnset;
	delete m_ifset;
	delete m_include;
	delete m_path;
	delete m_set;
	delete m_unset;
}

/*************************************************************************/

void ConfGlobal::add_default()
{
	m_else = new ConfElse("else");
	m_directives->add(*m_else);
	m_ifnset = new ConfIfnset("ifnset");
	m_directives->add(*m_ifnset);
	m_ifset = new ConfIfset("ifset");
	m_directives->add(*m_ifset);
	m_include = new ConfInclude("include");
	m_directives->add(*m_include);
	m_path = new ConfPath("path");
	m_directives->add(*m_path);
	m_set = new ConfVarSet("set");
	m_directives->add(*m_set);
	m_unset = new ConfUnset("unset");
	m_directives->add(*m_unset);
}

/*************************************************************************/

bool ConfGlobal::incl_paths_add(std::string const &path,
		PathList::const_iterator *pi)
{
	if (path.empty())
		return false;
	
	string real_path = path;
	
	/* Remove trailing / */
	while (real_path[real_path.size()-1] == '/')
		real_path.erase(real_path.size()-1);
	
	struct stat st;
	
	if (real_path[0] == '/') 
	{
		/* If path is absolute, just check for its existence. */
		if (stat(real_path.c_str(), &st) != 0)
			return false;
		if (!S_ISDIR(st.st_mode))
		{
#ifdef ENOTDIR
			errno = ENOTDIR;
#endif
			return false;
		}
	} else {
		/* If path is relative, try to find it in the current
		 * working directory. */
		char wd[PATH_MAX];
		if (!getcwd(wd, sizeof(wd)))
			return false;
		
		char buf[PATH_MAX];
		absolute_path(buf, sizeof(buf) - 1, wd, real_path.c_str());
		
		if (stat(buf, &st) == 0)
		{
			if (!S_ISDIR(st.st_mode))
			{
#ifdef ENOTDIR
				errno = ENOTDIR;
#endif
				return false;
			}
			/* We found it */
			real_path = buf;
		} else {
			/* Now check if it's relative to an already added
			 * path. */
			if (m_paths.empty())
				return false;
			
			PathList::const_iterator it, end;
			
			for (it = m_paths.begin(), end = m_paths.end(); it != end; ++it)
			{
				absolute_path(buf, sizeof(buf) - 1, it->c_str(),
						real_path.c_str());
				
				if (stat(buf, &st) == 0)
				{
					if (!S_ISDIR(st.st_mode))
					{
#ifdef ENOTDIR
						errno = ENOTDIR;
#endif
						return false;
					} else {
						break; /* We found it */
					}
				}
			}
			
			/* If it == end, then we didn't find the path at all. */
			if (it == end)
				return false;
			
			real_path = buf;
		}	
	}
	
	PathList::const_iterator pos;
	
	/* Don't add the same path twice. */
	if (!m_paths.empty())
	{
		if ((pos = find(m_paths.begin(), m_paths.end(), real_path)) !=
				 m_paths.end())
		{
			if (pi)
				*pi = pos;
			return true;
		}
	}
	
	pos = m_paths.insert(m_paths.end(), real_path);
	if (pi)
		*pi = pos;
	
	return true;
}

/*************************************************************************/

bool ConfGlobal::incl_paths_remove(std::string const &path)
{
	if (m_paths.empty())
		return false;
	
	PathList::iterator end = m_paths.end(), 
			pos = find(m_paths.begin(), end, path);
	
	if (pos == end)
		return false;
	
	m_paths.erase(pos);
	return true;
}

/*************************************************************************/

bool ConfGlobal::var_get(std::string const &name, std::string *value) const
{
	if (name.empty() || m_vars.empty())
		return false;
	
	VarList::const_iterator pos;
	
	if ((pos = m_vars.find(name)) == m_vars.end())
		return false;
	
	if (value)
		*value = pos->second;
	return true;
}

/*************************************************************************/

bool ConfGlobal::var_set(std::string const &name, std::string const &value)
{
	if (name.empty())
		return false;
	
	m_vars[name] = value;
	return true;
}

/*************************************************************************/

bool ConfGlobal::var_set(std::string const &assignment)
{
	string::const_iterator end = assignment.end();
	string::const_iterator it = find(assignment.begin(), end, '=');
	
	if (it == end)
		return var_set(assignment, "");
	
	string name(assignment.begin(), it), value(++it, end);
	return var_set(name, value);
}

/*************************************************************************/

bool ConfGlobal::var_unset(std::string const &name)
{
	if (name.empty())
		return false;
	
	if (m_vars.erase(name) == 0)
		return false;
	
	return true;
}

/*************************************************************************/

} /* namespace conf */

/*************************************************************************/
