/* Config file parser.
 *
 * PegSoft configuration parser library (c) 2004 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

#include <system/headers.h>

#include <algorithm>
#include <fstream>
#include <iostream>
#include <iterator>
#include <sstream>
#include <utility>

#include <misc/misc.h>

#include "conf/parser.h"
#include "conf/directives.h"
#include "conf/global.h"

using namespace std;
using conf::Conf;
using misc::absolute_path;

/*************************************************************************/

namespace
{

/*************************************************************************/

unsigned int const conf_maxparams = 32;

char const quote_chars[256] =
{
	  0,   1,   2,   3,   4,   5,   6,   7,
	  8,   9,  10,  11,  12,  13,  14,  15,
	 16,  17,  18,  19,  20,  21,  22,  23,
	 24,  25,  26,  27,  28,  29,  30,  31,
	 32,  33,  34,  35,  36,  37,  38,  39,
	 40,  41,  42,  43,  44,  45,  46,  47,
	 48,  49,  50,  51,  52,  53,  54,  55,
	 56,  57,  58,  59,  60,  61,  62,  63,
	 64,  65,   2,   3,  68,  69,  70,  71,
	 72,  73,  74,   3,  76,  77,  78,  79,
	 80,  81,  22,  83,  84,  31,  86,  87,
	 88,  89,  90,  91,  92,  93,  94,  95,
	 96,  97,  98,  99, 100, 101, 102, 103,
	104, 105, 106, 107, 108, 109, '\n', 111,
	112, 113, 114, 115, 116, 117, 118, 119,
	120, 121, 122, 123, 124, 125, 126, 127,
	128, 129, 130, 131, 132, 133, 134, 135,
	136, 137, 138, 139, 140, 141, 142, 143,
	144, 145, 146, 147, 148, 149, 150, 151,
	152, 153, 154, 155, 156, 157, 158, 159,
	160, 161, 162, 163, 164, 165, 166, 167,
	168, 169, 170, 171, 172, 173, 174, 175,
	176, 177, 178, 179, 180, 181, 182, 183,
	184, 185, 186, 187, 188, 189, 190, 191,
	192, 193, 194, 195, 196, 197, 198, 199,
	200, 201, 202, 203, 204, 205, 206, 207,
	208, 209, 210, 211, 212, 213, 214, 215,
	216, 217, 218, 219, 220, 221, 222, 223,
	224, 225, 226, 227, 228, 229, 230, 231,
	232, 233, 234, 235, 236, 237, 238, 239,
	240, 241, 242, 243, 244, 245, 246, 247,
	248, 249, 250, 251, 252, 253, 254, 255
};

inline bool next_param(Conf &conf, unsigned int &count, string sa[]);

/*************************************************************************/

inline bool next_param(Conf &conf, unsigned int &count, string sa[])
{
	if (count == conf_maxparams)
	{
		conf.errorf("Cannot process directive %s: no more than %u parameters "
				"allowed.", sa[0].c_str(), conf_maxparams); 
		return false;
	}
	
	++count;
	
	/* Reserve some space */
	if (sa[count - 1].capacity() < 1024)
		sa[count - 1].reserve(1024);
	/* Clear any previous data. */
	sa[count - 1].clear();

	return true;
}

/*************************************************************************/

} /* anonymous namespace */

/*************************************************************************/

namespace conf
{

/*************************************************************************/

bool Conf::add(ConfDir &dir)
{
	if (m_directives.insert(make_pair(dir.name(), &dir)).second == false)
		return false;
	return true;
}

/*************************************************************************/

bool Conf::begin(ConfGlobal &global, bool reload)
{
	/* If we're already ready, ignore the call quietly. */
	if (m_ready)
		return true;
	
	m_ready = true;
	m_reloading = reload;
	
	DirList::iterator it, end;
	
	m_global = &global;
	global.m_refcount++;
	if ((global.m_refcount == 1) && !global.directives().begin(global, reload))
		goto error;
	
	for (it = m_directives.begin(), end = m_directives.end(); it != end; ++it)
	{
		if (reload && !it->second->reloadable())
			continue;
		if (!it->second->begin(*this))
			goto error;
	}
	
	return true;

error:
	m_global = NULL;
	m_reloading = false;
	m_ready = false;
	return false;
}

/*************************************************************************/

#if DEBUG

void Conf::debug(int type, std::string const &err) const
{
	m_logger->debug(m_context, m_line, type, err);
}

#endif

/*************************************************************************/

#if DEBUG

void Conf::debugf(int type, char const *format, ...) const
{
	char buf[4096];
	
	va_list args;
	va_start(args, format);
	vsnprintf(buf, sizeof(buf), format, args);
	va_end(args);
	
	debug(type, buf);
}

#endif

/*************************************************************************/

bool Conf::end()
{
	/* Don't do anything if we're not ready, but still return true. */
	if (!m_ready)
		return true;
	
	/* Anti-infinite loop */
	m_ready = false;
	
	m_context = "";
	m_line = 0;
	
	DirList::iterator it = m_directives.begin(), end = m_directives.end();
	
	for (; it != end; ++it)
	{
		if (m_reloading && !it->second->reloadable())
			continue;
		if (!it->second->end(*this))
			goto error;
	}
	
	m_global->m_refcount--;
	if ((m_global->m_refcount == 0) && !m_global->directives().end())
		goto error;
	m_global = 0;
	
	m_reloading = false;
	
	return true;

error:
	m_ready = true;
	return false;
}

/*************************************************************************/

void Conf::error(std::string const &err) const
{
	m_logger->error(m_context, m_line, err);
}

/*************************************************************************/

void Conf::errorf(char const *format, ...) const
{
	char buf[4096];
	
	va_list args;
	va_start(args, format);
	vsnprintf(buf, sizeof(buf), format, args);
	va_end(args);
	
	error(buf);
}

/*************************************************************************/

ConfDir *Conf::find(std::string const &name)
{
	if (name.empty() || m_directives.empty())
		return NULL;
	
	DirList::const_iterator pos;
	
	if ((pos = m_directives.find(name)) == m_directives.end())
		return NULL;
	
	return pos->second;
}

/*************************************************************************/

bool Conf::parse(std::istream &is, std::string const &context)
{
	m_context = "";
	m_line = 0;
	
	/* Check if we're ready first. */
	if (!m_ready)
	{
		error("Parser not ready");
		return false;
	}
	
	return parse_real(is, context);
}

/*************************************************************************/

bool Conf::parse_file(std::string const &filename)
{
	m_context = "";
	m_line = 0;
	
	/* Check if we're ready first. */
	if (!m_ready)
	{
		error("Parser not ready");
		return false;
	}
	
	/* We can't open NULL files ;) */
	if (filename.empty()) 
	{
		error("Can't parse a file with no name");
		return false;
	}
	
	ifstream in(filename.c_str(), ios::in);
	
	if (!in && filename[0] != '/')
	{
		char buf[PATH_MAX];
		
		ConfGlobal::PathList::const_iterator it, end;
		
		for (it = m_global->incl_paths().begin(), 
				end = m_global->incl_paths().end(); it != end; ++it)
		{
			absolute_path(buf, sizeof(buf) - 1, it->c_str(), filename.c_str());
			
			in.clear();
			in.open(buf, ios::in);
			
			if (in)
				break;
		}
		
		if (!in)
		{
			if (errno)
				errorf("Can't open file %s: %s", filename.c_str(), strerror(errno));
			else
				errorf("Can't open file %s", filename.c_str());
			return false;
		}
	}
		
	return parse_real(in, filename.c_str());
}

/*************************************************************************/

bool Conf::parse_real(std::istream &is, std::string const &context)
{	
	istreambuf_iterator<char> it(is), end;
	
	/* Empty stream! Don't care about it. */
	if (it == end)
		return true;

	unsigned int count = 0;
	string params[conf_maxparams];
	
	m_context = context;
	m_line = 1;

	unsigned int braces = 0, save_ln = 0;

	while (it != end)
	{
		switch (*it)
		{
			case ' ':
			case '\f':
			case '\n':
			case '\r':
			case '\t':
			case '\v':
				do {
					if (*it == '\n')
						m_line++;
					++it;
				} while (it != end && isspace(*it));
				break;
			
			case '#':
				/* Ignore all characters until next newline. */
				do {
					++it;
				} while (it != end && *it != '\n' && *it != '\r');
				break;
			
			case '/':
				if (++it != end) 
				{
					switch (*it) 
					{
						case '/':
							do {
								++it;
							} while (it != end && *it != '\n' && *it != '\r');
							break;
						case '*':
							++it;
							
							/* Save current line number to report in case
							 * a syntax error occurs. */
							save_ln = m_line;
							
							while (it != end) 
							{
								if (*it == '\n') {
									m_line++;
								} else if (*it == '*') {
									if ((++it != end) && *it == '/')
										break;
									else
										continue;
								}
								
								++it;
							}
							
							/* If we're at end of stream, then the comment was
							 * not finished. Since it may be a user mistake we
							 * treat it as an error. */
							if (it == end)
							{
								errorf("Unfinished comment (possible start at "
										"line %u)", save_ln);
								return false;
							}
							
							/* *it == '/' so increment it. */
							++it;
							break;
						default:
							error("Invalid token /");
							return false;
					}
				} else {
					error("Invalid token /");
					return false;
				}
				break;
			
			case '"':
				++it;
				
				if (!next_param(*this, count, params))
					return false;

				/* Save current line number to report in case
				 * a syntax error occurs. */
				save_ln = m_line;
				
				while (it != end && *it != '"')
				{
					switch (*it)
					{
						case '\\':
							++it;
							
							if (it == end)
							{
								error("Unfinished character quote");
								return false;
							}
							
							params[count - 1] += quote_chars[static_cast
									<unsigned char>(*it)];
							++it;
							continue;
						case '$': {
							unsigned int subst_ln = m_line;
							string subst_name;
							
							++it;
							
							if (it == end)
							{
								error("Unfinished substitution");
								return false;
							}
							
							if (*it != '(')
							{
								error("$ not followed by a directive name "
								      "between parenthesis");
								return false;
							}
							
							++it;
							
							while (it != end && *it != ')') 
							{
								if (*it == '\n')
									++m_line;
								subst_name += *it;
								++it;
							}
							
							/* We should never be at end of stream, unless of course
				 			 * the substitution is unfinished. So we yell to user
							 * in that case. */
							if (it == end)
							{
								errorf("Unfinished substitution (possible "
									   "start at line %u)", subst_ln);
								return false;
							}
							
							string subst_value;
							if (!m_global->var_get(subst_name, &subst_value))
							{
								errorf("Unknown variable name %s in "
										"substitution", subst_name.c_str());
								return false;
							}
#if DEBUG
							debugf(dbgSubstitution,
									"Substitution of $(%s) to %s",
									subst_name.c_str(), subst_value.c_str());
#endif
							params[count - 1] += subst_value;
							
							/* *it == ')' so we increment it. */
							++it;
							continue;	
						}
						case '\n':
						case '\r':
							/* Ignore newlines and all subsequent white spaces */
							do {
								if (*it == '\n')
									++m_line;
								++it;
							} while (it != end && isspace(*it));
							
							continue;
						default:
							 break;
					}
					
					params[count - 1] += *it;
					++it;
				}
				
				/* We should never be at end of stream, unless of course
				 * the quoted string is unfinished. So we yell to user
				 * in that case. */
				if (it == end)
				{
					errorf("Unfinished quoted string (possible start at line "
							"%u)", save_ln);
					return false;
				}
				
				/* *it == '"' so we increment it. */
				++it;
				break;
			
			case '{':
				++it;
				
				if (!next_param(*this, count, params))
					return false;
				
				/* We track the number of open braces to know
				 * where this block stops. */
				braces = 1;
				
				/* Save current line number to report in case
				 * a syntax error occurs. */
				save_ln = m_line;
				
				while (it != end && braces)
				{
					switch (*it)
					{
						case '\n':
							++m_line;
							break;
						case '#':
							/* Ignore braces in comments */
							do {
								params[count - 1] += *it;
								++it;
							} while (it != end && *it != '\n' && *it != '\r');
							continue;
						case '/':
							params[count - 1] += *it;
							if (++it == end)
								continue;
								
							/* Ignore braces in comments */
							switch (*it) 
							{
								case '/':
									do {
										params[count - 1] += *it;
										++it;
									} while (it != end && *it != '\n' && *it != '\r');
									continue;
								case '*':
									params[count - 1] += *it;
									++it;
							
									while (it != end) 
									{
										if (*it == '\n') {
											m_line++;
										} else if (*it == '*') {
											params[count - 1] += *it;
											if ((++it != end) && *it == '/')
												break;
											else
												continue;
										}
								
										params[count - 1] += *it;
										++it;
									}
							
									if (it == end)
										continue;
							
									/* *it == '"' and will be stored and
							         * incremented at end of the while block. */
									break;
								default:
									continue;
							}
							break;
						case '"':
							params[count - 1] += *it;
							++it;
							
							/* Ignore braces in quoted strings */
							while (it != end && *it != '"')
							{
								switch (*it)
								{
									case '\n':
										++m_line;
										break;
									case '\\':
										params[count - 1] += *it;
										++it;
										if (it != end && *it == '"') {
											params[count - 1] += *it;
											++it;
										}
										continue;
									default:
										break;
								}
								
								params[count - 1] += *it;
								++it;
							}
							
							/* *it == '"' and will be stored and
							 * incremented at end of the while block. */
							break;
						case '{':
							++braces;
							break;
						case '}':
							--braces;
							if (!braces)
								continue;
							break;
						default:
							break;
					}
					
					params[count - 1] += *it;
					++it;
				}
				
				/* We should never be at end of stream, unless of course
				 * the block is unfinished. So we yell to user in that case. */
				if (it == end)
				{
					errorf("Unfinished block (possible start at line %u)",
							save_ln);
					return false;
				}
				
				/* *it == '}' so we increment it. */
				++it;
				
				/* End of directive ! */
				if (!process_directive(count, params)) 
					return false;
				
				break;
				
			case ';':
				++it;

				if (count == 0)
					break;

				if (!process_directive(count, params)) 
					return false;
				
				break;
			default:
				if (!next_param(*this, count, params))
					return false;
				
				do {
					params[count - 1] += *it;
					++it;
				} while (it != end && *it != ';' && *it != '{' && *it != '#' &&
						*it != '/' && *it != '"' && !isspace(*it));
				
				/* If this is true, then we're at end of stream but the
				 * last directive has not been completed by a semi-colon. */
				if (it == end) {
					errorf("Unfinished directive %s", params[0].c_str());
					return false;
				}
				
				break;
		}
	}
	
	if (count)
	{
		errorf("Unfinished directive %s", params[0].c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool Conf::parse_string(std::string const &str, std::string const &context)
{
	m_context = "";
	m_line = 0;
	
	/* Check if we're ready first. */
	if (!m_ready)
	{
		error("Parser not ready");
		return false;
	}
	
	/* Don't bother with empty strings. */
	if (str.empty())
		return true;
	
	stringstream in(str);
	return parse_real(in, context);
}

/*************************************************************************/

#if DEBUG

void Conf::print_directive(unsigned int count, string const params[]) const
{
	for (unsigned int i = 0; i < count; ++i)
		debugf(dbgParse, "params[%d]=%s", i, params[i].c_str());
}

#endif

/*************************************************************************/

bool Conf::process_directive(unsigned int &count, string params[])
{	
#if DEBUG
	print_directive(count, params);
#endif

	ConfDir *dir;
	
	if (!(dir = find(params[0])) &&
			!(dir = m_global->directives().find(params[0])))
	{
		errorf("Unknown directive %s", params[0].c_str());
		return false;
	}
	
	/* Ensure the directive can be set. */
	if (!m_reloading || dir->reloadable())
	{
		if (!dir->set(*this, count - 1 , (count - 1 ? params + 1 : NULL)))
			return false; /* It's up to the set() function to send an
			               * appropriate error message. */
	}
	
	count = 0;
	return true;
}

/*************************************************************************/

bool Conf::remove(ConfDir &dir)
{
	if (m_directives.erase(dir.name()) < 1)
		return false;
	return true;
}

/*************************************************************************/

void Conf::warning(std::string const &msg) const
{
	m_logger->warning(m_context, m_line, msg);
}

/*************************************************************************/

void Conf::warningf(char const *format, ...) const
{
	char buf[4096];
	
	va_list args;
	va_start(args, format);
	vsnprintf(buf, sizeof(buf), format, args);
	va_end(args);
	
	warning(buf);
}

/*************************************************************************/

#if DEBUG

void ConfStdLogger::debug(std::string const &context, unsigned int line,
		int type, std::string const &msg)
{
	if (context.empty()) {
		cout << "Debug: " << msg << endl;
	} else {
		cout << "Debug: " << context << ":" << line <<
				": " << msg << endl;
	}
}

#endif

/*************************************************************************/

void ConfStdLogger::error(std::string const &context, unsigned int line,
		std::string const &err)
{
	if (context.empty()) {
		cerr << "Configuration error: " << err << endl;
	} else {
		cerr << "Configuration error in " << context << ":" << line <<
				": " << err << endl;
	}
}

/*************************************************************************/

void ConfStdLogger::warning(std::string const &context, unsigned int line,
		std::string const &msg)
{
	if (context.empty()) {
		cerr << "Configuration warning: " << msg << endl;
	} else {
		cerr << "Configuration warning in " << context << ":" << line <<
				": " << msg << endl;
	}
}

/*************************************************************************/

} /* namespace conf */

/*************************************************************************/
