/* Data fields.
 *
 * PegSoft databases library (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file databases/field.h
  * \brief Database field class header file. */

#ifndef DATABASES_FIELD_H
#define DATABASES_FIELD_H

/*************************************************************************/

#include <databases/record.h>
#include <misc/string.h>

/*************************************************************************/
	
namespace databases
{

/*************************************************************************/

/// Database field.

/** This is a base class for fields used by databases to have some basic
  * information about what they store as well as to let them set values
  * in records in a generic way.
  *
  * For each field you create, you need to set its name, version information,
  * type, size, attributes and default value.
  *
  * Note that you can't change the type, size, attribute or default value
  * of a field once it has been created in a database. Therefore, changing
  * the type of a field and recompiling won't work because the database
  * driver is NOT required to change existing fields in the database.
  *
  * The setup() function may be overriden in derived classes. */

class Field
{
	public:
		/// Constructor.
		Field() : m_name(0), m_minver(0), m_maxver(0), m_type(0), m_size(0),
				m_attrs(0), m_default(0), m_utdefault(0), m_tag(0),
				m_setup(false) { }
		/// Destructor.
		virtual ~Field() { }
	
		/// Sets the field up.
		/** Sets the field up. This is called by the Database::add_field()
		  * function before the field is added to the field list of a
		  * database.
		  *
		  * The default implementation computes the integer default
		  * value if needed, and ensures all required information about
		  * the field has been set.
		  * 
		  * \return true if successful, false otherwise. */
		virtual bool setup();
	
		/// Returns the field name.
		char const *name() const { return m_name; }
		/// Sets the field name.
		/** Sets the field name.
		  *
		  * \param[in] name The name, which must only contain characters
		  * 0-9A-Za-z_ (no checks are made so you must make sure it
		  * is correct).
		  * \return The field name. */
		char const *name(char const *name);
		
		/// Returns version numbers.
		/** Returns version numbers that have been set for the field.
		  * A field's value is imported when the database version is
		  * between the minimum and maximum import version (but only if
		  * the value exists in the database), and exported when the
		  * database version is between the minimum and maximum export
		  * version.
		  *
		  * \param[in] type The type of version to return, which
		  * can be one of the following values:
		  *   - Database::vtMinimum: The minimum version number required.
		  *     If the database has a lower import or export version, then
		  *     the value of this field will respectively not be imported when
		  *     importing record data or not be exported to the database.
		  *     The minimum version can't be zero.
		  *   - Database::vtMaximum: The maximum version number required.
		  *     If the database has a higher import or export version, then
		  *     the value of this field will respectively not be imported when
		  *     importing record data or not be exported to the database.
		  *     If the maximum version is zero, then there won't be
		  *     any check for the maximum version number.
		  * \return The requested version number. */
		uint16_t version(int type) const;
		
		/// Sets version numbers.
		/** Sets version numbers for the field.
		  *
		  * Setting version numbers after the field has been added to
		  * a database has no effect.
		  *
		  * \param[in] type The type of version to set, which
		  * can be either Database::vtMinimum or Database::vtMaximum.
		  * \param[in] ver The version number.
		  * \return The new version number for the given type.
		  * \sa version(int) */
		uint16_t version(int type, uint16_t ver);
		
		/// String type
		static uint16_t const ftString = 1;
		/// Integer type
		static uint16_t const ftInteger = 2;
		/// Unsigned integer type
		static uint16_t const ftUnsigned = 3;
		/// Unix timestamp type
		static uint16_t const ftUnixTime = 4;
		/// Record identifier type
		static uint16_t const ftIdentifier = 5;	
		/// The number of types
		static uint16_t const ftCount = 5;
		
		/// Returns the type of the field.
		uint16_t type() const { return m_type; }
		/// Sets the type of the field.
		/** Sets the type of the field.
		  *
		  * \param[in] type The type of the field, which can be one
		  * of the following:
		  *   - ftString: The string type. Beware that database drivers
		  *     consider that a NULL string and an empty string ("") are
		  *     the same thing, and they automatically convert all empty
		  *     strings to NULL strings on return.
		  *   - ftInteger: The signed integer type, which may hold up to
		  *     32 bits.
		  *   - ftUnsigned: The unsigned integer type, which may hold up
		  *     to 32 bits.
		  *   - ftUnixTime: The Unix timestamp type (basically, time_t
		  *     in C/C++).
		  *   - ftIdentifier: The record identifier type (i.e.
		  *     Record::identifier_type), used for fields that need
		  *     to reference to identifiers.
		  * \return The type of the field. */
		uint16_t type(uint16_t type);
		
		/// Returns the size of the field.
		size_t size() const { return m_size; }
		/// Sets the size of the field.
		/** Sets the size of the field.
		  *
		  * \param[in] size The size of the field. If the type of the
		  * field is ftString, the size must be set to the maximum length
		  * allowed for the string, and if of the ftInteger or ftUnsigned
		  * types, it must be set to 1, 2 or 4 depending on the intended
		  * size for the integer (respectively 8-bit, 16-bit or 32-bit).
		  * For all other field types, the size is ignored.
		  * \return The size of the field. */
		size_t size(size_t size);
		
		/// Field attributes
		enum
		{
			/// Field value is case-sensitive.
			faCase = 0x0001,
			/// Field may be passed to Database::delete_owned().
			faOwner = 0x0002,
			/// Field may be passed to Database::delete_owned(), unique owner.
			faOwnerUnique = 0x0004
		};
		
		/// Returns the attributes of the field.
		/** Returns the attributes of the field.
		  *
		  * \return The attributes of the field, which consists of an
		  * or'ed set of the following:
		  *   - faCase: if the field type if ftString, the string should
		  *     be compared case-sensitively.
		  *   - faOwner: if the field type is ftIdentifier, the field
		  *     may be used with the Database::delete_owned() function.
		  *   - faOwnerUnique: if the field type is ftIdentifier, the
		  *     field may be used with the Database::delete_owned function.
		  *     Each owner ID is only set for one record in the database. */
		uint16_t attrs() const { return m_attrs; }
		/// Adds an attribute to the field.
		uint16_t attrs_add(uint16_t attrs);
		/// Removes an attribute from the field.
		uint16_t attrs_remove(uint16_t attrs);
		
		/// Returns the default value, as a string.
		char const *default_value() const { return m_default; }
		/// Sets the default value of the field.
		/** Sets the default value of the field. The default value are
		  * used by database drivers to assign the initial values for 
		  * newly-created fields in the existing records. This should
		  * prevent a lot of extra work from database drivers users
		  * because otherwise they would always have to set the intended
		  * initial value for new fields themselves.
		  *
		  * \param[in] defval The default value, which may be NULL if
		  * the field type is ftString.
		  * \return The default value of the field. */
		char const *default_value(char const *defval); 
		
		/// Returns the default value, as an integer.
		/** Returns the default value, as an integer. This function should be
		  * called only if the field is of the ftInteger type. Else, its
		  * result is unspecified.
		  *
		  * \return The default value of the field, as an integer. */
		int32_t idefault_value() const { return m_idefault; }
		
		/// Returns the default value, as an unsigned integer.
		/** Returns the default value, as an unsigned integer. This function
		  * should be called only if the field is of the ftUnsigned type. Else,
		  * its result is unspecified.
		  *
		  * \return The default value of the field, as an unsigned integer. */
		uint32_t uidefault_value() const { return m_uidefault; }
		
		/// Returns the default value, as a Unix timestamp.
		/** Returns the default value, as a Unix timestamp. This function
		  * should be called only if the field is of the ftUnixTime type. Else,
		  * its result is unspecified.
		  *
		  * \return The default value of the field, as a Unix timestamp. */
		time_t utdefault_value() const { return m_utdefault; }
		
		/// Returns the default value, as a record identifier.
		/** Returns the default value, as a record identifier. This function
		  * should be called only if the field is of the ftIdentifier type. Else,
		  * its result is unspecified.
		  *
		  * \return The default value of the field, as a record identifier. */
		Record::identifier_type ridefault_value() const { return m_ridefault; }
		
		/// Returns the tag of the field.
		uint16_t tag() const { return m_tag; }
		/// Sets the tag of the field.
		/** Sets the tag of the field, which may be used by other modules to
		  * quickly identify a field in a database by looking up a
		  * numeric value rather than comparing a string.
		  *
		  * The number assigned should be unique at all times in a database.
		  *
		  * \param[in] tag The tag number (should be different than zero).
		  * \return The tag of the field. */
		uint16_t tag(uint16_t tag);
	protected:
		/// The name of the field.
		misc::cstring m_name;
		
		/// Minimum version required.
		uint16_t m_minver;
		/// Maximum version allowed.
		uint16_t m_maxver;
		
		/// Type of the field.
		uint16_t m_type;
		/// Size of the field.
		size_t m_size;
		/// Attributes of the field
		uint16_t m_attrs;
		
		/// Default value of the field (as a string).
		misc::cstring m_default;
		
		/// The default values as other types.
		union
		{
			/// Default value of the field (as an integer -- if of ftInteger type.)
			int32_t m_idefault;
			/// Default value of the field (as an unsigned integer -- if of ftUnsigned type.)
			uint32_t m_uidefault;
			/// Default value of the field (as a unix timestamp -- if of FtUnixTime type.)
			time_t m_utdefault;
			/// Default value of the field (as a record identifier -- if of ftIdentifier type.)
			Record::identifier_type m_ridefault;
		};
		
		/// Field tag.
		uint16_t m_tag;
		
		/// Has setup() been called?
		bool m_setup;
	private:
		Field(Field const &);
		Field &operator=(Field const &);
};

/*************************************************************************/

} /* namespace databases */

/*************************************************************************/

#endif /* DATABASES_FIELD_H */
