/* Log levels.
 *
 * PegSoft log subsystem library (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

/** 
 * \file log/levels.h
 * \brief Log levels header file.
 */

#ifndef LOG_LEVELS_H
#define LOG_LEVELS_H

/*************************************************************************/

#include <string>

/*************************************************************************/

namespace xlog
{

/*************************************************************************/

/// Log level base class.

/** This class is the ancestor to all log level classes.
  *
  * The operator==() method must be overriden in derived classes.
  * The header() method may be overriden too. */

class LogLevel
{
	public:
		/// Destructor.
		virtual ~LogLevel() { }

		/// Checks for log level match.
		/** Checks whether a given name matches with the name of this log
		  * level or any of the class ancestor names.
		  *
		  * Derived classes must call the inherited method (the LogLevel
		  * implementation matches the name "*").
		  *
		  * \param[in] name The name to check.
		  * \return true if the name matches, false otherwise. */
		virtual bool operator==(std::string const &name) const
				{ return (!strcasecmp(name.c_str(), "All")); };
		
		/// Returns the log level header.
		/** Returns the log level header, that identifies the log level
		  * in a human-friendly way and may or may not be logged as
		  * a header to log messages by log modules.
		  *
		  * \return The log level header (must be non-NULL). */
		virtual char const *header() const { return ""; }
};

/*************************************************************************/

/// Declares a new log level.

/** Declares a new log level.
  *
  * \param classname The name of the new class.
  * \param ancestor The ancestor of the new class.
  * \param llname The name of the log level (as a char *). */

#define LOGLEVEL(classname, ancestor, llname) \
	class classname : public ancestor \
	{ \
		public: \
			virtual bool operator==(std::string const &name) const \
			{ \
				if (!strcasecmp(name.c_str(), llname)) \
					return true; \
				return ancestor::operator==(name); \
			} \
	}

/*************************************************************************/

/// Declares a new log level with a header.

/** Declares a new log level with a header.
  *
  * \param classname The name of the new class.
  * \param ancestor The ancestor of the new class.
  * \param llname The name of the log level (as a char *).
  * \param head The header of the log level (as a char *). */

#define LOGLEVELH(classname, ancestor, llname, head) \
	class classname : public ancestor \
	{ \
		public: \
			virtual bool operator==(std::string const &name) const \
			{ \
				if (!strcasecmp(name.c_str(), (llname))) \
					return true; \
				return ancestor::operator==(name); \
			} \
			\
			virtual char const *header() const { return (head); } \
	}

/*************************************************************************/

#if DEBUG
LOGLEVELH(LogDebug, LogLevel, "Debug", "debug: ");
#endif
LOGLEVEL(LogInfo, LogLevel, "Info");
LOGLEVEL(LogNotice, LogLevel, "Notice");
LOGLEVELH(LogWarning, LogLevel, "Warning", "warning: ");
LOGLEVELH(LogError, LogLevel, "Error", "Error: ");
LOGLEVELH(LogCritical, LogLevel, "Critical", "Critical error: ");
LOGLEVELH(LogFatal, LogLevel, "Fatal", "FATAL: ");

/*************************************************************************/

} /* namespace xlog */

/*************************************************************************/

#endif /* LOG_LEVELS_H */
