/* Logging everything.
 *
 * PegSoft log subsystem library (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

/** 
 * \file log/logger.h
 * \brief Logger header file.
 */

#ifndef LOG_LOGGER_H
#define LOG_LOGGER_H

/*************************************************************************/

#include <list>

/*************************************************************************/

namespace modules { class Core; }

/*************************************************************************/

namespace xlog
{

/*************************************************************************/

class Log;
class LogLevel;
class LogRules;

/*************************************************************************/

/// The logger class.

/** A logger is used to log stuff to standard output (when the
  * program runs in foreground) and forward messages to log interfaces so
  * they can log whatever they want the way they want. */

class Logger
{
	public:
		/// The container type.
		typedef std::list<Log *> container_type;
		
		/// Maximum length for log messages.
		static int const msg_length = 2048;
		
		/// Constructor.
		/** Constructor.
		  *
		  * \param[in] core The core module (must remain valid until
		  * the logger is destructed).
		  * \param[in] rules The LogRules directive that determines
		  * what gets logged to the standard output. */
		explicit Logger(modules::Core &core, LogRules const &rules) :
				m_core(&core), m_interfaces(), m_rules(&rules) { }
		/// Copy constructor.
		Logger (Logger const &l) : m_core(l.m_core),
				m_interfaces(l.m_interfaces), m_rules(l.m_rules) { }
		/// Assignment operator.
		Logger &operator=(Logger const &right);
		
		/// Logs a message.
		/** Logs a message. The message will be forwarded to all log
		  * interfaces and, if the program doesn't run in background,
		  * printed to either stdout or stderr.
		  *
		  * \param[in] level The log level of the message.
		  * \param[in] format A format string for the message, similar
		  * to the one used by the printf() family of functions.
		  * \param[in] ... Parameters for the format string. */
		void log(LogLevel const &level, char const *format, ...)
				FORMAT(printf, 3, 4);
		
		/// Registers a log interface.
		/** Registers a log interface. The interface must remain
		  * valid until it is registered (it is NOT copied).
		  *
		  * Don't forget to unregister the interface before it is 
		  * destructed, or there'll be troubles.
		  *
		  * \param[in] interface The log interface to register.
		  * \sa unregister_interface() */
		void register_interface(Log &interface);
		
		/// Unregisters a log interface.
		/** Unregisters a previously registered log interface.
		  *
		  * \param[in] interface The log interface to unregister.
		  * \sa register_interface() */
		void unregister_interface(Log &interface);
		
		/// Same as log(), but using variable argument lists.
		void vlog(LogLevel const &level, char const *format,
				va_list args);
	private:
		modules::Core *m_core;
		container_type m_interfaces;
		LogRules const *m_rules;
};

/*************************************************************************/

} /* namespace xlog */

/*************************************************************************/

#endif /* LOG_LOGGER_H */
