/* Logging everything.
 *
 * PegSoft log subsystem library (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

#include <system/headers.h>

#include <iostream>

#include <modules/core.h>

#include "log/conf.h"
#include "log/levels.h"
#include "log/log.h"
#include "log/logger.h"

using namespace std;
using namespace modules;

/*************************************************************************/

namespace xlog
{

/*************************************************************************/

Logger &Logger::operator=(Logger const &right)
{
	if (&right != this)
	{
		m_core = right.m_core;
		m_interfaces = right.m_interfaces;
		m_rules = right.m_rules;
	}
	
	return *this;
}

/*************************************************************************/

void Logger::log(LogLevel const &level, char const *format, ...)
{
	va_list args;
	
	va_start(args, format);
	vlog(level, format, args);
	va_end(args);
}

/*************************************************************************/

void Logger::register_interface(Log &interface)
{
	m_interfaces.push_back(&interface);
}

/*************************************************************************/

void Logger::unregister_interface(Log &interface)
{
	m_interfaces.remove(&interface);
}

/*************************************************************************/

void Logger::vlog(LogLevel const &level, char const *format,
				va_list args)
{
	/* Check if we need to log something at all */
	if (m_interfaces.empty() && (m_core->flags() & Core::pfBackground))
		return;
	
	/* Prepare the message */
	char msg[msg_length];
	vsnprintf(msg, sizeof(msg), format, args);
	
	/* Send it to log modules */
	if (!m_interfaces.empty())
	{
		container_type::iterator it, end;
		
		for (it = m_interfaces.begin(), end = m_interfaces.end();
				it != end; ++it)
			(*it)->log(level, msg);
	}
	
	/* Log message to stdout/stderr if not running in background. */
	if (!(m_core->flags() & (Core::pfBackground | Core::pfVeryQuiet)))
	{
		if ((m_core->flags() & Core::pfQuiet) && !(level == "Error") &&
				!(level == "Critical") && !(level == "Fatal"))
			return;
		
		/* Check for log rules, but only if the program is already
		 * configured (else it has no sense). */
		if (m_core->flags() & Core::pfConfigured)
		{
			LogRules::Rule const *rule = m_rules->match(level);
			if (!rule || (rule->type() == LogRules::lrReject))
				return;
		}
		
		char timebuf[128];
		
		time_t t = time(0);
		struct tm tm = *localtime(&t);
		
#if DEBUG && HAVE_GETTIMEOFDAY
		if (m_core->debuglevel())
		{
			strftime(timebuf, sizeof(timebuf), "[%H:%M:%S", &tm);
			
			struct timeval tv;
			gettimeofday(&tv, 0);
			
			char *end = timebuf + strlen(timebuf);
			snprintf(end, (sizeof(timebuf) - (end - timebuf)), ".%06ld] ", 
					(long) tv.tv_usec);
		} else {
#endif
			strftime(timebuf, sizeof(timebuf), "[%H:%M] ", &tm);
#if DEBUG && HAVE_GETTIMEOFDAY
		}
#endif
		
		if (level == "Error" || level == "Critical" || level == "Fatal")
			cerr << timebuf << level.header() << msg << endl;
		else
			cout << timebuf << level.header() << msg << endl;
	}	
}

/*************************************************************************/

} /* namespace xlog */

/*************************************************************************/
