/* Mail agent.
 *
 * PegSoft mail library (c) 2006 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file mail/mail.h
  * \brief Mail subsystem header file. */

#ifndef MAIL_MAIL_H
#define MAIL_MAIL_H

/*************************************************************************/

#include <string>
#include <vector>

#include <misc/ratio.h>

/*************************************************************************/

/// Namespace of the mail classes.

namespace mail 
{

/*************************************************************************/

class Transport;

/*************************************************************************/

/// E-mail representation
/** This class represents an e-mail that needs to be sent. */

class Mail
{
	public:
		/* Typedefs */
		typedef std::vector<std::string> text_type;
		
		/* Constants */
		
		/// Send flags
		enum
		{
			/// The target is authenticated
			mfAuthenticated = 0x00000001,
			/// No quota restriction
			mfNoQuota = 0x00000002
		};
		
		/// Error constants
		enum
		{
			/// No error
			meSuccess = 0x00000000,
			/// The e-mail is incomplete
			meIncomplete = 0x00000001,
			/// Quota exceeded
			meQuota = 0x00000002,
			/// Delivery error
			meDelivery = 0x0000004,
			/// Mail subsystem not configured
			meConfig = 0x00000008,
			/// Is a temporary error.
			meTemporary = 0x80000000
		};
		
		/* Functions */
		
		/// Determines whether the mail subsystem has been configured.
		static bool configured();
		
		/// Checks for address validity.
		/** Checks whether an e-mail address is syntaxically valid. The
		  * check does not entirely follow RFC 822; we're allowing only
		  * a subset of what is (theorically) allowed. However, this
		  * should be enough to cover 99,99% of the e-mail addresses
		  * commonly in use today while ensuring people don't type in
		  * subtly invalid addresses that somehow would still get
		  * accepted.
		  *
		  * \param[in] address The address that must be checked.
		  * \param[in] noplus Set this to true if you want to reject
		  * e-mail addresses containing a '+' character in the user part.
		  * \return true if the address is syntaxically valid, false  */
		static bool valid_address(char const *address, bool noplus);
		
		/// Constructor.
		Mail() : m_from_name(), m_from_address(), m_subject(), m_sender(),
				m_to_name(), m_to_address(), m_origin(), m_text(),
				m_data() { }
		/// Copy constructor.
		Mail(Mail const &m) : m_from_name(m.m_from_name),
				m_from_address(m.m_from_address), m_subject(m.m_subject),
				m_sender(m.m_sender), m_to_name(m.m_to_name),
				m_to_address(m.m_to_address), m_origin(m.m_origin),
				m_text(m.m_text), m_data(m.m_data) { }
		/// Assignment operator.
		Mail &operator=(Mail const &right);
		
		/// Sends the mail.
		/** Sends the mail. Please note that the e-mail is not
		  * guaranteed to have been effectively sent when the function
		  * returns; mail transport modules are not required to process
		  * the mails synchronously. However, the mail object may
		  * be deleted once this call has been done, as it is copied as
		  * necessary if needed.
		  *
		  * The only required field that needs to have been filled before
		  * calling this function is to_address(). All other fields are
		  * either optional or they have default values.
		  *
		  * \param[in] flags The flags qualifying this sending, which can
		  * be one or more of the following:
		  *   - mfAuthenticated: The recipient is an authenticated.
		  *   e-mail address, so apply limits accordingly.
		  * \param[out] err If non-NULL, this will be set to a set of
		  * the error constants if the function returns false. It is
		  * set to meSuccess if there are no errors.
		  * \param[out] errdesc If non-NULL, is set to a description
		  * of the error that just happened.
		  * \return true if the e-mail has been approved, false
		  * otherwise. */
		bool send(uint32_t flags, uint32_t *err = 0, std::string *errdesc = 0);
		
		/// Returns the name part of the From: field.
		std::string const &from_name() const { return m_from_name; }
		/// Sets the name part of the From: field.
		std::string const &from_name(std::string const &from_name)
				{ return (m_from_name = from_name); }
		
		/// Returns the address part of the From: field.
		std::string const &from_address() const { return m_from_address; }
		/// Sets the address part of the From: field.
		std::string const &from_address(std::string const &from_address)
				{ return (m_from_address = from_address); }
		
		/// Returns the Subject: field.
		std::string const &subject() const { return m_subject; }
		/// Sets the Subject: field.
		std::string const &subject(std::string const &subject)
				{ return (m_subject = subject); }
		
		/// Returns the Sender: field.
		std::string const &sender() const { return m_sender; }
		/// Sets the Sender: field.
		std::string const &sender(std::string const &sender)
				{ return (m_sender = sender); }
		
		/// Returns the name part of the To: field.
		std::string const &to_name() const { return m_to_name; }
		/// Sets the name part of the To: field.
		std::string const &to_name(std::string const &to_name)
				{ return (m_to_name = to_name); }
		
		/// Returns the address part of the To: field.
		std::string const &to_address() const { return m_to_address; }
		/// Sets the address part of the To: field.
		std::string const &to_address(std::string const &to_address)
				{ return (m_to_address = to_address); }
		
		/// Returns the X-Origin: field.
		std::string const &origin() const { return m_origin; }
		/// Sets the X-Origin: field.
		/** Sets the X-Origin field. This field is also used by the
		  * send() function to determine whether the sender limit
		  * has been exceeded.
		  *
		  * \param[in] origin The origin of the e-mail.
		  * \return The new origin of the e-mail. */
		std::string const &origin(std::string const &origin)
				{ return (m_origin = origin); }
		
		/// Returns the vector containing the current text lines.
		text_type const &text() const { return m_text; }
		/// Adds a line to the e-mail text.
		/** Adds a line to the e-mail text. A newline is automatically
		  * added at the end of the text if it is missing. Also, if there
		  * are newlines in the middle of the text, they will be 
		  * treated accordingly.
		  *
		  * \param[in] line The new line of text to add.
		  * \return The vector containing the current text lines. */
		text_type const &text_add(std::string const &line);
		/// printf-like version of text_add().
		text_type const &text_addf(char const *format, ...) FORMAT(printf, 2, 3);
		
		/// Returns the mail data.
		/** Returns the mail data. The data is generated in the send()
		  * function and consists of the mail headers and text formatted
		  * in a suitable way.
		  *
		  * \return The mail data. */
		std::string const &data() const { return m_data; }
	private:
		std::string m_from_name;
		std::string m_from_address;
		std::string m_subject;
		std::string m_sender;
		std::string m_to_name;
		std::string m_to_address;
		std::string m_origin;
		text_type m_text;
		std::string m_data;
		
		void generate_data();
		
		void set_error(uint32_t *err, std::string *errdesc, uint32_t myerr,
				std::string const &mydesc) const;
};

/*************************************************************************/

/// Configures options of the mail library (for use by the core).
extern void conf_options(std::string const &sname, std::string const &saddress,
		misc::Ratio const &lauth, misc::Ratio const &lunauth,
		misc::Ratio const &slimit); 

/// Sets the transport interface (for use by the core).
extern void conf_transport(Transport *transport);

/*************************************************************************/

} /* namespace mail */

/*************************************************************************/

#endif /* MAIL_MAIL_H */
