/* Various popular CRC.
 *
 * PegSoft miscellaneous library (c) 2004 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

#include <system/headers.h>

#include "misc/misc.h"

using namespace std;

/*************************************************************************/

namespace misc
{

/*************************************************************************/
/*************************************************************************/

/* The following code was written by Lammert Bies.
   http://www.lammertbies.nl/comm/software/index.html */

/*************************************************************************/

static uint16_t const P_16 = 0xA001;
static uint32_t const P_32 = 0xEDB88320;
static uint16_t const P_CCITT = 0x1021;
static uint16_t const P_DNP = 0xA6BC;

/*************************************************************************/

static bool crc_tab16_init = false;
static bool crc_tab32_init = false;
static bool crc_tabccitt_init = false;
static bool crc_tabdnp_init = false;

static uint16_t crc_tab16[256];
static uint32_t crc_tab32[256];
static uint16_t crc_tabccitt[256];
static uint16_t crc_tabdnp[256];

/*************************************************************************/

static void init_crc16_tab(void) 
{
	uint16_t crc, c;
	
	for (int i = 0; i < 256; ++i) 
	{
		crc = 0;
		c = (uint16_t) i;
	
		for (int j = 0; j < 8; ++j) 
		{
			if ((crc ^ c) & 0x0001)
				crc = (crc >> 1) ^ P_16;
			else 
				crc = crc >> 1;
			c = c >> 1;
		}
	
		crc_tab16[i] = crc;
	}
	
	crc_tab16_init = true;
}

/*************************************************************************/

static void init_crc32_tab(void) 
{
	uint32_t crc;
	
	for (int i = 0; i < 256; ++i) 
	{
		crc = (uint32_t) i;
	
		for (int j = 0; j < 8; ++j) 
		{
			if (crc & 0x00000001)
				crc = (crc >> 1) ^ P_32;
			else 
				crc = crc >> 1;
		}
	
		crc_tab32[i] = crc;
	}
	
	crc_tab32_init = true;
}

/*************************************************************************/

static void init_crcccitt_tab(void) 
{
	uint16_t crc, c;
	
	for (int i = 0; i < 256; ++i) 
	{
		crc = 0;
		c = ((uint16_t) i) << 8;
	
		for (int j = 0; j < 8; ++j) 
		{
			if ((crc ^ c) & 0x8000)
				crc = (crc << 1) ^ P_CCITT;
			else 
				crc = crc << 1;
			c = c << 1;
		}
	
		crc_tabccitt[i] = crc;
	}
	
	crc_tabccitt_init = true;
}

/*************************************************************************/

static void init_crcdnp_tab(void) 
{
	uint16_t crc, c;
	
	for (int i = 0; i < 256; ++i) 
	{
		crc = 0;
		c = (uint16_t) i;
	
		for (int j = 0; j < 8; ++j) 
		{
			if ((crc ^ c) & 0x0001)
				crc = (crc >> 1) ^ P_DNP;
			else 
				crc = crc >> 1;
			c = c >> 1;
		}
	
		crc_tabdnp[i] = crc;
	}
	
	crc_tabdnp_init = true;
}

/*************************************************************************/

static uint16_t update_crc16(uint16_t crc, char c) 
{
	uint16_t tmp, short_c;
	
	short_c = 0x00FF & (uint16_t) c;
	
	if (!crc_tab16_init) 
		init_crc16_tab();
	
	tmp = crc ^ short_c;
	crc = (crc >> 8) ^ crc_tab16[tmp & 0xFF];
	
	return crc;
}

/*************************************************************************/

static uint32_t update_crc32(uint32_t crc, char c) 
{
	uint32_t tmp, long_c;
	
	long_c = 0x000000FF & (uint32_t) c;
	
	if (!crc_tab32_init) 
		init_crc32_tab();
	
	tmp = crc ^ long_c;
	crc = (crc >> 8) ^ crc_tab32[tmp & 0xFF];
	
	return crc;
}

/*************************************************************************/

static uint16_t update_crc_ccitt(uint16_t crc, char c) 
{
	uint16_t tmp, short_c;
	
	short_c = 0x00FF & (uint16_t) c;
	
	if (!crc_tabccitt_init) 
		init_crcccitt_tab();
	
	tmp = (crc >> 8) ^ short_c;
	crc = (crc << 8) ^ crc_tabccitt[tmp];
	
	return crc;
}

/*************************************************************************/

static uint16_t update_crc_dnp(uint16_t crc, char c) 
{
	uint16_t tmp, short_c;
	
	short_c = 0x00FF & (uint16_t) c;
	
	if (!crc_tabdnp_init) 
		init_crcdnp_tab();
	
	tmp = crc ^ short_c;
	crc = (crc >> 8) ^ crc_tabdnp[tmp & 0xFF];
	
	return crc;
}

/*************************************************************************/
/*************************************************************************/

uint16_t crc16(char const *buf, size_t len, uint16_t start)
{
	uint16_t crc = start;
	
	while (len--)
		crc = update_crc16(crc, *buf++);
	
	return crc;
}

/*************************************************************************/

uint16_t crc16_str(char const *s, Trans const &trans)
{
	uint16_t crc = 0;
	
	while (*s)
		crc = update_crc16(crc, trans(*s++));
	
	return crc;
}

/*************************************************************************/

uint32_t crc32(char const *buf, size_t len, uint32_t start, bool last)
{
	uint32_t crc = start;
	
	while (len--)
		crc = update_crc32(crc, *buf++);
	
	if (last)
		return (crc ^ 0xFFFFFFFF);
	
	return crc;
}

/*************************************************************************/

uint32_t crc32_str(char const *s, Trans const &trans)
{
	uint32_t crc = 0xFFFFFFFF;
	
	while (*s)
		crc = update_crc32(crc, trans(*s++));
	
	return (crc ^ 0xFFFFFFFF);
}

/*************************************************************************/

uint16_t crc_ccitt(char const *buf, size_t len, uint16_t start)
{
	uint16_t crc = start;
	
	while (len--)
		crc = update_crc_ccitt(crc, *buf++);
	
	return crc;
}

/*************************************************************************/

uint16_t crc_ccitt_str(char const *s, Trans const &trans)
{
	uint16_t crc = 0xFFFF;
	
	while (*s)
		crc = update_crc_ccitt(crc, trans(*s++));
	
	return crc;
}

/*************************************************************************/

uint16_t crc_dnp(char const *buf, size_t len, uint16_t start, bool last)
{
	uint16_t crc = start;
	
	while (len--)
		crc = update_crc_dnp(crc, *buf++);
	
	if (last)
	{
		crc = ~crc;
		uint16_t lo = ((crc & 0xFF00) >> 8), hi = ((crc & 0x00FF) << 8);
	
		return (lo | hi);
	}
	
	return crc;
}

/*************************************************************************/

uint16_t crc_dnp_str(char const *s, Trans const &trans)
{
	uint16_t crc = 0;
	
	while (*s)
		crc = update_crc_dnp(crc, trans(*s++));
	
	crc = ~crc;
	uint16_t lo = ((crc & 0xFF00) >> 8), hi = ((crc & 0x00FF) << 8);
	
	return (lo | hi);
}

/*************************************************************************/

} /* namespace misc */

/*************************************************************************/
