/* Mysterious treasure box.
 *
 * PegSoft miscellaneous library (c) 2004 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

/** 
 * \file misc/misc.h
 * \brief Miscellaneous stuff.
 */

#ifndef MISC_MISC_H
#define MISC_MISC_H

/*************************************************************************/

#if HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif

#if HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <misc/string.h>

/*************************************************************************/

/// Namespace for all miscellaneous declarations.

namespace misc
{

/*************************************************************************/
/*************************************************************************/

/// Functor base class for string transformation
class Trans
{
	public:
		/// Destructor.
		virtual ~Trans() { }
		
		/// What is called by the actual functions.
		virtual int operator() (int ch) const = 0;
};

/// No transformation functor.
class TransNone : public Trans
{
	public:
		/// Does no transformation.
		virtual int operator() (int ch) const { return ch; }
};

/// Lowercase CRC transformation functor.
class TransLowerCase : public Trans
{
	public:
		/// Converts character to lowercase
		virtual int operator() (int ch) const { return tolower(ch); }
};

/// Uppercase CRC transformation functor.
class TransUpperCase : public Trans
{
	public:
		/// Converts character to lowercase
		virtual int operator() (int ch) const { return toupper(ch); }
};

/*************************************************************************/
/*************************************************************************/

/// Fills buf with the absolute path to filename.
/** Fills buf with the absolute path to filename (up to len characters).
  * If filename is not an absolute path already, default_dir is prepended
  * to it.
  * \param[out] buf The buffer to write to.
  * \param[in] len Maximum number of characters to write.
  * \param[in] default_dir Default directory to prepend if filename
  * is not an absolute path. If this is NULL or an empty string, then
  * the filename is copied as is to the buffer.
  * \param[in] filename Filename for which absolute path is needed.
  * \return Nothing.
  */ 
extern void absolute_path(char *buf, size_t len, char const *default_dir,
		 char const *filename);

/// Generates a string with the absolute path to filename.
/** Generates a string with the absolute path to filename. If filename is
  * not an absolute path already, default_dir is prepended to it.
  *
  * \param[in] default_dir The default directory to prepend if filename
  * is not an absolute path. If this is NULL or an empty string, then
  * the generated string is a copy of filename.
  * \param[in] filename The filename for which absolute path is needed.
  * \return The absolute path to filename. */
extern cstring absolute_path(char const *default_dir, char const *filename);

/// Same as absolute_path(char const *, char const *), but with std::string's.
extern std::string absolute_path(std::string const &default_dir,
		std::string const &filename);

/*************************************************************************/

/// Decodes the input from base64.
/** Decodes the input from base64.
  *
  * \param[in] input The input buffer.
  * \param[in] input_len The number of bytes in the input buffer.
  * \param[out] output The output buffer, whose length must be the
  * minimum for base64 decoding of the input buffer.
  * \return The length of the data written to the output buffer, or
  * zero if there was an error in the input.
  * \sa base64_declen() */
size_t base64_decode(char const *input, size_t input_len,
		char *output);

/// Encodes the input in base64.
/** Encodes the input in base64.
  *
  * \param[in] input The input buffer.
  * \param[in] input_len The number of bytes in the input buffer.
  * \param[out] output The output buffer, whose length must be the
  * minimum for base64 encoding of the input buffer.
  * \return The length of the data written to the output buffer.
  * \sa base64_enclen() */
size_t base64_encode(char const *input, size_t input_len,
		char *output);

/// Like base64_decode, but the output buffer will be NULL terminated.
size_t base64_decode_str(char const *input, size_t input_len,
		char *output);
/// Like base64_encode, but the output buffer will be NULL terminated.
size_t base64_encode_str(char const *input, size_t input_len,
		char *output);

/// Returns the minimum length for a base64 decoding output buffer.
/** Returns the minimum length for a base64 decoding output buffer for a
  * given input buffer length.
  *
  * \param[in] The size of the input buffer (which is a multiple of 4).
  * \return The minimum size of the output buffer. */

static inline size_t base64_declen(size_t input_len)
{
	return ((input_len / 4) * 3);
}

/// Returns the minimum length for a base64 encoding output buffer.
/** Returns the minimum length for a base64 encoding output buffer for a
  * given input buffer length.
  *
  * \param[in] The size of the input buffer.
  * \return The minimum size of the output buffer. */

static inline size_t base64_enclen(size_t input_len)
{
	return (((input_len % 3) == 0) ? ((input_len / 3) * 4) : (((input_len / 3) + 1) * 4));
}

/*************************************************************************/

/// Converts CIDR4 notation to a netmask.
static inline uint32_t cidr4_to_netmask(uint8_t cidr)
{
	return (cidr ? (0xFFFFFFFF - (1 << (32 - cidr)) + 1) : 0);
}

/// Converts a netmask to CIDR4 notation.
static inline uint8_t netmask_to_cidr4(uint32_t mask) 
{
   uint8_t tmp = 0;

   while (!(mask & (1 << tmp)) && tmp < 32) 
      tmp++;

   return (32 - tmp);
}

/*************************************************************************/

/// No transformation
static TransNone const ctNone;
/// All characters in lowercase
static TransLowerCase const ctLowerCase;
/// All characters in uppercase
static TransUpperCase const ctUpperCase;

/// Computes the CRC16 checksum of the given buffer.
/** Computes the CRC16 checksum of the given buffer.
  *
  * \param[in] buf The buffer for which a CRC is needed.
  * \param[in] len The number of bytes in the buffer.
  * \param[in] start The value used as the base CRC (use either
  * the default value if you just want to compute the CRC of the
  * buffer, or a previously generated CRC if this is a followup
  * call).
  * \return The computed CRC. */
extern uint16_t crc16(char const *buf, size_t len, uint16_t start = 0);

/// Computes the CRC16 checksum of a NULL-terminated string.
/** Computes the CRC16 checksum of the given NULL-terminated string.
  *
  * \param[in] s The string for which a CRC is needed.
  * \param[in] trans A function that does a transformation to each
  * character in the string before it is used to compute the CRC. The
  * constants ctNone, ctLowerCase and ctUpperCase are provided for
  * respectively no, lowercase and uppercase transformation, but you
  * may write your own function too.
  * \return The computed CRC. */
extern uint16_t crc16_str(char const *s, Trans const &trans = ctNone);

/// Computes the CRC32 checksum of the given buffer.
/** Computes the CRC32 checksum of the given buffer.
  *
  * \param[in] buf The buffer for which a CRC is needed.
  * \param[in] len The number of bytes in the buffer.
  * \param[in] start The value used as the base CRC (use either
  * the default value if you just want to compute the CRC of the
  * buffer, or a previously generated CRC if this is a followup
  * call).
  * \param[in] last Whether to apply the extra computation required
  * to complete the CRC.
  * \return The computed CRC. */
extern uint32_t crc32(char const *buf, size_t len,
		uint32_t start = 0xFFFFFFFF, bool last = true);

/// Computes the CRC32 checksum of a NULL-terminated string.
/** Computes the CRC32 checksum of the given NULL-terminated string.
  *
  * \param[in] s The string for which a CRC is needed.
  * \param[in] trans A function that does a transformation to each
  * character in the string before it is used to compute the CRC. The
  * constants ctNone, ctLowerCase and ctUpperCase are provided for
  * respectively no, lowercase and uppercase transformation, but you
  * may write your own function too.
  * \return The computed CRC. */
extern uint32_t crc32_str(char const *s, Trans const &trans = ctNone);

/// Computes the CRC-CCITT checksum of the given buffer.
/** Computes the CRC-CCITT checksum of the given buffer.
  *
  * \param[in] buf The buffer for which a CRC is needed.
  * \param[in] len The number of bytes in the buffer.
  * \param[in] start The value used as the base CRC (use either
  * the default value if you just want to compute the CRC of the
  * buffer, or a previously generated CRC if this is a followup
  * call).
  * \return The computed CRC. */
extern uint16_t crc_ccitt(char const *buf, size_t len,
		uint16_t start = 0xFFFF);

/// Computes the CRC-CCITT checksum of a NULL-terminated string.
/** Computes the CRC-CCITT checksum of the given NULL-terminated string.
  *
  * \param[in] s The string for which a CRC is needed.
  * \param[in] trans A function that does a transformation to each
  * character in the string before it is used to compute the CRC. The
  * constants ctNone, ctLowerCase and ctUpperCase are provided for
  * respectively no, lowercase and uppercase transformation, but you
  * may write your own function too.
  * \return The computed CRC. */
extern uint16_t crc_ccitt_str(char const *s, Trans const &trans = ctNone);

/// Computes the CRC-DNP checksum of the given buffer.
/** Computes the CRC-DNP checksum of the given buffer.
  *
  * \param[in] buf The buffer for which a CRC is needed.
  * \param[in] len The number of bytes in the buffer.
  * \param[in] start The value used as the base CRC (use either
  * the default value if you just want to compute the CRC of the
  * buffer, or a previously generated CRC if this is a followup
  * call).
  * \param[in] last Whether to apply the extra computation required
  * to complete the CRC.
  * \return The computed CRC. */
extern uint16_t crc_dnp(char const *buf, size_t len, uint16_t start = 0,
		bool last = true);

/// Computes the CRC-DNP checksum of a NULL-terminated string.
/** Computes the CRC-DNP checksum of the given NULL-terminated string.
  *
  * \param[in] s The string for which a CRC is needed.
  * \param[in] trans A function that does a transformation to each
  * character in the string before it is used to compute the CRC. The
  * constants ctNone, ctLowerCase and ctUpperCase are provided for
  * respectively no, lowercase and uppercase transformation, but you
  * may write your own function too.
  * \return The computed CRC. */
extern uint16_t crc_dnp_str(char const *s, Trans const &trans = ctNone);

/*************************************************************************/

/// Converts a number of seconds to a duration expression.
/** Converts the given number of seconds to a duration expression
  * that is suitable for the duration_scan() function.
  * \param[out] buf  The buffer where the expression will be written.
  * \param[in]  len  The maximum length of the buffer.
  * \param[in]  secs The number of seconds to convert.
  * \return The buf pointer. */
char *duration_expr(char *buf, size_t len, time_t secs);

/// Scans a string expression of a duration.
/** Scans a string expression of a duration to find the number of
  * seconds in it, which is then copied to a caller-submitted variable.
  * \param[in] expr The expression of a duration. It's a number followed
  * by one one optional letter (w for weeks, d for days, h for hours,
  * m for minutes, s for seconds -- the default unit). Combinations
  * are allowed, as are the use of negative numbers.
  * \param[out] secs The number of seconds found. May be 0 (in that case
  * it just checks whether the expression is valid).
  * \return true if the string representation of the duration is valid,
  * false otherwise (in that case the value of *secs is undefined). */

extern bool duration_scan(char const *expr, time_t *secs); 

/*************************************************************************/

/// Determines whether the given directory exists.
/** Determines whether the given name is an existing directory.
  *
  * \param[in] dirname The name of the directory to check for existence.
  * \return true if the directory exists, false otherwise. errno might
  * have been set in the latter case. */
extern bool dir_exists(char const *dirname);

/// Determines whether the given file exists.
/** Determines whether the given name is an existing regular file.
  *
  * \param[in] filename The name of the file to check for existence.
  * \return true if the file exists, false otherwise. errno might
  * have been set in the latter case. */
extern bool file_exists(char const *filename);

/*************************************************************************/

/// Matches a string against a pattern.
/** Matches a string against a pattern.
  *
  * \param[in] pattern The pattern. The ? character in the pattern matches
  * a single character in the string, while the * character matches any
  * number of characters in the string.
  * \param[in] str The string to match.
  * \param[in] trans The transformation to apply to each character of the
  * string and pattern before they are being matched. The default,
  * TransUpperCase(), has the effect of matching case-insensitively.
  * \return true if the string matches the pattern, false otherwise. */
extern bool match(char const *pattern, char const *str, 
		Trans const &trans = TransUpperCase());

/// Like match(), but optimized for IRC.
extern bool irc_match(char const *pattern, char const *str, 
		Trans const &trans = TransUpperCase());

/*************************************************************************/

/// Converts a simple range expression.
/** Converts a simple range expression (in the form n1[-n2] where n1 and n2
  * are two non-zero positive numbers) to two unsigned long.
  *
  * \param[in]  expr The range expression. If it is NULL, then *first
  * will be set to 1 and *last to ULONG_MAX.
  * \param[out] first A pointer to an integer variable that will be set to
  * the first number of the range.
  * \param[out] last A pointer to an integer variable that will be set to
  * the last number of the range.
  * \param[in]  desc Whether *last may be less than *first.
  * \return true if the range expression is valid, false otherwise (the
  * values of first and last are undefined in that case). */
extern bool simple_range(char const *expr, unsigned long *first,
		unsigned long *last, bool desc = false);

/// Converts a complex range expression.
/** This function is similar to simple_range(), except that it provides
  * support for multiple ranges separated by commas (i.e. n1[-n2],n3[-n4]).
  *
  * \param[in] expr A pointer to the multiple ranges expression. Each call
  * to the function will only convert one range, but will make expr point
  * to the start of the next range (if any), so you can call the function
  * again to convert the next range. If expr or *expr is NULL, then 
  * *first will be set to 1 and last to ULONG_MAX.
  * \param[out] first A pointer to an integer variable that will be set to
  * the first number of the first range.
  * \param[out] last A pointer to an integer variable that will be set to
  * the last number of the first range.
  * \param[in] desc Whether *last may be less than *first.
  * \return true if the first range expression is valid, false otherwise
  * (the values of *first and *last are undefined in that case, and the
  * value of *expr is left untouched). */
extern bool complex_range(char const **expr, unsigned long *first,
		unsigned long *last, bool desc = false);

/*************************************************************************/

/// Creates a directory recursively.
/** Creates a directory recursively (equivalent to the mkdir -p shell
  * command).
  * \param[in] pathname The directory.
  * \param[in] mode The permission bits of the created directories
  * (modified by the process umask).
  * \return true if successful, false otherwise (errno is set
  * accordingly in that case). */
extern bool recursive_mkdir(char const *pathname, mode_t mode);

/*************************************************************************/

/// Gets resolve error message.
/** Gets an error message from an error number received by a previous
  * call to a resolve_* function.
  * \param[out] buf Buffer receiving the error.
  * \param[in]  len Size of the buffer.
  * \param[in]  err The error number.
  * \return true if an error message is available, false otherwise. */
extern bool resolve_error(char *buf, size_t len, int err);  

/// Resolves a hostname.
/** Resolves a hostname, and filling a struct in_addr with the resulting
  * IP, and an int with the error number if applicable.
  *
  * \param[in]  host The hostname to resolve.
  * \param[out] in Pointer to the struct in_addr that will receive the
  * IP address.
  * \param[out] err Pointer to the int that will receive the error number
  * (0 if no error number is needed).
  * \return true if host has been successfully resolved, false otherwise
  * (the contents of in is undefined in that case). */
extern bool resolve_host(char const *host, struct in_addr *in, int *err);

/*************************************************************************/

} /* namespace misc */

/*************************************************************************/

#endif /* MISC_MISC_H */
