/* String-related treasures.
 *
 * PegSoft miscellaneous library (c) 2004 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

/** 
 * \file misc/string.h
 * \brief String utilities.
 */

#ifndef MISC_STRING_H
#define MISC_STRING_H

/*************************************************************************/

#include <string>

/*************************************************************************/

namespace misc
{

/*************************************************************************/

/// A smart pointer class for C strings.

/** This smart pointer class encapsulates a C null-terminated string,
  * makes sure it is freed when the class is destructed and automatically
  * duplicates strings passed to it through operator=.
  *
  * This class is suitable for small memory requirements. */

class cstring
{
	public:
		/// Default onstructor.
		cstring() : m_str(0) { }
		/// Constructor initializing the encapsulated string.
		/** This constructor initializes the encapsulated string
		  * to a duplicate of the given string.
		  *
		  * \param[in] str The string to copy. */
		explicit cstring(char const *str) : m_str(0) { copy(str); }
		/// Constructor setting the encapsulated string to a portion of a string.
		/** This constructor initializes the encapsulated string to a
		  * duplicate of a limited portion of the given string.
		  *
		  * \param[in] str The string to copy. If it is NULL, then
		  * the encapsulated string is set to NULL and len is ignored.
		  * \param[in] len The maximum number of characters to copy. */
		cstring(char const *str, size_t len) : m_str(0) { copy(str, len); }
		/// Copy constructor.
		cstring(cstring const &cs) : m_str(0) { copy(cs.m_str); }
		/// Assignment operator (for cstring r-values).
		cstring &operator=(cstring const &right);
		
		/// Assigns a new string.
		/** Assigns a new string to the encapsulated string. The
		  * old string is automatically freed.
		  *
		  * \param[in] right The string to copy.
		  * \return The cstring object. */
		cstring &operator=(char const *right) { copy(right); return *this; }
		
		/// Checks two strings for equality, byte-per-byte.
		/** Does a byte-per-byte comparison of the encapsulated string
		  * and the given one.
		  *
		  * \param[in] right The string to compare to the encapsulated string.
		  * \return true if the two strings are the same or are both NULL, false if
		  * they're different or if only one of the strings is NULL. */
		bool operator==(char const *right) const;
		/// Checks two strings for difference, byte-per-byte
		/** Does a byte-per-byte comparison of the encapsulated string
		  * and the given one.
		  *
		  * \param[in] right The string to compare to the encapsulated string.
		  * \return true if the two strings are different or if at least one of
		  * the strings is NULL, false otherwise. */
		bool operator!=(char const *right) const
				{ return !(operator==(right)); }
		
		/// Conversion operator (to char *)
		operator char *() { return m_str; }
		/// Conversion operator (to const char *).
		operator char const *() const { return m_str; }
		
		/// Assigns a new string.
		/** Assigns a new string to the encapsulated string. The
		  * old string is automatically freed.
		  *
		  * \param[in] str The string to copy. May be NULL.
		  * \return The cstring object. */
		cstring &assign(char const *str) { copy(str); return *this; }
		
		/// Assigns a limited portion of a string.
		/** Assigns a limited portion of a string to the encapsulated string.
		  * The old string is automatically freed.
		  *
		  * \param[in] str The string to copy. If it is NULL, then the
		  * encapsulated string is set to NULL and len is ignored.
		  * \param[in] len The maximum number of characters to copy.
		  * \return The cstring object */
		cstring &assign(char const *str, size_t len) 
				{ copy(str, len); return *this; }
		
		/// Compares to a string, byte-per-byte.
		/** Compares a string with the encapsulated string, byte-per-byte.
		  *
		  * \param[in] str The string to compare.
		  * \return < 0, 0 or > 0 if the encapsulated string is respectively
		  * lexographically less than, equal to or greater than the given
		  * string. If the encapsulated string is NULL, then -1 is returned,
		  * if the given string is NULL, then 1 is returned and if both
		  * strings are NULL, then 0 is returned. */
		int compare(char const *str) const;
		
		/// Compares to a limited portion of a string, byte-per-byte.
		/** Compares a limited portion of a string with the encapsulated string,
		  * byte-per-byte.
		  *
		  * \param[in] str The string to compare.
		  * \param[in] len The maximum number of characters to compare in
		  * the strings.
		  * \return < 0, 0 or > 0 if the encapsulated string is respectively
		  * lexographically less than, equal to or greater than the given
		  * string. If one or both strings are false, then -1 is returned. */
		int compare(char const *str, size_t len) const;
		
		/// Reserves a minimum size of writable memory.
		/** Reserves a minimum size of writable memory for the enclosed
		  * string (this includes the NULL termination character). The
		  * contents of the string currently assigned is lost.
		  *
		  * \param[in] size The size of the memory to reserve. Note that
		  * if this parameter has the value 0, then this function is
		  * an no-op.
		  * \return a pointer to the enclosed string. */
		char *reserve(size_t size);
		
		/// Swaps encapsulated strings of two cstring objects.
		/** Swaps encapsulated strings of two cstring objects. This is
		  * more optimized than normal swapping since no reallocation
		  * is done, the encapsulated pointers are simply exchanged.
		  *
		  * \param[in] str The string to swap. */
		void swap(misc::cstring &str);
		
		/// Returns The encapsulated string.
		char *str() { return m_str; }
		/// Returns the encapsulated string (const version).
		char const *str() const { return m_str; }
	private:
		char *m_str;
		
		void copy(char const *s);
		void copy(char const *s, size_t len);
		void free();
};

/*************************************************************************/
/*************************************************************************/

/// Static string encapsulation

/** This template class encapsulates a statically-allocated C string. */

template <size_t LEN>
class ststring
{
	public:
		/// Maximum length of the string
		static size_t const max_len = LEN;
		
		/// Default onstructor.
		ststring() : m_str() 
		{ 
			memset(m_str, 0, max_len + 1);
		}
		/// Constructor initializing the encapsulated string.
		/** This constructor initializes the encapsulated string
		  * to a copy of the given string. 
		  *
		  * \param[in] str The string to copy. If it is longer than what the
		  * encapsulated string can contain, it is truncated. */
		explicit ststring(char const *str) : m_str() 
		{
			m_str[max_len] = 0;
			strncpy(m_str, str, max_len); 
		}	 
		/// Copy constructor.
		ststring(ststring const &sts) : m_str() { strcpy(m_str, sts.m_str); }
		/// Assignment operator (for cstring r-values).
		ststring &operator=(ststring const &right)
		{
			if (this != &right)
				strcpy(m_str, right.m_str);
			
			return *this;
		}
		
		/// Assigns a new string.
		/** Assigns a new string to the encapsulated string.
		  *
		  * \param[in] right The string to copy.
		  * \return The ststring object. */
		ststring &operator=(char const *right) 
				{ strncpy(m_str, right, max_len); return *this; }
		
		/// Checks two strings for equality, byte-per-byte.
		/** Does a byte-per-byte comparison of the encapsulated string
		  * and the given one. The comparison is done only on the characters
		  * that would actually fit in the encapsulated string.
		  *
		  * \param[in] right The string to compare to the encapsulated string.
		  * \return true if the two strings are the same or are both NULL,
		  * false if they're different or if only one of the strings is NULL. 
		  */
		bool operator==(char const *right) const
		{
			if (!str() && !right)
				return true;
			
			if (!str() || !right)
				return false;
			
			return (strncmp(m_str, right, max_len) == 0);
		}
			
		/// Checks two strings for difference, byte-per-byte
		/** Does a byte-per-byte comparison of the encapsulated string
		  * and the given one. The comparison is done only on the characters
		  * that would actually fit in the encapsulated string.
		  *
		  * \param[in] right The string to compare to the encapsulated string.
		  * \return true if the two strings are different or if at least one of
		  * the strings is NULL, false otherwise. */
		bool operator!=(char const *right) const
				{ return !(operator==(right)); }
		
		/// Conversion operator (to char *)
		operator char *() { return str(); }
		/// Conversion operator (to const char *).
		operator char const *() const { return str(); }
		
		/// Returns The encapsulated string, or NULL if it is empty.
		char *str() { return (*m_str ? m_str : 0); }
		/// Returns the encapsulated string, or NULL if it is empty (const version).
		char const *str() const { return (*m_str ? m_str : 0); }
	private:
		char m_str[LEN + 1];
};

/*************************************************************************/
/*************************************************************************/

/// Functor used for case-insensitive comparison between two strings. 
class insensitive_less
{
	public:
		bool operator()(std::string const &s1, std::string const &s2)
				const;
};

/*************************************************************************/
/*************************************************************************/

/// A stpncpy replacement that ensures the string is NULL-terminated.

static inline char *stpncpyzt(char *dest, const char *src, size_t n)
{
	char *end = stpncpy(dest, src, n);
	dest[n-1] = 0;
	return end;
}

/*************************************************************************/

/// A strncpy replacement that ensures the string is NULL-terminated.

static inline char *strncpyzt(char *dest, const char *src, size_t n)
{
	strncpy(dest, src, n);
	dest[n-1] = 0;
	return dest;
}

/*************************************************************************/
/*************************************************************************/

/// Duplicates a string using operator new [].
/** Duplicates a string using operator new [] (instead of malloc for the
  * standard strdup() function.
  *
  * \param[in] str The string to duplicate.
  * \return A pointer to the newly-allocated string, which can be
  * freed with operator delete []. */
extern char *nstrdup(char const *str);

/*************************************************************************/

/// Range checking version of strtol.
/** This function converts the given character string in the same way as
  * the C library strtol() function would, then ensures that the entire
  * string is valid and that the computed number is in a given range.
  *
  * \param[out] num A pointer to the variable that will hold the
  * computed number. The variable is changed only if the function
  * returns true.
  * \param[in] str The string to convert.
  * \param[in] min The minimum value allowed for the computed number.
  * \param[in] max The maximum value allowed for the computed number.
  * \param[in] base The base to use when converting the string.
  * \return true if the entire string is a valid number and the
  * computed number is in range, false otherwise (in this case, errno is set
  * to ERANGE if the computed number was out of range, and to whatever
  * error was set by strtol() otherwise). */
extern bool strtol_range(long *num, char const *str, long min, long max,
		int base = 10);
/// Range checking version of strtoul (see strtol_range).
extern bool strtoul_range(unsigned long *num, char const *str,
		unsigned long min, unsigned long max, int base = 10);


/*************************************************************************/

} /* namespace misc */

/*************************************************************************/

#endif /* MISC_STRING_H */
