/* URI-related treasures.
 *
 * PegSoft miscellaneous library (c) 2004 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

/** 
 * \file misc/uri.h
 * \brief URI processing utilities.
 */

#ifndef MISC_URI_H
#define MISC_URI_H

/*************************************************************************/

#include <map>
#include <string>

/*************************************************************************/

namespace misc
{

/*************************************************************************/

/// URI parser class
/** This class parses a URI and allows subsequent access to each of its
  * parts.
  */
class URI
{
	public:
		/// The query map type.
		typedef std::map<std::string, std::string> parameters_t;
		
		/// Constructor.
		URI() : m_uri(), m_scheme(), m_contents(), m_authority(), m_path(), 
				m_query(), m_hostname(), m_port(0), m_username(), m_password(), 
				m_parameters() { }
		/// Constructor assigning a URI.
		URI(std::string const &uri) : m_uri(), m_scheme(), m_contents(), 
				m_authority(), m_path(), m_query(), m_hostname(), m_port(0),
				m_username(),  m_password(), m_parameters() { assign(uri); }
		/// Constructor assigning a URI, char const * version
		URI(char const *uri) : m_uri(), m_scheme(), m_contents(), 
				m_authority(), m_path(), m_query(), m_hostname(), m_port(0),
				m_username(),  m_password(), m_parameters() { assign(uri); }
		/// Copy constructor.
		URI(URI const &uri) : m_uri(uri.m_uri), m_scheme(uri.m_scheme), 
				m_contents(uri.m_contents), m_authority(uri.m_authority),
				m_path(uri.m_path), m_query(uri.m_query), 
				m_hostname(uri.m_hostname), m_port(uri.m_port),
				m_username(uri.m_username), m_password(uri.m_password),
				m_parameters(uri.m_parameters) { }
		/// Assignment operator.
		URI &operator=(URI const &right);
		/// == operator.
		bool operator==(URI const &right)
				{ return (m_uri == right.m_uri); }
		/// != operator.
		bool operator!=(URI const &right)
				{ return !(operator==(right)); }
		
		/// Assigns a URI.
		/** Assigns a URI.
		  *
		  * \param[in] uri The URI that will be parsed.
		  * \return true if the URI has been parsed successfully, false
		  * otherwise.
		  */
		bool assign(std::string const &uri) 
				{ clear(); return parse_uri(uri.begin(), uri.end()); }
		/// Same as assign(std::string const &) but with a char const * parameter
		bool assign(char const *uri) 
				{ clear(); return parse_uri(uri, uri + strlen(uri)); }
		
		/// Clears the object from its URI.
		void clear();
		
		/// Returns the full URI
		std::string const &uri() const { return m_uri; }
		
		/// Returns the URI scheme
		std::string const &scheme() const { return m_scheme; }
		/// Returns the URI scheme-specific part
		std::string const &contents() const { return m_contents; }
		
		/// Returns the URI authority
		std::string const &authority() const { return m_authority; }
		/// Returns the URI path
		std::string const &path() const { return m_path; }
		/// Returns the URI query
		std::string const &query() const { return m_query; }
		
		/// Returns the URI hostname
		std::string const &hostname() const { return m_hostname; }
		/// Returns the URI port
		uint16_t port() const { return m_port; }
		/// Returns the URI username
		std::string const &username() const { return m_username; }
		/// Returns the URI password
		std::string const &password() const { return m_password; }
		
		/// Returns the URI parameters as a map
		parameters_t const &parameters() const { return m_parameters; }
	private:
		std::string m_uri;
		
		std::string m_scheme;
		std::string m_contents;
		
		std::string m_authority;
		std::string m_path;
		std::string m_query;
			
		std::string m_hostname;
		uint16_t m_port;
		std::string m_username;
		std::string m_password;
			
		parameters_t m_parameters;
			
		template <typename T> bool parse_uri(T beg, T end)
		{
			T send = end;
			
			for (T it = beg; it != end; ++it)
			{
				if (*it == ':')
				{
					send = it;
					
					if (++it == end)
						return false;
					
					break;
				}
			}
			
			if (send == end)
				return false;
			
			if (!parse_contents(send + 1, end))
				return false;
			
			m_scheme.assign(beg, send);
			m_contents.assign(send + 1, end);
			m_uri.assign(beg, end);
			
			return true;
		}
		
		template <typename T> bool parse_contents(T beg, T end)
		{	
			if (*beg == '/' && (beg + 1) != end && *(beg + 1) == '/')
				beg += 2;
			else
				return true;
			
			T aend = end, pend = end;
			
			for (T it = beg; it != end; ++it)
			{	
				if (*it == '/')
				{
					if (aend != end)
						return false;
					
					aend = it;
				}
				else if (*it == '?')
				{
					if (aend == end)
						return false;
					
					if (it + 1 == end)
						pend = end = it;
					else
						pend = it;
					break;
				}
			}
			
			if (beg != aend && !parse_authority(beg, aend))
				return false;
			
			if (pend != end && (pend + 1) != end && 
					!parse_query(pend + 1, end))
				return false;
			
			if (beg != aend)
				m_authority.assign(beg, aend);
			if (aend != end)
				m_path.assign(aend, pend);
			if (pend != end)
				m_query.assign(pend + 1, end);
			return true;
		}
		
		template <typename T> bool parse_authority(T beg, T end)
		{	
			T cend = beg;
			
			for (T it = beg; it != end; ++it)
			{	
				if (*it == '@')
				{
					cend = it;
					
					if (++it == end)
						return false;
					
					break;
				}
			}
			
			if (cend != beg) 
				return parse_credentials(beg, cend, end);
			else
				return parse_host(cend, end);
		}
		
		template <typename T> bool parse_credentials(T beg, T end,
				T hend)
		{	
			T uend = end;
			
			for (T it = beg; it != end; ++it)
			{	
				if (*it == ':')
				{
					if (it == beg)
						return false;
					
					if (it + 1 == end)
						uend = end = it;
					else
						uend = it;
					break;
				}
			}
			
			if (!parse_host(end + 1, hend))
				return false;
			
			m_username.assign(beg, uend);
			if (uend != end)
				m_password.assign(uend + 1, end);
			
			return true;
		}
		
		template <typename T> bool parse_host(T beg, T end)
		{	
			T nend = end;
			
			for (T it = beg; it != end; ++it)
			{	
				if (*it == ':')
				{
					if (it == beg)
						return false;
					
					if (it + 1 == end)
						nend = end = it;
					else
						nend = it;
					break;
				}
			}
			
			m_hostname.assign(beg, nend);
			if (nend != end)
			{
				std::string port(nend + 1, end);
				m_port = atoi(port.c_str());
			}
			
			return true;
		}
		
		template <typename T> bool parse_query(T beg, T end)
		{	
			while (beg != end && *beg == '&')				
				++beg;
			
			T pbeg = beg;
			
			for (T it = beg; it != end; ++it)
			{	
				if (*it == '&')
				{
					if (!parse_parameter(pbeg, it))
						return false;
					
					while (it != end && *it == '&')
						++it;
					
					pbeg = it;
					
					if (it == end)
						break;
				}
			}
			
			if (end - pbeg >= 1 && !parse_parameter(pbeg, end))
				return false;
			
			return true;
		}
		
		template <typename T> bool parse_parameter(T beg, T end)
		{	
			T nend = end;
			
			for (T it = beg; it != end; ++it)
			{	
				if (*it == '=')
				{
					if (it == beg)
						return false;
					
					if (it + 1 == end)
						nend = end = it;
					else
						nend = it;
					break;
				}
			}
			
			if (nend != end)
				m_parameters[std::string(beg, nend)] = 
						std::string(nend + 1, end);
			else
				m_parameters[std::string(beg, nend)] = "";
			
			return true;
		}
};

/*************************************************************************/

} /* namespace misc */

/*************************************************************************/

#endif /* MISC_URI_H */
