/* Gotta load 'em all.
 *
 * PegSoft modules subsystem library (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

#include <system/headers.h>

#include <fstream>
#include <iostream>

#include <ltdl.h>

#include <conf/quick.h>
#include <log/levels.h>

#include "modules/conf.h"
#include "modules/core.h"
#include "modules/manager.h"
#include "modules/module.h"

using namespace std;
using namespace conf;

/*************************************************************************/

static char const *path_to_func(char const *path, char const *func)
{
	static char buf[PATH_MAX];
	char npath[PATH_MAX], *end = npath;
	
	memset(npath, 0, sizeof(npath));
	strncpy(npath, path, PATH_MAX-1);
	
	while (*end)
	{
		if (!isalnum(*end))
			*end++ = '_';
		else
			++end;
	} 
	
	snprintf(buf, sizeof(buf), "%s_%s", npath, func);
	return buf;
}

/*************************************************************************/

/* Trick to allow LTDL_SET_PRELOADED_SYMBOLS to be called outside of
 * the modules namespace. */

static inline void SET_PRELOADED_SYMBOLS()
{
	LTDL_SET_PRELOADED_SYMBOLS();
}

/*************************************************************************/

namespace modules
{

/*************************************************************************/

typedef Module *(*module_get_t)(Core &);
typedef void (*module_release_t)(Core &, Module *);

/*************************************************************************/

LOGLEVEL(LogModule, xlog::LogNotice, "Notice::Module");

/*************************************************************************/

ModManager::ModManager(Core &core, ConfModuleA const &confmodule) :
		m_modules(), m_core(&core), m_error(), m_errorcode(0),
		m_confmodule(&confmodule), m_global(0), m_started(false)
{
	SET_PRELOADED_SYMBOLS();
	
	if (lt_dlinit())
		m_core->fatalf("Unable to initialize module subsystem: "
				"libltdl error: %s", lt_dlerror());
}

/*************************************************************************/

ModManager::~ModManager()
{
	/* Ignore errors; we're probably shutting down anyway. */
	lt_dlexit();
}

/*************************************************************************/

void ModManager::broadcast(int state, Module &sender, int msg, void *arg)
{
	if (m_core != &sender)
		m_core->notify(sender, msg, arg);
	
	list_type::iterator it, end;
	
	for (it = m_modules.begin(), end = m_modules.end(); it != end; ++it)
	{
		/* Send notification messages to loaded modules only. */
		if (*it != &sender && ((*it)->state() & (Module::msLoaded | state)))
			(*it)->notify(sender, msg, arg);
	}
}

/*************************************************************************/

void ModManager::broadcast_reverse(int state, Module &sender, int msg, void *arg)
{
	if (m_core != &sender)
		m_core->notify(sender, msg, arg);
	
	list_type::reverse_iterator it, end;
	
	for (it = m_modules.rbegin(), end = m_modules.rend(); it != end; ++it)
	{
		/* Send notification messages to loaded modules only. */
		if (*it != &sender && ((*it)->state() & (Module::msLoaded | state)))
			(*it)->notify(sender, msg, arg);
	}
}

/*************************************************************************/

std::string ModManager::confblock(std::string const &name)
{
	string block;
	
	ConfModuleA::container_type::const_iterator it, end;
	
	for (it = m_confmodule->value().begin(), 
			end = m_confmodule->value().end(); it != end; ++it)
	{
		if (it->mname.value() == name)
		{
			block += it->config.value();
			block += "\n";
		}
	}
	
	return block;
}

/*************************************************************************/

void ModManager::errorf(int code, const char *format, ...)
{
	char buf[4096];
	
	va_list args;
	va_start(args, format);
	vsnprintf(buf, sizeof(buf), format, args);
	va_end(args);
	
	m_error = buf;
	m_errorcode = code;
}

/*************************************************************************/

void ModManager::feed(Module &recipient, int state, int msg, void *arg)
{
	list_type::iterator it, end;
	
	for (it = m_modules.begin(), end = m_modules.end(); it != end; ++it)
	{
		if (*it != &recipient && ((*it)->state() & state))
			recipient.notify(**it, msg, arg);
	}
}

/*************************************************************************/

void ModManager::feed_reverse(Module &recipient, int state, int msg, void *arg)
{
	list_type::reverse_iterator it, end;
	
	for (it = m_modules.rbegin(), end = m_modules.rend(); it != end; ++it)
	{
		if (*it != &recipient && ((*it)->state() & state))
			recipient.notify(**it, msg, arg);
	}
}

/*************************************************************************/

Module *ModManager::find(std::string const &name) const
{
	if (m_core->name() == name)
		return m_core;
	
	list_type::const_iterator it, end;
	
	for (it = m_modules.begin(), end = m_modules.end(); it != end; ++it)
	{
		if ((*it)->name() == name && ((*it)->state() & Module::msLoaded))
			return *it;
	}
	
	return 0;
}

/*************************************************************************/

Module *ModManager::find(std::string const &name, int state) const
{
	if (m_core->name() == name)
		return m_core;
	
	list_type::const_iterator it, end;
	
	for (it = m_modules.begin(), end = m_modules.end(); it != end; ++it)
	{
		if ((*it)->name() == name && (!state || ((*it)->state() & state)))
			return *it;
	}
	
	return 0;
}

/*************************************************************************/

Interface *ModManager::find_interface(PSIID iid, Module **from) const
{
	Interface *ret = 0;
	
	if ((ret = m_core->interface(iid)))
	{
		if (from)
			*from = m_core;
		return ret;
	}
	
	list_type::const_iterator it, end;
	
	for (it = m_modules.begin(), end = m_modules.end(); it != end; ++it)
	{
		if (((*it)->state() & Module::msLoaded) && 
				(ret = (*it)->interface(iid)))
		{
			if (from)
				*from = *it;
			return ret;
		}
	}
	
	return 0;
}

/*************************************************************************/

bool ModManager::load(std::string const &name)
{	
	Module *m;
	ConfQ *parser;
	list_type::iterator it;
	
	errorf(errSuccess, "Success");
	
	/* If the module is already (or is going to be) loaded,
	 * don't try to do it twice! */
	if (find(name, 0))
		return true;
	
	/* Compute the module path */
	
	std::string modpath = misc::absolute_path(LIBDIR, name);
		
	/* Load dependencies */
	
	if (!load_deps(modpath))
		return false;
	
	/* Open the module */
	lt_dlhandle h = lt_dlopenext(modpath.c_str());
	
	if (!h)
	{
		errorf(errOpen, "While loading %s: libltdl error: %s", name.c_str(),
				lt_dlerror());
		return false;
	}
	
	/* Ensure we have all the symbols we might need. */
	module_get_t get = (module_get_t) lt_dlsym(h,
			path_to_func(name.c_str(), "get"));
	module_release_t release = (module_release_t) lt_dlsym(h,
			path_to_func(name.c_str(), "release"));
	
	if (!get || !release)
	{
		errorf(errLookup, "While loading %s: libltdl error: %s", name.c_str(),
				lt_dlerror());
		goto closelib;
	}
	
	/* Retrieve the module */
	m = get(*m_core);
	
	if (!m)
	{
		errorf(errUnavailable, "While loading %s: module is unavailable",
				name.c_str());
		goto closelib;
	}
	
	/* Save the handle in the module */
	m->data(h);
	
	/* Configure the module */
	parser = m->config(*m_core, *m_global);
	
	if (parser && !parser->parse_string(*m_global, false,
			confblock(name), name))
	{
		errorf(errConfig, "While loading %s: configuration error",
				m->name().c_str());
		goto release;
	}
	
	/* Add the module to the list now so that it doesn't get loaded
	 * twice. But we'll have to remove it later and reinsert it so
	 * save its position. */
	it = m_modules.insert(m_modules.end(), m);
	
	/* Load the module */
	if (!m->load(*m_core))
	{
		errorf(errLoad, "While loading %s: load error", m->name().c_str());
		goto listcleanup;
	}
	
	/* Now remove the module from the list and reinsert it. */
	m_modules.erase(it);
	m_modules.push_back(m);
	
	/* Broadcast the loading of the module to all *already loaded* modules. */
	broadcast(Module::msLoaded, *m, Module::mnLoaded, 0);
	
	/* Say to the module which modules are already *loaded*. */
	feed(*m, Module::msLoaded, Module::mnLoaded, 0);
	
	/* Log the load */
	m_core->log(LogModule(), "Loaded module: %s", m->name().c_str());
	
	/* If we're started, start the module */
	if (m_started)
		start(*m);
	
	return true;

listcleanup:
	m_modules.erase(it);
	goto release;

release:
	release(*m_core, m);
	goto closelib;

closelib:
	lt_dlclose(h);
	return false;
}

/*************************************************************************/

bool ModManager::load(LoadModule const &lm)
{
	LoadModule::container_type::const_iterator it, end;
	
	for (it = lm.value().begin(), end = lm.value().end(); it != end; ++it)
	{
		if (!load(it->value()))
			return false;
	}
	
	return true;
}

/*************************************************************************/

bool ModManager::load_deps(std::string const &path)
{
	ifstream in(std::string(path + ".deps").c_str());
	
	/* If we can't open the file, it's a GOOD thing. */
	if (!in)
		return true;
	
	std::string line;
	
	while (getline(in, line))
		if (!line.empty() && !isspace(line[0]) && !load(line))
			return false;
	
	return true;
}

/*************************************************************************/

bool ModManager::reload()
{
	ConfQ *parser;
	list_type::iterator it, end;
	
	for (it = m_modules.begin(), end = m_modules.end(); it != end; ++it)
	{
		if ((parser = (*it)->config(*m_core, *m_global)))
		{
			if (!parser->parse_string(*m_global, true,
					confblock((*it)->name()), (*it)->name()))
				return false;
		}
	}
	
	return true;
}

/*************************************************************************/

void ModManager::start()
{
	m_started = true;
	
	list_type::iterator it, end;
	
	for (it = m_modules.begin(), end = m_modules.end(); it != end; ++it)
	{
		/* If it is loaded but NOT already started */
		if (((*it)->state() & Module::msLoaded) &&
				!((*it)->state() & Module::msStarted))
			start(**it);
	}
}

/*************************************************************************/

void ModManager::start(Module &module)
{
	/* Start the module */
	module.start(*m_core);
	
	/* Broadcast the loading of the module to all *already started* modules. */
	broadcast(Module::msStarted, module, Module::mnStarted, 0);
	
	/* Say to the module which modules are already *started*. */
	feed(module, Module::msStarted, Module::mnStarted, 0);
	
	/* Log what we did */
	m_core->log(LogModule(), "Started module: %s", module.name().c_str());
}

/*************************************************************************/

void ModManager::stop()
{
	m_started = false;
	
	list_type::reverse_iterator it, end;
	
	for (it = m_modules.rbegin(), end = m_modules.rend(); it != end; ++it)
	{
		if ((*it)->state() & Module::msStarted)
		{
			(*it)->stop(*m_core);
			m_core->log(LogModule(), "Stopped module: %s",
					(*it)->name().c_str());
		}
	}
}

/*************************************************************************/

bool ModManager::unload(Module &module)
{
	errorf(errSuccess, "Success");
	
	/* Don't unload the core! */
	if (&module == m_core)
	{
		errorf(errCore, "While unloading %s: can't unload the core!",
				module.name().c_str());
		return false;
	}
	
	/* Don't unload modules that have an non-zero reference count. */
	if (module.references() != 0)
	{
		errorf(errNonZeroRefs,
				"While unloading %s: the module is still referenced (%d)",
				module.name().c_str(), module.references());
		return false;
	}
	
	/* If the module is started... */
	if (module.state() & Module::msStarted)
	{
		/* ... stop it first */
		if (!module.stop(*m_core))
		{
			errorf(errStop, "While unloading %s: the module can't be stopped",
					module.name().c_str());
			return false;
		}
		
		/* Tell the module all other modules have stopped (not true,
		 * but it won't know :P) */
		feed_reverse(module, Module::msStarted, Module::mnStopped, 0);
		
		/* Tell other modules this module has stopped */
		broadcast_reverse(Module::msStarted, module, Module::mnStopped, 0);
		
		/* Log what we did */
		m_core->log(LogModule(), "Stopped module: %s", module.name().c_str());
	}
	
	/* Retrieve the module handle */
	lt_dlhandle h = (lt_dlhandle) module.data();
	
	/* Unload the module */
	module.unload(*m_core);
	
	/* Remove it from the list */
	m_modules.remove(&module);
	
	/* Find the release function */
	module_release_t release = (module_release_t) lt_dlsym(h,
			path_to_func(module.name().c_str(), "release"));
	
	/* We already ensured that the function exists before. If it doesn't
	 * exist anymore, then something really wrong has happened (maybe
	 * the handle is invalid, maybe the system call failed, whatever..),
	 * so we report a fatal error. */
	if (!release)
		m_core->fatalf("While unloading %s: libltdl error: %s",
				module.name().c_str(), lt_dlerror());
	
	/* Log the unloading */
	m_core->log(LogModule(), "Unloaded module: %s", module.name().c_str());
	
	/* Call the release function */
	release(*m_core, &module);
	
	/* Finally, close the handle (don't care about the return value) */
	lt_dlclose(h);
	
	return true;
}

/*************************************************************************/

} /* namespace modules */

/*************************************************************************/
