/* Gotta load 'em all.
 *
 * PegSoft modules subsystem library (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

/** 
 * \file modules/manager.h
 * \brief Module manager header file.
 */

#ifndef MODULES_MANAGER_H
#define MODULES_MANAGER_H

/*************************************************************************/

#include <list>
#include <string>

/*************************************************************************/

namespace conf { class ConfGlobalQ; }

/*************************************************************************/

namespace modules
{

/*************************************************************************/

class Module;

class Core;

class ConfModuleA;
class LoadModule;

/*************************************************************************/

/// Module manager

/** The module manager is responsible to load, unload and maintain a list
  * of modules used by the program. */

class ModManager
{
	public:
		/// Type of the list of modules.
		typedef std::list<Module *> list_type;
	
		/// Constructor.
		/** Constructor.
		  *
		  * \param[in] core The core module (must be valid until the
		  * module manager is destroyed).
		  * \param[in] confmodule The conf-module directive (must be
		  * valid until the module manager is destroyed). */
		ModManager(Core &core, ConfModuleA const &confmodule);
		/// Destructor.
		~ModManager();
		
		/// Sends a notification message to all modules in load order.
		/** Calls the Module::notify() function for all modules in a given
		  * state, in the order they were loaded, using the given
		  * parameters. 
		  *
		  * \param[in] state An or'ed set of states the module must have
		  * to receive the notification (the msLoaded state is implied). 
		  * \param[in] sender The module sending the notification.
		  * This module, which can be NULL, won't get the notification.
		  * \param[in] msg The notification message.
		  * \param[in] arg An optional parameter that may be used 
		  * to pass additional data.
		  *
		  * \sa Module::notify() */
		void broadcast(int state, Module &sender, int msg, void *arg);
		
		/// Sends a notification message to all modules in reverse load order.
		/** Calls the Module::notify() function for all modules in a given
		  * state, in the reverse order they were loaded, using the given
		  * parameters. 
		  *
		  * \param[in] state An or'ed set of states the module must have
		  * to receive the notification (the msLoaded state is implied). 
		  * \param[in] sender The module sending the notification.
		  * This module, which can be NULL, won't get the notification.
		  * \param[in] msg The notification message.
		  * \param[in] arg An optional parameter that may be used 
		  * to pass additional data.
		  *
		  * \sa Module::notify() */
		void broadcast_reverse(int state, Module &sender, int msg, void *arg);
		
		/// Finds a loaded module by name.
		/** Finds a loaded module with the given name.
		  * \param[in] name The name of the module to find (case-sensitive).
		  * \return A pointer to the matching Module object, or 0
		  * if no module by this name is loaded. */
		Module *find(std::string const &name) const;
		
		/// Finds a module by name and state.
		/** Finds a module by name and state.
		  * \param[in] name The name of the module to find (case-sensitive).
		  * \param[in] state An or'ed set of states the module must have
		  * to be found.
		  * \return A pointer to the matching Module object, or 0
		  * if no module by this name and state is found. */
		Module *find(std::string const &name, int state) const;
			
		/// Finds an exposed interface in one of the loaded modules.
		/** Finds an exposed interface in one of the loaded (possibly not started)
		  * modules. This calls Module::interface() on all loaded modules until it
		  * finds it.
		  *
		  * \param[in] iid The interface identifier
		  * \param[out] from If non-NULL, will point to the module that exposes
		  * the interface on return.
		  * \return A pointer to the interface returned by a module, or NULL if
		  * no modules expose that interface.
		  */
		Interface *find_interface(PSIID iid, Module **from = NULL) const;
		
		/// Loads a module.
		/** Loads a module. 
		  *
		  * This method first looks for a .deps file in the module
		  * path and will load each module listed in it (one per line).
		  * If the file doesn't exist, no errors are returned.
		  *
		  * A successful load does NOT guarantee that the module
		  * is in the Module::msLoaded state yet. Other modules
		  * calling this function should wait to get the
		  * Module::mnLoaded notification before using the module.
		  *
		  * The module is started by this function as well, but only
		  * if the start() function has already been called by the
		  * core.
		  *
		  * \param[in] name The module name.
		  * \return true if the module has been loaded successfully,
		  * false otherwise. Modules can only be loaded once; all
		  * subsequent loading attempts will do nothing but will
		  * still return true. */
		bool load(std::string const &name);
		
		/// Loads modules as specified by the load-module directive.
		/** Loads modules listed through the given load-module directive.
		  * 
		  * \param[in] lm The load-module directive.
		  * \return true if all modules have been loaded successfully,
		  * false otherwise.
		  */
		bool load(LoadModule const &lm);
		
		/// Reloads the configuration for all modules.
		/** Reloads the configuration for all modules.
		  *
		  * This function is for use by the core module only. */
		bool reload();
		
		/// Starts all modules.
		/** Starts all modules, and sets the module manager to the
		  * started state. 
		  *
		  * This function is for use by the core module only.
		  *
		  * \sa started() */
		void start();
		
		/// Stops all modules.
		/** Stops all modules. 
		  *
		  * This function is for use by the core module only. */
		void stop();
		
		/// Unloads a module.
		/** Unloads a module. The module is first stopped if it was
		  * started. If it fails, then the module is NOT unloaded. 
		  *
		  * A module can't be unloaded if its reference count is
		  * non-zero and you can't unload the core module. Attempting
		  * to do so will result in a failure.
		  *
		  * \param[in] module The module to be unloaded.
		  * \return true on success, false otherwise. */
		bool unload(Module &module);
		
		/// Returns the core.
		Core &core() const { return *m_core; }
		
		/// Returns a description of the last error that occured.
		/** Returns a description of the last error that occured,
		  * after a failed call to load() or unload().
		  *
		  * \return The last error.
		  * \sa load(std::string const &), unload() */
		std::string const &error() const { return m_error; }
		
		/// Possible error codes for the error_code() function.
		enum
		{
			/// No error.
			errSuccess = 0,
			/// Library open error.
			errOpen = 1,
			/// Library lookup error (can't find *_get or *_release).
			errLookup = 2,
			/// Module unavailable (*_get failed).
			errUnavailable = 3,
			/// Module configuration error.
			errConfig = 4,
			/// Module::load() failed.
			errLoad = 5,
			/// Forbidden operation on the core module.
			errCore = 6,
			/// Reference count is non-zero.
			errNonZeroRefs = 7,
			/// Can't stop module.
			errStop = 8
		};
		
		/// Returns the code of the last error that occured.
		/** Returns the code of the last error that occured, after
		  * a failed call to load() or unload().
		  *
		  * \return The last error code.
		  * \sa load(std::string const &), unload() */
		int error_code() const { return m_errorcode; }
		
		/// Sets the ConfGlobalQ object used to configure modules.
		/** Sets the ConfGlobalQ object used to configure modules.
		  *
		  * This function is for use by the core module only. */
		void global(conf::ConfGlobalQ &global) { m_global = &global; }
		
		/// Returns the list of loaded modules.
		/** Returns the list of loaded modules. The core is NOT in this
		  * list.
		  * \return The list of loaded modules. */
		list_type const &modules() const { return m_modules; }
		
		/// Returns whether the modules manager is started.
		bool started() const { return m_started; }
	private:
		list_type m_modules;
	
		Core *m_core;
		
		std::string m_error;
		int m_errorcode;
		
		ConfModuleA const *m_confmodule;
		conf::ConfGlobalQ *m_global;
		
		bool m_started;
	
		ModManager(ModManager const &);
		ModManager &operator=(ModManager const &);
		
		std::string confblock(std::string const &name);
		void errorf(int code, const char *fmt, ...) FORMAT(printf, 3, 4);
		void feed(Module &recipient, int state, int msg, void *arg);
		void feed_reverse(Module &recipient, int state, int msg, void *arg);
		bool load_deps(std::string const &path);
		void start(Module &module);
};

/*************************************************************************/

} /* namespace modules */

/*************************************************************************/

#endif /* MODULES_MANAGER_H */
