/* select()-based monitor class.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <algorithm>

#include "select.h"

using namespace std;

/*************************************************************************/

namespace sockets
{

/*************************************************************************/

Select::Select(size_t hint) : Monitor("select", hint), m_rfds(), m_wfds(),
		m_rds(), m_wds()
{
	FD_ZERO(&m_rds);
	FD_ZERO(&m_wds);
}

/*************************************************************************/

bool Select::add(Socket *s, int fd, unsigned int events)
{
	if (fd >= FD_SETSIZE)
		return false;
	
	if (!Monitor::add(s, fd, events))
		return false;
	
	if (events & evRead)
	{
		FD_SET(fd, &m_rds);
		m_rfds.insert(fd);
	}
	
	if (events & evWrite)
	{
		FD_SET(fd, &m_wds);
		m_wfds.insert(fd);
	}
	
	return true;
}

/*************************************************************************/

void Select::do_events(time_t wait)
{
	int nfds = last_fd() + 1;
	
	if (nfds == 0)
		return;
	
	fd_set rds, wds;
	memcpy(&rds, &m_rds, sizeof(fd_set));
	memcpy(&wds, &m_wds, sizeof(fd_set));
	
	struct timeval tv;
	
	tv.tv_sec = wait / 1000;
	wait -= (tv.tv_sec * 1000);
	tv.tv_usec = wait * 1000;
	
	int ret;
	
	while ((ret = select(nfds, (m_rfds.empty() ? 0 : &rds),
			(m_wfds.empty() ? 0 : &wds), 0, &tv)) == -1)
	{
		if (errno == EINTR)
		{
			memcpy(&rds, &m_rds, sizeof(fd_set));
			memcpy(&wds, &m_wds, sizeof(fd_set));
			
			tv.tv_sec = tv.tv_usec = 0;
			continue;
		}
		
		/* A bad file descriptor means our program is flawed. */
		if (errno == EBADF)
			abort();
		
		return;
	}
	
	/* Do we need to bother ? */
	if (ret == 0)
		return;
	
	/* Now, send the event messages. */
	for (int i = 0; i < nfds && ret; ++i)
	{
		/* Do the write event first so the sockets can be set
		 * to connected state if read_event() is then called. */
		if (FD_ISSET(i, &wds))
		{
			--ret;

			Socket *s = get_socket(i);

			if (s && !s->write_event())
				continue;
		}
		
		if (FD_ISSET(i, &rds))
		{
			--ret;
			
			Socket *s = get_socket(i);
			if (s)
				s->read_event();
		}
	}
}

/*************************************************************************/

int Select::last_fd() const
{
	if (m_rfds.empty())
	{
		if (m_wfds.empty())
			return -1;
		
		return *m_wfds.rbegin();
	}
	
	if (m_wfds.empty())
		return *m_rfds.rbegin();
	
	int last_rfd = *m_rfds.rbegin(), last_wfd = *m_wfds.rbegin();
	
	return max(last_rfd, last_wfd);
}

/*************************************************************************/

void Select::remove(Socket *s, int fd)
{
	FD_CLR(fd, &m_rds);
	m_rfds.erase(fd);
	
	FD_CLR(fd, &m_wds);
	m_wfds.erase(fd);
	
	Monitor::remove(s, fd);
}

/*************************************************************************/

bool Select::update(Socket *s, int fd, unsigned int events)
{
	if (!Monitor::update(s, fd, events))
		return false;
	
	if ((events & evRead) && !FD_ISSET(fd, &m_rds))
	{
		FD_SET(fd, &m_rds);
		m_rfds.insert(fd);
	}
	else if (!(events & evRead) && FD_ISSET(fd, &m_rds))
	{
		FD_CLR(fd, &m_rds);
		m_rfds.erase(fd);
	}
	
	if ((events & evWrite) && !FD_ISSET(fd, &m_wds))
	{
		FD_SET(fd, &m_wds);
		m_wfds.insert(fd);
	}
	else if (!(events & evWrite) && FD_ISSET(fd, &m_wds))
	{
		FD_CLR(fd, &m_wds);
		m_wfds.erase(fd);
	}
	
	return true;
}

/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/
