/* Sockets base class.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include "sockets/monitor.h"
#include "sockets/socket.h"
#include "sockets/ssl.h"

using namespace std;

/*************************************************************************/

namespace
{

/*************************************************************************/

std::string local_hostname;

/*************************************************************************/

} /* anonymous namespace */

/*************************************************************************/

namespace sockets
{

/*************************************************************************/

Socket::Socket(int family, int type, int protocol) : m_fd(-1),
		m_family(family), m_type(type), m_protocol(protocol), m_status(0),
		m_monitor(default_monitor) 
{ 
	
}

/*************************************************************************/

Socket::~Socket()
{
	close();
}

/*************************************************************************/

Socket &Socket::operator=(Socket const &right)
{
	if (this != &right)
	{
		if (status() & ssCreated)
			close();
		
		m_fd = -1;
		m_family = right.m_family;
		m_type = right.m_type;
		m_protocol = right.m_protocol;
		m_status = (right.m_status & ssCopyFlags);
		m_monitor = right.m_monitor;
	}
	
	return *this;
}

/*************************************************************************/

bool Socket::assign(int fd, uint32_t status)
{
	if (this->status() & ssCreated)
		return true;
	
	if (fd < 0)
		return false;
	
	m_fd = fd;
	status_add(ssCreated | status);
	
	if (/*!(status & ssMonitored) &&*/ !watch())
		return false;
	
	return true;
}

/*************************************************************************/

bool Socket::close()
{
	if (!(status() & ssCreated))
		return true;
	
	unwatch();
	
	while (::close(fd()) == -1)
	{
		/* Don't consider this as an error */
		if (errno == EINTR)
			continue;
		
		/* Don't take other errors too seriously; they are
		 * unlikely on sockets and absolutely not fatal. */
	}
	
	m_fd = -1;
	status_remove(ssCloseFlags);
	
	return true;
}

/*************************************************************************/

bool Socket::create()
{
	if (status() & ssCreated)
		return true;
	
	int fd = socket(m_family, m_type, m_protocol);
	
	if (fd == -1)
		return false;
	
	m_fd = fd;
	status_add(ssCreated);
	
	return true;
}

/*************************************************************************/

int Socket::family(int family)
{
	if (status() & ssCreated)
		return status();
	
	return (m_family = family);
}

/*************************************************************************/

char const *Socket::hostname()
{
	if (local_hostname.empty())
	{
		char host[256];
		
		if (gethostname(host, sizeof(host)) == -1)
		{
			local_hostname = "localhost";
		} else
		{
			host[255] = 0;
			if (*host)
				local_hostname = host;
			else
				local_hostname = "localhost";
		}
	}
	
	return local_hostname.c_str();
}

/*************************************************************************/

void Socket::init(size_t hint)
{
	SSLContext::init();
	setup_monitors(hint);
}

/*************************************************************************/

std::string const &Socket::monitor() const
{
	return m_monitor->name();
}

/*************************************************************************/

bool Socket::monitor(std::string const &name)
{
	if (status() & ssMonitored)
		return true;
	
	monitors_list::const_iterator it, end;
	
	for (it = monitors.begin(), end = monitors.end(); it != end; ++it)
	{
		if ((*it)->name() == name)
		{
			m_monitor = (*it);
			return true;
		}
	}
	
	return false;
}


/*************************************************************************/

int Socket::protocol(int protocol)
{
	if (status() & ssCreated)
		return status();
	
	return (m_protocol = protocol);
}

/*************************************************************************/

int Socket::type(int type)
{
	if (status() & ssCreated)
		return status();
	
	return (m_type = type);
}

/*************************************************************************/

void Socket::unwatch()
{
	if ((status() & ssCreated) && (status() & ssMonitored))
	{
		m_monitor->remove(this, fd());
		status_remove(ssMonitored);
	}
}

/*************************************************************************/

bool Socket::watch()
{
	if (!(status() & ssCreated))
		return false;
	
	unsigned int events = 0;
	
	if (!(status() & ssReadable))
		events |= Monitor::evRead;
	if (!(status() & ssWritable))
		events |= Monitor::evWrite;
	
	if (status() & ssMonitored)
		return m_monitor->update(this, fd(), events);
	
	if (!m_monitor->add(this, fd(), events))
		return false;
	
	status_add(ssMonitored);
	return true;
}

/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/
