/* Buffers base class.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file sockets/buffer.h
  * \brief Buffers base class header file. */

#ifndef SOCKETS_BUFFER_H
#define SOCKETS_BUFFER_H

/*************************************************************************/

#include <iterator>

/*************************************************************************/

namespace sockets
{

/*************************************************************************/

/// Base class for buffers.

/** Base class for buffers. Buffers are used by sockets to store and
  * retrieve bytes read from the socket or bytes that need to be written
  * to the socket. */

class Buffer
{
	public:
		/* Typedefs */
		/// Value type.
		typedef char value_type;
		/// Reference type
		typedef char & reference;
		/// Const reference type
		typedef char const & const_reference;
		/// Pointer type
		typedef char * pointer;
		/// Const pointer type
		typedef char const * const_pointer;
		/// Size type.
		typedef size_t size_type;
		/// Difference type.
		typedef ptrdiff_t difference_type;
		
		/* Classes */
		
		/// The buffer iterator class.
		/** This class implements the buffer iterator. Buffer iterators
		  * are always constant.
		  *
		  * Note that buffer iterators become invalid when data is added
		  * or removed from the buffer. */
		class Iterator;
		friend class Iterator;
		class Iterator : public std::iterator<std::bidirectional_iterator_tag,
				Buffer::value_type, Buffer::difference_type,
				Buffer::const_pointer, Buffer::const_reference>
		{	
			public:
				/* Constants */
				
				/// Possible values for the constructor state parameter.
				enum
				{
					/// An iterator suitable for the begin() function.
					isBegin = 0,
					/// An iterator suitable for the end() function.
					isEnd = 1
				};
				
				/* Functions */
			
				/// Default constructor.
				Iterator() : m_buffer(0), m_idx(0), m_ch(0) { }
				/// Constructor setting the buffer and state (isBegin or isEnd).
				Iterator(Buffer const *buffer, int state);
				/// Constructor setting the buffer and index.
				Iterator(Buffer const *buffer, Buffer::size_type idx);
				/// Constructor setting the buffer, index and character.
				Iterator(Buffer const *buffer, Buffer::size_type idx, char ch) :
						m_buffer(buffer), m_idx(idx), m_ch(ch) { }
				/// Copy constructor.
				Iterator(Iterator const &it) : m_buffer(it.m_buffer),
						m_idx(it.m_idx), m_ch(it.m_ch) { }
				/// Assignment operator.
				Iterator &operator=(Iterator const &right);
				
				/// * unary operator.
				reference operator*() const { return m_ch; }
				/// -> operator.
				pointer operator->() const { return &m_ch; }
				
				/// == operator
				bool operator==(Iterator const &right) const;
				/// != operator
				bool operator!=(Iterator const &right) const
						{ return !(*this == right); }
				
				/// ++ operator (prefix)
				Iterator &operator++();
				/// ++ operator (suffix)
				Iterator operator++(int);
				
				/// -- operator (prefix)
				Iterator &operator--();
				/// -- operator (suffix)
				Iterator operator--(int);
				
				/// - operator(iter1, iter2)
				difference_type operator-(Iterator const &right) const
						{ return (m_idx - right.m_idx); }
			private:
				Buffer const *m_buffer;
				Buffer::size_type m_idx;
				value_type m_ch;
		};
		
		/* Typedefs */
		
		/// Normal iterator type.
		typedef Iterator iterator;
		/// Const iterator type.
		typedef iterator const_iterator;
		/// Reverse iterator
		typedef std::reverse_iterator<iterator> reverse_iterator;
		/// Const reverse iterator
		typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
		
		/* Functions */
		
		/// Destructor.
		virtual ~Buffer() { }
		
		/// Reads data from the buffer.
		/** Reads data from the buffer. This function does not
		  * copy data from the buffer; instead, it returns a pointer
		  * to the memory chunk the buffer uses to store its data.
		  *
		  * This function does NOT erase data stored in the buffer,
		  * and consecutives calls to read() will return the same
		  * sequence of bytes. See erase() to discard data.
		  *
		  * \param[in] max The maximum length of data, in bytes, that
		  * need to be retrieved, or zero if no limit is wanted.
		  * \param[out] data A pointer to the pointer that will be
		  * set to the beginning of the data stored in the buffer.
		  * \param[out] len A pointer to an integer that will be set
		  * to the length of the data that has been returned. Please
		  * note that this may be less than max and less than the
		  * actual number of bytes stored in the buffer as returned
		  * by the size() function.
		  * \return true if successful, false otherwise. Possible
		  * reasons of failure include an empty buffer. */
		virtual bool read(size_t max,
				char const **data, size_t *len) const = 0;
		
		/// Writes data in the buffer.
		/** Writes data in the buffer. How the data is stored
		  * internally is up to the implementer.
		  *
		  * \param[in] data The data to write. This does NOT have
		  * to be a NULL-terminated string.
		  * \param[in] len The length of the data, in bytes. The
		  * data is guaranteed to be written entirely to the buffer;
		  * if there's no more room available in the buffer, the
		  * function fails.
		  * \return true if successful, false otherwise. */
		virtual bool write(char const *data, size_t len) = 0;
		
		/// Erases a portion of the data in the buffer.
		/** Erases a portion of the data in the buffer. This
		  * portion always starts at the first byte stored in the
		  * buffer and spans the given number of bytes. The
		  * first byte of hte buffer (as returned by the read()
		  * function) will then be the byte immediately
		  * after the last character erased, if any.
		  *
		  * \param[in] len The number of bytes to erase. That
		  * can't be more than the size of the buffer (as returned
		  * by the size() function).
		  * \return true if successful, false otherwise. */
		virtual bool erase(size_t len) = 0;
		
		/// Same as erase(size_t) but with an iterator to mark the first character that is not erased.
		virtual bool erase(iterator end) { return erase(end - begin()); }
		
		/// Clears all data in the buffer.
		/** Clears all data in the buffer.
		  *
		  * \return true if successful, false otherwise. */
		virtual bool clear() = 0;
		
		/// Returns an iterator for the beginning of the buffer.
		iterator begin() { return iterator(this, iterator::isBegin); }
		/// Returns an iterator to the end of the buffer.
		iterator end() { return iterator(this, iterator::isEnd); }
		/// Returns a const iterator for the beginning of the buffer.
		const_iterator begin() const 
				{ return const_iterator(this, const_iterator::isBegin); }
		/// Returns a const iterator to the end of the buffer.
		const_iterator end() const 
				{ return const_iterator(this, const_iterator::isEnd); }
		
		/// Returns a reverse iterator to the beginning of the buffer.
		reverse_iterator rbegin() { return reverse_iterator(end()); }
		/// Returns a reverse iterator to the end of the buffer.
		reverse_iterator rend() { return reverse_iterator(begin()); }
		/// Returns a const reverse iterator to the beginning of the buffer.
		const_reverse_iterator rbegin() const
				{ return const_reverse_iterator(end()); }
		/// Returns a const reverse iterator to the end of the buffer.
		const_reverse_iterator rend() const
				{ return const_reverse_iterator(begin()); }
		
		/// Returns whether the buffer is empty.
		virtual bool empty() const = 0;
		/// Returns the number of bytes stored in the buffer.
		virtual size_type size() const = 0;
		/// Returns the maximum number of bytes that can be stored in the buffer.
		virtual size_type max_size() const = 0;
		/// Returns the number of bytes that can still be written to the buffer.
		virtual size_type remaining() const { return max_size() - size(); }
	protected:
		/// Reads a specific character in the buffer.
		/** Reads a specific character in the buffer.
		  *
		  * \param[in] pos The position of the character to read
		  * (0 is the position of the first character).
		  * \param[out] ch On return, this will be set to the
		  * character that was requested.
		  * \return true if successful, false otherwise. */
		virtual bool read_byte(size_type pos, char *ch) const = 0;
	private:
};

/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/

#endif /* SOCKETS_BUFFER_H */
