/* Monitors base class.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file sockets/monitor.h
  * \brief Monitors base class header file. */

#ifndef SOCKETS_MONITOR_H
#define SOCKETS_MONITOR_H

/*************************************************************************/

#include <string>

#include <containers/vectorsp.h>

/*************************************************************************/

namespace sockets
{

/*************************************************************************/

class Socket;

/*************************************************************************/

/// Base class for monitors.

/** Base class for monitors. A monitor is used to watch requested events
  * on sockets registered with it.
  *
  * The do_events(), add(), update() and remove() function must be
  * overriden in derived classes. */

class Monitor
{
	public:
		/* Constants */
		enum
		{
			/// Read event
			evRead = 0x0001,
			/// Write event
			evWrite = 0x0002
		};
		
		/* Functions */
		
		/// Constructor.
		/** Constructor.
		  *
		  * \param[in] name The name of the monitor.
		  * \param[in] hint An estimate of the number of maximum sockets
		  * that will be in use at a given time. This is NOT an actual
		  * maximum value. */
		Monitor(std::string const &name, size_t hint);
			
		/// Destructor.
		virtual ~Monitor() { }
		
		/// Do events.
		/** This function checks for new events on the monitored sockets.
		  *
		  * The default implementation does nothing. 
		  *
		  * \param[in] wait The number of milliseconds to wait for events on
          * sockets, or 0 if no waiting must be done.
		  */
		virtual void do_events(time_t wait) { }
			
		/// Monitors a socket
		/** Adds a socket to the monitor. This socket will be
		  * watched for the given events and receive
		  * notifications when they happen.
		  *
		  * The default implementation calls set_socket() for
		  * the given socket/fd pair.
		  *
		  * \param[in] socket The socket to add.
		  * \param[in] fd The socket file descriptor.
		  * \param[in] events The events to watch for initially.
		  * \return true if successful, false otherwise. */
		virtual bool add(Socket *socket, int fd, unsigned int events);
		
		/// Updates events monitored for a socket.
		/** Updates events monitored for a socket. The socket must
		  * have been added previously through the add() function.
		  *
		  * The default implementation does nothing.
 		  *
		  * \param[in] socket The socket to update.
		  * \param[in] fd The socket file descriptor.
		  * \param[in] events The events to watch for.
		  * \return true if successful, false otherwise. */
		virtual bool update(Socket *socket, int fd, unsigned int events);
		
		/// Unmonitors a socket.
		/** Unmonitors a socket. The socket must have been added
		  * previously through the add() function.
		  *
		  * The default implementation removes the socket from
		  * the internal socket list.
		  *
		  * \param[in] socket The socket to remove.
		  * \param[in] fd The socket file descriptor. */
		virtual void remove(Socket *socket, int fd);
		
		/// Returns the name of the monitor.
		std::string const &name() const { return m_name; }
	protected:
		/// Type of the list of sockets.
		typedef containers::VectorSP<Socket *, size_t> sockets_list;
		
		/// The sockets list.
		sockets_list m_sockets;
		
		/// Gets a socket for a given fd.
		Socket *get_socket(int fd) const;
		/// Sets a socket for a given fd.
		void set_socket(int fd, Socket *s);
		
		/// Returns the hint.
		size_t hint() const { return m_hint; }
	private:
		std::string m_name;
		size_t m_hint;
		
		Monitor(Monitor const &);
		Monitor &operator=(Monitor const &);
};

/*************************************************************************/

typedef containers::VectorSP<Monitor *, size_t> monitors_list;

extern monitors_list monitors;
extern Monitor *default_monitor;

/*************************************************************************/

/// Sets monitors up.
/** This function, which must be called by the core at program startup,
  * registers and configures all usable monitors.
  *
  * \param[in] hint The number of sockets that is estimated to
  * be in concurrent use at a given time. This is only a hint, not
  * the maximum number that will be allowed. */
extern void setup_monitors(size_t hint);

/// Does events.
/** This function, which must be called by the core in its main loop,
  * checks for events on all sockets that need to be monitored.
  *
  * \param[in] wait The number of milliseconds to wait for events on
  * sockets, or 0 if no waiting must be done. 
  */
extern void do_events(time_t wait);

/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/

#endif /* SOCKETS_MONITOR_H */
