/* Sockets base class.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file sockets/socket.h
  * \brief Sockets base class header file. */

#ifndef SOCKETS_SOCKET_H
#define SOCKETS_SOCKET_H

/*************************************************************************/

#if HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif

#if HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#if HAVE_SYS_UN_H
# include <sys/un.h>
#endif

#include <string>

/*************************************************************************/

/// Namespace of the sockets classes.

namespace sockets
{

/*************************************************************************/

class Monitor;

/*************************************************************************/

/// Base class for all sockets.

/** This is the base class for all sockets class. It must not be
  * instantiated directly; rather, use a derived class or create
  * your own. 
  *
  * It handles socket creation and destruction, and provides 
  * means to track socket status. */

class Socket
{
	public:
		/* Friend classes */
		
		/// Add SSLInfo as a friend
		friend class SSLInfo;
		
		/* Constants */
		
		/// Possible flags for socket status
		enum
		{
			/// The socket has been created.
			ssCreated = 0x00000001,
			/// Local info has been set
			ssLocalInfo = 0x00000002,
			/// Remote info has been set
			ssRemoteInfo = 0x00000004,
			/// Socket is bound
			ssBound = 0x00000008,
			/// Connection in progress
			ssConnecting = 0x00000010,
			/// Socket is connected
			ssConnected = 0x00000020,
			/// Socket is listening
			ssListening = 0x00000040,
			/// Socket is monitored
			ssMonitored = 0x00000080,
			/// Socket is readable
			ssReadable = 0x00000100,
			/// Socket is writable
			ssWritable = 0x00000200,
			/// SSL support requested
			ssSSL = 0x00000400,
			/// SSL handshake has been done
			ssSSLHandshake = 0x00000800,
			
			/// Flags removed by the close() function.
			ssCloseFlags = (ssCreated | ssBound | ssConnecting | ssConnected |
					ssListening | ssMonitored | ssReadable | ssWritable |
					ssSSLHandshake),
			/// Flags that can be copied
			ssCopyFlags = (ssLocalInfo | ssRemoteInfo | ssSSL)
		};
		
		/* Static functions */
		
		/// Initializes the sockets library.
		/** Initializes the sockets library.
		  *
		  * This calls the setup_monitors() function and does any other required
		  * initialization.
		  *
		  * \param[in] hint A hint passed to the setup_monitors() function.
		  */ 
		static void init(size_t hint);
		
		/// Returns the hostname of the current processor.
		static char const *hostname();
		
		/* Functions */
		
		/// Constructor.
		/** Constructor.
		  *
		  * \param[in] family The socket protocol family (one of the
		  * PF_... constants defined in the system header files). May
		  * be set to PF_UNSPEC if you want to set it later.
		  * \param[in] type The socket type (one of the SOCK_...
		  * constants defined in the system header files). May be set to
		  * 0 if you want to set it later.
		  * \param[in] protocol The socket protocol that needs to be
		  * used. This is seldom used, as there is generally only one
		  * protocol for a given family/type combination; use the value
		  * 0 to use the default protocol. */
		explicit Socket(int family = PF_UNSPEC, int type = 0,
				int protocol = 0);
		/// Copy constructor.
		/** This copy constructor copies the socket family, type, protocol,
		  * monitor and part of the status only.
		  *
		  * \param[in] s The socket to copy. */
		Socket(Socket const &s) : m_fd(-1), m_family(s.m_family),
				m_type(s.m_type), m_protocol(s.m_protocol),
				m_status(s.m_status & ssCopyFlags), m_monitor(s.m_monitor) { }
		/// Assignment operator.
		/** This assignment operator copies the socket family, type, protocol,
		  * monitor and part of the status only.
		  *
		  * \param[in] right The socket to be assigned. */
		virtual Socket &operator=(Socket const &right);
		/// Destructor.
		/** This destructor calls the close() function. */
		virtual ~Socket();
		
		/// Assigns a socket file descriptor.
		/** Assigns a socket file descriptor to the socket object,
		  * that is then used by the object for all operations on
		  * the socket.
		  *
		  * The ssCreated status flag must not be set prior to calling
		  * this function; it is set once the function returns
		  * successfully.
		  *
		  * \param[in] fd The socket file descriptor to assign.
		  * \param[in] status The status flags that are added to
		  * the socket. If the ssMonitored flag is among them,
		  * the watch() function is automatically called.
		  * \returns true if successful, false otherwise. errno
		  * is set accordingly in the latter case. */
		virtual bool assign(int fd, uint32_t status);
		
		/// Read event -- For call by the Monitor-derived classes.
		/** This function is called by the Monitor-derived classes
		  * when a read event happens.
		  *
		  * \return false if the socket has been closed or is no longer
		  * valid, true otherwise. */
		virtual bool read_event() { return true; }
		
		/// Write event -- For call by the Monitor and derived classes.
		/** This function is called by the Monitor-derived classes
		  * when a write event happens.
		  *
		  * \return false if the socket has been closed or is no longer
		  * valid, true otherwise. */
		virtual bool write_event() { return true; }
		
		/// Returns the socket protocol family.
		int family() const { return m_family; }
		/// Returns the socket type
		int type() const { return m_type; }
		/// Returns the socket protocol
		int protocol() const { return m_protocol; }
		
		/// Returns the socket status
		/** Returns the socket status, which is an or'ed set of
		  * the ss* flags defined in the Socket class.
		  *
		  * \return The socket status. */
		uint32_t status() const { return m_status; }
		
		/// Returns the monitor name.
		std::string const &monitor() const;
		/// Sets the monitor used by this socket.
		/** Sets the monitor used by this socket. This can't be
		  * changed once the ssMonitored status flag is set on the
		  * socket.
		  *
		  * \param[in] name The name of the monitor to use for
		  * this socket.
		  * \return true if successful, false if no monitor with
		  * this name is registered. */
		bool monitor(std::string const &name);
	protected:
		/// Creates the socket.
		/** Creates the socket. The socket protocol family, type
		  * and protocol must have been set prior to calling this
		  * function. Upon successful creation, the ssCreated flag
		  * is added to the socket status.
		  *
		  * This function may be overriden to change the socket
		  * options, etc. as needed.
		  *
		  * \return true if the creation succeeded or the socket
		  * was already created, false otherwise. errno is set
		  * accordingly in the latter case (see the socket(2)
		  * manpage for possible errors). */
		virtual bool create();
		
		/// Closes the socket.
		/** Closes the socket, if it has been created. The ssCloseFlags
		  * flags are removed from the socket status on success.
		  *
		  * If the socket is monitored, the unwatch() function is
		  * called as well.
		  *
		  * \return true if successful or the socket wasn't created,
		  * false otherwise. errno is set accordingly in the latter
		  * case (see the close(2) manpage for possible errors). */
		virtual bool close();
		
		/// Monitors a socket.
		/** Monitors a socket. The Monitor::add(), Monitor::update()
		  * function is called with an appropriate event set depending
		  * on whether the ssMonitored, ssReadable and ssWritable status
		  * flags are set.
		  *
		  * The ssMonitored flag is always set after this function returns
		  * successfully.
		  *
		  * \return true if successful, false otherwise. */
		virtual bool watch();
		
		/// Unmonitors a socket.
		/** Unmonitors a socket.
		  *
		  * The ssMonitored flag is always unset after this function
		  * returns. */
		virtual void unwatch();
		
		/// Returns the socket file descriptor.
		/** Returns the socket file descriptor.
		  *
		  * \return The socket file descriptor, or -1 if the
		  * socket creation has not been done yet or failed. */
		int fd() const { return m_fd; }
		
		/// Sets the socket protocol family.
		/** Sets the socket protocol family. No change is done when the
		  * ssCreated status flag is set on the socket.
		  *
		  * \param[in] family The socket protocol family.
		  * \return The new socket protocol family. */
		int family(int family);
		
		/// Sets the socket type.
		/** Sets the socket type. No change is done when the ssCreated
		  * status flag is set on the socket.
		  *
		  * \param[in] type The socket type.
		  * \return The new socket type. */
		int type(int type);
		
		/// Sets the socket protocol.
		/** Sets the socket protocol. No change is done when the
		  * ssCreated status flag is set on the socket.
		  *
		  * \param[in] protocol The socket protocol.
		  * \return The new socket protocol. */
		int protocol(int protocol);
		
		/// Adds a flag to the socket status.
		/** Adds a flag to the socket status. Possible flags include
		  * all ss* flags defined in the Socket class.
		  *
		  * \param[in] flags The flags to add.
		  * \return The new socket status. */
		uint32_t status_add(uint32_t flags) { return (m_status |= flags); }
		
		/// Removes a flag from the socket status.
		/** Removes a flag from the socket status. Possible flags include
		  * all ss* flags defined in the Socket class.
		  *
		  * \param[in] flags The flags to remove.
		  * \return The new socket status. */
		uint32_t status_remove(uint32_t flags) { return (m_status &= ~flags); }
	private:
		/// The socket file descriptor
		int m_fd;
		
		/// The socket protocol family
		int m_family;
		/// The socket type
		int m_type;
		/// The socket protocol
		int m_protocol;
		
		/// The socket status
		uint32_t m_status;
		
		/// The socket monitor
		Monitor *m_monitor;
};

/*************************************************************************/



/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/

#endif /* SOCKETS_SOCKET_H */
