/* OpenSSL encapsulation.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file sockets/ssl.h
  * \brief OpenSSL SSL library encapsulation classes header file. */

#ifndef SOCKETS_SSL_H
#define SOCKETS_SSL_H

/*************************************************************************/

#if HAVE_OPENSSL
# include <openssl/ssl.h>
#endif

/*************************************************************************/

namespace sockets
{

/*************************************************************************/

class SSLInfo;
class TCPClient;

/*************************************************************************/

/// X509_STORE_CTX utility class

class X509StoreContext
{
#if HAVE_OPENSSL
	public:
		/// Constructor
		X509StoreContext(X509_STORE_CTX *x509_ctx) : 
				m_x509_ctx(x509_ctx) { }
		
		/// Returns the X509_STORE_CONTEXT structure
		X509_STORE_CTX *x509_ctx() const { return m_x509_ctx; }
		
		/// Returns the error that happened, if any
		int error() const;
	private:
		X509_STORE_CTX *m_x509_ctx;
#endif
};

/*************************************************************************/

/// SSL context class
/** This class encapsulates a SSL_CTX structure as created by OpenSSL.
  *
  * Like all SSL classes, it is designed to be instantiatable even if
  * there is OpenSSL libraries and header files have not been found
  * on the system. The factory() function in particular allows 
  * to build a SSLContext if and only if such support is provided.
  */
class SSLContext
{
	public:
		/* Classes */
		
		/// Add SSLInfo as a friend
		friend class SSLInfo;
		
		/// Events callback class
		/** This class is an event sink allowing customized behaviour to
		  * certain */
		class Events
		{
			public:
				/// Constructor
				Events(SSLContext &ctx) : m_context(ctx) { }
				/// Destructor.
				virtual ~Events() { }
				
				/// An event called when a password is requested.
				/** An event called when a password is requested, for a
				  * key, certificate, etc.
				  *
				  * The default implementation always returns 0.
				  *
				  * \param[in] buf The buffer to copy the password to.
				  * \param[in] size The length of the buffer, in bytes.
				  * \param[in] rwflag 1 if the password is used for
				  * encryption, 0 if it is used for decryption.
				  * \return The length of the password copied to the
				  * buffer. 
				  */
				virtual int on_getpass(char *buf, int size, int rwflag)
						{ return 0; }
				
				/// An event called during verification.
				/** An event called during verification.
				  *
				  * The default implementation always returns preverify_ok.
				  *
				  * \param[in] preverify_ok 1 if verification succeeded,
				  * 0 otherwise.
				  * \param[in] si The SSLInfo object this verification was
				  * made for.
				  * \param[in] x509_ctx See the man page.
				  * \return 1 if verification must succeed, 0 otherwise.
				  */
				virtual int on_verify(int preverify_ok, SSLInfo *si,
						X509StoreContext &x509_ctx) { return preverify_ok; }
				
				/// Called when initial handshake is completed
				virtual void on_handshake(SSLInfo &si) { }
				
				/// Called when an SSL library error happens on a SSL connection
				virtual void on_error(SSLInfo &si, char const *error) { }
				
				/// Returns whether the SSL context is the owner of the object (and must free it)
				virtual bool is_context_owner() const { return false; }
				
				/// Returns our SSL context
				SSLContext const &context() const { return m_context; }
			private:
				SSLContext &m_context;
		};
		
		/* Constants */
		
		/// Connection types handled by a context
		enum
		{
			/// Server connection
			ctServer = 1,
			/// Client connection
			ctClient = 2,
			
			/// Minimum type allowed
			ctMin = ctServer,
			/// Maximum type allowed
			ctMax = ctClient
		};
		
		/// Connection methods handled by a context
		enum
		{
			/// SSLv2
			cmSSLv2 = 0x1,
			/// SSLv3
			cmSSLv3 = 0x2,
			/// TLSv1
			cmTLSv1 = 0x4,
			/// Any method
			cmAny = (cmSSLv2 | cmSSLv2 | cmTLSv1)
		};
		
		/// Verify modes
		enum
		{
			/// Do a verification
			vmVerify = 0x0001,
			/// Fail immediately if the client did not return a certificate (server only)
			vmFailIfNoPeerCert = 0x0004,
			/// Only request client certificate at initial handshake (server only)
			vmClientOnce = 0x0008
		};
			
		/* Functions */
		
		/// Initializes the SSL library.
		static void init();
		
		/// Instantiates a SSL context if OpenSSL support is available.
		/** Instantiates a SSL context if OpenSSL support is available.
		  * This is the only way of creating a SSLContext object.
		  *
		  * \param[in] type The connection type for sockets using this
		  * context, that might be either ctServer or ctClient.
		  * \param[in] methods The connection method(s) allowed for 
		  * sockets using this context, that might be either cmAny, cmSSLv2,
		  * cmSSLv3 or cmTLSv1.
		  * \return A new SSL context object.
		  */
		static SSLContext *factory(uint8_t type, uint16_t methods);
		
		/* Methods */
		
		/// Destructor.
		~SSLContext();
		
		/// Creates a new SSLInfo object from this context
		/** Creates a new SSLInfo object from this context.
		  */
		SSLInfo *create(TCPClient &client) const;
		
		/// Rejects some connection methods.
		/** Rejects some connection methods. This is supported when
		  * the SSL context was created with cmAny as the chosen method.
		  *
		  * Note that attempts to disable all remaining methods are
		  * ignored.
		  */
		void reject_methods(uint16_t methods);
		
		/// Configures verification
		/** Configures verification.
		  *
		  * Note that the Events::verify() method is called once verification
		  * is done, if you set the events handler.
		  * 
		  * \param[in] mode An or'ed set of vmVerify, vmFailIfNoPeerCert and
		  * vmClientOnce.
		  */
		void verify(uint16_t mode);
		
		/// Loads a list of CA certificates from a file
		/** Loads a list of CA certificates for the certificate verification
		  * process from a file.
		  *
		  * \param[in] file The file to load the certificates from, in PEM
		  * format.
		  * \return true if successful, false otherwise.
		  */
		bool verify_certificate_file(char const *file);
		
		/// Returns the events handler.
		Events *events() const { return m_events; }
		/// Sets the events handler.
		/** Sets the events handler. The methods of this object are called when
		  * the corresponding event happens on the context and the connections
		  * initiated by the context.
		  */
		void events(Events *events) { m_events = events; }
		
		/// Returns whether session caching is enabled.
		/** Returns whether session caching is enabled. By default, it
		  * is enabled for servers but not for clients.
		  *
		  * \return true if session caching is enabled, false otherwise.
		  */
		bool is_session_caching() const;
		/// Sets whether session caching is enabled.
		void set_session_caching(bool value);
		
		/// Sets the certificate chain file.
		/** Sets the certificate chain file.
		  *
		  * \param[in] file The certificate chain file, in PEM format 
		  * \return true if successful, false otherwise.
		  */
		bool certificate_chain(char const *file);
		
		/// Sets the private key.
		/** Sets the private key. 
		  *
		  * \param[in] file The certificate key file, in PEM format.
		  * \return true if successful, false otherwise.
		  */
		bool private_key_file(char const *file);
		
		/// Ensures the consistency of the private key with the certificate.
		bool private_key_check() const;
		
		/// Returns the type of connection handled by this context
		uint8_t type() const { return m_type; }
		/// Returns an or'ed set of connection methods accepted in this context
		uint16_t methods() const { return m_methods; }
		
		/// Returns the last error of the thread.
		char const *error() const;
	private:
#if HAVE_OPENSSL
		mutable SSL_CTX *m_ctx;
#endif
		Events *m_events;

		uint8_t m_type;
		uint16_t m_methods;

		SSLContext(uint8_t type, uint16_t methods);
		
		SSLContext(SSLContext const &);
		SSLContext &operator=(SSLContext const &);
};

/*************************************************************************/

/// SSL encapsulation
/** This class encapsulates the SSL structure. It also creates the underlying
  * BIO whenever the assign method is called. This class may only be created
  * by SSLContext.
  */

class SSLInfo
{
	public:
		/// Add SSLContext as a friend
		friend class SSLContext;
		
		/// Destructor.
		~SSLInfo();
		
		/// Transforms the given socket file descriptor into a SSL connection
		/** Transforms the given socket file descriptor into a SSL connection.
		  * 
		  * \param[in] fd The socket file descriptor
		  * \return true if successful, false otherwise.
		  */
		bool assign(int fd);
		
		/// Reads data from the SSL connection
		/** Reads data from the SSL connection.
		  *
		  * \param[in] buf The buffer that receives the data.
		  * \param[in] len The length of the buffer, in bytes.
		  * \return -1 if there was an error, 0 if the (possibly SSL only) 
		  * connection has been shutdown gracefully, else the number of
		  * bytes read.
		  */
		ssize_t read(void *buf, size_t len);
		
		/// Writes data to the SSL connection
		/** Writes data to the SSL connection.
		  *
		  * \param[in] buf The buffer to write.
		  * \param[in] len The length of the buffer, in bytes.
		  * \return -1 if there was an error, or the number of bytes
		  * written to the socket.
		  */
		ssize_t write(void const *buf, size_t len);
		
		/// Sends the close-notify alert to the peer.
		/** Sends the close-notify alert to the peer. See SSL_shutdown(3)
		  * for more information.
		  *
		  * This method automatically clears the ssReadable or ssWritable
		  * flag of the socket and calls Socket::watch() if needed.
		  *
		  * \return 1 if the shutdown is complete, 0 if it is in progress,
		  * -1 if there was a fatal error.
		  */
		int shutdown();
		
		/// Did we get the close-alert?
		bool got_shutdown() const;
		/// Did we send the close alert?
		bool sent_shutdown() const;
		
		/// Returns the version of this SSL connection protocol
		char const *version() const;
		
		/// Returns the cipher version
		char const *cipher_version() const;
		/// Returns the cipher name
		char const *cipher_name() const;
		/// Returns the cipher description
		char const *cipher_desc() const;
		
		/// Returns the peer certificate key info
		char const *peer_cert_key() const;
		
		/// Returns the context associated with this object
		SSLContext &context() const { return m_context; }
		/// Returns the client associated with this object
		TCPClient &client() const { return m_client; }
	private:
		/* Constants */
		
		enum
		{
			uwfRead = 1,
			uwfWrite = 2,
			uwfShutdown = 3
		};
		
		/* Data */
#if HAVE_OPENSSL
		mutable SSL *m_ssl;
#endif
		SSLContext &m_context;
		TCPClient &m_client;
		
		/* Methods */		
		SSLInfo(SSLContext &context, TCPClient &client);

		SSLInfo(SSLInfo const &);
		SSLInfo &operator=(SSLInfo const &);
		
		void set_handshake();
		bool update_watch(int ret, uint8_t from);
};

/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/

#endif /* SOCKETS_SSL_H */
