/* TCP client class.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file sockets/tcpclient.h
  * \brief TCP client class header file. */

#ifndef SOCKETS_TCPCLIENT_H
#define SOCKETS_TCPCLIENT_H

/*************************************************************************/

#include <memory>
#include <string>

#include <sockets/ipsocket.h>
#include <sockets/ssl.h>

/*************************************************************************/

namespace sockets
{

/*************************************************************************/

class Buffer;

/*************************************************************************/

/// TCP client class
/** This class is used to represent connected TCP clients, both
  * those initiated through client sockets and server sockets. It has
  * therefore no way of actually connecting to a socket.
  *
  * This class provides functions for reading and writing from/to
  * the socket, and event handlers. */

class TCPClient : public IPSocket
{
	public:
		/* Constants */
		
		/// printf() options
		enum
		{
			/// Always write data to the write buffer
			poInBuf = 0x0001,
			/// Add line separator to the output
			poLine = 0x0002
		};
		
		/* Functions */
		
		/// Constructor.
		/** Constructor.
		  *
		  * \param[in] rbuf If non-zero, points to the read buffer of
		  * the socket.
		  * \param[in] wbuf If non-zero, points to the write buffer of
		  * the socket.
		  * \sa rbuf(), wbuf() */
		explicit TCPClient(Buffer *rbuf = 0, Buffer *wbuf = 0) :
				IPSocket(SOCK_STREAM), m_rbuf(rbuf), m_wbuf(wbuf),
				m_ssl(0), m_linesep("\r\n") { }
		/// Copy constructor
		/** This copy constructor copies the line separator but NOT the read
		  * and write buffer. */
		TCPClient(TCPClient const &tcpc);
		/// Assignment operator
		virtual TCPClient &operator=(TCPClient const &right);
		
		/// Assigns a socket file descriptor.
		/** Assigns a socket file descriptor to the socket object,
		  * that is then used by the object for all operations on
		  * the socket.
		  *
		  * This function overrides IPSocket::assign() to call
		  * the on_connect(), on_read() and on_write() functions
		  * as needed.
		  *
		  * \param[in] fd The socket file descriptor to assign.
		  * \param[in] status The status flags that are added to
		  * the socket. If the ssMonitored flag is among them,
		  * the watch() function is automatically called. 
		  * on_connect(), on_write() and on_read() are called
		  * respectively when the ssConnected, ssWritable and
		  * ssReadable flags are among them.
		  * \returns true if successful, false otherwise. errno
		  * is set accordingly in the latter case. */
		virtual bool assign(int fd, uint32_t status);
		
		/// Overrides Socket::close().
		/** This function overrides Socket::close to call the shutdown()
		  * function on the socket. It also clears the read and write
		  * buffers, if set.
		  *
		  * \return the same values as Socket::close(). */
		virtual bool close();
		
		/// Reads data from socket directly.
		/** Reads data from socket directly, without touching to the
		  * eventual read buffer set to the socket. Reading is
		  * done regardless of whether the ssReadable status flag
		  * is set, but the ssConnected flag needs to be set.
		  *
		  * \param[in] buf The buffer that will receive the data
		  * read from the socket.
		  * \param[in] len The length of the buffer, in bytes.
		  * \return The number of bytes read from the socket, or
		  * -1 on error. A return value of 0 means that an orderly
		  * shutdown has been made on the other end of the socket,
		  * or that the ssConnected status flag wasn't set. */
		virtual ssize_t read(char *buf, size_t len);
		
		/// Reads data from socket in the read buffer.
		/** Reads data from socket in the read buffer. Reading is
		  * done regardless of whether the ssReadable status flag
		  * is set, but the ssConnected flag needs to be set.
		  *
		  * \param[in] max The maximum number of bytes to read from
		  * the socket. Set this to zero if you want to fill the
		  * buffer with all the data available (provided the socket
		  * is big enough, of course!)
		  * \return The number of bytes that were put in the buffer.
		  * On error, -1 is returned if errno is set to anything
		  * else than EAGAIN (or EWOULDBLOCK), else either the
		  * number of bytes that were put in the buffer (if non-zero)
		  * or -1 is returned. If the other end shut down the
		  * connection, 0 is returned, unless there were bytes put
		  * in the buffer. 0 is also returned if the ssConnected flag
		  * is not set or if the buffer is already full.  */
		virtual ssize_t read_inbuf(size_t max = 0);
		
		/// Reads a line from the read buffer.
		/** Reads a line from the read buffer. The read buffer must
		  * have been filled by a call to the read_inbuf() function.
		  * Line separator is the LF character, any CR character
		  * found before or after is ignored.
		  *
		  * \param[in] buf A buffer that will receive a copy of the
		  * line that has been found on success. The copy does not
		  * include the trailing CR or LF characters but will be
		  * NULL terminated.
		  * \param[in] len The length of the buffer, in bytes. This
		  * also determines the maximum length allowed for lines
		  * before giving up.
		  * \return the length of the line if an entire line has been
		  * found in the buffer, 0 if no line was found, -1 on error
		  * (most likely, a line longer than len in the buffer). */
		virtual ssize_t read_line(char *buf, size_t len) const;
		
		/// Writes data to the socket, falling back on buffer if it blocks.
		/** Writes data to the socket, falling back on the buffer if it
		  * blocks. Writing is done regardless of whether the
		  * ssWritable status flag is set, but the ssConnecting or
		  * ssConnected flags need to be set.
		  *
		  * The on_writing() function is called before doing anything
		  * if the ssConnected flag is set and inbuf is false.
		  *
		  * This function will write to the write buffer instead of the
		  * socket when one of the following conditions is true:
		  *   - writing to the socket would block;
		  *   - there is already data in the write buffer;
		  *   - the ssConnecting status flag is set;
		  *   - the inbuf parameter is set to true.
		  * Attempts to send the data from the write buffer to the
		  * socket can then be made through the write_outbuf()
		  * function, that is generally called from the on_write()
		  * and on_writing() functions or after a few calls to write()
		  * with inbuf set to true.
		  *
		  * \param[in] buf A buffer containing the data to write to the
		  * socket.
		  * \param[in] len The length of the buffer, in bytes.
		  * \param[in] inbuf If true, the data is always written to
		  * the write buffer even if sending it directly through the
		  * socket could be tried.
		  * \return The len parameter if everything has either been
		  * sent through the socket or written in the buffer, 0 if
		  * the socket is not connecting or connected, between
		  * 0 and len-1 if the buffer is full, and -1 if there was an
		  * error on the socket other than EAGAIN or EWOULDBLOCK. */
		virtual ssize_t write(char const *buf, size_t len, bool inbuf = false);
		/// Same as write() but uses a NUL-terminated string as buffer.
		virtual ssize_t write_str(char const *str, bool inbuf = false)
				{ return write(str, strlen(str), inbuf); }
		/// Same as write_str() but uses a std::string as buffer.
		virtual ssize_t write_string(std::string const &str, bool inbuf = false)
				{ return write(str.data(), str.size(), inbuf); }
		
		/// Writes a line to the socket.
		/** Writes a line to the socket. This is done by writing the
		  * given line and the socket line separator to the write
		  * buffer, and then call write_outbuf().
		  *
		  * \param[in] line The line to write, without the line
		  * separators.
		  * \param[in] inbuf If true, write_outbuf() is NOT called
		  * after the line has been written to the socket.
		  * \return The total number of characters written to the
		  * buffer on success, 0 if the socket is not connecting or
		  * connected or if the buffer could not contain the
		  * whole data, -1 if the write_outbuf() function failed. */
		virtual ssize_t write_line(char const *line, bool inbuf = false);
		
		/// Writes to the socket, printf()-like style.
		/** Writes to the socket, printf()-like style. The actual
		  * writing is done through the write() function by default.
		  *
		  * \param[in] options An or'ed set of either
		  *    - pfInBuf: always write to the write buffer even if
		  *      sending it directly through the socket could be tried.
		  *    - pfLine: writing is done through write_line().
		  * \param[in] format The format string.
		  * \param[in] ... Args for the format string.
		  * \return whatever is returned by the write() or write_line()
		  * function. */
		virtual ssize_t printf(uint32_t options, char const *format,
				...) FORMAT(printf,3,4);
		
		/// Writes to the socket, vprintf()-like style.
		/** Writes to the socket, vprintf()-like style. The actual
		  * writing is done through the write() function by default.
		  *
		  * \param[in] options An or'ed set of either
		  *    - pfInBuf: always write to the write buffer even if
		  *      sending it directly through the socket could be tried.
		  *    - pfLine: writing is done through write_line().
		  * \param[in] format The format string.
		  * \param[in] args Args for the format string.
		  * \return whatever is returned by the write() or write_line()
		  * function. */
		virtual ssize_t vprintf(uint32_t options, char const *format,
				va_list args);
		
		/// Writes data from the write buffer to the socket
		/** Writes as much data as possible from the write buffer to
		  * the socket. Writing is done regardless of whether the
		  * ssWritable status flag is set, but the ssConnected flag
		  * needs to be set.
		  *
		  * \param[in] max The maximum number of bytes from the buffer
		  * to write. Set this to zero if you don't want to set a
		  * limit.
		  * \return The number of bytes that were written to the
		  * socket, 0 if the socket is not connected or if the write
		  * buffer is empty, or -1 if there was an error and errno was
		  * not set to EAGAIN or EWOULDBLOCK. */
		virtual ssize_t write_outbuf(size_t max = 0);
		
		/// Writes data to the socket directly.
		/** Writes data to the socket directly, without touching to the
		  * eventual write buffer set to the socket. Writing is
		  * done regardless of whether the ssWritable status flag
		  * is set, but the ssConnected flag needs to be set.
		  *
		  * \param[in] buf A buffer containing the data to write to
		  * the socket.
		  * \param[in] len The length of the buffer, in bytes.
		  * \return The number of bytes that were sent, 0 if the
		  * socket is not connected, or -1 on error. */
		virtual ssize_t write_unbuffered(char const *buf, size_t len);
		
		/// Returns the read buffer.
		/** Returns the read buffer. This buffer is mainly used by
		  * the read_inbuf() function, with other functions expecting
		  * to find data in it (of course, you are free to use the
		  * function provided by the Buffer class as well).
		  *
		  * \return The read buffer. */
		Buffer *rbuf() const { return m_rbuf; }
		/// Sets the read buffer.
		/** Sets the read buffer. By default, there is no read buffer,
		  * and it is not required to set one unless you intend to
		  * use functions that manipulate it.
		  *
		  * Please note that the socket does not try to free the
		  * buffer once it is deleted; you have to take care of that
		  * by yourself.
		  *
		  * \param[in] rbuf The read buffer.
		  * \return The new read buffer. */
		Buffer *rbuf(Buffer *rbuf) { return (m_rbuf = rbuf); }
		
		/// Returns the write buffer.
		/** Returns the write buffer. This buffer is mainly used to
		  * store data that needs to be written to the socket, in case
		  * it is not possible to do it immediately. 
		  *
		  * \return The write buffer. */
		Buffer *wbuf() const { return m_wbuf; }
		/// Sets the write buffer.
		/** Sets the write buffer. By default, there is no write
		  * buffer, but it is highly recommended to set one.
		  *
		  * Please note that the socket does not try to free the
		  * buffer once it is deleted; you have to take care of that
		  * by yourself.
		  *
		  * \param[in] wbuf The write buffer.
		  * \return The new write buffer. */
		Buffer *wbuf(Buffer *wbuf) { return (m_wbuf = wbuf); }
		
		/// Configures the SSL option for the socket.
		/** Configures the SSL option for the socket. When enabled, the
		  * connection is encrypted. This requires OpenSSL support.
		  *
		  * Changing this option is only allowed on non-connected 
		  * sockets. 
		  *
		  * \param[in] context The SSLContext that is used to create
		  * our SSLInfo object. If this is 0, the connection won't be
		  * encrypted.
		  * \return true if successful, false otherwise (including when
		  * the socket is already connected).
		  * \sa SSLContext, SSLInfo
		  */
		bool set_ssl(SSLContext const *context);
		
		/// Returns the SSL connection
		SSLInfo const *ssl() const { return m_ssl.get(); }
		
		/// Returns the line separator of the socket.
		/** Returns the line separator of the socket, that is used by
		  * the write_line() function.
		  *
		  * \return The line separator of the socket. */
		std::string const &linesep() const { return m_linesep; }
		/// Sets the line separator of the socket.
		std::string const &linesep(std::string const linesep)
				{ return (m_linesep = linesep); } 
	protected:
		/// Overrides IPSocket::create().
		/** This adds support for SSL connections */
		virtual bool create();
		
		/// Actually calls the recv() function.
		/** Actually calls the recv() function. The ssReadable
		  * flag is removed and watch() is called if the read
		  * would block (i.e. errno == EAGAIN).
		  *
		  * \param[in] buf The buffer that will receive the data
		  * read from the socket.
		  * \param[in] len The length of the buffer, in bytes.
		  * \return The number of bytes read from the socket, or
		  * -1 on error. A return value of 0 means that an orderly
		  * shutdown has been made on the other end of the socket. */
		virtual ssize_t read_real(void *buf, size_t len);
		
		/// Actually calls the send() function.
		/** Actually calls the send() function. The ssWritable flag is
		  * removed and watch() is called if the write would block
		  * (i.e. errno == EAGAIN).
		  *
		  * \param[in] buf A buffer containing the data to write to
		  * the socket.
		  * \param[in] len The length of the buffer, in bytes.
		  * \return The number of bytes that were sent, or -1 on
		  * on error. */
		virtual ssize_t write_real(void const *buf, size_t len);
		
		/// Read event handling
		/** This function sets the ssReadable status flag, then calls
		  * the on_read() function and finishes by calling watch()
		  * to update the events being monitored, if ssReadable is
		  * still set on return.
		  *
		  * \return false if the socket has been closed or is no longer
		  * valid, true otherwise. */
		virtual bool read_event();
		
		/// Write event handling
		/** This function sets the ssConnected and ssWritable status flags
		  * then calls the on_connect() function if the ssConnecting
		  * status flag is set (this flag is removed prior to calling the
		  * function). Note that the ssConnected flag is NOT set if the
		  * connection failed.
		  *
		  * Then, it'll call the on_write() function if the ssWritable
		  * flag is still set. If this flag is then still set, the watch()
		  * function is called to update the events being monitored.
		  *
		  * \return false if the socket has been closed or is no longer
		  * valid, true otherwise. */
		virtual bool write_event();
		
		/// Connect event
		/** This function is called whenever a successful connection
		  * attempt finishes, or an asynchronous connection attempt
		  * fails.
		  *
		  * The default implementation just returns true.
		  *
		  * \param[in] error This is set to zero if the connection
		  * is established, or is the error number that occured otherwise.
		  * \return false if the socket has been closed or is no longer
		  * valid, true otherwise. */
		virtual bool on_connect(int error) { return true; }
		
		/// Read event
		/** This function is called whenever the socket becomes readable
		  * (there is new data to be read).
		  *
		  * The default implementation just returns true.
		  * \return false if the socket has been closed or is no longer
		  * valid, true otherwise. */
		virtual bool on_read() { return true; }
		
		/// Write event
		/** This function is called whenever the socket becomes writable
		  * (new data can be sent through the socket).
		  *
		  * The default implementation just returns true. It is highly
		  * recommended that you ouverride this function to call the
		  * write_outbuf() if you use buffered output functions of the
		  * socket.
		  *
		  * \return false if the socket has been closed or is no longer
		  * valid, true otherwise. */
		virtual bool on_write() { return true; }
		
		/// Writing event
		/** This event is called by the write() function whenever
		  * there is data in the buffer, which means the write()
		  * would normally write to the buffer without trying to
		  * actually write data to the socket.
		  *
		  * You may therefore want to call the write_outbuf() function
		  * from here if you want your data to be sent faster than
		  * waiting for the on_write() event to be called.
		  *
		  * The default implementation just returns true.
		  *
		  * \return true if no error other than EAGAIN or EWOULDBLOCK
		  * happened on the socket, false otherwise. */
		virtual bool on_writing() { return true; }
		
		/// Returns the ssl connection
		SSLInfo *ssl() { return m_ssl.get(); }
	private:
		Buffer *m_rbuf;
		Buffer *m_wbuf;
		
		std::auto_ptr<SSLInfo> m_ssl;
		
		std::string m_linesep;
};

/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/

#endif /* SOCKETS_TCPCLIENT_H */
