/* OpenSSL encapsulation.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#if HAVE_OPENSSL
# include <openssl/bio.h>
# include <openssl/err.h>
#endif

#include <sockets/tcpclient.h>

using namespace std;

/*************************************************************************/

#if HAVE_OPENSSL

namespace
{

/*************************************************************************/

int SSLContext_idx = -1;
int SSLInfo_idx = -1;

/*************************************************************************/

int pem_passwd_cb(char *buf, int size, int rwflag, void *userdata)
{
	sockets::SSLContext *ctx = (sockets::SSLContext *)userdata;
	
	if (ctx && ctx->events())
		return ctx->events()->on_getpass(buf, size, rwflag);
	else
		return 0;
}

/*************************************************************************/

int verify_cb(int preverify_ok, X509_STORE_CTX *x509_ctx)
{	
	SSL *ssl = (SSL *)X509_STORE_CTX_get_ex_data(x509_ctx, 
			SSL_get_ex_data_X509_STORE_CTX_idx());
	
	if (!ssl)
		return preverify_ok;
			
	sockets::SSLInfo *si = (sockets::SSLInfo *)
			SSL_get_ex_data(ssl, SSLInfo_idx);
	sockets::X509StoreContext x509_ctx_(x509_ctx);
	
	if (si && si->context().events())
		return si->context().events()->on_verify(preverify_ok, si,
				x509_ctx_);
	else
		return preverify_ok;
}

/*************************************************************************/

} /* anonymous namespace */

#endif

/*************************************************************************/

namespace sockets
{

/*************************************************************************/

SSLContext::SSLContext(uint8_t type, uint16_t methods) :
#if HAVE_OPENSSL
	m_ctx(0),
#endif
	m_type(type), m_methods(methods)
{
#if HAVE_OPENSSL
	switch (type)
	{
		case ctServer:
			switch (methods)
			{
				case cmSSLv2:
					m_ctx = SSL_CTX_new(SSLv2_server_method());
					break;
				case cmSSLv3:
					m_ctx = SSL_CTX_new(SSLv3_server_method());
					break;
				case cmTLSv1:
					m_ctx = SSL_CTX_new(TLSv1_server_method());
					break;
				default:
					m_ctx = SSL_CTX_new(SSLv23_server_method());
					break;
			}
			break;
		case ctClient:
			switch (methods)
			{
				case cmSSLv2:
					m_ctx = SSL_CTX_new(SSLv2_client_method());
					break;
				case cmSSLv3:
					m_ctx = SSL_CTX_new(SSLv3_client_method());
					break;
				case cmTLSv1:
					m_ctx = SSL_CTX_new(TLSv1_client_method());
					break;
				default:
					m_ctx = SSL_CTX_new(SSLv23_client_method());
					break;
			}
			break;
	}
	
	if (m_ctx)
	{
		SSL_CTX_set_mode(m_ctx, (/*SSL_MODE_ENABLE_PARTIAL_WRITE |*/
				SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER));
		
		SSL_CTX_set_default_passwd_cb(m_ctx, pem_passwd_cb);
		SSL_CTX_set_default_passwd_cb_userdata(m_ctx, this);
	}
#endif
}

/*************************************************************************/

SSLContext::~SSLContext()
{
#if HAVE_OPENSSL
	if (m_events && m_events->is_context_owner())
		delete m_events;
	m_events = 0;

	if (m_ctx)
		SSL_CTX_free(m_ctx);
	m_ctx = 0;
#endif
}

/*************************************************************************/


bool SSLContext::certificate_chain(char const *file)
{
#if HAVE_OPENSSL
	return SSL_CTX_use_certificate_chain_file(m_ctx, file);
#else
	return false;
#endif
}

/*************************************************************************/

SSLInfo *SSLContext::create(TCPClient &client) const
{
#if HAVE_OPENSSL
	SSLInfo *si = new SSLInfo((SSLContext &)*this, client);
	
	if (!si->m_ssl || SSL_set_ex_data(si->m_ssl, SSLInfo_idx, si) == 0)
	{
		delete si;
		si = 0;
	}

	return si;
#else
	return 0;
#endif
}

/*************************************************************************/

char const *SSLContext::error() const
{
#if HAVE_OPENSSL
	return ERR_error_string(ERR_peek_last_error(), 0);
#else
	return 0;
#endif
}

/*************************************************************************/

SSLContext *SSLContext::factory(uint8_t type, uint16_t methods)
{	
#if HAVE_OPENSSL
	if (type < ctMin || type > ctMax)
		return 0;
	
	SSLContext *sc = new SSLContext(type, methods);
	
	/* Ensure we don't return an impossible to use context */
	if (!sc->m_ctx || SSL_CTX_set_ex_data(sc->m_ctx, SSLContext_idx, sc) == 0)
	{
		delete sc;
		sc = 0;
	}
	
	return sc;
#else
	return 0;
#endif
}

/*************************************************************************/

void SSLContext::init()
{
#if HAVE_OPENSSL
	SSL_library_init();
	SSL_load_error_strings();
	
	SSLContext_idx = SSL_CTX_get_ex_new_index(0, 0, 0, 0, 0);
	SSLInfo_idx = SSL_get_ex_new_index(0, 0, 0, 0, 0);
#endif
}

/*************************************************************************/

bool SSLContext::is_session_caching() const
{
#if HAVE_OPENSSL
	switch (m_type)
	{
		case ctServer:
			return (SSL_CTX_get_session_cache_mode(m_ctx) & 
					SSL_SESS_CACHE_CLIENT);
		case ctClient:
			return (SSL_CTX_get_session_cache_mode(m_ctx) & 
					SSL_SESS_CACHE_SERVER);
		default:
			return false;
	}
#else
	return false;
#endif
}

/*************************************************************************/

bool SSLContext::private_key_check() const
{
#if HAVE_OPENSSL
	return SSL_CTX_check_private_key(m_ctx);
#else
	return false;
#endif
}

/*************************************************************************/

bool SSLContext::private_key_file(char const *file)
{
#if HAVE_OPENSSL
	return SSL_CTX_use_PrivateKey_file(m_ctx, file, SSL_FILETYPE_PEM);
#else
	return false;
#endif
}

/*************************************************************************/

void SSLContext::reject_methods(uint16_t methods)
{
#if HAVE_OPENSSL
	/* Don't allow all methods to be disabled */
	if ((m_methods & ~methods) == 0)
		return;
	
	long options = 0;
	
	if ((m_methods & cmSSLv2) && (methods & cmSSLv2))
		options |= SSL_OP_NO_SSLv2;
	if ((m_methods & cmSSLv3) && (methods & cmSSLv3))
		options |= SSL_OP_NO_SSLv3;
	if ((m_methods & cmTLSv1) && (methods & cmTLSv1))
		options |= SSL_OP_NO_TLSv1;
	
	if (options)
		SSL_CTX_set_options(m_ctx, options);
#endif
}

/*************************************************************************/

void SSLContext::set_session_caching(bool value)
{
#if HAVE_OPENSSL
	switch (m_type)
	{
		case ctServer:
			SSL_CTX_set_session_cache_mode(m_ctx, 
					(value ? SSL_SESS_CACHE_CLIENT : SSL_SESS_CACHE_OFF));
		case ctClient:
			SSL_CTX_set_session_cache_mode(m_ctx, 
					(value ? SSL_SESS_CACHE_SERVER : SSL_SESS_CACHE_OFF));
	}
#endif
}

/*************************************************************************/

void SSLContext::verify(uint16_t mode)
{
#if HAVE_OPENSSL
	long vmode = 0;
	
	if (mode & vmVerify)
	{
		vmode |= SSL_VERIFY_PEER;
		
		if (type() == ctServer)
		{
			if (mode & vmFailIfNoPeerCert)
				vmode |= SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
			if (mode & vmClientOnce)
				vmode |= SSL_VERIFY_CLIENT_ONCE;
		}
	} else {
		vmode |= SSL_VERIFY_NONE;
	}
	
	SSL_CTX_set_verify(m_ctx, vmode, verify_cb);
#endif
}

/*************************************************************************/

bool SSLContext::verify_certificate_file(char const *file)
{
#if HAVE_OPENSSL
	return SSL_CTX_load_verify_locations(m_ctx, file, 0);
#else
	return false;
#endif
}

/*************************************************************************/

SSLInfo::SSLInfo(SSLContext &context, TCPClient &client) : 
#if HAVE_OPENSSL
		m_ssl(0),
#endif
		m_context(context), m_client(client)
{
#if HAVE_OPENSSL
	m_ssl = SSL_new(m_context.m_ctx);
#endif
}

/*************************************************************************/

SSLInfo::~SSLInfo()
{
#if HAVE_OPENSSL
	if (m_ssl)
		SSL_free(m_ssl);
#endif
}

/*************************************************************************/

bool SSLInfo::assign(int fd)
{
#if HAVE_OPENSSL
	BIO *bio = BIO_new_socket(fd, BIO_NOCLOSE);
	if (!bio)
		return false;
	
	SSL_set_bio(m_ssl, bio, bio);

	if (context().type() == SSLContext::ctServer)
		SSL_set_accept_state(m_ssl);
	else if (context().type() == SSLContext::ctClient)
		SSL_set_connect_state(m_ssl);

	return true;
#else
	return false;
#endif
}

/*************************************************************************/

char const *SSLInfo::cipher_desc() const
{
#if HAVE_OPENSSL
	SSL_CIPHER *ciph = SSL_get_current_cipher(m_ssl);
	
	if (!ciph)
		return "none";
	
	static char desc[128];
	return SSL_CIPHER_description(ciph, desc, sizeof(desc));
#else
	return "unknown";
#endif
}

/*************************************************************************/

char const *SSLInfo::cipher_name() const
{
#if HAVE_OPENSSL
	SSL_CIPHER *ciph = SSL_get_current_cipher(m_ssl);
	
	if (!ciph)
		return "none";
	
	return SSL_CIPHER_get_name(ciph);
#else
	return "unknown";
#endif
}

/*************************************************************************/

char const *SSLInfo::cipher_version() const
{
#if HAVE_OPENSSL
	SSL_CIPHER *ciph = SSL_get_current_cipher(m_ssl);
	
	if (!ciph)
		return "none";
	
	return SSL_CIPHER_get_version(ciph);
#else
	return "unknown";
#endif
}

/*************************************************************************/

bool SSLInfo::got_shutdown() const
{
#if HAVE_OPENSSL
	return (SSL_get_shutdown(m_ssl) & SSL_RECEIVED_SHUTDOWN);
#else
	return false;
#endif
}

/*************************************************************************/

char const *SSLInfo::peer_cert_key() const
{
#if HAVE_OPENSSL
	static char key[32];

	X509 *cert = SSL_get_peer_certificate (m_ssl);
	if (cert != NULL)
	{
		EVP_PKEY *pkey = X509_get_pubkey (cert);
		if (pkey != NULL)
		{
			if (pkey->type == EVP_PKEY_RSA && pkey->pkey.rsa != NULL &&
					pkey->pkey.rsa->n != NULL)
			{
				snprintf (key, sizeof(key), "%d bit RSA",
						BN_num_bits (pkey->pkey.rsa->n));
				return key;
			}
			else if (pkey->type == EVP_PKEY_DSA && pkey->pkey.dsa != NULL &&
					pkey->pkey.dsa->p != NULL)
			{
				snprintf (key, sizeof(key), ", %d bit DSA",
						BN_num_bits (pkey->pkey.dsa->p));
				return key;
			}
			EVP_PKEY_free(pkey);
		}
		X509_free (cert);
	}
	
	return "unknown";
#else
	return "unknown";
#endif
}

/*************************************************************************/

ssize_t SSLInfo::read(void *buf, size_t len)
{
#if HAVE_OPENSSL
	errno = 0;
	ERR_clear_error();
	
	int ret;
	
	while ((ret = SSL_read(m_ssl, buf, len)) < 0)
	{
		if (errno != EINTR)
			break;
	}
	
	if (ret > 0)
	{
		set_handshake();
		return ret;
	}
	
	if (!update_watch(ret, uwfRead))
		return -1;
	
	return ((ret < 0) ? -1 : 0);
#else
	return -1;
#endif
}

/*************************************************************************/

bool SSLInfo::sent_shutdown() const
{
#if HAVE_OPENSSL
	return (SSL_get_shutdown(m_ssl) & SSL_SENT_SHUTDOWN);
#else
	return false;
#endif
}

/*************************************************************************/

void SSLInfo::set_handshake()
{
	if (!(client().status() & Socket::ssSSLHandshake))
	{
		client().status_add(Socket::ssSSLHandshake);
		if (context().events())
			context().events()->on_handshake(*this);
	}
}

/*************************************************************************/

int SSLInfo::shutdown()
{
#if HAVE_OPENSSL
	errno = 0;
	ERR_clear_error();
	
	int ret;
	
	while ((ret = SSL_shutdown(m_ssl)) < 0)
	{
		if (errno != EINTR)
			break;
	}
	
	if (!update_watch(ret, uwfShutdown))
		return -1;
	
	if (ret > 0)
		return 1;
	else if (ret < 0)
		return -1;
	else
		return 0;
#else
	return -1;
#endif
}

/*************************************************************************/

bool SSLInfo::update_watch(int ret, uint8_t from)
{
#if HAVE_OPENSSL
	int error;

	switch ((error = SSL_get_error(m_ssl, ret)))
	{
		case SSL_ERROR_ZERO_RETURN:
			if (got_shutdown() && !sent_shutdown())
				shutdown();
			return true;
		case SSL_ERROR_WANT_READ:
		case SSL_ERROR_WANT_WRITE:
			if (error == SSL_ERROR_WANT_READ)
			{
				client().status_remove(Socket::ssReadable);
				if (from == uwfWrite)
					client().status_remove(Socket::ssWritable);
			}
			if (error == SSL_ERROR_WANT_WRITE)
			{
				client().status_remove(Socket::ssWritable);
				if (from == uwfRead)
					client().status_remove(Socket::ssReadable);
			}
			
			client().watch();
			
			/* Set this so that the socket class knows we need to retry */
			errno = EWOULDBLOCK;
			return true;
		case SSL_ERROR_SSL:
			if (context().events())
				context().events()->on_error(*this, context().error());
			return true;
		default:
			return true;
	}
#else
	return -1;
#endif
}

/*************************************************************************/

char const *SSLInfo::version() const
{
#if HAVE_OPENSSL
	return SSL_get_version(m_ssl);
#else
	return "unknown";
#endif
}

/*************************************************************************/

ssize_t SSLInfo::write(void const *buf, size_t len)
{
#if HAVE_OPENSSL
	errno = 0;
	ERR_clear_error();
	
	int ret;
	
	while ((ret = SSL_write(m_ssl, buf, len)) < 0)
	{
		if (errno != EINTR)
			break;
	}
	
	if (ret > 0)
	{
		set_handshake();
		return ret;
	}
	
	if (!update_watch(ret, uwfWrite))
		return -1;
	
	return ((ret < 0) ? -1 : 0);
#else
	return -1;
#endif
}

/*************************************************************************/

int X509StoreContext::error() const
{
#if HAVE_OPENSSL
	return X509_STORE_CTX_get_error(m_x509_ctx);
#else
	return 0;
#endif
}

/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/
