/* Essential tools.
 *
 * PegSoft compatibility library (c) 2004 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

/** \file system/compat.h
  * \brief Compatible function replacements header file.
  *
  * This header file provides replacements for missing functions, types
  * and other declarations in the C library headers. */

#ifndef SYSTEM_COMPAT_H
#define SYSTEM_COMPAT_H

/*************************************************************************/

#include <system/headers.h>

/*************************************************************************/

#if !HAVE_INET_ADDR || !HAVE_INET_ATON || !HAVE_INET_NTOA
# if HAVE_SYS_SOCKET_H
#  include <sys/socket.h>
# endif
# if HAVE_NETINET_IN_H
#  include <netinet/in.h>
# endif
#endif

/*************************************************************************/

#if defined(__cplusplus)
# define EXT extern "C"
#else
# define EXT extern
#endif

#if defined(__cplusplus)
namespace std 
{
#endif

#if !HAVE_SNPRINTF
EXT int compat_snprintf(char *buf, size_t len, const char *format, ...);
# define snprintf compat_snprintf
#endif

#if !HAVE_VSNPRINTF
EXT int compat_vsnprintf(char *buf, size_t len, const char *format,
		va_list ap);
# define vsnprintf compat_vsnprintf
#endif

#if defined(__cplusplus)
} /* namespace std */
#endif

#if !HAVE_INET_ADDR
EXT in_addr_t compat_inet_addr(char const *cp);
# define inet_addr compat_inet_addr
#endif

#if !HAVE_INET_ATON
EXT int compat_inet_aton(const char *cp, struct in_addr *inp);
# define inet_aton compat_inet_aton
#endif

#if !HAVE_INET_NTOA
EXT char *compat_inet_ntoa(struct in_addr in);
# define inet_ntoa compat_inet_ntoa
#endif

#if !HAVE_STPCPY
EXT char *compat_stpcpy(char *dest, char const *src);
# define stpcpy compat_stpcpy
#endif

#if !HAVE_STPNCPY
EXT char *compat_stpncpy(char *dest, char const *src, size_t n);
# define stpncpy compat_stpncpy
#endif

#if !HAVE_STRCASECMP
EXT int compat_strcasecmp(char const *s1, char const *s2);
# define strcasecmp compat_strcasecmp
#endif

#if !HAVE_STRCSPN
EXT size_t compat_strcspn(char const *s, char const *reject);
# define strcspn compat_strcspn
#endif

#if !HAVE_STRDUP
EXT char *compat_strdup(char const *s);
# define strdup compat_strdup
#endif

#if !HAVE_STRERROR
EXT char *compat_strerror(int errnum);
# define strerror compat_strerror
#endif

#if !HAVE_STRICMP
# define stricmp strcasecmp
#endif

#if !HAVE_STRNCASECMP
EXT int compat_strncasecmp(char const *s1, char const *s2, size_t n);
# define strncasecmp compat_strncasecmp
#endif

#if !HAVE_STRNICMP
# define strnicmp strncasecmp
#endif

#if !HAVE_STRSIGNAL
EXT char *compat_strsignal(int signum);
# define strsignal compat_strsignal
#endif

#if !HAVE_STRSPN
EXT size_t compat_strspn(char const *s, char const *accept);
# define strspn compat_strspn
#endif

#undef EXT

/*************************************************************************/

#if !HAVE_GID_T
typedef int gid_t;
#endif

#if !HAVE_MODE_T
typedef int mode_t;
#endif

#if !HAVE_OFF_T
typedef long off_t;
#endif

#if !HAVE_PID_T
typedef int pid_t;
#endif

#if !HAVE_PTRDIFF_T
typedef signed ptrdiff_t;
#endif

#if !HAVE_SIZE_T
typedef unsigned size_t;
#endif

#if !HAVE_SOCKLEN_T
typedef int socklen_t;
#endif

#if !HAVE_SSIZE_T
typedef signed ssize_t;
#endif

#if !HAVE_UID_T
typedef int uid_t;
#endif

/*************************************************************************/

#if !NO_STDINT

#if !HAVE_INT8_T || !HAVE_INT16_T || !HAVE_INT32_T || !HAVE_UINT8_T || !HAVE_UINT16_T || !HAVE_UINT32_T
# if SIZEOF_CHAR == 1
#  if !HAVE_INT8_T
typedef signed char int8_t;
#  endif
#  if !HAVE_UINT8_T
typedef unsigned char uint8_t;
#  endif
# endif
# if SIZEOF_CHAR == 2
#  if !HAVE_INT16_T
typedef signed char int16_t;
#   undef HAVE_INT16_T
#   define HAVE_INT16_T 1
#  endif
#  if !HAVE_UINT16_T
typedef unsigned char uint16_t;
#   undef HAVE_UINT16_T
#   define HAVE_UINT16_T 1
#  endif
# endif
# if SIZEOF_SHORT == 2
#  if !HAVE_INT16_T
typedef signed short int16_t;
#   undef HAVE_INT16_T
#   define HAVE_INT16_T 1
#  endif
#  if !HAVE_UINT16_T
typedef unsigned short uint16_t;
#   undef HAVE_UINT16_T
#   define HAVE_UINT16_T 1
#  endif
# endif
# if SIZEOF_SHORT == 4
#  if !HAVE_INT32_T
typedef signed short int32_t;
#   undef HAVE_INT32_T
#   define HAVE_INT32_T 1
#  endif
#  if !HAVE_UINT32_T
typedef unsigned short uint32_t;
#   undef HAVE_UINT32_T
#   define HAVE_UINT32_T 1
#  endif
# endif
# if SIZEOF_INT == 2
#  if !HAVE_INT16_T
typedef signed int int16_t;
#   undef HAVE_INT16_T
#   define HAVE_INT16_T 1
#  endif
#  if !HAVE_UINT16_T
typedef unsigned int uint16_t;
#   undef HAVE_UINT16_T
#   define HAVE_UINT16_T 1
#  endif
# endif
# if SIZEOF_INT == 4
#  if !HAVE_INT32_T
typedef signed int int32_t;
#   undef HAVE_INT32_T
#   define HAVE_INT32_T 1
#  endif
#  if !HAVE_UINT32_T
typedef unsigned int uint32_t;
#   undef HAVE_UINT32_T
#   define HAVE_UINT32_T 1
#  endif
# endif
# if SIZEOF_LONG == 4
#  if !HAVE_INT32_T
typedef signed long int32_t;
#   undef HAVE_INT32_T
#   define HAVE_INT32_T 1
#  endif
#  if !HAVE_UINT32_T
typedef unsigned long uint32_t;
#   undef HAVE_UINT32_T
#   define HAVE_UINT32_T 1
#  endif
# endif
#endif

#endif

/*************************************************************************/

#ifndef NAME_MAX
# define NAME_MAX	255
#endif

#ifndef PATH_MAX
# define PATH_MAX   4096
#endif

/*************************************************************************/

#ifdef __GNUC__
# define FORMAT(type,fmt,start) __attribute__((format(type,fmt,start)))
#else 
# define FORMAT(type,fmt,start)
#endif

/*************************************************************************/

#endif /* SYSTEM_COMPAT_H */
