/* Hostname blacklist.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * Based on the original code of Services by Andy Church.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#if HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif
#if HAVE_ARPA_INET_H
# include <arpa/inet.h>
#endif

#include <fstream>

#include <core/directives.h>
#include <core/epona.h>
#include <log/levels.h>
#include <misc/string.h>

#include "blacklist.h"
#include "epona.h"
#include "language.h"
#include "nickserv.h"
#include "services.h"

using namespace std;
using namespace core;

/*************************************************************************/

class BLHostHash : public containers::Hash<BLHost *, char const *>
{
	public:
		virtual int compare(char const *const &key,
				BLHost *const &value) const 
				{ return stricmp(key, value->mask()); }	
		virtual uint32_t hash(char const *const &key, uint32_t max) const
				{ return (misc::crc16_str(key) % max); }
		virtual bool match(char const *const &key,
				BLHost *const &value) const
				{ return !stricmp(key, value->mask()); }
};

/*************************************************************************/

/* Blacklisted hosts hash */
static BLHostHash host_hash;

/* Blacklisted hosts list */
static BLHost::list_type blhosts(host_hash);

/* The network banner used by the blacklist for hosts. */
static NetBanner *hbanner = 0;
/* The network banner used by the blacklist for IPs */
static NetBanner *ipbanner = 0;

/*************************************************************************/

LOGLEVELH(DebugBlacklist, xlog::LogDebug, "Debug::Blacklist",
		"debug: blacklist: ");
LOGLEVELH(ErrorBlacklist, xlog::LogError, "Error::Blacklist",
		"Error: blacklist: ");

/*************************************************************************/

static bool blacklist_load(char const *filename, bool open_error);

/*************************************************************************/

BLHost::BLHost() : m_ban(0), m_mask(0), m_iter(blhosts.end())
{
}

/*************************************************************************/

BLHost::~BLHost()
{
	if (m_iter != blhosts.end())
		blhosts.erase(m_iter);
	
	if (m_mask)
	{
		if (epona->debuglevel())
			epona->log(DebugBlacklist(), "Removing mask %s from blacklist",
					m_ban->mask());
		
		if (!m_ban->banner()->remove(m_ban))
			if (epona->debuglevel())
				epona->log(DebugBlacklist(), "Blacklist mask %s is no longer "
						"in its network ban list (BUG?)", m_ban->mask());
	}
	
	if (m_ban)
		delete m_ban;
}

/*************************************************************************/

bool BLHost::mask(char const *mask)
{
	bool ip_mask = false;
	
	if (ipbanner)
	{	
		struct in_addr in;
		
		if (inet_aton(mask, &in))
		{
			if (!(m_ban = ipbanner->create()))
				return false;
			ip_mask = true;
		}
	}
	
	if (!m_ban && !(m_ban = hbanner->create()))
		return false;
	
	bool result = m_ban->mask(mask);
	
	if (result)
	{
		result = m_ban->banner()->add(m_ban);
		
		if (result)
		{
			m_ban->creator("Blacklist");
			
			char reason[256];
			snprintf(reason, sizeof(reason), epona->cd().blacklist.ban_reason,
					mask, mask, mask);
			m_ban->reason(reason);
			
			m_mask = mask;
			m_iter = blhosts.add_sorted(mask, this);
		}
	}
	
	if (epona->debuglevel())
	{
		if (result)
			epona->log(DebugBlacklist(), "Added %s mask %s to blacklist",
					(ip_mask ? "IP" : "host"), mask);
		else
			epona->log(DebugBlacklist(), "Invalid %s mask %s",
					(ip_mask ? "IP" : "host"), mask);
	}
	
	return result;
}

/*************************************************************************/

void get_blacklist_stats(long *nrec, long *memuse)
{
	*nrec = blhosts.size();
	*memuse = sizeof(BLHost) * *nrec;
}

/*************************************************************************/

/* Initializes the blacklist */
 
void blacklist_init()
{	
	if (epona->cd().blacklist.enable)
	{
		if (!(hbanner = netforce.ban_supports(UmaskNetBanner::IID)))
			epona->fatal("This build of Epona doesn't support usermask network"
					" banner -- please disable the blacklist feature");
		
		ipbanner = netforce.ban_supports(IPNetBanner::IID);
		
		conf::ConfStrCharA::container_type::const_iterator it, end;
			
		for (it = epona->cd().blacklist.diff->begin(),
				end = epona->cd().blacklist.diff->end(); it != end; ++it)
		{
			if (remove((char const *)*it) == -1 && errno != ENOENT)
				epona->fatalf("Can't delete blacklist file %s: %s",
						(char const *)*it, strerror(errno));
		}
		
		for (it = epona->cd().blacklist.master->begin(),
				end = epona->cd().blacklist.master->end(); it != end; ++it)
		{
			if (!blacklist_load((char const *)*it, true))
				epona->fatalf("Can't load blacklist file %s",
						(char const *)*it);
		}
		
		if (epona->debuglevel())
			epona->log(DebugBlacklist(), "Blacklist initialized");
	}
}

/*************************************************************************/

static bool blacklist_load(char const *filename, bool open_error)
{	
	ifstream in(filename, ios::in);
	
	if (!in) 
	{
		if (open_error)
		{
			epona->log_perror(ErrorBlacklist(), "Can't open blacklist file %s",
					filename);
		}
		return false;
	}
	
	char maskbuf[USERMAX+HOSTMAX+3], *mask; /* Prefix + user + @ + host + null */
	BLHost *bh;
	
	while (in)
	{	
		if (!in.getline(maskbuf, sizeof(maskbuf))) 
		{
			if (!in.eof()) {
				epona->log_perror(ErrorBlacklist(),
						"Read error from blacklist file %s", filename);
				return false;
			}
			continue;
		}
		
		mask = maskbuf;
		
		if (*mask == '-') 
		{
			++mask;
			
			BLHost::list_type::iterator it = blhosts.find_sorted(mask);
			
			if (it != blhosts.end())
				delete (*it);
		} else {
			if (*mask == '+')
				++mask;
			
			if (blhosts.find_sorted(mask) == blhosts.end())
			{
				bh = new BLHost;
				if (!bh->mask(mask))
					delete bh;
			}
		}
	}
	
	return true;
}

/*************************************************************************/

void blacklist_update()
{
	if (epona->cd().blacklist.enable)
	{
		conf::ConfStrCharA::container_type::const_iterator it, end;
			
		for (it = epona->cd().blacklist.diff->begin(),
				end = epona->cd().blacklist.diff->end(); it != end; ++it)
		{
			blacklist_load((char const*)*it, false);
			
			if (remove((char const*)*it) == -1 && errno != ENOENT)
				epona->log_perror(ErrorBlacklist(),
						"Can't delete blacklist file %s", (char const*)*it);
		}
	}
}

/*************************************************************************/

/* OperServ BLACKLIST */

void do_blacklist(User *u)
{
	char *cmd = strtok(NULL, " ");
	char *pattern = strtok(NULL, " ");
	
	if (!epona->cd().blacklist.enable) 
	{
		notice_lang(s_OperServ, u, OPER_BLACKLIST_DISABLED);
	    return;
	}
	
	if (!cmd || (stricmp(cmd, "LIST") && !pattern)) 
	{
		syntax_error(s_OperServ, u, "BLACKLIST", OPER_BLACKLIST_SYNTAX);
	} else if (!stricmp(cmd, "DEL")) 
	{
		BLHost::list_type::iterator it = blhosts.find_sorted(pattern);
		
		if (it == blhosts.end())
		{
			notice_lang(s_OperServ, u, OPER_BLACKLIST_NOT_FOUND, pattern);
			return;
		}
		
		delete (*it);
		notice_lang(s_OperServ, u, OPER_BLACKLIST_REMOVED, pattern);		
	} else if (!stricmp(cmd, "LIST")) 
	{	
		notice_lang(s_OperServ, u, OPER_BLACKLIST_HEADER);
		
		int count = 0, total = 0;
		
		BLHost::list_type::iterator it, end;
		
		for (it = blhosts.begin(), end = blhosts.end(); it != end; ++it)
		{
			if (pattern && !match_wild_nocase(pattern, (*it)->mask()))
				continue;
			++total;
			
			if (count >= epona->cd().blacklist.max)
				continue;
			++count;
			
			notice_lang(s_OperServ, u, OPER_BLACKLIST_LIST, (*it)->mask());
		}
		
		notice_lang(s_OperServ, u, OPER_BLACKLIST_FOOTER, count, total);
		
	} else {
		syntax_error(s_OperServ, u, "BLACKLIST", OPER_BLACKLIST_SYNTAX);
	}
}

/*************************************************************************/
