/* BotServ functions
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#include "tweak.h"

#include <system/headers.h>

#include <core/directives.h>
#include <core/epona.h>
#include <databases/database.h>
#include <log/levels.h>
#include <modules/manager.h>
 
#include "botserv.h"
#include "channels.h"
#include "chanserv.h"
#include "epona.h"
#include "nickserv.h"
#include "operserv.h"

#include "services.h"
#include "pseudo.h"

using namespace std;
using namespace core;

/*************************************************************************/

static UserData *get_user_data(Channel *c, User *u);

static void check_ban(ChannelInfo *ci, User *u, int kbbtype);
static void bot_kick(ChannelInfo *ci, User *u, int message, ...);
static void bot_raw_ban(User *requester, ChannelInfo *ci, char const *nick);
static void bot_raw_kick(User *requester, ChannelInfo *ci, char const *nick, char const *reason);
static void bot_raw_mode(User *requester, ChannelInfo *ci, char const *mode, char const *nick);
static void bot_raw_unban(ChannelInfo *ci, char const *nick);

static void do_help(User *u);
static void do_bot(User *u);
static void do_botlist(User *u);
static void do_assign(User *u);
static void do_unassign(User *u);
static void do_info(User *u);
static void do_set(User *u);
static void do_kickcmd(User *u);
static void do_badwords(User *u);
static void do_say(User *u);
static void do_act(User *u);

/*************************************************************************/

static int badwords_msgs[ListCmd::cmListCmdTotal];

/*************************************************************************/

class BadWordsCmd : public ListCmd
{
	public:
		/// Constructor.
		BadWordsCmd();
	protected:
		/// Overrides ListCmd::context().
		virtual char const *context(User *u) const { return m_context; }
		
		/// Overrides ListCmd::container().
		virtual container_type *container(User *u);
		
		/// Overrides ListCmd::allow_dispatch().
		virtual bool allow_dispatch(User *u, container_type &ctn,
				char const *sub);
		
		/// Overrides ListCmd::changeable().
		virtual bool changeable(User *u, container_type &ctn,
				char const *sub) const;
		
		/// Overrides ListCmd::add().
		virtual int add(User *u, container_type &ctn, char const *sub);
		
		/// Adds a record to the access list.
		int add_record(User *u, ChannelInfo *owner, char const *word,
				uint8_t type);
		
		/// Overrides ListCmd::del_record().
		virtual bool del_record(User *u, container_type &ctn,
				char const *sub, databases::Record *rec);
		
		/// Overrides ListCmd::list_record().
		virtual bool list_record(User *u, container_type &ctn,
				char const *sub, uint16_t num, databases::Record const *rec,
				uint16_t count);
	private:
		char *m_context;
		ChannelInfo *ci;
		
		BadWordsCmd(BadWordsCmd const &);
		BadWordsCmd &operator=(BadWordsCmd const &);
};

/*************************************************************************/

static Command cmds[] = {
    { "HELP",			    do_help,     NULL,  -1,                          -1,-1,-1,-1 },
    { "BOTLIST",		    do_botlist,  NULL,  BOT_HELP_BOTLIST,            -1,-1,-1,-1 },
    { "ASSIGN",			    do_assign,   NULL,  BOT_HELP_ASSIGN,             -1,-1,-1,-1 },
    { "UNASSIGN",		    do_unassign, NULL,  BOT_HELP_UNASSIGN,           -1,-1,-1,-1 },
    { "INFO",               do_info,     NULL,  BOT_HELP_INFO,               -1,-1,-1,-1 },
    { "SET",                do_set,      NULL,  BOT_HELP_SET,                
        -1, BOT_SERVADMIN_HELP_SET, 
        BOT_SERVADMIN_HELP_SET, BOT_SERVADMIN_HELP_SET },
    { "SET DONTKICKOPS",    NULL,        NULL,  BOT_HELP_SET_DONTKICKOPS,    -1,-1,-1,-1 },
    { "SET DONTKICKVOICES", NULL,        NULL,  BOT_HELP_SET_DONTKICKVOICES, -1,-1,-1,-1 },
    { "SET FANTASY",		NULL,		 NULL,	BOT_HELP_SET_FANTASY,		 -1,-1,-1,-1 },
    { "SET GREET",			NULL,		 NULL,	BOT_HELP_SET_GREET,		 	 -1,-1,-1,-1 },
    { "SET SYMBIOSIS",		NULL,		 NULL,	BOT_HELP_SET_SYMBIOSIS,		 -1,-1,-1,-1 },
    { "KICK",               do_kickcmd,  NULL,  BOT_HELP_KICK,               -1,-1,-1,-1 },
    { "KICK BADWORDS",      NULL,        NULL,  BOT_HELP_KICK_BADWORDS,      -1,-1,-1,-1 },
    { "KICK BOLDS",         NULL,        NULL,  BOT_HELP_KICK_BOLDS,         -1,-1,-1,-1 },
    { "KICK CAPS",          NULL,        NULL,  BOT_HELP_KICK_CAPS,          -1,-1,-1,-1 },
    { "KICK COLORS",        NULL,        NULL,  BOT_HELP_KICK_COLORS,        -1,-1,-1,-1 },
    { "KICK FLOOD",         NULL,        NULL,  BOT_HELP_KICK_FLOOD,         -1,-1,-1,-1 },
    { "KICK REPEAT",        NULL,        NULL,  BOT_HELP_KICK_REPEAT,        -1,-1,-1,-1 },
    { "KICK REVERSES",      NULL,        NULL,  BOT_HELP_KICK_REVERSES,      -1,-1,-1,-1 },
    { "KICK UNDERLINES",    NULL,        NULL,  BOT_HELP_KICK_UNDERLINES,    -1,-1,-1,-1 },
    { "BADWORDS",           do_badwords, NULL,  BOT_HELP_BADWORDS,           -1,-1,-1,-1 },
    { "SAY",				do_say,		 NULL,  BOT_HELP_SAY,				 -1,-1,-1,-1 },
    { "ACT",				do_act,		 NULL,  BOT_HELP_ACT,				 -1,-1,-1,-1 },
    
    /* Services admins commands */
    { "BOT",		do_bot,   is_services_admin,  -1,
		           	-1, BOT_SERVADMIN_HELP_BOT,
		           	BOT_SERVADMIN_HELP_BOT, BOT_SERVADMIN_HELP_BOT },
	{ "SET NOBOT", 	NULL,	 NULL,	-1,		     
	               	-1, BOT_SERVADMIN_HELP_SET_NOBOT,
	               	BOT_SERVADMIN_HELP_SET_NOBOT, BOT_SERVADMIN_HELP_SET_NOBOT },
	{ "SET PRIVATE", NULL,	 NULL,	-1,		     
	               	-1, BOT_SERVADMIN_HELP_SET_PRIVATE,
	               	BOT_SERVADMIN_HELP_SET_PRIVATE, BOT_SERVADMIN_HELP_SET_PRIVATE },
		
    { NULL }
};

/*************************************************************************/
/*************************************************************************/

/* BotServ initialization. */

void bs_init(void)
{
	Command *cmd;
    
    cmd = lookup_cmd(cmds, "SET SYMBIOSIS");
    if (cmd)
    	cmd->help_param1 = s_ChanServ;
}

/*************************************************************************/

bool botserv_load()
{
	if (!epona->modules().load(epona->cd().botserv.database_bots))
	{
		epona->log(ErrorBotServ(), "Can't load database module %s: %s",
				epona->cd().botserv.database_bots.value().c_str(),
				epona->modules().error().c_str());
		return false;
	}
	
	badwords_msgs[ListCmd::cmSyntax] = BOT_BADWORDS_SYNTAX;
	badwords_msgs[ListCmd::cmDisabled] = BOT_BADWORDS_DISABLED;
	badwords_msgs[ListCmd::cmEmpty] = BOT_BADWORDS_EMPTY;
	badwords_msgs[ListCmd::cmSyntaxDel] = BOT_BADWORDS_DEL_SYNTAX;
	badwords_msgs[ListCmd::cmNoSuchEntry] = BOT_BADWORDS_NO_SUCH_ENTRY;
	badwords_msgs[ListCmd::cmNoMatch] = BOT_BADWORDS_NO_MATCH;
	badwords_msgs[ListCmd::cmDeletedOne] = BOT_BADWORDS_DELETED_ONE;
	badwords_msgs[ListCmd::cmDeletedSeveral] = BOT_BADWORDS_DELETED_SEVERAL;
	badwords_msgs[ListCmd::cmNotFound] = BOT_BADWORDS_NOT_FOUND;
	badwords_msgs[ListCmd::cmDeleted] = BOT_BADWORDS_DELETED;
	badwords_msgs[ListCmd::cmSyntaxList] = -1; /* Not used for BADWORDS */
	badwords_msgs[ListCmd::cmListHeader] = BOT_BADWORDS_LIST_HEADER;
	badwords_msgs[ListCmd::cmListedOne] = BOT_BADWORDS_LISTED_ONE;
	badwords_msgs[ListCmd::cmListedSeveral] = BOT_BADWORDS_LISTED_SEVERAL;
	badwords_msgs[ListCmd::cmClearPartially] = BOT_BADWORDS_CLEAR_PARTIALLY;
	badwords_msgs[ListCmd::cmClear] = BOT_BADWORDS_CLEAR;
	
	return true;
}

/*************************************************************************/

void botserv_notify(modules::Module &sender, int msg, void *arg)
{
	switch (msg)
	{
		case modules::Module::mnLoaded:
			if (sender.name() == epona->cd().botserv.database_bots.value())
			{
				databases::Database *db = (databases::Database *)
						sender.factory(databases::Database::IID);
				if (!db)
				{
					epona->fatalf("Can't create database botserv_bots: "
							"module %s does not expose the Database interface",
							sender.name().c_str());	
				}
				
				/* Add a reference to the module */
				sender.references_add();
				/* Setup the database */
				bots.setup_database(db);
			}
		default:
			break;
	}
}

/*************************************************************************/

void botserv_start()
{
	if (!bots.db() || !bots.db()->start())
		epona->fatal("Can't start database botserv_bots");
}

/*************************************************************************/

/* Main BotServ routine. */

void botserv(User *u, char *buf)
{
    char *cmd, *s;

    cmd = strtok(buf, " ");

    if (!cmd) {
		return;
    } else if (stricmp(cmd, "\1PING") == 0) {
		s = strtok(NULL, "\1");
	    
		notice(s_BotServ, u->nick, "\1PING%s%s\1", (s ? " " : ""),
				(s ? s : ""));
    } else {
		run_cmd(s_BotServ, u, cmds, cmd);
    }

}

/*************************************************************************/

/* Handles all messages sent to bots. (Currently only answers to pings ;) */

void botmsgs(User *u, BotInfo *bi, char *buf)
{
	char *cmd = strtok(buf, " ");
	char *s;
	
	if (!cmd || !u) return;
	
	if (!stricmp(cmd, "\1PING")) 
	{
		s = strtok(NULL, "\1");
	    
		notice(bi->nick(), u->nick, "\1PING%s%s\1", (s ? " " : ""),
				(s ? s : ""));
    }
}

/*************************************************************************/

/* Handles all messages that are sent to registered channels where a 
 * bot is on.
 *
 */

void botchanmsgs(User *u, ChannelInfo *ci, char *buf)
{
	int c;
	int16_t cstatus = 0;
	char *cmd;
	UserData *ud;
	
	if (!u)
		return;
	
	/* Answer to ping if needed, without breaking the buffer. */
	if (!strnicmp(buf, "\1PING", 5))
		notice(ci->bi()->nick(), u->nick, buf);
	
	/* If it's a /me, cut the CTCP part at the beginning (not
	 * at the end, because one character just doesn't matter,
	 * but the ACTION may create strange behaviours with the
	 * caps or badwords kickers */
	if (!strnicmp(buf, "\1ACTION ", 8))
		buf += 8;
	
	/* Filter color and other control characters so that they don't
	 * get in the way of what we intend to do. */
	enum
	{
		ibBold  =     0x1,
		ibColor =     0x2,
		ibReverse =   0x4,
		ibUnderline = 0x8
	};
	
	uint8_t cchar = 0;
	char *end = buf, *nend = buf;
	
	while (*end)
	{
		switch (*end++)
		{
			case 2:
				cchar |= ibBold;
				continue;
			case 3:
				cchar |= ibColor;
				if (isdigit(*end) && isdigit(*(++end)))
					++end;
				continue;
			case 15:
				continue;
			case 22:
				cchar |= ibReverse;
				continue;
			case 31:
				cchar |= ibUnderline;
				continue;
			default:
				break;
		}
		
		*nend++ = *(end - 1);
	}
	
	*nend = 0;
	
	/* Now we can make kicker stuff. We try to order the checks
	 * from the fastest one to the slowest one, since there's
	 * no need to process other kickers if an user is kicked before
	 * the last kicker check.
	 *
	 * But FIRST we check whether the user is protected in any
	 * way.
	 */
	
	/* We first retrieve the user status on the channel if needed */
	if (ci->botflags() & (ChannelInfo::bfDontKickOps | ChannelInfo::bfDontKickVoices))
		cstatus = chan_get_user_status(ci->c, u);
	
	if (buf && !check_access(u, ci, CA_NOKICK) &&
#ifdef HAS_HALFOP
# ifdef IRC_UNREAL
			(!(ci->botflags() & ChannelInfo::bfDontKickOps) || !(cstatus & (CUS_HALFOP | CUS_OP | CUS_OWNER | CUS_PROTECT)))
# else
			(!(ci->botflags() & ChannelInfo::bfDontKickOps) || !(cstatus & (CUS_HALFOP | CUS_OP)))
# endif
#else
			(!(ci->botflags() & ChannelInfo::bfDontKickOps) || !(cstatus & CUS_OP))
#endif
			&& (!(ci->botflags() & ChannelInfo::bfDontKickVoices) || !(cstatus & CUS_VOICE)))
	{
		/* Bolds kicker */
		if ((ci->botflags() & ChannelInfo::bfKickBolds) && (cchar & ibBold))
		{
			check_ban(ci, u, KBB_BOLDS);
			bot_kick(ci, u, BOT_REASON_BOLD);
			return;
		}
		
		/* Color kicker */
		if ((ci->botflags() & ChannelInfo::bfKickColors) && (cchar & ibColor))
		{
			check_ban(ci, u, KBB_COLORS);
			bot_kick(ci, u, BOT_REASON_COLOR);
			return;
		}

		/* Reverses kicker */
		if ((ci->botflags() & ChannelInfo::bfKickReverses) && (cchar & ibReverse))
		{
			check_ban(ci, u, KBB_REVERSES);
			bot_kick(ci, u, BOT_REASON_REVERSE);
			return;	
		}
		
		/* Underlines kicker */
		if ((ci->botflags() & ChannelInfo::bfKickUnderlines) && (cchar & ibUnderline))
		{
			check_ban(ci, u, KBB_UNDERLINES);
			bot_kick(ci, u, BOT_REASON_UNDERLINE);	
			return;
		}	
		
		/* Caps kicker */
		if ((ci->botflags() & ChannelInfo::bfKickCaps) && ((c = strlen(buf)) >= ci->caps_min()))
		{
			int i = 0;
			char *s = buf;
			
			do {
				if (isupper(*s)) 
					i++;
			} while (*s++);
			  
			if (i >= ci->caps_min() && i*100/c>=ci->caps_percent())
			{
				check_ban(ci, u, KBB_CAPS);
				bot_kick(ci, u, BOT_REASON_CAPS);
				return;	
			}	
		}
		
		/* Bad words kicker */
		if (ci->botflags() & ChannelInfo::bfKickBadWords)
		{
			int mustkick = 0;
			BadWord *bw;
			
			ChannelInfo::badwords_type::const_iterator it, end;
				
			for (it = ci->badwords().begin(), end = ci->badwords().end();
					it != end; ++it)
			{
				bw = (BadWord *) (*it);
				
				if (!bw)
					continue; 
			
				if (bw->type() == BadWord::bwAny && stristr(buf, bw->word())) {
					mustkick = 1;
				} else if (bw->type() == BadWord::bwSingle) {
					int len = strlen(bw->word());
					
					if (!stricmp(buf, bw->word())) {
						mustkick = 1;
					/* two next if are quite odd isn't it? =) */
					} else if ((strchr(buf, ' ') == buf+len) && (stristr(buf, bw->word()) == buf)) {
						mustkick = 1;
					} else if ((strrchr(buf, ' ') == buf+strlen(buf)-len-1) && (stristr(buf, bw->word()) == buf+strlen(buf)-len)) {
						mustkick = 1;
					} else {
						char wordbuf[BadWord::word_len + 3];
						snprintf(wordbuf, sizeof(wordbuf), " %s ", bw->word());
						
						if (stristr(buf, wordbuf)) 
							mustkick = 1;
					}
				} else if (bw->type() == BadWord::bwStart) {
					int len = strlen(bw->word());
					
					if (!strnicmp(buf, bw->word(), len)) { 
						mustkick = 1;
					} else {
						char wordbuf[BadWord::word_len + 2];
						snprintf(wordbuf, sizeof(wordbuf), " %s", bw->word());
						
						if (stristr(buf, wordbuf)) 
							mustkick = 1;
					}
				} else if (bw->type() == BadWord::bwEnd) {
					int len = strlen(bw->word());
					
					if (!strnicmp(buf+strlen(buf)-len, bw->word(), len)) {
						mustkick = 1; 
					} else {
						char wordbuf[BadWord::word_len + 2];
						snprintf(wordbuf, sizeof(wordbuf), "%s ", bw->word());
						
						if (stristr(buf, wordbuf)) 
							mustkick = 1;
					}
				}
				
				if (mustkick)
				{
					check_ban(ci, u, KBB_BADWORDS);
					if (epona->cd().botserv.soft_reason)
					bot_kick(ci, u, BOT_REASON_BADWORD_GENTLE);
					else
					bot_kick(ci, u, BOT_REASON_BADWORD, bw->word());
					return;
				}
			} 
		}
		
		/* Flood kicker */
		if (ci->botflags() & ChannelInfo::bfKickFlood)
		{
			time_t now = time(NULL);
			
			ud = get_user_data(ci->c, u);
			if (!ud) 
				return;
			
			if (now - ud->last_start > ci->flood_secs()) 
			{
				ud->last_start = time(NULL);
				ud->lines = 0;
			}
			
			ud->lines++;
			if (ud->lines >= ci->flood_lines())
			{
				check_ban(ci, u, KBB_FLOOD);
				bot_kick(ci, u, BOT_REASON_FLOOD);
				return;	
			}
		}
		
		/* Repeat kicker */
		if (ci->botflags() & ChannelInfo::bfKickRepeats)
		{
			ud = get_user_data(ci->c, u);
			if (!ud) return;
			
			if (ud->lastline && stricmp(ud->lastline, buf)) {
				free(ud->lastline);
				ud->lastline = sstrdup(buf);
				ud->times = 0;	
			} else {
				if (!ud->lastline) 
					ud->lastline = sstrdup(buf);
				ud->times++;
			}
			
			if (ud->times >= ci->repeats())
			{
				check_ban(ci, u, KBB_REPEAT);
				bot_kick(ci, u, BOT_REASON_REPEAT);
				return;	
			}
		}
	}
	
	/* Fantaisist commands */
	
	/* First remove trailing spaces */
	while (isspace(*buf))
		++buf;
	
	if ((ci->botflags() & ChannelInfo::bfFantasy) && *buf == '!' && check_access(u, ci, CA_FANTASIA)) 
	{
		cmd = strtok(buf, " ");
		
		if (cmd) {
#ifdef IRC_UNREAL
			if (!stricmp(cmd, "!deowner") || !stricmp(cmd, "!-q")) {
				if (is_owner(u, ci))
					bot_raw_mode(u, ci, "-q", u->nick);
			} else
#endif
			if (!stricmp(cmd, "!kb") || !stricmp(cmd, "!kickban")) {
				char *target = strtok(NULL, " ");
				char *reason = strtok(NULL, "");
				
				if (!target && check_access(u, ci, CA_BANME)) {
					bot_raw_ban(u, ci, u->nick);
					bot_raw_kick(u, ci, u->nick, "Requested");
				} else if (target && check_access(u, ci, CA_BAN)) {
					if (!stricmp(target, ci->bi()->nick())) {
						bot_raw_ban(u, ci, u->nick);
						bot_raw_kick(u, ci, u->nick, "Oops!");
					} else {
						bot_raw_ban(u, ci, target);
						if (!reason)
							bot_raw_kick(u, ci, target, "Requested");
						else
							bot_raw_kick(u, ci, target, reason);	
					}
				}
			} else if (!stricmp(cmd, "!kick") || !stricmp(cmd, "!k")) {
				char *target = strtok(NULL, " ");
				char *reason = strtok(NULL, "");
				
				if (!target && check_access(u, ci, CA_KICKME)) {
					bot_raw_kick(u, ci, u->nick, "Requested");
				} else if (target && check_access(u, ci, CA_KICK)) {
					if (!stricmp(target, (ci->bi())->nick()))
						bot_raw_kick(u, ci, u->nick, "Oops!");
					else if (!reason)
						bot_raw_kick(u, ci, target, "Requested");
					else
						bot_raw_kick(u, ci, target, reason);	
				}
#ifdef IRC_UNREAL
			} else if (!stricmp(cmd, "!owner") || !stricmp(cmd, "!+q")) {
				if (is_owner(u, ci))
					bot_raw_mode(u, ci, "+q", u->nick);
#endif
			} else if (!stricmp(cmd, "!seen")) {
				char *target = strtok(NULL, " ");
				char buf[BUFSIZE];
				
				if (target) {
					User *u2;
					NickAlias *na;
					Access *access;
					
					if (!stricmp(ci->bi()->nick(), target))
					{
						/* If we look for the bot */
						snprintf(buf, sizeof(buf), getstring(u->na, BOT_SEEN_BOT), u->nick);
						send_cmd(ci->bi()->nick(), "PRIVMSG %s :%s", ci->name(), buf);
					} else if (!(na = aliases.find(target))) {
						/* If the nick is not registered or forbidden */
						snprintf(buf, sizeof(buf), getstring(u->na, BOT_SEEN_UNKNOWN), target);
						send_cmd(ci->bi()->nick(), "PRIVMSG %s :%s", ci->name(), buf);
					} else if ((u2 = ng_on_chan(ci->c, na->ng()))) {
						/* If the nick we're looking for is on the channel,
						 * there are three possibilities: it's yourself,
						 * it's the nick we look for, it's an alias of the
						 * nick we look for.
						 */
						if (u == u2 || (u->na && u->na->ng() == na->ng())) 
						snprintf(buf, sizeof(buf), getstring(u->na, BOT_SEEN_YOU), u->nick);
						else if (!stricmp(u2->nick, target))
						snprintf(buf, sizeof(buf), getstring(u->na, BOT_SEEN_ON_CHANNEL), u2->nick);
						else
						snprintf(buf, sizeof(buf), getstring(u->na, BOT_SEEN_ON_CHANNEL_AS), target, u2->nick);
						send_cmd(ci->bi()->nick(), "PRIVMSG %s :%s", ci->name(), buf);
					} else if ((access = get_access_entry(na->ng(), ci))) {
						/* User is on the access list but not present actually.
						   Special case: if access->last_seen is 0 it's that we
						   never seen the user.
						 */
						if (access->last_seen()) {
							char durastr[192];
							duration(u->na, durastr, sizeof(durastr), 
									time(NULL) - access->last_seen());
							snprintf(buf, sizeof(buf), getstring(u->na, BOT_SEEN_ON), target, durastr);
						} else {
							snprintf(buf, sizeof(buf), getstring(u->na, BOT_SEEN_NEVER), target);
						}
						send_cmd(ci->bi()->nick(), "PRIVMSG %s :%s", ci->name(), buf);
					} else {
						/* All other cases */
						snprintf(buf, sizeof(buf), getstring(u->na, BOT_SEEN_UNKNOWN), target);
						send_cmd(ci->bi()->nick(), "PRIVMSG %s :%s", ci->name(), buf);
					}
				}
			} else if (!stricmp(cmd, "!unban") && check_access(u, ci, CA_OPDEOP)) {
				char *target = strtok(NULL, " ");
				
				if (!target)
					bot_raw_unban(ci, u->nick);
				else
					bot_raw_unban(ci, target);
			} else {
				CSModeUtil *util = csmodeutils;
				
				do {
					if (!stricmp(cmd, util->bsname) ||
							!stricmp(cmd+1, util->mode)) {
						char *target = strtok(NULL, " ");
						
						if (!target && check_access(u, ci, util->levelself))
							bot_raw_mode(u, ci, util->mode, u->nick);
						else if (target && check_access(u, ci, util->level))
							bot_raw_mode(u, ci, util->mode, target);	
					}
				} while ((++util)->name != NULL);
			}
		}	
	}
}

/*************************************************************************/

/* Unassign a bot from a channel. Assumes u, ci and ci->bi() are not NULL */

/* static void unassign(User *u, ChannelInfo *ci) 
{
	if (ci->c && ci->c->usercount >= epona->cd().botserv.min_users)
		send_cmd(ci->bi()->nick(), "PART %s :UNASSIGN from %s", ci->name(), u->nick);

	ci->bi()->chancount(ci->bi()->chancount() - 1);
	ci->bi(NULL);
}*/

/*************************************************************************/

/* Returns ban data associated with an user if it exists, allocates it
   otherwise. */

static BanData *get_ban_data(Channel *c, User *u)
{
	char mask[BUFSIZE];
	BanData *bd, *next;
	time_t now = time(NULL);
	
	if (!c || !u) 
		return NULL;
	
	snprintf(mask, sizeof(mask), "%s@%s", u->username, GetHost(u));
	
	for (bd = c->bd; bd; bd = next) {
		if (now - bd->last_use > epona->cd().botserv.keep_data)
		{
			if (bd->next)
				bd->next->prev = bd->prev;
			if (bd->prev)
				bd->prev->next = bd->next;
			else
				c->bd = bd->next;
			if (bd->mask) free(bd->mask);
			next = bd->next;
			free(bd);
			continue;
		}
		if (!stricmp(bd->mask, mask)) 
		{
			bd->last_use = now;
			return bd;
		}
		next = bd->next;
	}
		
	/* If we fall here it is because we haven't found the record */
	bd = static_cast<BanData *>(scalloc(sizeof(BanData), 1));
	bd->mask = sstrdup(mask);
	bd->last_use = now;
	
	bd->prev = NULL;
	bd->next = c->bd;
	if (bd->next) 
		bd->next->prev = bd;
	c->bd = bd;
	
	return bd;
}

/*************************************************************************/

/* Returns BotServ data associated with an user on a given channel. 
 * Allocates it if necessary.
 */

static UserData *get_user_data(Channel *c, User *u)
{
	Channel::c_userlist *user;
	
	if (!c || !u) return NULL;
	
	for (user = c->users; user; user = user->next) {
		if (user->user == u) {
			if (user->ud) 
			{
				time_t now = time(NULL);
				
				/* Checks whether data is obsolete */
				if (now - user->ud->last_use > epona->cd().botserv.keep_data)
				{
					if (user->ud->lastline) 
						free(user->ud->lastline);
					/* We should not free and realloc, but reset to 0
					   instead. */
					memset(user->ud, 0, sizeof(UserData));
					user->ud->last_use = now;
				}
				
				return user->ud;	
			} else {
				user->ud = static_cast<UserData *>(scalloc(sizeof(UserData), 1));
				user->ud->last_use = time(NULL);
				return user->ud;	
			}	
		}
	}
	
	return NULL;
}

/*************************************************************************/

/* Makes the bot join a channel and op himself. */

void bot_join(ChannelInfo *ci)
{	
	int i;
	
	if (!ci || !ci->c || !ci->bi()) return;	
	
	if (epona->cd().botserv.smart_join)
	{	
		/* We check for bans */
		int count = ci->c->bancount;
		if (count) {
		char botmask[BUFSIZE];
		
		char **bans = static_cast<char **>(smalloc(sizeof(char *) * count));
		char const *av[3];
		
		memcpy(bans, ci->c->bans, sizeof(char *) * count);
		snprintf(botmask, sizeof(botmask), "%s!%s@%s", ci->bi()->nick(),
				ci->bi()->username(), ci->bi()->hostname());

		av[0] = ci->c->name;
		av[1] = "-b";
		for (i = 0; i < count; i++) {
	    	if (match_wild_nocase(ci->c->bans[i], botmask)) {
				s_mode(ci->bi()->nick(), ci->name(), "-b", bans[i]);
				av[2] = misc::nstrdup(bans[i]);
				do_cmode(ci->bi()->nick(), 3, av);
				delete [] av[2];
	    	}
		}
		free(bans);
		}
		
		/* Should we be invited? */
		if ((ci->c->mode & CMODE_i) || (ci->c->limit && ci->c->usercount >= ci->c->limit))
			send_cmd(NULL, "NOTICE @%s :%s invited %s into the channel.", 
					ci->c->name, ci->bi()->nick(), ci->bi()->nick());
	}
	
#ifdef IRC_BAHAMUT
	send_cmd(ci->bi()->nick(), "SJOIN %ld %s", (long) ci->c->creation_time,
			ci->c->name);
#else
	send_cmd(ci->bi()->nick(), "JOIN %s", ci->c->name);
#endif

#ifdef IRC_UNREAL
	send_cmd(ci->bi()->nick(), "MODE %s +ao %s %s", ci->c->name, ci->bi()->nick(),
			ci->bi()->nick());
#else
	s_mode(ci->bi()->nick(), ci->c->name, "+o", ci->bi()->nick());
#endif
}

/*************************************************************************/

/* This makes the bot rejoin all channel he is on when he gets killed
 * or changed.
 */
 
void bot_rejoin_all(BotInfo *bi)
{
	ChannelInfo *ci;
	
	Channels::const_iterator it, end;
	
	for (it = channels.begin(), end = channels.end(); it != end; ++it)
	{
		ci = (ChannelInfo *)(*it);
	
		if (ci->bi() == bi && ci->c &&
	 			(ci->c->usercount >= epona->cd().botserv.min_users))
	    	bot_join(ci);
	}
}

/*************************************************************************/

/* This makes a ban if the user has to have one. In every cases it increments
   the kick count for the user. */

static void check_ban(ChannelInfo *ci, User *u, int kbbtype)
{
	BanData *bd = get_ban_data(ci->c, u);
	
	if (!bd) 
		return;
	
	bd->kbb[kbbtype]++;
	if (bd->kbb[kbbtype] == ci->kbb(kbbtype))
	{
		char const *av[3];
		char mask[BUFSIZE];
		
		bd->kbb[kbbtype] = 0;
		
		av[0] = ci->name();
		av[1] = "+b";
		get_idealban(ci, u, mask, sizeof(mask));
		av[2] = mask;
		s_mode(ci->bi()->nick(), av[0], "+b", av[2]);
		do_cmode(ci->bi()->nick(), 3, av);
	} 
}

/*************************************************************************/

/* This makes a bot kick an user. Works somewhat like notice_lang in fact ;) */

static void bot_kick(ChannelInfo *ci, User *u, int message, ...)
{
    va_list args;
    char buf[1024];
    const char *fmt;
    char const *av[3];

    if (!ci || !ci->bi() || !ci->c || !u) return;
    
    va_start(args, message);
    fmt = getstring(u->na, message);
    if (!fmt) return;
    vsnprintf(buf, sizeof(buf), fmt, args);
    va_end(args);
    
    av[0] = ci->name();
    av[1] = u->nick;
    av[2] = buf;
    send_cmd(ci->bi()->nick(), "KICK %s %s :%s", av[0], av[1], av[2]);
    do_kick(ci->bi()->nick(), 3, av);
}

/*************************************************************************/

/* Makes a simple ban. */

static void bot_raw_ban(User *requester, ChannelInfo *ci, char const *nick)
{
	char const *av[3];
	char mask[BUFSIZE];
	User *u = finduser(nick);
		
	if (!u) 
		return;
	
	if ((ci->flags() & ChannelInfo::cfPeace) && stricmp(requester->nick, nick) && (get_access(u, ci) >= get_access(requester, ci)))
		return;
		
	av[0] = ci->name();
	av[1] = "+b";
	get_idealban(ci, u, mask, sizeof(mask));
	av[2] = mask;
	s_mode(ci->bi()->nick(), av[0], "+b", av[2]);
	do_cmode(ci->bi()->nick(), 3, av);
}

/*************************************************************************/

/* Makes a kick with a "dynamic" reason ;) */

static void bot_raw_kick(User *requester, ChannelInfo *ci, char const *nick, 
		char const *reason)
{
    char const *av[3];
	User *u = finduser(nick);
    
    if (!u || !is_on_chan(ci->c, u)) 
    	return;

	if ((ci->flags() & ChannelInfo::cfPeace) && stricmp(requester->nick, nick) && 
			(get_access(u, ci) >= get_access(requester, ci)))
		return;

    av[0] = ci->name();
    av[1] = nick;
	
	char rbuf[301];
	
	if (!reason) {
		av[2] = (ci->bi())->nick();
	} else {
		if (strlen(reason) > 300)
		{
			strncpy(rbuf, reason, 300);
			rbuf[300] = '\0';
			av[2] = rbuf;
		} else
		{
			av[2] = reason;
		}
	}
	
    send_cmd((ci->bi())->nick(), "KICK %s %s :%s", av[0], av[1], av[2]);
    do_kick((ci->bi())->nick(), 3, av);
}

/*************************************************************************/

/* Makes a mode operation on a channel for a nick */

static void bot_raw_mode(User *requester, ChannelInfo *ci, char const *mode,
		char const *nick)
{
	char const *av[3];
	User *u = finduser(nick);
	
	if (!u || !is_on_chan(ci->c, u))
		return;
		
	if (*mode == '-' && (ci->flags() & ChannelInfo::cfPeace) && stricmp(requester->nick, nick) && (get_access(u, ci) >=  get_access(requester, ci)))
		return;
	
	av[0] = ci->name();
	av[1] = mode;
	av[2] = nick;
	
	s_mode(ci->bi()->nick(), av[0], av[1], av[2]);
	do_cmode(ci->bi()->nick(), 3, av);
}

/*************************************************************************/

/* Removes all bans for a nick on a channel */

static void bot_raw_unban(ChannelInfo *ci, char const *nick)
{
#if !defined(HAS_SVSMODE_UNBAN) || defined(HAS_VHOST)
	int count, i;
	char const *av[3]; 
	char **bans;
	User *u;
#endif
	
	if (!ci || !ci->c || !ci->bi() || !nick) 
		return;
#if defined(HAS_SVSMODE_UNBAN) && !defined(HAS_VHOST)
	if (!finduser(nick)) 
		return;
#else
	if (!(u = finduser(nick))) 
		return;
#endif

#if defined(HAS_SVSMODE_UNBAN) && !defined(HAS_VHOST)
	send_cmd(epona->cd().services.sname, "SVSMODE %s -b %s",
			ci->c->name, nick);
#else
	av[0] = ci->name();
	av[1] = "-b";
	
	count = ci->c->bancount;
	
	bans = new char *[count];
	memcpy(bans, ci->c->bans, sizeof(char *) * count);
	
	for (i = 0; i < count; i++) 
	{
	    if (match_usermask(bans[i], u, true)) 
	    {
			s_mode(ci->bi()->nick(), ci->name(), "-b", bans[i]);
			av[2] = misc::nstrdup(bans[i]);
			do_cmode(ci->bi()->nick(), 3, av);
			delete [] av[2];
		}
	}
	
	delete [] bans;
#endif
}

/*************************************************************************/
/*************************************************************************/

static void do_help(User *u)
{
    char *cmd = strtok(NULL, "");

    if (!cmd) {
		notice_help(s_BotServ, u, BOT_HELP,
				(uint16_t) epona->cd().botserv.min_users);
		if (is_services_oper(u))
	    	notice_help(s_BotServ, u, BOT_SERVADMIN_HELP);
    } else {
		help_cmd(s_BotServ, u, cmds, cmd);
    }
}

/*************************************************************************/

static void do_bot(User *u)
{
	BotInfo *bi;
	char *cmd = strtok(NULL, " ");
	
	if (!cmd)
		syntax_error(s_BotServ, u, "BOT", BOT_BOT_SYNTAX);
	else if (!stricmp(cmd, "ADD"))
	{
		char *nick = strtok(NULL, " ");	
		char *user = strtok(NULL, " ");
		char *host = strtok(NULL, " ");
		char *real = strtok(NULL, "");
		
		if (!nick || !user || !host || !real)
			syntax_error(s_BotServ, u, "BOT", BOT_BOT_SYNTAX);
		else if (!bots.writable())
			notice_lang(s_BotServ, u, BOT_BOT_READONLY);
		else if (bots.find(nick))
			notice_lang(s_BotServ, u, BOT_BOT_ALREADY_EXISTS, nick);
		else {
			NickAlias *na;
			
			/* We check whether the nick is registered, and drop it if so. */
			if ((na = aliases.find(nick))) 
			{
				if (epona->cd().nickserv.protect_admins &&
						nick_is_services_admin(na->ng()) &&
						!is_services_root(u)) 
				{
					notice_lang(s_BotServ, u, PERMISSION_DENIED);
					return;
				}
				
				if (!bots.nick_cleanup(nick))
				{
					notice_lang(s_BotServ, u, BOT_BOT_READONLY);
					return;
				}
			}
			
			bi = (BotInfo *)bots.create();
			
			if (!bi) 
			{
				notice_lang(s_BotServ, u, BOT_BOT_CREATION_FAILED);
				return;	
			}
			
			bi->nick(nick);
			bi->username(user);
			bi->hostname(host);
			bi->realname(real);
			
			if (!bots.add(bi))
			{
				notice_lang(s_BotServ, u, BOT_BOT_CREATION_FAILED);
				delete bi;
				
				return;
			}
			
			notice_lang(s_BotServ, u, BOT_BOT_ADDED, bi->nick(), bi->username(),
					bi->hostname(), bi->realname());
		}
	}
	else if (!stricmp(cmd, "CHANGE"))
	{
		char *oldnick = strtok(NULL, " ");
		char *nick = strtok(NULL, " ");	
		char *user = strtok(NULL, " ");
		char *host = strtok(NULL, " ");
		char *real = strtok(NULL, "");
		
		if (!oldnick || !nick)
			syntax_error(s_BotServ, u, "BOT", BOT_BOT_SYNTAX);
		else if (!bots.writable())
			notice_lang(s_BotServ, u, BOT_BOT_READONLY);
		else if (!(bi = bots.find(oldnick)))
			notice_lang(s_BotServ, u, BOT_DOES_NOT_EXIST, oldnick);
		else
		{
			/* Check if there *are* changes.
			 * Case sensitive because we may want to change just the case.
			 * And finally, we must ensure the nick is not already
			 * used by another bot.
			 */
			if (!strcmp(bi->nick(), nick) 
					&& ((user) ? !strcmp(bi->username(), user) : 1)
					&& ((host) ? !strcmp(bi->hostname(), host) : 1)
					&& ((real) ? !strcmp(bi->realname(), real) : 1))
			{
				notice_lang(s_BotServ, u, BOT_BOT_ANY_CHANGES);
				return;
			}
			
			if (stricmp(bi->nick(), nick) && bots.find(nick))
			{
				notice_lang(s_BotServ, u, BOT_BOT_ALREADY_EXISTS, nick);
				return;
			}
			
			if (stricmp(bi->nick(), nick))
			{	
				NickAlias *na = aliases.find(nick);
				
				if (na && epona->cd().nickserv.protect_admins &&
						nick_is_services_admin(na->ng()) &&
						!is_services_root(u)) 
				{
					notice_lang(s_BotServ, u, PERMISSION_DENIED);
					return;
				}
				
				if (!bots.nick_cleanup(nick))
				{
					notice_lang(s_BotServ, u, BOT_BOT_READONLY);
					return;
				}
			}
			
			bi->nick(nick);
				
			if (user)
				bi->username(user);
			if (host)
				bi->hostname(host);
			if (real)
				bi->realname(real);
			
			bi->update_client();
			
			notice_lang(s_BotServ, u, BOT_BOT_CHANGED, oldnick, bi->nick(),
					bi->username(), bi->hostname(), bi->realname());
		}
	}
	else if (!stricmp(cmd, "DEL"))
	{
		char *nick = strtok(NULL, " ");
		
		if (!nick)
			syntax_error(s_BotServ, u, "BOT", BOT_BOT_SYNTAX);
		else if (!bots.writable())
			notice_lang(s_BotServ, u, BOT_BOT_READONLY);
		else if (!(bi = bots.find(nick)))
			notice_lang(s_BotServ, u, BOT_DOES_NOT_EXIST, nick);
		else
		{
			if (!bots.remove(bi))
			{
				notice_lang(s_BotServ, u, BOT_BOT_DELETION_FAILED);
				return;
			}
			
			notice_lang(s_BotServ, u, BOT_BOT_DELETED, nick);
		}
	}
	else if (!stricmp(cmd, "LIST"))
		do_botlist(u);
	else 
		syntax_error(s_BotServ, u, "BOT", BOT_BOT_SYNTAX);
}

/*************************************************************************/

static void do_botlist(User *u)
{
	size_t count = 0;
	BotInfo *bi;
	
	if (bots.empty()) 
	{
		notice_lang(s_BotServ, u, BOT_BOTLIST_EMPTY);
		return;
	}
	
	Bots::const_niterator nit, nend;
	
	for (nit = bots.nbegin(), nend = bots.nend(); nit != nend; ++nit)
	{
		bi = (BotInfo *)(*nit);
		
		if (!(bi->flags() & BotInfo::bfPrivate)) 
		{
			if (!count)
				notice_lang(s_BotServ, u, BOT_BOTLIST_HEADER);
			count++;
			notice_user(s_BotServ, u, "   %-15s  (%s@%s)", bi->nick(),
					bi->username(), bi->hostname());
		}
	}
	
	if (is_oper(u) && count < bots.size()) 
	{
		notice_lang(s_BotServ, u, BOT_BOTLIST_PRIVATE_HEADER);
		
		for (nit = bots.nbegin(), nend = bots.nend(); nit != nend; ++nit)
		{
			bi = (BotInfo *)(*nit);
			
			if (bi->flags() & BotInfo::bfPrivate) 
			{
				notice_user(s_BotServ, u, "   %-15s  (%s@%s)",
						bi->nick(), bi->username(), bi->hostname());
				count++;
			}
		}
	}
	
	if (!count) 
		notice_lang(s_BotServ, u, BOT_BOTLIST_EMPTY);
	else
		notice_lang(s_BotServ, u, BOT_BOTLIST_FOOTER, count);
}

/*************************************************************************/

static void do_assign(User *u)
{
	char *chan = strtok(NULL, " ");
	char *nick = strtok(NULL, " ");
	BotInfo *bi;
	ChannelInfo *ci;
	
	if (!channels.writable())
		notice_lang(s_BotServ, u, BOT_ASSIGN_READONLY); 
	else if (!chan || !nick)
		syntax_error(s_BotServ, u, "ASSIGN", BOT_ASSIGN_SYNTAX);
	else if (!(bi = bots.find(nick)))
		notice_lang(s_BotServ, u, BOT_DOES_NOT_EXIST, nick);
	else if ((bi->flags() & BotInfo::bfPrivate) && !is_oper(u))
		notice_lang(s_BotServ, u, PERMISSION_DENIED);
	else if (!(ci = channels.find(chan)))
		notice_lang(s_BotServ, u, CHAN_X_NOT_REGISTERED, chan);
	else if (ci->bi())
		notice_lang(s_BotServ, u, BOT_ASSIGN_ALREADY, chan);
	else if ((ci->botflags() & ChannelInfo::bfNoBot) || (!check_access(u, ci, CA_ASSIGN) && !is_services_admin(u)))
		notice_lang(s_BotServ, u, PERMISSION_DENIED);
	else
	{
		ci->bi(bi);
		notice_lang(s_BotServ, u, BOT_ASSIGN_ASSIGNED, bi->nick(), ci->name());
	}
}

/*************************************************************************/

static void do_unassign(User *u)
{
	char *chan = strtok(NULL, " ");
	ChannelInfo *ci;
	
	if (!channels.writable())
		notice_lang(s_BotServ, u, BOT_ASSIGN_READONLY); 
	else if (!chan)
		syntax_error(s_BotServ, u, "UNASSIGN", BOT_UNASSIGN_SYNTAX);
	else if (!(ci = channels.find(chan)))
		notice_lang(s_BotServ, u, CHAN_X_NOT_REGISTERED, chan);
    else if (!is_services_admin(u) && !check_access(u, ci, CA_ASSIGN))
		notice_lang(s_BotServ, u, ACCESS_DENIED);
	else
	{
		if (ci->bi()) 
			ci->bi(0);
		notice_lang(s_BotServ, u, BOT_UNASSIGN_UNASSIGNED, ci->name());
	}
}

/*************************************************************************/

static void send_bot_channels(User *u, BotInfo *bi)
{
	ChannelInfo *ci;
	char buf[307], *end;
	
	*buf = 0;
	end = buf;
	
	Channels::const_niterator it, nend;
	
	for (it = channels.nbegin(), nend = channels.nend(); it != nend; ++it)
	{
		ci = (ChannelInfo *)(*it);
	
    	if (ci->bi() == bi) {
    		if (strlen(buf) + strlen(ci->name()) > 300) {
    			notice_user(s_BotServ, u, buf);
    			*buf = 0;
				end = buf;
			}
			end += snprintf(end, sizeof(buf)-(end-buf), " %s ", ci->name());
    	}
    }
    
    if (*buf) 
		notice_user(s_BotServ, u, buf);
}

static void do_info(User *u)
{
	BotInfo *bi;
	ChannelInfo *ci;
	char *query = strtok(NULL, " ");
	
	int need_comma = 0, is_servadmin = is_services_admin(u);
	char buf[BUFSIZE], *end;
	const char *commastr = getstring(u->na, COMMA_SPACE);
		
	if (!query)
		syntax_error(s_BotServ, u, "INFO", BOT_INFO_SYNTAX);
	else if ((bi = bots.find(query))) {
		char buf[BUFSIZE];
		struct tm *tm;
		time_t t;
		
		notice_lang(s_BotServ, u, BOT_INFO_BOT_HEADER, bi->nick());
		notice_lang(s_BotServ, u, BOT_INFO_BOT_MASK, bi->username(), bi->hostname());
		notice_lang(s_BotServ, u, BOT_INFO_BOT_REALNAME, bi->realname());
		t = bi->created();
		tm = localtime(&t);
		strftime_lang(buf, sizeof(buf), u, STRFTIME_DATE_TIME_FORMAT, tm);
		notice_lang(s_BotServ, u, BOT_INFO_BOT_CREATED, buf);
		notice_lang(s_BotServ, u, BOT_INFO_BOT_OPTIONS, getstring(u->na,
				(bi->flags() & BotInfo::bfPrivate) ? BOT_INFO_OPT_PRIVATE :
				BOT_INFO_OPT_NONE));
		notice_lang(s_BotServ, u, BOT_INFO_BOT_USAGE, bi->chancount());
		
		if (is_services_admin(u))
			send_bot_channels(u, bi);
	} else if ((ci = channels.find(query))) {
		if (!is_servadmin && !is_owner(u, ci)) {
			notice_lang(s_BotServ, u, PERMISSION_DENIED);
			return;	
		}
		
		notice_lang(s_BotServ, u, BOT_INFO_CHAN_HEADER, ci->name());
		if (ci->bi()) notice_lang(s_BotServ, u, BOT_INFO_CHAN_BOT, ci->bi()->nick());
		else notice_lang(s_BotServ, u, BOT_INFO_CHAN_BOT_NONE);
		
		if (ci->botflags() & ChannelInfo::bfKickBadWords) 
		{
			if (ci->kbb(KBB_BADWORDS)) 
				notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_BADWORDS_BAN, 
						getstring(u->na, BOT_INFO_ACTIVE), 
						ci->kbb(KBB_BADWORDS));
			else 
				notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_BADWORDS, 
						getstring(u->na, BOT_INFO_ACTIVE));
		} else {
			notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_BADWORDS, 
					getstring(u->na, BOT_INFO_INACTIVE));
		}
		if (ci->botflags() & ChannelInfo::bfKickBolds)
		{
			if (ci->kbb(KBB_BOLDS)) 
				notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_BOLDS_BAN, 
						getstring(u->na, BOT_INFO_ACTIVE), ci->kbb(KBB_BOLDS));
			else 
				notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_BOLDS, 
						getstring(u->na, BOT_INFO_ACTIVE));
		} else {
			notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_BOLDS, 
					getstring(u->na, BOT_INFO_INACTIVE));
		}
		if (ci->botflags() & ChannelInfo::bfKickCaps)
		{
			if (ci->kbb(KBB_CAPS)) 
				notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_CAPS_BAN, 
						getstring(u->na, BOT_INFO_ACTIVE), ci->kbb(KBB_CAPS),
						ci->caps_min(), ci->caps_percent());
			else 
				notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_CAPS_ON,
						getstring(u->na, BOT_INFO_ACTIVE), ci->caps_min(),
						ci->caps_percent());
		} else {
			notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_CAPS_OFF,
					getstring(u->na, BOT_INFO_INACTIVE));
		}
		if (ci->botflags() & ChannelInfo::bfKickColors)
		{
			if (ci->kbb(KBB_COLORS)) 
				notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_COLORS_BAN,
						getstring(u->na, BOT_INFO_ACTIVE), 
						ci->kbb(KBB_COLORS));
			else 
				notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_COLORS,
						getstring(u->na, BOT_INFO_ACTIVE));
		} else { 
			notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_COLORS,
					getstring(u->na, BOT_INFO_INACTIVE));
		}
		if (ci->botflags() & ChannelInfo::bfKickFlood)
		{
			if (ci->kbb(KBB_FLOOD)) 
				notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_FLOOD_BAN,
						getstring(u->na, BOT_INFO_ACTIVE), ci->kbb(KBB_FLOOD),
						ci->flood_lines(), ci->flood_secs());
			else 
				notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_FLOOD_ON, 
						getstring(u->na, BOT_INFO_ACTIVE), ci->flood_lines(),
						ci->flood_secs());
		} else { 
			notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_FLOOD_OFF, 
					getstring(u->na, BOT_INFO_INACTIVE));
		}
		if (ci->botflags() & ChannelInfo::bfKickRepeats)
		{
			if (ci->kbb(KBB_REPEAT)) 
					notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_REPEAT_BAN,
							getstring(u->na, BOT_INFO_ACTIVE), 
							ci->kbb(KBB_REPEAT), ci->repeats());
			else 
				notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_REPEAT_ON,
						getstring(u->na, BOT_INFO_ACTIVE), ci->repeats());
		} else { 
			notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_REPEAT_OFF,
					getstring(u->na, BOT_INFO_INACTIVE));
		}
		if (ci->botflags() & ChannelInfo::bfKickReverses)
		{
			if (ci->kbb(KBB_REVERSES)) 
				notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_REVERSES_BAN, 
						getstring(u->na, BOT_INFO_ACTIVE), 
						ci->kbb(KBB_REVERSES));
			else 
				notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_REVERSES,
						getstring(u->na, BOT_INFO_ACTIVE));
		} else { 
			notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_REVERSES,
					getstring(u->na, BOT_INFO_INACTIVE));
		}
		if (ci->botflags() & ChannelInfo::bfKickUnderlines)
		{
			if (ci->kbb(KBB_UNDERLINES)) 
				notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_UNDERLINES_BAN,
						getstring(u->na, BOT_INFO_ACTIVE), 
						ci->kbb(KBB_UNDERLINES));
			else 
				notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_UNDERLINES,
						getstring(u->na, BOT_INFO_ACTIVE));
		} else { 
			notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_UNDERLINES,
					getstring(u->na, BOT_INFO_INACTIVE));
		}
		
		end = buf;
		*end = 0;
		if (ci->botflags() & ChannelInfo::bfDontKickOps) {
	    	end += snprintf(end, sizeof(buf)-(end-buf), "%s",
				getstring(u->na, BOT_INFO_OPT_DONTKICKOPS));
	    	need_comma = 1;
		}
		if (ci->botflags() & ChannelInfo::bfDontKickVoices) {
	    	end += snprintf(end, sizeof(buf)-(end-buf), "%s%s",
				need_comma ? commastr : "",
				getstring(u->na, BOT_INFO_OPT_DONTKICKVOICES));
	    	need_comma = 1;
		}
		if (ci->botflags() & ChannelInfo::bfFantasy) {
	    	end += snprintf(end, sizeof(buf)-(end-buf), "%s%s",
				need_comma ? commastr : "",
				getstring(u->na, BOT_INFO_OPT_FANTASY));
	    	need_comma = 1;
		}
		if (ci->botflags() & ChannelInfo::bfGreet) {
	    	end += snprintf(end, sizeof(buf)-(end-buf), "%s%s",
				need_comma ? commastr : "",
				getstring(u->na, BOT_INFO_OPT_GREET));
	    	need_comma = 1;
		}
		if (ci->botflags() & ChannelInfo::bfNoBot) {
	    	end += snprintf(end, sizeof(buf)-(end-buf), "%s%s",
				need_comma ? commastr : "",
				getstring(u->na, BOT_INFO_OPT_NOBOT));
	    	need_comma = 1;
		}
		if (ci->botflags() & ChannelInfo::bfSymbiosis) {
	    	end += snprintf(end, sizeof(buf)-(end-buf), "%s%s",
				need_comma ? commastr : "",
				getstring(u->na, BOT_INFO_OPT_SYMBIOSIS));
	    	need_comma = 1;
		}
		notice_lang(s_BotServ, u, BOT_INFO_CHAN_OPTIONS,
			*buf ? buf : getstring(u->na, BOT_INFO_OPT_NONE));
			
	} else notice_lang(s_BotServ, u, BOT_INFO_NOT_FOUND, query);
}

/*************************************************************************/

static void do_set(User *u)
{
	char *chan = strtok(NULL, " ");
	char *option = strtok(NULL, " ");
	char *value = strtok(NULL, " ");
	int is_servadmin = is_services_admin(u);
	
	ChannelInfo *ci;
	
	if (!channels.writable())
		notice_lang(s_BotServ, u, BOT_SET_DISABLED);
	else if (!chan || !option || !value)
		syntax_error(s_BotServ, u, "SET", BOT_SET_SYNTAX);
	else if (is_servadmin && !stricmp(option, "PRIVATE")) {
		BotInfo *bi;
		
		if ((bi = bots.find(chan))) {
			if (!stricmp(value, "ON")) {
				bi->flags_add(BotInfo::bfPrivate);
				notice_lang(s_BotServ, u, BOT_SET_PRIVATE_ON, bi->nick());
			} else if (!stricmp(value, "OFF")) {
				bi->flags_remove(BotInfo::bfPrivate);
				notice_lang(s_BotServ, u, BOT_SET_PRIVATE_OFF, bi->nick());	
			} else {
				syntax_error(s_BotServ, u, "SET PRIVATE", BOT_SET_PRIVATE_SYNTAX);	
			}
		} else {
			notice_lang(s_BotServ, u, BOT_DOES_NOT_EXIST, chan);
		}
		return;
	} else if (!(ci = channels.find(chan)))
		notice_lang(s_BotServ, u, CHAN_X_NOT_REGISTERED, chan);
    else if (!is_servadmin && !check_access(u, ci, CA_SET))
		notice_lang(s_BotServ, u, ACCESS_DENIED);
	else
	{
		if (!stricmp(option, "DONTKICKOPS")) {
			if (!stricmp(value, "ON")) {
				ci->botflags_add(ChannelInfo::bfDontKickOps);
				notice_lang(s_BotServ, u, BOT_SET_DONTKICKOPS_ON, ci->name());
			} else if (!stricmp(value, "OFF")) {
				ci->botflags_remove(ChannelInfo::bfDontKickOps);
				notice_lang(s_BotServ, u, BOT_SET_DONTKICKOPS_OFF, ci->name());	
			} else {
				syntax_error(s_BotServ, u, "SET DONTKICKOPS", BOT_SET_DONTKICKOPS_SYNTAX);	
			}
		} else if (!stricmp(option, "DONTKICKVOICES")) {
			if (!stricmp(value, "ON")) {
				ci->botflags_add(ChannelInfo::bfDontKickVoices);
				notice_lang(s_BotServ, u, BOT_SET_DONTKICKVOICES_ON, ci->name());
			} else if (!stricmp(value, "OFF")) {
				ci->botflags_remove(ChannelInfo::bfDontKickVoices);
				notice_lang(s_BotServ, u, BOT_SET_DONTKICKVOICES_OFF, ci->name());	
			} else {
				syntax_error(s_BotServ, u, "SET DONTKICKVOICES", BOT_SET_DONTKICKVOICES_SYNTAX);	
			}
		} else if (!stricmp(option, "FANTASY")) {
			if (!stricmp(value, "ON")) {
				ci->botflags_add(ChannelInfo::bfFantasy);
				notice_lang(s_BotServ, u, BOT_SET_FANTASY_ON, ci->name());
			} else if (!stricmp(value, "OFF")) {
				ci->botflags_remove(ChannelInfo::bfFantasy);
				notice_lang(s_BotServ, u, BOT_SET_FANTASY_OFF, ci->name());	
			} else {
				syntax_error(s_BotServ, u, "SET FANTASY", BOT_SET_FANTASY_SYNTAX);	
			}
		} else if (!stricmp(option, "GREET")) {
			if (!stricmp(value, "ON")) {
				ci->botflags_add(ChannelInfo::bfGreet);
				notice_lang(s_BotServ, u, BOT_SET_GREET_ON, ci->name());
			} else if (!stricmp(value, "OFF")) {
				ci->botflags_remove(ChannelInfo::bfGreet);
				notice_lang(s_BotServ, u, BOT_SET_GREET_OFF, ci->name());	
			} else {
				syntax_error(s_BotServ, u, "SET GREET", BOT_SET_GREET_SYNTAX);	
			}
		} else if (is_servadmin && !stricmp(option, "NOBOT")) {
			if (!stricmp(value, "ON")) {
				ci->botflags_add(ChannelInfo::bfNoBot);
				if (ci->bi())
					ci->bi(0);
				notice_lang(s_BotServ, u, BOT_SET_NOBOT_ON, ci->name());
			} else if (!stricmp(value, "OFF")) {
				ci->botflags_remove(ChannelInfo::bfNoBot);
				notice_lang(s_BotServ, u, BOT_SET_NOBOT_OFF, ci->name());	
			} else {
				syntax_error(s_BotServ, u, "SET NOBOT", BOT_SET_NOBOT_SYNTAX);	
			}
		} else if (!stricmp(option, "SYMBIOSIS")) {
			if (!stricmp(value, "ON")) {
				ci->botflags_add(ChannelInfo::bfSymbiosis);
				notice_lang(s_BotServ, u, BOT_SET_SYMBIOSIS_ON, ci->name());
			} else if (!stricmp(value, "OFF")) {
				ci->botflags_remove(ChannelInfo::bfSymbiosis);
				notice_lang(s_BotServ, u, BOT_SET_SYMBIOSIS_OFF, ci->name());
			} else {
				syntax_error(s_BotServ, u, "SET SYMBIOSIS", BOT_SET_SYMBIOSIS_SYNTAX);	
			}
		} else {
			notice_help(s_BotServ, u, BOT_SET_UNKNOWN, option);
		}
	}
}

/*************************************************************************/

static void do_kickcmd(User *u)
{
	char *chan = strtok(NULL, " ");
	char *option = strtok(NULL, " ");
	char *value = strtok(NULL, " ");
	char *kbb = strtok(NULL, " ");
	
	ChannelInfo *ci;
	
	if (!channels.writable())
		notice_lang(s_BotServ, u, BOT_KICK_DISABLED);
	else if (!chan || !option || !value)
		syntax_error(s_BotServ, u, "KICK", BOT_KICK_SYNTAX);
	else if (stricmp(value, "ON") && stricmp(value, "OFF"))
		syntax_error(s_BotServ, u, "KICK", BOT_KICK_SYNTAX);
	else if (!(ci = channels.find(chan)))
		notice_lang(s_BotServ, u, CHAN_X_NOT_REGISTERED, chan);
    else if (!is_services_admin(u) && !check_access(u, ci, CA_SET))
		notice_lang(s_BotServ, u, ACCESS_DENIED);
	else
	{
		if (!stricmp(option, "BADWORDS")) {
			if (!stricmp(value, "ON")) {
				if (kbb) {
					ci->kbb(KBB_BADWORDS, atol(kbb));
					/*if (ci->kbb(KBB_BADWORDS) < 0) {
						notice_lang(s_BotServ, u, BOT_KICK_BAD_TTB, kbb);
						ci->kbb(KBB_BADWORDS, 0);
						return;
					}*/
				} else 
					ci->kbb(KBB_BADWORDS, 0);
				ci->botflags_add(ChannelInfo::bfKickBadWords);
				if (ci->kbb(KBB_BADWORDS)) 
					notice_lang(s_BotServ, u, BOT_KICK_BADWORDS_ON_BAN,
							ci->kbb(KBB_BADWORDS));
				else 
					notice_lang(s_BotServ, u, BOT_KICK_BADWORDS_ON);
			} else {
				ci->botflags_remove(ChannelInfo::bfKickBadWords);
				notice_lang(s_BotServ, u, BOT_KICK_BADWORDS_OFF);	
			}	
		} else if (!stricmp(option, "BOLDS")) {
			if (!stricmp(value, "ON")) {
				if (kbb) {
					ci->kbb(KBB_BOLDS, atol(kbb));
					/*if (ci->kbb(KBB_BOLDS) < 0) {
						notice_lang(s_BotServ, u, BOT_KICK_BAD_TTB, kbb);
						ci->kbb(KBB_BOLDS, 0);
						return;
					}*/
				} else 
					ci->kbb(KBB_BOLDS, 0);
				ci->botflags_add(ChannelInfo::bfKickBolds);
				if (ci->kbb(KBB_BOLDS)) 
					notice_lang(s_BotServ, u, BOT_KICK_BOLDS_ON_BAN, 
							ci->kbb(KBB_BOLDS));
				else 
					notice_lang(s_BotServ, u, BOT_KICK_BOLDS_ON);
			} else {
				ci->botflags_remove(ChannelInfo::bfKickBolds);
				notice_lang(s_BotServ, u, BOT_KICK_BOLDS_OFF);	
			}	
		} else if (!stricmp(option, "CAPS")) {
			if (!stricmp(value, "ON")) {
				char *min = strtok(NULL, " ");
				char *percent = strtok(NULL, " ");
				
				if (kbb) {
					ci->kbb(KBB_CAPS, atol(kbb));
					/*if (ci->kbb(KBB_CAPS) < 0) {
						notice_lang(s_BotServ, u, BOT_KICK_BAD_TTB, kbb);
						ci->kbb(KBB_CAPS, 0);
						return;
					}*/
				} else 
					ci->kbb(KBB_CAPS, 0);
				
				long lmin = 10, lpercent = 25;
				
				if (min && !misc::strtol_range(&lmin, min, 1, 255))
				{
					notice_lang(s_BotServ, u, BOT_KICK_CAPS_ON_INVALID_MIN,
							1, 255);
					return;
				}
				
				if (percent && !misc::strtol_range(&lpercent, percent, 1, 100))
				{
					notice_lang(s_BotServ, u, BOT_KICK_CAPS_ON_INVALID_PERCENT,
							1, 100);
					return;
				}
				
				ci->change_caps(lmin, lpercent);
				
				ci->botflags_add(ChannelInfo::bfKickCaps);
				if (ci->kbb(KBB_CAPS)) 
					notice_lang(s_BotServ, u, BOT_KICK_CAPS_ON_BAN,
							ci->caps_min(), ci->caps_percent(), ci->kbb(KBB_CAPS));
				else 
					notice_lang(s_BotServ, u, BOT_KICK_CAPS_ON, ci->caps_min(),
							ci->caps_percent());
			} else {
				ci->botflags_remove(ChannelInfo::bfKickCaps);
				notice_lang(s_BotServ, u, BOT_KICK_CAPS_OFF);	
			}	
		} else if (!stricmp(option, "COLORS")) {
			if (!stricmp(value, "ON")) {
				if (kbb) {
					ci->kbb(KBB_COLORS, atol(kbb));
					/*if (ci->kbb(KBB_COLORS) < 0) {
						notice_lang(s_BotServ, u, BOT_KICK_BAD_TTB, kbb);
						ci->kbb(KBB_COLORS, 0);
						return;
					}*/
				} else 
					ci->kbb(KBB_COLORS, 0);
				ci->botflags_add(ChannelInfo::bfKickColors);
				if (ci->kbb(KBB_COLORS)) 
					notice_lang(s_BotServ, u, BOT_KICK_COLORS_ON_BAN,
							ci->kbb(KBB_COLORS));
				else notice_lang(s_BotServ, u, BOT_KICK_COLORS_ON);
			} else {
				ci->botflags_remove(ChannelInfo::bfKickColors);
				notice_lang(s_BotServ, u, BOT_KICK_COLORS_OFF);	
			}
		} else if (!stricmp(option, "FLOOD")) {
			if (!stricmp(value, "ON")) {
				char *lines = strtok(NULL, " ");
				char *secs = strtok(NULL, " ");
				
				if (kbb) {
					ci->kbb(KBB_FLOOD, atol(kbb));
					/*if (ci->kbb(KBB_FLOOD) < 0) {
						notice_lang(s_BotServ, u, BOT_KICK_BAD_TTB, kbb);
						ci->kbb(KBB_FLOOD, 0);
						return;
					}*/
				} else 
					ci->kbb(KBB_FLOOD, 0);
				
				long llines = 6, lsecs = 10;
				
				if (lines && !misc::strtol_range(&llines, lines, 2, 255))
				{
					notice_lang(s_BotServ, u, BOT_KICK_FLOOD_ON_INVALID_LINES,
							2, 255);
					return;
				}
				
				if (secs && !misc::strtol_range(&lsecs, secs, 2, 255))
				{
					notice_lang(s_BotServ, u, BOT_KICK_FLOOD_ON_INVALID_SECONDS,
							1, (epona->cd().botserv.keep_data < 255 ? 
								epona->cd().botserv.keep_data : 255));
					return;
				}
				
				ci->change_flood(llines, lsecs);
				
				ci->botflags_add(ChannelInfo::bfKickFlood);
				if (ci->kbb(KBB_FLOOD)) 
					notice_lang(s_BotServ, u, BOT_KICK_FLOOD_ON_BAN, 
							ci->flood_lines(), ci->flood_secs(), ci->kbb(KBB_FLOOD));
				else 
					notice_lang(s_BotServ, u, BOT_KICK_FLOOD_ON, 
							ci->flood_lines(), ci->flood_secs());
			} else {
				ci->botflags_remove(ChannelInfo::bfKickFlood);
				notice_lang(s_BotServ, u, BOT_KICK_FLOOD_OFF);	
			}
		} else if (!stricmp(option, "REPEAT")) {
			if (!stricmp(value, "ON")) {
				char *times = strtok(NULL, " ");
				
				if (kbb) {
					ci->kbb(KBB_REPEAT, atol(kbb));
					/*if (ci->kbb(KBB_REPEAT) < 0) {
						notice_lang(s_BotServ, u, BOT_KICK_BAD_TTB, kbb);
						ci->kbb(KBB_REPEAT, 0);
						return;
					}*/
				} else 
					ci->kbb(KBB_REPEAT, 0);
				
				long ltimes = 3;
				
				if (times && !misc::strtol_range(&ltimes, times, 2, 255))
				{
					notice_lang(s_BotServ, u, BOT_KICK_REPEAT_ON_INVALID_TIMES,
							2, 255);
					return;
				}
				
				ci->repeats(ltimes);
				ci->botflags_add(ChannelInfo::bfKickRepeats);

				if (ci->kbb(KBB_REPEAT)) 
					notice_lang(s_BotServ, u, BOT_KICK_REPEAT_ON_BAN,
							ci->repeats(), ci->kbb(KBB_REPEAT));
				else 
					notice_lang(s_BotServ, u, BOT_KICK_REPEAT_ON, 
							ci->repeats());
			} else {
				ci->botflags_remove(ChannelInfo::bfKickRepeats);
				notice_lang(s_BotServ, u, BOT_KICK_REPEAT_OFF);	
			}
		} else if (!stricmp(option, "REVERSES")) {
			if (!stricmp(value, "ON")) {
				if (kbb) {
					ci->kbb(KBB_REVERSES, atol(kbb));
					/*if (ci->kbb(KBB_REVERSES) < 0) {
						notice_lang(s_BotServ, u, BOT_KICK_BAD_TTB, kbb);
						ci->kbb(KBB_REVERSES, 0);
						return;
					}*/
				} else 
					ci->kbb(KBB_REVERSES, 0);
				ci->botflags_add(ChannelInfo::bfKickReverses);
				if (ci->kbb(KBB_REVERSES)) 
					notice_lang(s_BotServ, u, BOT_KICK_REVERSES_ON_BAN,
							ci->kbb(KBB_REVERSES));
				else 
					notice_lang(s_BotServ, u, BOT_KICK_REVERSES_ON);
			} else {
				ci->botflags_remove(ChannelInfo::bfKickReverses);
				notice_lang(s_BotServ, u, BOT_KICK_REVERSES_OFF);	
			}
		} else if (!stricmp(option, "UNDERLINES")) {
			if (!stricmp(value, "ON")) {
				if (kbb) {
					ci->kbb(KBB_UNDERLINES, atol(kbb));
					/*if (ci->kbb(KBB_UNDERLINES) < 0) {
						notice_lang(s_BotServ, u, BOT_KICK_BAD_TTB, kbb);
						ci->kbb(KBB_UNDERLINES, 0);
						return;
					}*/
				} else 
					ci->kbb(KBB_UNDERLINES, 0);
				ci->botflags_add(ChannelInfo::bfKickUnderlines);
				if (ci->kbb(KBB_UNDERLINES)) 
					notice_lang(s_BotServ, u, BOT_KICK_UNDERLINES_ON_BAN,
							ci->kbb(KBB_UNDERLINES));
				else 
					notice_lang(s_BotServ, u, BOT_KICK_UNDERLINES_ON);
			} else {
				ci->botflags_remove(ChannelInfo::bfKickUnderlines);
				notice_lang(s_BotServ, u, BOT_KICK_UNDERLINES_OFF);	
			}
		} else notice_help(s_BotServ, u, BOT_KICK_UNKNOWN, option);
	}
}

/*************************************************************************/

BadWordsCmd::BadWordsCmd() : ListCmd("BADWORDS", s_BotServ, badwords_msgs, true),
		m_context(0), ci(0)
{ 

}

/*************************************************************************/

int BadWordsCmd::add(User *u, container_type &ctn, char const *sub)
{
	/* Get the word */
	char *word = strtok(NULL, "");
	
	if (!word)
	{
		syntax_error(m_service, u, m_name, BOT_BADWORDS_ADD_SYNTAX);
		return crStop;
	}
	
	/* Check for the option */
	uint8_t type = BadWord::bwAny;
	char *pos = strrchr(word, ' ');
	
	if (pos) 
	{
		char *opt = pos + 1;
		
		if (*opt) 
		{
			if (!stricmp(opt, "SINGLE"))
				type = BadWord::bwSingle;
			else if (!stricmp(opt, "START")) 
				type = BadWord::bwStart;
			else if (!stricmp(opt, "END"))
				type = BadWord::bwEnd;
			
			if (type != BadWord::bwAny)
				*pos = 0;
		}
	}
	
	/* Validates the word */
	
	size_t word_len = strlen(word);
	
	if (word_len < 3 && type != BadWord::bwSingle && strcmp(word, "#"))
	{
		notice_lang(m_service, u, BOT_BADWORDS_EXPR_TOO_SHORT, word);
		return crStop;
	}
	
	if (word_len > BadWord::word_len)
	{
		notice_lang(m_service, u, BOT_BADWORDS_EXPR_TOO_LONG,
				BadWord::word_len);
		return crStop;
	}
	
	/* Add the record */
	
	if (add_record(u, ci, word, type) < 0)
		return crStop;

	return crStop;
}

/*************************************************************************/

int BadWordsCmd::add_record(User *u, ChannelInfo *owner, char const *word,
		uint8_t type)
{
	/* Check whether we can add more entries. */
	if (owner->badwords().size() >= epona->cd().botserv.max_badwords)
	{
		ChannelInfo::badwords_type::const_iterator it, end;
		
		for (it = owner->badwords().begin(), end = owner->badwords().end();
				it != end; ++it)
		{
			if ((*it) == 0)
				break;
		}
		
		if (it == end)
		{
			if (u)
				notice_lang(m_service, u, BOT_BADWORDS_REACHED_LIMIT,
						epona->cd().botserv.max_badwords.value());
			return -1;
		}
	}
	
	BadWord *bw = (BadWord *)badwords.create();
	
	if (!bw)
	{
		if (u)
			notice_lang(m_service, u, BOT_BADWORDS_CREATION_FAILED);
		return -1;
	}
	
	bw->owner(owner);
	bw->word(word);
	bw->type(type);
	
	/* Try to add the access entry to the list */
	
	int error;
	
	if (!badwords.add(bw, &error))
	{
		if (u)
		{
			if (error == BadWords::ceName ||
					error == databases::Collection::ceList)
				notice_lang(m_service, u, BOT_BADWORDS_ALREADY_PRESENT,
				 		bw->word(), owner->name());
			else
				notice_lang(m_service, u, m_msgs[cmDisabled]);
		}
				
		delete bw;
		return -1;
	} else {
		if (u)
			notice_lang(m_service, u, BOT_BADWORDS_ADDED,
					bw->word(), owner->name());
	}

	return 0;
}

/*************************************************************************/

bool BadWordsCmd::allow_dispatch(User *u, container_type &ctn, char const *sub)
{
	/* Allow services admin access to DEL/LIST/CLEAR but not ADD. */
	if (stricmp(sub, "ADD") && is_services_admin(u))
		return true;
	
	return check_access(u, ci, CA_BADWORDS);
}

/*************************************************************************/

bool BadWordsCmd::changeable(User *u, container_type &ctn,
		char const *sub) const
{
	return badwords.writable();
}

/*************************************************************************/

BadWordsCmd::container_type *BadWordsCmd::container(User *u)
{
	m_context = strtok(NULL, " ");
	
	if (!m_context)
	{
		syntax_error(m_service, u, m_name, m_msgs[cmSyntax]);
		return 0;
	}
	
	if (!(ci = channels.find(m_context)))
	{
		notice_lang(m_service, u, CHAN_X_NOT_REGISTERED, m_context);
		return 0;
	}
	
	return &ci->badwords();
}

/*************************************************************************/

bool BadWordsCmd::del_record(User *u, container_type &ctn, char const *sub,
		databases::Record *rec)
{
	return badwords.remove(rec);
}

/*************************************************************************/

bool BadWordsCmd::list_record(User *u, container_type &ctn, char const *sub,
		uint16_t num, databases::Record const *rec, uint16_t count)
{
	BadWord *bw = (BadWord *)rec;
	
	if (count == 0)
		notice_lang(m_service, u, BOT_BADWORDS_LIST_COLHEAD);
	
	notice_lang(m_service, u, BOT_BADWORDS_LIST_FORMAT, num, bw->word(),
			((bw->type() == BadWord::bwSingle) ? "(SINGLE)" : 
    		((bw->type() == BadWord::bwStart) ? "(START)" : 
    		((bw->type() == BadWord::bwEnd) ? "(END)" : ""))));
	return true;
}

/*************************************************************************/

static void do_badwords(User *u)
{
	static BadWordsCmd badwords_cmd;
	badwords_cmd.execute(u);
}

/*************************************************************************/

static void do_say(User *u)
{
	ChannelInfo *ci;
	
	char *chan = strtok(NULL, " ");
	char *text = strtok(NULL, "");
	
	if (!chan || !text)
		syntax_error(s_BotServ, u, "SAY", BOT_SAY_SYNTAX);
	else if (!(ci = channels.find(chan)))
		notice_lang(s_BotServ, u, CHAN_X_NOT_REGISTERED, chan);
    else if (!ci->bi())
		notice_help(s_BotServ, u, BOT_NOT_ASSIGNED);
	else if (!ci->c || ci->c->usercount < epona->cd().botserv.min_users)
		notice_lang(s_BotServ, u, BOT_NOT_ON_CHANNEL, ci->name());
    else if (!check_access(u, ci, CA_SAY))
		notice_lang(s_BotServ, u, ACCESS_DENIED);
	else {
		send_cmd(ci->bi()->nick(), "PRIVMSG %s :%s", ci->name(), text);	
		ci->bi()->lastmsg(time(NULL));
	}
}

/*************************************************************************/

static void do_act(User *u)
{
	ChannelInfo *ci;
	
	char *chan = strtok(NULL, " ");
	char *text = strtok(NULL, "");
	
	if (!chan || !text)
		syntax_error(s_BotServ, u, "ACT", BOT_ACT_SYNTAX);
	else if (!(ci = channels.find(chan)))
		notice_lang(s_BotServ, u, CHAN_X_NOT_REGISTERED, chan);
    else if (!ci->bi())
		notice_help(s_BotServ, u, BOT_NOT_ASSIGNED);
	else if (!ci->c || ci->c->usercount < epona->cd().botserv.min_users)
		notice_lang(s_BotServ, u, BOT_NOT_ON_CHANNEL, ci->name());
    else if (!check_access(u, ci, CA_SAY))
		notice_lang(s_BotServ, u, ACCESS_DENIED);
	else {
		send_cmd(ci->bi()->nick(), "PRIVMSG %s :%cACTION %s%c", ci->name(), 1, text, 1);
		ci->bi()->lastmsg(time(NULL));	
	}
}

/*************************************************************************/
