/* You decide.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * Based on the original code of Services by Andy Church.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#if HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif

#if HAVE_SOCKET_H
# include <socket.h>
#endif
#if HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <string>
#include <utility>

#include <cmdline/cmdline.h>
#include <core/directives.h>
#include <core/epona.h>
#include <core/version.h>
#include <log/levels.h>
#include <mail/mail.h>
#include <mail/transport.h>
#include <modules/manager.h>

#include "conf.h"
#include "epona.h"

#include "blacklist.h"
#include "chanserv.h"
#include "nickserv.h"

using namespace std;
using namespace cmdline;
using namespace conf;
using namespace core;

using core::hidden::Directives;

/*************************************************************************/

static int const dbgGlobal = Conf::dbgUserDefined + 0;

static in_addr const inaddr_any = { 0 };
static in_addr const inaddr_loopback = { 0x7f000001 };

/*************************************************************************/

LOGLEVELH(DebugConf, xlog::LogDebug, "Debug::Conf", "debug: config: ");
LOGLEVELH(WarningConf, xlog::LogWarning, "Warning::Conf", "warning: config: ");
LOGLEVELH(CriticalConf, xlog::LogCritical, "Critical::Conf",
		"Critical error: config: ");

/*************************************************************************/
/*************************************************************************/

#define CHECK_ADD(x, y, z) if (!(x).add((z))) { \
	(y).errorf("Can't add directive %s: out of memory?", (z).ConfDir::name().c_str()); \
	return false; }

/*************************************************************************/

Directives::Directives(conf::ConfLogger &logger) : DirectivesQ(logger),
		allow_root(), directory(), dump_core(), pidfile(), umask(),
		strict_passwords("strict-passwords", true, true), bad_passwords(), 
		expire_timeout("expire-timeout", 1800, true, 300, 604800),
		warning_timeout("warning-timeout", 3600, true, 300, 604800),
		timeout_check("timeout-check", 3, true, 0, 10),
		log("log", true), module(), load_module(), 
		lock_module("lock-module", ConfStringA::container_type(), false),
		mail(), services(), session(), blacklist(), nickserv(),
		chanserv(), memoserv(), botserv(),
		helpserv("helpserv", "HelpServ", "Help Service"), operserv(), global()
{
	
}

/*************************************************************************/

bool Directives::add(Conf &conf)
{
	CHECK_ADD(conf, conf, allow_root);
	CHECK_ADD(conf, conf, directory);
	CHECK_ADD(conf, conf, dump_core);
	CHECK_ADD(conf, conf, pidfile);
	CHECK_ADD(conf, conf, umask);
	CHECK_ADD(conf, conf, strict_passwords);
	CHECK_ADD(conf, conf, bad_passwords);
	CHECK_ADD(conf, conf, expire_timeout);
	CHECK_ADD(conf, conf, warning_timeout);
	CHECK_ADD(conf, conf, timeout_check);
	
	CHECK_ADD(conf, conf, log);
	
	CHECK_ADD(conf, conf, module);
	CHECK_ADD(conf, conf, load_module);
	CHECK_ADD(conf, conf, lock_module);
	
	CHECK_ADD(conf, conf, mail);
	
	CHECK_ADD(conf, conf, services);
	CHECK_ADD(conf, conf, session);
	CHECK_ADD(conf, conf, blacklist);
	
	CHECK_ADD(conf, conf, nickserv);
	CHECK_ADD(conf, conf, chanserv);
	CHECK_ADD(conf, conf, memoserv);
	CHECK_ADD(conf, conf, botserv);
	CHECK_ADD(conf, conf, helpserv);
	CHECK_ADD(conf, conf, operserv);
	CHECK_ADD(conf, conf, global);
	
	return true;
}

/*************************************************************************/

Directives::AllowRoot::AllowRoot() : ConfBool("allow-root", false, true)
{
	
}

/*************************************************************************/

bool Directives::AllowRoot::apply(conf::Conf &conf)
{
	if ((m_value == false) && (geteuid() == 0))
	{
		epona->fatal("Epona is now allowed to execute as root");
		return false;
	}
	
	return ConfBool::apply(conf);
}

/*************************************************************************/

Directives::BadPasswords::BadPasswords() :
		ConfBoolBlock("bad-passwords", true), limit("limit", 5, true, 1, 100),
		timeout("timeout", 3600 /* seconds */, true, 0, 604800)
{
	
}

/*************************************************************************/

bool Directives::BadPasswords::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, limit);
	CHECK_ADD(new_conf, conf, timeout);
	
	return true;
}

/*************************************************************************/

Directives::Blacklist::Blacklist() : ConfBlock("blacklist", true),
		enable("enable", false, false),
		master("master", ConfStrCharA::container_type(), false, 256),
		diff("diff", ConfStrCharA::container_type(), true, 256),
		update("update", 3600, true, 60, 315360000 /* 10 years */),
		ban_reason("ban-reason", "Blacklisted user", true, 1, 256, ""),
		max("max", 50, true, 1, 1024)
{
	
}

/*************************************************************************/

bool Directives::Blacklist::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, enable);
	CHECK_ADD(new_conf, conf, master);
	CHECK_ADD(new_conf, conf, diff);
	CHECK_ADD(new_conf, conf, update);
	CHECK_ADD(new_conf, conf, ban_reason);
	CHECK_ADD(new_conf, conf, max);
	
	return true;
}

/*************************************************************************/

Directives::BotServ::BotServ() : ConfBlock("botserv", true),
		service("service", "BotServ", "Bot service"),
		database_bots("database-bots", false, 1, NAME_MAX, ""),
		database_kbb("database-kbb", false, 1, NAME_MAX, ""),
		database_badwords("database-badwords", false, 1, NAME_MAX, ""),
		min_users("min-users", 1, false, 1, 32767),
		smart_join("smart-join", true, true),
		keep_data("keep-data", 600 /* 10m */, true, 60, 7200),
		default_options(), max_badwords("max-badwords", 32, true, 1, 1024),
		soft_reason("soft-reason", true, true)
{
	
}

/*************************************************************************/

bool Directives::BotServ::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, service);
	
	CHECK_ADD(new_conf, conf, database_bots);
	CHECK_ADD(new_conf, conf, database_kbb);
	CHECK_ADD(new_conf, conf, database_badwords);
	
	CHECK_ADD(new_conf, conf, min_users);
	CHECK_ADD(new_conf, conf, smart_join);
	CHECK_ADD(new_conf, conf, keep_data);
	
	CHECK_ADD(new_conf, conf, default_options);
	CHECK_ADD(new_conf, conf, max_badwords);
	CHECK_ADD(new_conf, conf, soft_reason);
	
	return true;
}

/*************************************************************************/

static uint32_t bfDefaultConf = ChannelInfo::bfDefaultConf;

Directives::BotServ::DefaultOptions::DefaultOptions() :
		ConfSet<uint32_t>("default-options", bfDefaultConf, true)
{
	
}

/*************************************************************************/

bool Directives::BotServ::DefaultOptions::add(Conf &conf)
{
	m_bits["DONTKICKOPS"]    = ChannelInfo::bfDontKickOps;
	m_bits["DONTKICKVOICES"] = ChannelInfo::bfDontKickVoices;
	m_bits["FANTASY"]        = ChannelInfo::bfFantasy;
	m_bits["GREET"]          = ChannelInfo::bfGreet;
	m_bits["SYMBIOSIS"]      = ChannelInfo::bfSymbiosis;

	if (m_bits.size() != 5)
	{
		conf.errorf("Can't initialize directive %s", m_name.c_str());
		return false;
	}

	return true;
}

/*************************************************************************/

Directives::ChanServ::ChanServ() : ConfBlock("chanserv", true),
		service("service", "ChanServ", "Channel service"),
		database_channels("database-channels", false, 1, NAME_MAX, ""),
		database_levels("database-levels", false, 1, NAME_MAX, ""),
		database_access("database-access", false, 1, NAME_MAX, ""),
		database_akicks("database-akicks", false, 1, NAME_MAX, ""),
		max_reg("max-reg", 15, true, 0, 1024),
		password_module(), default_options(),
		default_bantype("default-bantype", 2, true, 0, 3), restrict_auth(),
		expire(),
		max_access("max-access", 512, true, 1, 32767),
		max_akicks("max-akicks", 32, true, 1, 32767),
		akick_reason("akick-reason", "User has been banned from the channel",
			true, 1, 256, ""),
		inhabit("inhabit", 15, true, 5, 300),
		list_opers("list-opers", true, true),
		list_max("list-max", 50, true, 1, 1024),
		clist_opers("clist-opers", true, true),
		sendpass_opers("sendpass-opers", true, true)
{
	
}

/*************************************************************************/

bool Directives::ChanServ::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, service);
	
	CHECK_ADD(new_conf, conf, database_channels);
	CHECK_ADD(new_conf, conf, database_levels);
	CHECK_ADD(new_conf, conf, database_access);
	CHECK_ADD(new_conf, conf, database_akicks);
	
	CHECK_ADD(new_conf, conf, max_reg);
	CHECK_ADD(new_conf, conf, password_module);
	CHECK_ADD(new_conf, conf, default_options);
	CHECK_ADD(new_conf, conf, restrict_auth);
	CHECK_ADD(new_conf, conf, expire);
	CHECK_ADD(new_conf, conf, default_bantype);
	CHECK_ADD(new_conf, conf, max_access);
	CHECK_ADD(new_conf, conf, max_akicks);
	
	CHECK_ADD(new_conf, conf, akick_reason);
	CHECK_ADD(new_conf, conf, inhabit);
	
	CHECK_ADD(new_conf, conf, list_opers);
	CHECK_ADD(new_conf, conf, list_max);
	CHECK_ADD(new_conf, conf, clist_opers);
	CHECK_ADD(new_conf, conf, sendpass_opers);
	
	return true;
}

/*************************************************************************/

static uint32_t cfDefaultConf = ChannelInfo::cfDefaultConf;

Directives::ChanServ::DefaultOptions::DefaultOptions() :
		ConfSet<uint32_t>("default-options", cfDefaultConf, true)
{
	
}

/*************************************************************************/

bool Directives::ChanServ::DefaultOptions::add(Conf &conf)
{
	m_bits["KEEPTOPIC"]      = ChannelInfo::cfKeepTopic;
	m_bits["OPNOTICE"]       = ChannelInfo::cfOpNotice;
	m_bits["PEACE"]          = ChannelInfo::cfPeace;
	m_bits["PRIVATE"]        = ChannelInfo::cfPrivate;
	m_bits["RESTRICTED"]     = ChannelInfo::cfRestricted;
	m_bits["SECURE"]         = ChannelInfo::cfSecure;
	m_bits["SECUREFOUNDER"]  = ChannelInfo::cfSecureFounder;
	m_bits["SECUREOPS"]      = ChannelInfo::cfSecureOps;
	m_bits["SIGNKICK"]       = ChannelInfo::cfSignKick;
	m_bits["SIGNKICK LEVEL"] = ChannelInfo::cfSignKickLevel;
	m_bits["TOPICLOCK"]      = ChannelInfo::cfTopicLock;

	if (m_bits.size() != 11)
	{
		conf.errorf("Can't initialize directive %s", m_name.c_str());
		return false;
	}

	return true;
}

/*************************************************************************/

bool Directives::ChanServ::DefaultOptions::check(Conf &conf,
		uint32_t const &value)
{
	if ((value == ChannelInfo::cfSignKickLevel) &&
			(m_value & ChannelInfo::cfSignKick))
		m_value &= ~ChannelInfo::cfSignKick;
	if ((value == ChannelInfo::cfSignKick) &&
			(m_value & ChannelInfo::cfSignKickLevel))
		m_value &= ~ChannelInfo::cfSignKickLevel;
	
	return ConfSet<uint32_t>::check(conf, value);
}

/*************************************************************************/

Directives::ChanServ::Expire::Expire() : ConfBoolBlock("expire", true),
		delay("delay", 1209600 /* 2w */, true, 3600, 2147483647),
		grace("grace", 0 /* 0s */, true, 0, 2147483647)
{
	
}		

/*************************************************************************/

bool Directives::ChanServ::Expire::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, delay);
	CHECK_ADD(new_conf, conf, grace);
	
	return true;
}

/*************************************************************************/

Directives::ChanServ::PasswordModule::PasswordModule() :
			ConfStrChar("password-module", "", true, 0, NAME_MAX, "")
{
	
}

/*************************************************************************/

bool Directives::ChanServ::PasswordModule::apply_change(conf::Conf &conf)
{
	password::Encrypter *enc = 0;
		
	if (!value().empty() && !(enc = channels.load_encrypter(value().c_str())))
		return true;
	
	channels.release_encrypter(channels.encrypter());
	channels.encrypter(enc);
	
	return ConfStrChar::apply_change(conf);
}

/*************************************************************************/

Directives::ChanServ::RestrictAuth::RestrictAuth() :
		ConfSet<uint32_t>("restrict-auth", 0, true)
{
	
}

/*************************************************************************/

bool Directives::ChanServ::RestrictAuth::add(conf::Conf &conf)
{
	m_bits["FOUNDER"]  = rfFounder;
	m_bits["ACCESS"]   = rfAccess;
	m_bits["IDENTIFY"] = rfIdentify;
	m_bits["REGISTER"] = rfRegister;
	m_bits["ALL"]      = rfAll;
	
	if (m_bits.size() != 5)
	{
		conf.errorf("Can't initialize directive %s", m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

Directives::Directory::Directory() : 
		ConfStrChar("directory", EPONA_DIR, true, 1, PATH_MAX, "")
{
	
}

/*************************************************************************/

bool Directives::Directory::apply(conf::Conf &conf)
{
	if (chdir(m_value.c_str()) == -1)
	{
		
		epona->fatalf("Can't set current directory to %s: %s",
				m_value.c_str(), strerror(errno));
		return false;
	}
	
	return ConfStrChar::apply(conf);
}

/*************************************************************************/

Directives::DumpCore::DumpCore() : ConfBool("dump-core", false, true)
{
	
}

/*************************************************************************/

bool Directives::DumpCore::apply_change(conf::Conf &conf)
{
	epona->dump_core(m_value);
	return conf::ConfBool::apply_change(conf);
}

/*************************************************************************/

Directives::Global::Global() : ConfBlock("global", true),
		service("service", "Global", "Global noticer"),
		network_domain("network-domain", true, 1, HOSTMAX,
			"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz.-")
{
	
}

/*************************************************************************/

bool Directives::Global::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, service);
	CHECK_ADD(new_conf, conf, network_domain);
	
	return true;
}

/*************************************************************************/

Directives::Mail::Mail() : ConfBoolBlock("mail", true),
		transport(), sender(), limit(),
		sender_limit("sender-limit", misc::Ratio(5, 3600), true,
			1, 32768, 1, 604800)
{
	
}

/*************************************************************************/

bool Directives::Mail::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, transport);
	CHECK_ADD(new_conf, conf, sender);
	CHECK_ADD(new_conf, conf, limit);
	CHECK_ADD(new_conf, conf, sender_limit);
	
	return true;
}

/*************************************************************************/

bool Directives::Mail::apply_changes(conf::Conf &conf)
{
	mail::conf_options(sender.name, sender.address,
			limit.authentified.value(), limit.unauthentified.value(),
			sender_limit.value());
	
	return conf::ConfBoolBlock::apply_changes(conf);
}

/*************************************************************************/

Directives::Mail::Limit::Limit() : conf::ConfBlock("limit", true),
		authentified("authentified", misc::Ratio(10, 3600), true,
			1, 32768, 1, 604800),
		unauthentified("unauthentified", misc::Ratio(3, 7200), true,
			1, 32768, 1, 604800)
{
	
}

/*************************************************************************/

bool Directives::Mail::Limit::add(conf::Conf &conf, conf::Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, authentified);
	CHECK_ADD(new_conf, conf, unauthentified);
	
	return true;
}

/*************************************************************************/

Directives::Mail::Sender::Sender() : conf::ConfBlock("sender", true),
		name("name", "", true, 1, 255, ""), address()
{
	
}

/*************************************************************************/

bool Directives::Mail::Sender::add(conf::Conf &conf, conf::Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, name);
	CHECK_ADD(new_conf, conf, address);
	
	return true;
}

/*************************************************************************/

Directives::Mail::Sender::Address::Address() :
		conf::ConfString("address", true, 1, 255, "")
{
	
}

/*************************************************************************/

bool Directives::Mail::Sender::Address::check(conf::Conf &conf,
		std::string const &value) const
{
	if (!ConfString::check(conf, value))
		return false;
	
	return mail::Mail::valid_address(value.c_str(), false);
}

/*************************************************************************/

Directives::Mail::Transport::Transport() :
		conf::ConfString("transport", true, 1, NAME_MAX, "")
{
	
}

/*************************************************************************/

bool Directives::Mail::Transport::apply_change(conf::Conf &conf)
{
	modules::Module *oldm = epona->modules().find(m_oldvalue);
	if (oldm)
		oldm->references_remove();
	
	modules::Module *m = epona->modules().find(m_value);
	
	if (!m)
	{
		if (!epona->modules().load(m_value))
		{
			conf.errorf("Unable to change mail transport module: %s",
					epona->modules().error().c_str());
			return false;
		}
		
		return conf::ConfString::apply_change(conf);
	}
	
	mail::Transport *transport = (mail::Transport *)
			m->interface(mail::Transport::IID);
				
	if (!transport)
	{
		conf.errorf("%s is not a mail transport module", m->name().c_str());
		return false;
	}
				
	m->references_add();			
	mail::conf_transport(transport);
	
	return conf::ConfString::apply_change(conf);
}

/*************************************************************************/

Directives::MemoServ::MemoServ() : ConfBlock("memoserv", true),
		service("service", "MemoServ", "Memo service"), restrict_auth(),
		max_memos("max-memos", 50, true, 0, 32767),
		send_delay("send-delay", 5, true, 0, 31536000),
		channel_target("channel-target", false, true),
		forward("forward", false, true)
{
	
}

/*************************************************************************/

bool Directives::MemoServ::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, service);
	
	CHECK_ADD(new_conf, conf, restrict_auth);
	CHECK_ADD(new_conf, conf, max_memos);
	CHECK_ADD(new_conf, conf, send_delay);
	CHECK_ADD(new_conf, conf, channel_target);
	CHECK_ADD(new_conf, conf, forward);
	
	return true;
}

/*************************************************************************/

Directives::MemoServ::RestrictAuth::RestrictAuth() :
		ConfSet<uint32_t>("restrict-auth", 0, true)
{
	
}

/*************************************************************************/

bool Directives::MemoServ::RestrictAuth::add(conf::Conf &conf)
{
	m_bits["SEND"]  = rfSend;
	m_bits["GET"]   = rfGet;
	m_bits["ALL"]   = rfAll;
	
	if (m_bits.size() != 3)
	{
		conf.errorf("Can't initialize directive %s", m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

Directives::NickServ::NickServ() : ConfBlock("nickserv", true),
		service("service", "NickServ", "Nickname service"), password_module(),
		database_groups("database-groups", false, 1, NAME_MAX, ""),
		database_aliases("database-aliases", false, 1, NAME_MAX, ""),
		database_access("database-access", false, 1, NAME_MAX, ""),
		database_memos("database-memos", false, 1, NAME_MAX, ""),
		require_email("require-email", true, true), default_options(),
		default_language(0), real_deflang(), 
		reg_delay("reg-delay", 30, true, 0, 604800), auth(), expire(),
		max_aliases("max-aliases", 16, true, 1, 1024),
		max_access("max-access", 16, true, 1, 32767), enforcer(),
		release_timeout("release-timeout", 60 /* 1m */, true, 1, 3600),
		kill_immed("kill-immed", false, false),
		group_change("group-change", false, true),
		list_opers("list-opers", true, true),
		list_max("list-max", 50, true, 1, 1000),
		sendpass_opers("sendpass-opers", true, true),
		protect_admins("protect-admins", true, true),
		strict_privileges("strict-privileges", true, true), custom_host()
{
	
}

/*************************************************************************/

bool Directives::NickServ::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, service);
	
	CHECK_ADD(new_conf, conf, password_module);
	
	CHECK_ADD(new_conf, conf, database_groups);
	CHECK_ADD(new_conf, conf, database_aliases);
	CHECK_ADD(new_conf, conf, database_access);
	CHECK_ADD(new_conf, conf, database_memos);
	
	CHECK_ADD(new_conf, conf, require_email);
	CHECK_ADD(new_conf, conf, default_options);
	CHECK_ADD(new_conf, conf, real_deflang);
	CHECK_ADD(new_conf, conf, reg_delay);
	CHECK_ADD(new_conf, conf, auth);
	CHECK_ADD(new_conf, conf, expire);
	CHECK_ADD(new_conf, conf, max_aliases);
	CHECK_ADD(new_conf, conf, max_access);
	CHECK_ADD(new_conf, conf, enforcer);
	CHECK_ADD(new_conf, conf, release_timeout);
	CHECK_ADD(new_conf, conf, kill_immed);
	CHECK_ADD(new_conf, conf, group_change);
	CHECK_ADD(new_conf, conf, list_opers);
	CHECK_ADD(new_conf, conf, list_max);
	CHECK_ADD(new_conf, conf, sendpass_opers);
	CHECK_ADD(new_conf, conf, protect_admins);
	CHECK_ADD(new_conf, conf, strict_privileges);
	CHECK_ADD(new_conf, conf, custom_host);
	
	return true;
}

/*************************************************************************/

Directives::NickServ::Auth::Auth() : ConfBoolBlock("auth", true),
		required("required", false, true), 
		web_form("web-form", "", true, 1, 1024, ""),
		attempts("attempts", 10, true, 1, 100), restrict()
{
	
}		

/*************************************************************************/

bool Directives::NickServ::Auth::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, required);
	CHECK_ADD(new_conf, conf, web_form);
	CHECK_ADD(new_conf, conf, attempts);
	CHECK_ADD(new_conf, conf, restrict);
	
	return true;
}

/*************************************************************************/

Directives::NickServ::Auth::Restrict::Restrict() :
		ConfSet<uint32_t>("restrict", 0, true)
{
	
}

/*************************************************************************/

bool Directives::NickServ::Auth::Restrict::add(conf::Conf &conf)
{
	m_bits["GROUP"]    = rfGroup;
	m_bits["IDENTIFY"] = rfIdentifyImplied;
	m_bits["ALL"]      = rfAll;
	
	if (m_bits.size() != 3)
	{
		conf.errorf("Can't initialize directive %s", m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

Directives::NickServ::CustomHost::CustomHost() :
		ConfBlock("custom-host", true),
		enable("enable", false, false),
		restrict("restrict", true, true),
		age("age", 7257600 /* 12w */, true, 0, 2147483647),
		delay("delay", 2419200 /* 4w */, true, 0, 2147483647),
		enforce_strict("enforce-strict", true, true)
{
	
}		

/*************************************************************************/

bool Directives::NickServ::CustomHost::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, enable);
	CHECK_ADD(new_conf, conf, restrict);
	CHECK_ADD(new_conf, conf, age);
	CHECK_ADD(new_conf, conf, delay);
	CHECK_ADD(new_conf, conf, enforce_strict);
	
	return true;
}

/*************************************************************************/

Directives::NickServ::DefaultLanguage::DefaultLanguage() :
		ConfInt<int16_t>("default-language", 1, true, 1, USED_LANGS)
{
	
}

/*************************************************************************/

bool Directives::NickServ::DefaultLanguage::apply_change(conf::Conf &conf)
{
	if ((epona->cd().nickserv.default_language = langlist[m_value - 1]) < 0)
		epona->cd().nickserv.default_language = DEF_LANGUAGE;
	
	return ConfInt<int16_t>::apply_change(conf);
}

/*************************************************************************/

static uint32_t gfDefaultConf = NickGroup::gfDefaultConf;

Directives::NickServ::DefaultOptions::DefaultOptions() :
			ConfSet<uint32_t>("default-options", gfDefaultConf, true)
{
	
}

/*************************************************************************/

bool Directives::NickServ::DefaultOptions::add(Conf &conf)
{
	m_bits["SECURE"] = NickGroup::gfSecure;
	m_bits["KILL"] = NickGroup::gfKill;
	m_bits["KILL QUICK"] = NickGroup::gfKillQuick;
	m_bits["PRIVATE"] = NickGroup::gfPrivate;
	m_bits["HIDE EMAIL"] = NickGroup::gfHideEmail;
	m_bits["HIDE QUIT"] = NickGroup::gfHideQuit;
	m_bits["HIDE USERMASK"] = NickGroup::gfHideMask;
	m_bits["NOTIFY LOGON"] = NickGroup::gfMemoSignon;
	m_bits["NOTIFY NEW"] = NickGroup::gfMemoReceive;
	
	if (m_bits.size() != 9)
	{
		conf.errorf("Can't initialize directive %s", m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool Directives::NickServ::DefaultOptions::check(Conf &conf,
		uint32_t const &value)
{
	if ((value == NickGroup::gfKillQuick) && (m_value & NickGroup::gfKill))
		m_value &= ~NickGroup::gfKill;
	if ((value == NickGroup::gfKill) && (m_value & NickGroup::gfKillQuick))
		m_value &= ~NickGroup::gfKillQuick;
	
	return ConfSet<uint32_t>::check(conf, value);
}

/*************************************************************************/

Directives::NickServ::Enforcer::Enforcer() : ConfBlock("enforcer", true),
		username("username", "enforcer", true, 1, USERMAX,
			"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz~"),
		hostname("hostname", true, 1, HOSTMAX,
			"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz.-")
{
	
}

/*************************************************************************/

bool Directives::NickServ::Enforcer::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, username);
	CHECK_ADD(new_conf, conf, hostname);
	
	return true;
}

/*************************************************************************/

Directives::NickServ::Expire::Expire() : ConfBoolBlock("expire", true),
		delay("delay", 1814400 /* 3w */, true, 3600, 2147483647),
		grace("grace", 0 /* 0s */, true, 0, 2147483647)
{
	
}		

/*************************************************************************/

bool Directives::NickServ::Expire::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, delay);
	CHECK_ADD(new_conf, conf, grace);
	
	return true;
}

/*************************************************************************/

Directives::NickServ::PasswordModule::PasswordModule() :
			ConfStrChar("password-module", "", true, 0, NAME_MAX, "")
{
	
}

/*************************************************************************/

bool Directives::NickServ::PasswordModule::apply_change(conf::Conf &conf)
{
	password::Encrypter *enc = 0;
		
	if (!value().empty() && !(enc = groups.load_encrypter(value().c_str())))
		return true;
	
	groups.release_encrypter(groups.encrypter());
	groups.encrypter(enc);
	
	return ConfStrChar::apply_change(conf);
}

/*************************************************************************/

Directives::OperServ::OperServ() : ConfBlock("operserv", true),
		service("service", "OperServ", "Operator service"),
		database_stats("database-stats", false, 1, NAME_MAX, ""),
		database_akills("database-akills", false, 1, NAME_MAX, ""),
		database_sglines("database-sglines", false, 1, NAME_MAX, ""),
		database_sqlines("database-sqlines", false, 1, NAME_MAX, ""),
		database_szlines("database-szlines", false, 1, NAME_MAX, ""),
		database_news("database-news", false, 1, NAME_MAX, ""), root(), wall(),
		default_akilltype("default-akilltype", 2, true, 0, 3),
		expire_akills("expire-akills", 1800 /* 30m */, true, 0, 2147483647),
		expire_sglines("expire-sglines", 1800 /* 30m */, true, 0, 2147483647),
		expire_sqlines("expire-sqlines", 1800 /* 30m */, true, 0, 2147483647),
		expire_szlines("expire-szlines", 1800 /* 30m */, true, 0, 2147483647),
		list_max("list-max", 50, true, 1, 1000),
		allow_raw("allow-raw", false, true),
		allow_modules("allow-modules", false, true)
{
	
}

/*************************************************************************/

bool Directives::OperServ::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, service);
	
	CHECK_ADD(new_conf, conf, database_stats);
	CHECK_ADD(new_conf, conf, database_akills);
	CHECK_ADD(new_conf, conf, database_sglines);
	CHECK_ADD(new_conf, conf, database_sqlines);
	CHECK_ADD(new_conf, conf, database_szlines);
	CHECK_ADD(new_conf, conf, database_news);
	
	CHECK_ADD(new_conf, conf, root);
	
	CHECK_ADD(new_conf, conf, wall);
	
	CHECK_ADD(new_conf, conf, default_akilltype);
	CHECK_ADD(new_conf, conf, expire_akills);
	CHECK_ADD(new_conf, conf, expire_sglines);
	CHECK_ADD(new_conf, conf, expire_sqlines);
	CHECK_ADD(new_conf, conf, expire_szlines);
	
	CHECK_ADD(new_conf, conf, list_max);
	
	CHECK_ADD(new_conf, conf, allow_raw);
	CHECK_ADD(new_conf, conf, allow_modules);
	
	return true;
}

/*************************************************************************/

Directives::OperServ::Root::Root() : ConfStrCharA("root", true)
{
	
}

/*************************************************************************/

bool Directives::OperServ::Root::apply_change(Conf &conf)
{
	NickAlias *na;
	ConfStrCharA::container_type::const_iterator it, end;
	
	/* Remove old services roots */
	for (it = m_oldvalue.begin(), end = m_oldvalue.end(); it != end; ++it)
	{
		na = aliases.find(*it);
		if (na)
			na->ng()->flags_remove(NickGroup::gfServicesRoot);
	}
	
	/* And add new services roots */
	for (it = m_value.begin(), end = m_value.end(); it != end; ++it) 
	{
		na = aliases.find(*it);
		if (na)
			na->ng()->flags_add(NickGroup::gfServicesRoot);
	}
	
	return ConfStrCharA::apply_change(conf);
}

/*************************************************************************/

Directives::OperServ::Wall::Wall() : ConfSet<uint32_t>("wall", 0, true)
{
	
}

/*************************************************************************/

bool Directives::OperServ::Wall::add(conf::Conf &conf)
{
	m_bits["NEWOPER"]          = waNewOper;
	m_bits["NONOPER"]          = waNonOper;
	m_bits["GLOBAL"]           = waGlobal;
	m_bits["MODE"]             = waMode;
	m_bits["CLEARMODES"]       = waClearModes;
	m_bits["KICK"]             = waKick;
	m_bits["AKILL"]            = waAkill;
	m_bits["SGLINE"]           = waSGLine;
	m_bits["SQLINE"]           = waSQLine;
	m_bits["SZLINE"]           = waSZLine;
	m_bits["NOOP"]             = waNoOp;
	m_bits["JUPE"]             = waJupe;
	m_bits["RAW"]              = waRaw;
	m_bits["RELOAD"]           = waReload;
	m_bits["MODULES"]          = waModules;
	m_bits["AKILL EXPIRE"]     = waAkillExpire;
	m_bits["SGLINE EXPIRE"]    = waSGLineExpire;
	m_bits["SQLINE EXPIRE"]    = waSQLineExpire;
	m_bits["SZLINE EXPIRE"]    = waSZLineExpire;
	m_bits["EXCEPTION EXPIRE"] = waExceptionExpire;
	m_bits["DROP"]             = waDrop;
	m_bits["GETPASS"]          = waGetpass;
	m_bits["SETPASS"]          = waSetpass;
	m_bits["ALL"]              = waAll;
	
	if (m_bits.size() != 24)
	{
		conf.errorf("Can't initialize directive %s", m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

Directives::PIDFile::PIDFile() : 
		ConfStrChar("pid-file", PACKAGE_TARNAME ".pid", true, 0, NAME_MAX, "")
{
	
}

/*************************************************************************/

bool Directives::PIDFile::apply_change(Conf &conf)
{
	if (!m_oldvalue.empty())
		epona->pid_erase(m_oldvalue.c_str());
	if (!m_value.empty())
		epona->pid_write(m_value.c_str());
	
	return ConfStrChar::apply_change(conf);
}

/*************************************************************************/

Directives::Service::Service(std::string const &name,
		std::string const &defnick, std::string const &defreal) :
		conf::ConfBlock(name, false),
		nick("nick", defnick, false, 1, NICKMAX - 1,
			"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-_{}[]`^|\\"),
		username("username", "services", false, 1, USERMAX,
			"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz~"), 
		hostname("hostname", false, 1, HOSTMAX,
			"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz.-"),
		realname("realname", defreal, false, 1, REALMAX, "")
{
	
}
		

/*************************************************************************/

bool Directives::Service::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, nick);
	CHECK_ADD(new_conf, conf, username);
	CHECK_ADD(new_conf, conf, hostname);
	CHECK_ADD(new_conf, conf, realname);
	
	return true;
}

/*************************************************************************/

Directives::Services::Services() : conf::ConfBlock("services", true),
		sname("name", false, 1, SERVERMAX - 1,
				"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz.-"),
		description("description", string(program_name) + " IRC Services",
				false, 1, SDESCMAX - 1, ""), 
		connect("connect", inaddr_loopback, 6667, false),
		bind("bind", inaddr_any, 0, false),
		read_timeout("read-timeout", 5, true, 1, 60),
		password("password", false, 1, PASSMAX - 1, ""),
		network_name("network-name", true, 1, 65536, ""),
		log_users("log-users", false, true),
		log_maxusers("log-maxusers", true, true), 
		log_input("log-input", false, true),
		log_output("log-output", false, true),
		motd("motd", PACKAGE_TARNAME ".motd", true, 1, NAME_MAX, ""),
		guest_nick("guest-nick", "Guest", true, 1, NICKMAX - 8,
			"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-_{}[]`^|\\"),
		inhabit("inhabit", 30, true, 5, 300),
		help_channel("help-channel", "", true, 0, CHANMAX, "")
{

}

/*************************************************************************/

bool Directives::Services::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, sname);
	CHECK_ADD(new_conf, conf, description);
	CHECK_ADD(new_conf, conf, this->connect);
	CHECK_ADD(new_conf, conf, this->bind);
	CHECK_ADD(new_conf, conf, read_timeout);
	CHECK_ADD(new_conf, conf, password);
	CHECK_ADD(new_conf, conf, network_name);
	CHECK_ADD(new_conf, conf, log_users);
	CHECK_ADD(new_conf, conf, log_maxusers);
	CHECK_ADD(new_conf, conf, log_input);
	CHECK_ADD(new_conf, conf, log_output);
	CHECK_ADD(new_conf, conf, motd);
	CHECK_ADD(new_conf, conf, guest_nick);
	CHECK_ADD(new_conf, conf, inhabit);
	CHECK_ADD(new_conf, conf, help_channel);
	
	return true;
}

/*************************************************************************/

Directives::Session::Session() : conf::ConfBlock("session", true),
		enable("enable", true, false),
		default_limit("default-limit", 3, true, 0, 65535),
		message("message", ConfStrCharA::container_type(), true, 16),
		max_kills("max-kills", 15, true, 0, 1000),
		expire_akills("expire-akills", 1800 /* 30m */, true, 1, 31536000),
		database_exceptions("database-exceptions", false, 1, NAME_MAX, ""),
		max_limit("max-limit", 100, true, 1, 65535),
		expire_exceptions("expire-exceptions", 86400 /* 1d */, true, 0, 2147483647)
{
	
}

/*************************************************************************/

bool Directives::Session::add(Conf &conf, Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, enable);
	CHECK_ADD(new_conf, conf, default_limit);
	CHECK_ADD(new_conf, conf, message);
	CHECK_ADD(new_conf, conf, max_kills);
	CHECK_ADD(new_conf, conf, expire_akills);
	CHECK_ADD(new_conf, conf, database_exceptions);
	CHECK_ADD(new_conf, conf, max_limit);
	CHECK_ADD(new_conf, conf, expire_exceptions);
	
	return true;
}

/*************************************************************************/

Directives::Umask::Umask() : ConfUint<mode_t>("umask", 0077, true, 0, 0777)
{
	
}

/*************************************************************************/

bool Directives::Umask::apply(conf::Conf &conf)
{
	::umask(m_value);	
	return ConfUint<mode_t>::apply(conf);
}

/*************************************************************************/

#undef CHECK_ADD

/*************************************************************************/
/*************************************************************************/

#if DEBUG

void EConfLogger::debug(std::string const &context, unsigned int line,
		int type, std::string const &msg)
{
	switch (type)
	{
		case Conf::dbgParse:
			if (epona->debuglevel() < 3)
				return;
			break;
		case Conf::dbgProcessLow:
		case dbgGlobal:
			if (epona->debuglevel() < 1)
				return;
			break;
		default:
			if (epona->debuglevel() < 2)
				return;
			break;
	}
	
	if (context.empty())
		epona->log(DebugConf(), "%s", msg.c_str());
	else
		epona->log(DebugConf(), "%s:%u: %s", context.c_str(), line,
				msg.c_str());
}

#endif

/*************************************************************************/

void EConfLogger::error(std::string const &context, unsigned int line,
		std::string const &err)
{
	if (context.empty())
		epona->log(CriticalConf(), "%s", err.c_str());
	else
		epona->log(CriticalConf(), "%s:%u: %s", context.c_str(), line,
				err.c_str());
}

/*************************************************************************/

void EConfLogger::warning(std::string const &context, unsigned int line,
		std::string const &msg)
{
	if (context.empty())
		epona->log(WarningConf(), "%s", msg.c_str());
	else
		epona->log(WarningConf(), "%s:%u: %s", context.c_str(), line,
				msg.c_str());
}

/*************************************************************************/
/*************************************************************************/

EConfGlobalQ::EConfGlobalQ() : conf::ConfGlobalQ(epona->conf_logger())
{
	
}

/*************************************************************************/

bool EConfGlobalQ::add_paths(std::string const &p, PIType &beg, PIType &end)
{
	ConfGlobal::PathList::const_iterator np;
	
	for (PIType it = beg; it != end; ++it)
	{
		if (!m_global.incl_paths_add(it->second, &np))
		{
			if (errno != 0)
				epona->log(CriticalConf(),
						"Can't add directory %s to the configuration path: %s",
						it->second.c_str(), strerror(errno));
			else
				epona->log(CriticalConf(),
						"Can't add directory %s to the configuration path",
						it->second.c_str());
			return false;
		}
		
		if (epona->debuglevel())
			epona->log(DebugConf(), "Added %s to configuration path",
					np->c_str());
	}
	
	return true;
}

/*************************************************************************/

bool EConfGlobalQ::add_vars(std::string const &p, PIType &beg, PIType &end)
{
	for (PIType it = beg; it != end; ++it)
	{
		if (!m_global.var_set(it->second))
		{
			epona->log(CriticalConf(),
					"Can't add configuration variable %s: name/syntax error",
					it->second.c_str());
			return false;
		}
		
		if (epona->debuglevel())
			epona->log(DebugConf(), "Added configuration variable %s",
					it->second.c_str());
	}
	
	return true;
}

/*************************************************************************/

bool EConfGlobalQ::setup()
{
	/* Add potential paths for configuration files.
	 * Don't check for errors though because these directories may
	 * be irrelevant. */
	m_global.incl_paths_add(EPONA_CONFDIR);
	m_global.incl_paths_add(EPONA_DIR);
	
	/* Process the -I/--include parameters. */
	pair<PIType, PIType> inc_range;
	
	inc_range = epona->clp().params().equal_range("I");
	if (!add_paths("I", inc_range.first, inc_range.second))
		return false;
	
	inc_range = epona->clp().params().equal_range("include");
	if (!add_paths("include", inc_range.first, inc_range.second))
		return false;
	
	/* Add some variables that may be useful for the user. */
	m_global.var_set("confdir", EPONA_CONFDIR);
	m_global.var_set("program", program_name);
	m_global.var_set("program_lowercase", program_name_lc);
	m_global.var_set("program_uppercase", program_name_uc);
	m_global.var_set("installdir", EPONA_DIR);
	m_global.var_set("version", epona->version_string());
	
	if (epona->debuglevel())
	{
		char buf[16];
		snprintf(buf, sizeof(buf), "%d", epona->debuglevel());
		
		m_global.var_set("debug", buf);
	}
	
	/* Add environment variables without checking for errors,
	 * because they're not always user input. */
	for (VIType it = epona->clp().vars().begin(); it !=
			epona->clp().vars().end(); ++it)
	{
		m_global.var_set(it->first, it->second);
		if (epona->debuglevel() >= 2)
			epona->log(DebugConf(), "Added configuration variable %s",
					it->first.c_str());
	}
	
	/* Process the -D/--define parameters. */
	pair<PIType, PIType> def_range;
	
	def_range = epona->clp().params().equal_range("D");
	if (!add_vars("D", def_range.first, def_range.second))
		return false;
	
	def_range = epona->clp().params().equal_range("define");
	if (!add_vars("define", def_range.first, def_range.second))
		return false;
	
	return ConfGlobalQ::setup();
}

/*************************************************************************/
/*************************************************************************/

bool Epona::config(bool reload)
{	
	if (m_global)
		delete m_global;
	m_global = new EConfGlobalQ();
	
	if (!m_global->setup())
		return false;
	
	if (!m_cd->parser().parse_file(*m_global, reload, epona->conffile()))
		return false;
	
	return true;
}

/*************************************************************************/

conf::ConfQ *Epona::config(modules::Core &core, conf::ConfGlobalQ &global)
{
	return &m_cd->parser();
}

/*************************************************************************/
