/* File compression interfaces.
 *
 * File compression module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file filecomp/filecomp.h
  * \brief File compression interface header file. */

#ifndef FILECOMP_FILECOMP_H
#define FILECOMP_FILECOMP_H

/*************************************************************************/

#include <modules/interface.h>

/*************************************************************************/

/// Namespace of the file compression interface
namespace filecomp 
{

/*************************************************************************/

/// File compressor interface.

/** This interface class can be used as a base class by modules providing
  * a simple file compression feature.
  *
  * You must override the compress() and extension() functions in
  * derived classes.
  *
  * Modules must provide objects of this interface class through the
  * Module::factory() function. */

class Compressor : public modules::ErrorInterface
{
	public:
		/// Interface ID.
		static modules::PSIID const IID = 0x425f77ee;
		
		/// Destructor.
		virtual ~Compressor() { }
		
		/// Compresses a file.
		/** Compresses a file. 
		  *
		  * \param[in] path The absolute path to the file to compress.
		  * The file must not be modified or deleted. The compressed
		  * file will have the same name but with an appropriate
		  * extension for the compression scheme used.
		  * \param[in] level The compression level, from 1 (faster) 
		  * to 9 (better ratio). Use the value 0 to use an
		  * implementation-defined default value. Implementers
		  * will make their best to adapt the compression level
		  * as appropriate for the used compression scheme.
		  *
		  * \return true if successful, false otherwise. An error
		  * message should then be available through a call to
		  * the error() function in the latter case.
		  *
		  * \sa error() */
		virtual bool compress(char const *path, int level) = 0;
		
		/// Returns the extension appended to compressed files.
		/** Returns the extension for compressed files appended to
		  * the original file by the compress() function.
		  *
		  * \return The appended extension.
		  * \sa compress() */
		virtual char const *extension() const = 0;
};

/*************************************************************************/

} /* namespace filecomp */

/*************************************************************************/

#endif /* FILECOMP_FILECOMP_H */
