/* Database file handling routines.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * Based on the original code of Services by Andy Church.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#if HAVE_FCNTL_H
# include <fcntl.h>
#endif

#include <core/epona.h>
#include <log/levels.h>
#include <misc/string.h>

#include "decl.h"
#include "datafiles.h"
#include "module.h"

using namespace std;

/*************************************************************************/

namespace legacydb
{

/*************************************************************************/

/*************************************************************************/
/*************************************************************************/

/* Return the version number on the file.  Return 0 if there is no version
 * number or the number doesn't make sense (i.e. less than 1).
 */

int get_file_version(dbFILE *f)
{
	FILE *fp = f->fp;
	int version = fgetc(fp)<<24 | fgetc(fp)<<16 | fgetc(fp)<<8 | fgetc(fp);
	
	if (ferror(fp)) 
	{
		me->core().log_perror(ErrorLegacyDB(),
				"Error reading version number on %s", f->filename);
		return 0;
	} else if (feof(fp)) {
		me->core().log(ErrorLegacyDB(),
				"while reading version number on %s: End of file detected",
				f->filename);
		return 0;
	} else if (version < 1) {
		me->core().log(ErrorLegacyDB(), "Invalid version number (%d) on %s",
				version, f->filename);
		return 0;
	}
	
	return version;
}

/*************************************************************************/
/*************************************************************************/

/* Open a database file for reading. Return the stream pointer, or NULL on error. */

dbFILE *open_db(char const *filename)
{
	dbFILE *f;
	FILE *fp;
	
	f = new dbFILE;
	if (!f) {
		me->core().log_perror(ErrorLegacyDB(), "Can't read database %s", filename);
		return NULL;
	}
	
	misc::strncpyzt(f->filename, filename, sizeof(f->filename));
	
	fp = fopen(f->filename, "rb");
	if (!fp) 
	{
		int errno_save = errno;
		if (errno != ENOENT)
		    me->core().log_perror(ErrorLegacyDB(), "Can't read database %s", 
					f->filename);
		delete f;
		errno = errno_save;
		return NULL;
	}
	
	f->fp = fp;
	return f;
}

/*************************************************************************/

/* Close a database file. */

void close_db(dbFILE *f)
{
	fclose(f->fp);
	delete f;
}

/*************************************************************************/
/*************************************************************************/

/* Read 2- and 4-byte quantities and strings.  All multibyte values are
 * stored in big-endian order (most significant byte first). A string is
 * stored with a 2-byte unsigned length (including the trailing \0) first;
 * a length of 0 indicates that the string pointer is NULL.
 *
 * All routines return -1 on error, 0 otherwise.
 */

int read_int16(uint16_t *ret, dbFILE *f)
{
	int c1, c2;
	
	c1 = fgetc(f->fp);
	c2 = fgetc(f->fp);
	if (c1 == EOF || c2 == EOF)
		return -1;
	*ret = c1<<8 | c2;
	return 0;
}

/*************************************************************************/

int read_int32(uint32_t *ret, dbFILE *f)
{
	int c1, c2, c3, c4;
	
	c1 = fgetc(f->fp);
	c2 = fgetc(f->fp);
	c3 = fgetc(f->fp);
	c4 = fgetc(f->fp);
	if (c1 == EOF || c2 == EOF || c3 == EOF || c4 == EOF)
		return -1;
	*ret = c1<<24 | c2<<16 | c3<<8 | c4;
	return 0;
}

/*************************************************************************/

int read_string(char **ret, dbFILE *f)
{
	char *s;
	uint16_t len;
	
	if (read_int16(&len, f) < 0)
		return -1;
	
	if (len == 0) 
	{
		*ret = NULL;
		return 0;
	}
	
	s = new char[len];
	if (len != fread(s, 1, len, f->fp)) 
	{
		delete [] s;
		return -1;
	}
	
	*ret = s;
	return 0;
}

/*************************************************************************/

} /* namespace legacydb */

/*************************************************************************/
