/* Log collection.
 *
 * Log file module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef COLL_H
#define COLL_H

/*************************************************************************/

#include <list>
#include <string>

#include "conf.h"

/*************************************************************************/

namespace modules { class Module; }
namespace xlog { class LogLevel; }

/*************************************************************************/

namespace logfile
{

/*************************************************************************/

class File;

/*************************************************************************/

/// Manages a collection of files.

class Collection
{
	public:
		typedef std::list<File *> container_type;
		
		/// Constructor.
		Collection() : m_members() { }
		/// Destructor.
		~Collection();
		
		/// Adds a file.
		/** Creates a new File object using the given file
		  * configuration directive, adds it to the collection
		  * and mark the file so that it isn't deleted by a call
		  * to cleanup().
		  *
		  * \param[in] directive The file configuration directive
		  * object.
		  *
		  * \return The return value depends on the following:
		  *     - If a file with the same name is already in
		  *       the list, and the file hasn't been marked,
		  *       marks the file and returns true, else logs
		  *       an error and returns false.
		  *     - If the file can't be open, logs an error and
		  *       returns false (the file is not added to the
		  *       collection).
		  *     - In all other cases, returns true. */
		bool add(hidden::Directives::File const &directive);
		
		/// Clears the collection.
		void clear();
		
		/// Finds a file in the collection.
		File *find(std::string const &name);
		
		/// Removes a file from the collection.
		bool remove(File *file);
		
		/// Does a cleanup.
		/** Removes from the collection all files that weren't
		  * added (through add()) since the last cleanup() call. */
		void cleanup();
		
		/// Relocates all files (used when the log directory changes).
		void relocate();
		
		/// Updates file compressors as needed.
		void module_notify(modules::Module &sender, int msg);
		
		/// Forwards a log message to all files in the collection.
		void log(xlog::LogLevel const &level, char const *message);
		
		/// Returns the members list.
		container_type const &members() const { return m_members; }
	private:
		container_type m_members;
	
		Collection(Collection const &);
		Collection &operator=(Collection const &);
};

/*************************************************************************/

} /* namespace logfile */

/*************************************************************************/

#endif /* COLL_H */
