/* Log file configuration.
 *
 * Log file module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <log/levels.h>
#include <misc/misc.h>
#include <modules/core.h>
#include <modules/manager.h>

#include "coll.h"
#include "conf.h"
#include "file.h"
#include "decl.h"
#include "module.h"

using namespace std;
using namespace conf;

using misc::recursive_mkdir;

/*************************************************************************/

namespace logfile
{

/*************************************************************************/

namespace hidden
{

/*************************************************************************/

#define CHECK_ADD(x, y, z) if (!(x).add((z))) { \
	(y).errorf("Can't add directive %s: out of memory?", (z).name().c_str()); \
	return false; }

/*************************************************************************/

Directives::Directives(conf::ConfLogger &logger) : DirectivesQ(logger),
		directory(), file()
{
	
}

/*************************************************************************/

bool Directives::add(conf::Conf &conf)
{
	CHECK_ADD(conf, conf, directory);
	CHECK_ADD(conf, conf, file);
	
	return true;
}

/*************************************************************************/

Directives::Directory::Directory() :
		ConfStrChar("directory", PROGRAM_DIR "/logs", true, 1, PATH_MAX, "")
{
		
}

/*************************************************************************/

bool Directives::Directory::apply(conf::Conf &conf)
{
	if (!recursive_mkdir(m_value.c_str(), 0777))
	{
		conf.errorf("Unable to create directory %s: %s", m_value.c_str(),
				m_value.c_str());
		return false;
	}
	
	if (conf.reloading() && m_value_changed)
		me->collection().relocate();
	
	return ConfStrChar::apply(conf);
}

/*************************************************************************/

Directives::File::File(std::string const &name, bool reloadable) :
		ConfBlock(name, reloadable), 
		filename("filename", true, 1, PATH_MAX, ""), rules("rules", true),
		timestamp_format("timestamp-format", "[%Y/%m/%d %H:%M:%S]",
				true, 0, logfile::File::timestamp_length, ""),
		header("header", true, true), rotate()
{
	
}

/*************************************************************************/

Directives::File::File(File const &f) : ConfBlock(f),
		filename(f.filename), rules(f.rules),
		timestamp_format(f.timestamp_format), header(f.header),
		rotate(f.rotate)
{
	
}

/*************************************************************************/

Directives::File &Directives::File::operator=(File const &right)
{
	if (&right != this)
	{
		ConfBlock::operator=(right);
		
		filename = right.filename;
		rules = right.rules;
		timestamp_format = right.timestamp_format;
		header = right.header;
		rotate = right.rotate;
	}
	
	return *this;
}

/*************************************************************************/

bool Directives::File::add(conf::Conf &conf, conf::Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, filename);
	CHECK_ADD(new_conf, conf, rules);
	CHECK_ADD(new_conf, conf, timestamp_format);
	CHECK_ADD(new_conf, conf, header);
	CHECK_ADD(new_conf, conf, rotate);
	
	return true;
}

/*************************************************************************/

Directives::File::Rotate::Rotate() : ConfBoolBlock("rotate", true),
		span("span", 86400 /* 1d */, true, 0, 2147483647),
		keep("keep", 30, true, 0, 65535), suffix(), compress()
{
	
}

/*************************************************************************/

Directives::File::Rotate::Rotate(Rotate const &r) : ConfBoolBlock(r),
		span(r.span), keep(r.keep), suffix(r.suffix), compress(r.compress)
{
	
}

/*************************************************************************/

Directives::File::Rotate
		&Directives::File::Rotate::operator=(Rotate const &right)
{
	if (&right != this)
	{
		ConfBoolBlock::operator=(right);
		
		span = right.span;
		keep = right.keep;
		suffix = right.suffix;
		compress = right.compress;
	}
	
	return *this;
}

/*************************************************************************/

bool Directives::File::Rotate::add(conf::Conf &conf, conf::Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, span);
	CHECK_ADD(new_conf, conf, keep);
	CHECK_ADD(new_conf, conf, suffix);
	CHECK_ADD(new_conf, conf, compress);
	
	return true;
}

/*************************************************************************/

Directives::File::Rotate::Compress::Compress() :
		ConfBoolBlock("compress", true),
		module("module", true, 1, NAME_MAX, ""),
		level("level", 0, true, 0, 9)
{
	
}

/*************************************************************************/

Directives::File::Rotate::Compress::Compress(Compress const &c) :
		ConfBoolBlock(c), module(c.module), level(c.level)
{
	
}

/*************************************************************************/

Directives::File::Rotate::Compress
		&Directives::File::Rotate::Compress::operator=(Compress const &right)
{
	if (this != &right)
	{
		ConfBoolBlock::operator=(right);
		
		module = right.module;
		level = right.level;
	}
	
	return *this;
}

/*************************************************************************/

bool Directives::File::Rotate::Compress::add(conf::Conf &conf,
		conf::Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, module);
	CHECK_ADD(new_conf, conf, level);
	
	return true;
}

/*************************************************************************/

Directives::File::Rotate::Suffix::Suffix() : ConfBoolBlock("suffix", true),
		first_part("first-part", true, 1, NAME_MAX, ""),
		second_part("second-part", true, 1, NAME_MAX, "")
{
	
}

/*************************************************************************/

Directives::File::Rotate::Suffix::Suffix(Suffix const &s) : ConfBoolBlock(s),
		first_part(s.first_part), second_part(s.second_part)
{
	
}

/*************************************************************************/

Directives::File::Rotate::Suffix
		&Directives::File::Rotate::Suffix::operator=(Suffix const &right)
{
	if (&right != this)
	{
		ConfBoolBlock::operator=(right);
		
		first_part = right.first_part;
		second_part = right.second_part;
	}
	
	return *this;
}

/*************************************************************************/

bool Directives::File::Rotate::Suffix::add(conf::Conf &conf,
		conf::Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, first_part);
	CHECK_ADD(new_conf, conf, second_part);
	
	return true;
}

/*************************************************************************/

Directives::FileA::FileA() : ConfArray<File>("file", true, 32768)
{
	
}

/*************************************************************************/


bool Directives::FileA::apply_change(conf::Conf &conf)
{
	container_type::const_iterator it, end;
	
	for (it = m_value.begin(), end = m_value.end(); it != end; ++it)
	{
		if (it->rotate && it->rotate.compress &&
				!me->core().modules().load(it->rotate.compress.module))
		{
			conf.errorf("Can't load compression module %s: %s",
					it->rotate.compress.module.value().c_str(),
					me->core().modules().error().c_str());
		}
		
		if (!me->collection().add(*it))
			conf.errorf("Can't create log file %s",
					it->filename.value().c_str());
	}
	
	me->collection().cleanup();
	
	return ConfArray<File>::apply_change(conf);
}

/*************************************************************************/

} /* namespace hidden */

/*************************************************************************/

} /* namespace logfile */

/*************************************************************************/
