/* MySQL result set class.
 *
 * MySQL client module (c) 2006 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file mysql-cli/result.h
  * \brief MySQL result set header file. */

#ifndef MYSQL_CLI_RESULT_H
#define MYSQL_CLI_RESULT_H

/*************************************************************************/

#include <mysql.h>

/*************************************************************************/

namespace mysql_cli
{

/*************************************************************************/

/// MySQL result set class.

/** This class encapsulates an instance of MYSQL_RES. It automatically
  * frees it with mysql_free_result when it is not needed anymore. The
  * result is reference counted; you may copy the class as needed, they'll
  * all share the same result set. */

class Result
{
	public:
		/// Default constructor.
		Result() : m_sh(0) { }
		/// Constructor.
		/** Constructor.
		  *
		  * \param[in] res The MYSQL_RES instance to manage in this
		  * class. It may be NULL. */
		explicit Result(MYSQL_RES *res);
		/// Copy constructor.
		Result(Result const &r);
		/// Assignment operator.
		Result &operator=(Result const &right);
		/// Assignment operator for MYSQL_RES *.
		Result &operator=(MYSQL_RES *right);
		/// Destructor.
		~Result();
		
		/// Conversion operator.
		operator MYSQL_RES *() const { return (m_sh ? m_sh->res() : 0); }
		/// Returns pointer to the MYSQL_RES result set.
		MYSQL_RES *result() const { return (m_sh ? m_sh->res() : 0); }
		
		/// Returns the number of rows in the result set.
		my_ulonglong num_rows();
		/// Retrieves the next row of the result set (0 if no more rows).
		MYSQL_ROW fetch_row();
		/// Returns the lengths of the columns of the current row.
		unsigned long *fetch_lengths();
		/// Seeks to an arbitrary row in the result set.
		/** Seeks to an arbitrary row in the result set. Should not be used
		  * if the result set was obtained through mysql_use_result.
		  *
		  * \param[in] offset The row number, which must be between 0 and
		  * num_rows() - 1. */
		void data_seek(my_ulonglong offset);
		/// Returns the position of the row cursor used for the last fetch_row().
		MYSQL_ROW_OFFSET row_tell();
		/// Sets the row cursor to the given offset.
		MYSQL_ROW_OFFSET row_seek(MYSQL_ROW_OFFSET offset);
		
		/// Returns the number of columns in the result set.
		unsigned int num_fields();
		/// Returns the definition of one column of the result set.
		MYSQL_FIELD *fetch_field();
		/// Returns the definition of the given column of the result set.
		MYSQL_FIELD *fetch_field(unsigned int fieldnum);
		/// Returns an array of all MYSQL_FIELD structures of the result set.
		MYSQL_FIELD *fetch_fields();
		/// Returns the position of the field cursor used for the last fetch_field().
		MYSQL_FIELD_OFFSET field_tell();
		/// Sets the field cursor to the given offset.
		MYSQL_FIELD_OFFSET field_seek(MYSQL_FIELD_OFFSET offset);
	private:
		class Shared
		{
			public:
				/// Constructor.
				explicit Shared(MYSQL_RES *res) : m_refcount(1), m_res(res) { }
				/// Destructor.
				~Shared();
				
				/// Returns a reference to the reference count.
				unsigned int &refcount() { return m_refcount; }
				/// Returns the result.
				MYSQL_RES *res() const { return m_res; }
			private:
				unsigned int m_refcount;
				MYSQL_RES *m_res;
				
				Shared(Shared const &);
				Shared &operator=(Shared const &);
		};
		
		Shared *m_sh;
};

/*************************************************************************/

} /* namespace mysql_cli */

/*************************************************************************/

#endif /* MYSQL_CLI_RESULT_H */
