/* Configuration.
 *
 * MySQL database module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <core/version.h>
#include <log/levels.h>
#include <misc/constants.h>
#include <modules/core.h>

#include "conf.h"
#include "decl.h"
#include "module.h"

using namespace std;

/*************************************************************************/

namespace mysql
{

/*************************************************************************/

#define CHECK_ADD(x, y, z) if (!(x).add((z))) { \
	(y).errorf("Can't add directive %s: out of memory?", (z).name().c_str()); \
	return false; }

/*************************************************************************/

Directives::Directives(conf::ConfLogger &logger) : DirectivesQ(logger),
		connection(), reconnect_frequency(), sync(),
		select_max("select-max", 5000, true, 1, 0xFFFFFFFF),
		table_prefix("table-prefix", "", false, 0, 16, ""),
		table_charset("table-charset", "latin1", true, 1, 8, ""),
		table_collation("table-collation", "", true, 0, 32, "")
{
	
}

/*************************************************************************/

bool Directives::add(conf::Conf &conf)
{
	CHECK_ADD(conf, conf, connection);
	CHECK_ADD(conf, conf, reconnect_frequency);
	CHECK_ADD(conf, conf, sync);
	CHECK_ADD(conf, conf, select_max);
	CHECK_ADD(conf, conf, table_prefix);
	CHECK_ADD(conf, conf, table_charset);
	CHECK_ADD(conf, conf, table_collation);
	
	return true;
}

/*************************************************************************/

Directives::Connection::Connection() : ConfBlock("connection", true),
		hostname("hostname", "localhost", true, 1, misc::DOMAIN_MAX, ""),
		port("port", 0, true),
		unix_socket("unix-socket", "", true, 0, NAME_MAX, ""),
		username("username", core::program_name_lc, true, 0, 16, ""),
		password("password", "", true, 0, 255, ""),
		database("database", core::program_name_lc, true, 1, 64, ""),
		timeout("timeout", 5 /* 5s */, true, 1, 60),
		charset("charset", "latin1", true, 1, 8, ""),
		log_queries("log-queries", false, true),
		log_warnings("log-warnings", false, true)
{
	
}

/*************************************************************************/

bool Directives::Connection::add(conf::Conf &conf, conf::Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, hostname);
	CHECK_ADD(new_conf, conf, port);
	CHECK_ADD(new_conf, conf, unix_socket);
	CHECK_ADD(new_conf, conf, username);
	CHECK_ADD(new_conf, conf, password);
	CHECK_ADD(new_conf, conf, database);
	CHECK_ADD(new_conf, conf, timeout);
	CHECK_ADD(new_conf, conf, log_queries);
	CHECK_ADD(new_conf, conf, log_warnings);
	CHECK_ADD(new_conf, conf, charset);
	
	return true;
}

/*************************************************************************/

bool Directives::Connection::apply_changes(conf::Conf &conf)
{
	if (hostname.has_value_changed() || port.has_value_changed() ||
			unix_socket.has_value_changed() || timeout.has_value_changed())
	{
		me->conn().host(hostname);
		me->conn().port(port);
		me->conn().unix_socket(unix_socket);
		me->conn().user(username);
		me->conn().password(password);
		me->conn().db(database);
		me->conn().connect_timeout(timeout);
		
		if (!me->conn().reconnect())
			me->core().log(CriticalMySQLConn(),
					"Reconnection to the MySQL server failed");
	} 
	else
	{
		if (username.has_value_changed() || password.has_value_changed())
			me->conn().change_user(username, password, database);
		else if (database.has_value_changed())
			me->conn().select_db(database);
		
		if (charset.has_value_changed())
			me->conn().charset(charset);
	}
	
	me->conn().log_queries(log_queries);
	me->conn().log_warnings(log_warnings);
	
	return ConfBlock::apply_changes(conf);
}

/*************************************************************************/

Directives::ReconnectFrequency::ReconnectFrequency() :
		ConfDuration("reconnect-frequency", 120 /* 2m */, true, 10, 3600)
{
	
}

/*************************************************************************/

bool Directives::ReconnectFrequency::apply_change(conf::Conf &conf)
{
	me->recotask().timeout(value());
	
	return ConfDuration::apply_change(conf);
}

/*************************************************************************/

Directives::Sync::Sync() : ConfBlock("sync", true), frequency(),
		blacklist("blacklist", conf::ConfStrCharA::container_type(), true, 1000),
		mem_usage("mem-usage", false, true)
{
}

/*************************************************************************/

bool Directives::Sync::add(conf::Conf &conf, conf::Conf &new_conf)
{
	CHECK_ADD(new_conf, conf, frequency);
	CHECK_ADD(new_conf, conf, blacklist);
	CHECK_ADD(new_conf, conf, mem_usage);
	
	return true;
}

/*************************************************************************/

Directives::Sync::Frequency::Frequency() :
		ConfDuration("frequency", 300 /* 5m */, true, 30, 2147483647)
{
	
}

/*************************************************************************/

bool Directives::Sync::Frequency::apply_change(conf::Conf &conf)
{
	me->synctask().timeout(value());
	
	return ConfDuration::apply_change(conf);
}

/*************************************************************************/

} /* namespace mysql */

/*************************************************************************/
