/* Database interface.
 *
 * MySQL database module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef DATABASE_H
#define DATABASE_H

/*************************************************************************/

#include <databases/database.h>
#include <databases/field.h>
#include <databases/record.h>
#include <misc/string.h>
#include <mysql-cli/result.h>

/*************************************************************************/
	
namespace mysql 
{

class Field;

/*************************************************************************/

class Database : public databases::Database
{
	public:
		/// Free all databases.
		static void free_all();
		/// Sync all databases.
		static void sync_all();
	
		/// Constructor.
		explicit Database(modules::Core &core) : databases::Database(core),
				m_tname(0), m_eflist(0), m_iflist(0) { }
		/// Destructor.
		/** Unregisters the database and calls operator delete on each field
		  * that was added to the database. */
		virtual ~Database();
		
		/// Overrides databases::Database::field_create().
		virtual databases::Field *field_create(char const *name,
				uint16_t minver, uint16_t type, size_t size, uint16_t attrs,
				char const *def, uint16_t tag = 0);
		
		/// Overrides Database::insert().
		virtual bool insert(databases::Record *record);
		/// Overrides Database::update_s().
		virtual bool update_s(databases::Record const *record,
				databases::Field const *field, char const *value);
		/// Overrides Database::update_i().
		virtual bool update_i(databases::Record const *record,
				databases::Field const *field, int32_t value);
		/// Overrides Database::update_ui().
		virtual bool update_ui(databases::Record const *record,
				databases::Field const *field, uint32_t value);
		/// Overrides Database::update_ut().
		virtual bool update_ut(databases::Record const *record,
				databases::Field const *field, time_t value);
		/// Overrides Database::update_ri().
		virtual bool update_ri(databases::Record const *record,
				databases::Field const *field,
				databases::Record::identifier_type value);
		/// Overrides Database::update_multi().
		virtual bool update_multi(databases::Record const *record,
				unsigned int valcount, va_list args);
		/// Overrides Database::update_owner().
		virtual bool update_owned(databases::Field const *field,
				databases::Record::identifier_type id,
				databases::Record::identifier_type new_id);
		/// Overrides Database::delete_().
		virtual bool delete_(databases::Record const *record);
		/// Overrides Database::delete_owned().
		virtual bool delete_owned(databases::Field const *field,
				databases::Record::identifier_type id);
		/// Overrides Database::truncate().
		virtual bool truncate();
		
		/// Overrides databases::Database::supports().
		virtual modules::Interface *supports(modules::PSIID sid);
		
		/// Overrides databases::Database::readable().
		virtual bool readable() const;
		/// Overrides databases::Database::writable().
		virtual bool writable() const;
		
		/// Returns the table name.
		char const *tname() const { return m_tname; }
	protected:
		/// Overrides databases::Database::generate_expfields().
		virtual void generate_expfields();
		/// Overrides databases::Database::generate_impfields().
		virtual void generate_impfields();
		
		/// Overrides databases::Database::reg().
		virtual bool reg();
		/// Overrides databases::Database::start_real().
		virtual bool start_real();
	private:
		/// Database list type.
		typedef std::list<Database *> mycontainer_type;
		/// Database list.
		static mycontainer_type g_mylist;
		
		/// The table name (for mysql).
		misc::cstring m_tname;
		/// The exported fields (for mysql).
		misc::cstring m_eflist;
		/// The imported fields (for mysql).
		misc::cstring m_iflist;
		
		/* Initial import functions */
		bool initial_import();
		/* Import functions */
		bool import(bool initial = false);
		bool import_create(mysql_cli::Result &res, bool initial);
		bool import_delete(mysql_cli::Result &res);
		bool import_update(mysql_cli::Result &res);
		bool import_update_record(databases::Record *record, MYSQL_ROW row,
				bool &is_error);
		void sync();
		/* Export functions */
		bool real_update_s(databases::Record const *record, Field const *field,
				char const *value);
		bool real_update_i(databases::Record const *record, Field const *field,
				int32_t value);
		bool real_update_ui(databases::Record const *record,
				Field const *field, uint32_t value);
		bool real_update_ut(databases::Record const *record,
				Field const *field, time_t value);
		bool real_update_ri(databases::Record const *record,
				Field const *field, databases::Record::identifier_type value);
		bool real_delete(databases::Record const *record);
		bool real_delete_id(databases::Record::identifier_type id);
		bool real_truncate();
		bool update_table();
		/* Misc functions */
		void update_memusage();
};

/*************************************************************************/

class Field : public databases::Field
{
	public:
		/// Constructor.
		Field() : databases::Field(), m_fname(0) { }
		
		/// Overrides databases::Field::setup().
		virtual bool setup();
		
		/// Returns the field name for MySQL.
		char const *fname() const { return m_fname; }
	private:
		misc::cstring m_fname;
};

/*************************************************************************/

class DBRemove : public databases::DBRemove
{
	public:
		/// Implements DBRemove::remove().
		virtual bool remove(std::string const &database);
		/// Implements DBRemove::clear().
		virtual bool clear();
};

/*************************************************************************/

} /* namespace mysql */

/*************************************************************************/

#endif /* DATABASE_H */
