/* SMTP socket.
 *
 * SMTP mail transport module (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <modules/core.h>

#include "conf.h"
#include "decl.h"
#include "module.h"
#include "smtp.h"

using namespace std;

/*************************************************************************/

namespace smtp_cli
{

/*************************************************************************/

SMTPSocket::~SMTPSocket()
{
	me->remove_sock(this);
}

/*************************************************************************/

SMTPSocket &SMTPSocket::operator=(SMTPSocket const &right)
{
	return *this;
}

/*************************************************************************/

bool SMTPSocket::on_connect(int error)
{
	if (error)
	{
		me->core().log(ErrorSMTPCli(), "While sending e-mail to %s: "
				"connect error: %s", m_mail.to_address().c_str(),
				strerror(error));

		close();
		delete this;

		return false;
	}
	
	
	if (me->core().debuglevel())
		me->core().log(DebugSMTPCli(),
				"Socket %d is connected (mail %s)", fd(),
				m_mail.to_address().c_str());
	
	return true;
}

/*************************************************************************/

bool SMTPSocket::on_read()
{	
	while (1)
	{
		ssize_t ret;
		
		if ((ret = read_inbuf()) == -1)
		{
			if (errno != EAGAIN && errno != EWOULDBLOCK)
			{
				me->core().log(ErrorSMTPCli(), "While sending e-mail to %s: "
						"read error: %s", m_mail.to_address().c_str(),
						strerror(errno));
				
				close();
				delete this;
				
				return false;
			}
			
			break;
		}
		
		if (ret == 0)
		{
			if (m_state != msContents)
				me->core().log(ErrorSMTPCli(), "While sending e-mail to %s: "
						"remote end initiated shutdown during transmission",
						m_mail.to_address().c_str());

			close();
			delete this;
			
			return false;
		}
		
		while (1)
		{
			char line[1024];
			
			ret = read_line(line, sizeof(line));
			
			if (ret == -1)
			{
				me->core().log(ErrorSMTPCli(), "While sending e-mail to %s: "
						"an invalid answer was received from the server: "
						"line too long", m_mail.to_address().c_str());
				
				close();
				delete this;
				
				return false;
			}
				
			if (ret == 0)
				break;
			
			if (me->core().debuglevel())
				me->core().log(DebugSMTPCli(), "Socket %d: Received: %s",
						fd(), line);
			
			if (!process(line))
				return false;
		}
	}
	
	return true;
}

/*************************************************************************/

bool SMTPSocket::on_write()
{
	if (wbuf()->empty())
		return true;
	
	if (write_outbuf() == -1)
	{
		me->core().log(ErrorSMTPCli(), "While sending e-mail to %s: "
				"write error: %s", m_mail.to_address().c_str(),
				strerror(errno));

		close();
		delete this;

		return false;
	}
	
	return true;
}

/*************************************************************************/

bool SMTPSocket::on_writing()
{
	if (wbuf()->empty())
		return true;
	
	if (write_outbuf() == -1)
	{
		/* No need to close here; will be closed as needed by the
		 * one who called the write() function */
		/* close(); */
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool SMTPSocket::process(char const *line)
{
	char *end;
	unsigned long num = strtol(line, &end, 10);
	
	if (end - line != 3 || (*end != ' ' && *end != '-'))
	{
		me->core().log(ErrorSMTPCli(), "While sending e-mail to %s: "
				"an invalid answer was received from the server: "
				"syntax incorrect", m_mail.to_address().c_str());

		close();
		delete this;
		
		return false;
	}
	
	if (num < 100 || num > 599)
	{
		me->core().log(ErrorSMTPCli(), "While sending e-mail to %s: "
				"an invalid answer was received from the server: "
				"invalid numeric", m_mail.to_address().c_str());
		
		close();
		delete this;
		
		return false;
	}
	
	if (num >= 400)
	{
		me->core().log(ErrorSMTPCli(), "While sending e-mail to %s: "
				"server returned an error: %s",
				m_mail.to_address().c_str(), ++end);
		
		close();
		delete this;
		
		return false;
	}
	
	switch (m_state)
	{
		case msNothing:
			if (printf(poLine, "HELO %s", me->cd().fqdn.value().c_str()) <= 0)
				return false;
			m_state = msHelo;
			break;
		case msHelo:
			if (printf(poLine, "MAIL FROM: <%s>", m_mail.sender().c_str()) <= 0)
				return false;
			m_state = msMailFrom;
			break;
		case msMailFrom:
			if (printf(poLine, "RCPT TO: <%s>", m_mail.to_address().c_str()) <= 0)
				return false;
			m_state = msRcptTo;
			break;
		case msRcptTo:
			if (printf(poLine, "DATA") <= 0)
				return false;
			m_state = msData;
			break;
		case msData:
		{
			ssize_t ret = write(m_mail.data().c_str(), m_mail.data().size());
			
			if (ret == -1)
			{
				me->core().log(ErrorSMTPCli(), "While sending e-mail to %s: "
						"write error: %s", m_mail.to_address().c_str(),
						strerror(errno));
				
				close();
				delete this;
				
				return false;
			}
			else if ((size_t)ret < m_mail.data().size())
			{
				me->core().log(ErrorSMTPCli(), "While sending e-mail to %s: "
						"write error: buffer is full",
						m_mail.to_address().c_str());
				
				close();
				delete this;
				
				return false;
			}
			
			if (printf(poLine, ".") <= 0)
				return false;
			m_state = msContents;
		}
			break;
		case msContents:
			if (printf(poLine, "QUIT") <= 0)
				return false;

			close();
			delete this;
			
			return false; /* We've finished, so stop */
		default:
			break;
	}
	
	return true;
}

/*************************************************************************/

ssize_t SMTPSocket::write_line(char const *line, bool inbuf)
{
	ssize_t ret = TCPSocket::write_line(line, inbuf);
	
	if (ret == -1)
	{
		me->core().log(ErrorSMTPCli(), "While sending e-mail to %s: "
				"write error: %s", m_mail.to_address().c_str(),
				strerror(errno));
		
		close();
		delete this;
	}
	else if (ret == 0)
	{
		if (status() & ssConnected)
		{
			me->core().log(ErrorSMTPCli(), "While sending e-mail to %s: "
					"write error: buffer is full",
					m_mail.to_address().c_str());

			close();
			delete this;
		}
	}
	else
	{
		if (me->core().debuglevel())
			me->core().log(DebugSMTPCli(), "Socket %d: Sent: %s", fd(), line);
	}
	
	return ret;
}

/*************************************************************************/

} /* namespace smtp_cli */

/*************************************************************************/
