/* Database interface.
 *
 * Twilight database module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <databases/collection.h>
#include <databases/field.h>
#include <databases/record.h>
#include <log/levels.h>
#include <misc/misc.h>
#include <modules/core.h>

#include "conf.h"
#include "decl.h"
#include "database.h"
#include "dbfile.h"
#include "module.h"

using namespace std;

using databases::Collection;
using databases::Field;
using databases::Record;

/*************************************************************************/

namespace twilight
{

/*************************************************************************/

Database::twcontainer_type Database::g_twlist;

/*************************************************************************/

Database::~Database()
{
	g_twlist.remove(this);
}

/*************************************************************************/

void Database::free_all()
{
	twcontainer_type::iterator it, end;
	
	for (it = g_twlist.begin(), end = g_twlist.end(); it != end; ++it)
		delete (*it);
	
	g_twlist.clear();
}

/*************************************************************************/

bool Database::initial_import()
{	
	/* If the database doesn't exist, don't try to read it! ;) */
	bool exists = me->index().exists(m_name);
	
	/* Read header information */
	if (exists && (!readable() || !read_header()))
		return false;
	
	/* Check if we need to DO initial import, based on version numbers. */
	if (version(vtImport) >= version(vtMinimum) &&
			(!version(vtMaximum) || version(vtImport) <= version(vtMaximum)))
	{
		/* Say to the collection initial import is started. */
		if (!coll()->begin_initial())
		{
			me->core().log(CriticalTwilight(),
					"Aborting initial import (initialization failed)");
			return false;
		}
		
		if (exists && (!readable() || !read_data()))
			return false;
		
		/* Say to the collection initial import is finished. */
		if (!coll()->end_initial())
		{
			me->core().log(CriticalTwilight(),
					"Aborting initial import (finalization failed)");
			return false;
		}
	}
	
	/* Woohoo! Initial import is done, set the corresponding status flag
	 * (even if the version number didn't match). */
	m_status |= dsInitial;
	
	/* Generate the list of imported fields again now that initial
	 * loading is done. */
	generate_impfields();
	
	if (me->core().debuglevel())
		me->core().log(DebugTwilight(), "Loaded database %s", m_name.c_str());
	
	return true;
}

/*************************************************************************/

bool Database::reg()
{
	g_twlist.push_back(this);
	
	return databases::Database::reg();
}

/*************************************************************************/

bool Database::read_data()
{
	DBReader dbd(m_dbdname);
	
	if (!dbd.open())
		return false;
	
	uint32_t count;
	
	if (!(dbd.read_count(&count)))
		return false;
	
	Field *field;
	uint16_t fieldid, type;
	
	char const *svalue;
	int32_t ivalue;
	uint32_t uivalue;
	time_t utvalue;
	Record::identifier_type rivalue;
	
	for (uint32_t i = 0; i < count; ++i)
	{
		Record *record;
		
		record = coll()->create();
		
		if (record)
			record->begin_update();
		
		for (;;)
		{
			if (!dbd.read_valueinfo(&fieldid, &type))
				return false;
			
			if ((fieldid == DBFile::end_fieldid) && (type == DBFile::end_type))
				break;
			
			if (fieldid != DBFile::id_fieldid)
			{
				if (fieldid >= m_fieldids.size())
				{
					me->core().log(ErrorTwilightDB(),
							"Field ID %u is out of range in database %s"
							" (corrupted file?)", fieldid, m_name.c_str());
					return false;
				}
				
				field = m_fieldids[fieldid];
			} else {
				field = 0;
			}
			
			if (field && field->type() != type)
			{
				me->core().log(ErrorTwilightDB(),
						"Type mismatch for field %s in database %s "
						"(is %u, got %u)", field->name(), m_name.c_str(),
						field->type(), type);
				return false;
			}
			
			switch (type)
			{
				case Field::ftString:
					if (!dbd.read_string(&svalue))
						return false;
					
					if (field && record)
						record->set_s(field, svalue);
					break;
				case Field::ftInteger:
					if (!dbd.read_integer(&ivalue))
						return false;
					
					if (field && record)
						record->set_i(field, ivalue);
					break;
				case Field::ftUnsigned:
					if (!dbd.read_unsigned(&uivalue))
						return false;
					
					if (field && record)
						record->set_ui(field, uivalue);
					break;
				case Field::ftUnixTime:
					if (!dbd.read_time(&utvalue))
						return false;
					
					if (field && record)
						record->set_ut(field, utvalue);
					break;
				case Field::ftIdentifier:
					if (!dbd.read_id(&rivalue))
						return false;
					
					if (fieldid == DBFile::id_fieldid)
					{
						if (record)
							record->id(rivalue);
						else
							me->core().log(ErrorTwilight(),
									"Ignoring record #" RECORD_FORMAT
									" from database %s", rivalue,
									m_name.c_str());
					} else if (field && record)
						record->set_ri(field, rivalue);
					break;
				default:
					me->core().log(ErrorTwilightDB(),
							"Unsupported field type %u in database %s",
							type, m_name.c_str());
					return false;
			}
		}
		
		if (record && !record->end_update())
		{
			me->core().log(InfoTwilight(), "Deleting bad record #" RECORD_FORMAT
					" from database %s", record->id(), m_name.c_str());
			
			delete record;
			continue;
		}
		
		if (record && !coll()->add(record, 0))
		{
			me->core().log(InfoTwilight(), "Deleting record #" RECORD_FORMAT
					" from database %s (could not be added to collection)",
					record->id(), m_name.c_str());
			
			delete record;
		}
	}
	
	return true;
}

/*************************************************************************/

bool Database::read_header()
{
	DBReader dbh(m_dbhname);
	
	/* Open the file */
	if (!dbh.open())
		return false;
	
	/* Set the import version */
	uint16_t version;
	
	if (!dbh.read_version() || !dbh.read_dbinfo(&version))
		return false;
	
	this->version(vtImport, version);
	
	/* Generate the list of imported fields. */
	generate_impfields();
	
	/* Gets the list of fields */
	databases::Field *field;
	
	uint16_t fieldid;
	char const *name;
	
	for (;;)
	{
		if (!dbh.read_fieldinfo(&fieldid, &name))
			return false;
		
		if (fieldid == DBFile::end_fieldid)
			break;
		
		field = field_find(name, vtImport);
		/* Ignore fields that we don't import... */
		if (!field)
			continue;
		
		if (m_fieldids.size() < fields_type::size_type(fieldid + 1))
			m_fieldids.resize((fieldid + 1), 0);
		
		m_fieldids[fieldid] = field;
	}
	
	return true;
}

/*************************************************************************/

bool Database::save()
{
	/* Header file */
	DBWriter dbh(m_dbhname);
	
	if (!dbh.open())
		return false;
	
	/* Write database info */
	if (!dbh.write_version() || !dbh.write_dbinfo(version(vtExport)))
		return false;
	
	/* Write field info */
	for (uint16_t i = 0; i < m_expfields.size(); ++i)
		if (!dbh.write_fieldinfo(i, m_expfields[i]))
			return false;
	
	if (!dbh.write_endfields())
		return false;
	
	/* Data file */
	DBWriter dbd(m_dbdname);
	
	if (!dbd.open())
		return false;
	
	/* Write record count */
	if (!dbd.write_count(coll()->size()))
		return false;
	
	/* Write records */
	Collection::const_iterator it, end;
	
	for (it = coll()->begin(), end = coll()->end(); it != end; ++it)
	{
		/* Write the identifier first ;) */
		if (!dbd.write_valueinfo(DBFile::id_fieldid, Field::ftIdentifier) ||
				!dbd.write_id((*it)->id()))
			return false;
		
		for (uint16_t i = 0; i < m_expfields.size(); ++i)
		{
			if (!dbd.write_valueinfo(i, m_expfields[i]->type()))
				return false;
			
			switch (m_expfields[i]->type())
			{
				case Field::ftString:
					if (!dbd.write_string((*it)->get_s(m_expfields[i])))
						return false;
					break;
				case Field::ftInteger:
					if (!dbd.write_integer((*it)->get_i(m_expfields[i])))
						return false;
					break;
				case Field::ftUnsigned:
					if (!dbd.write_unsigned((*it)->get_ui(m_expfields[i])))
						return false;
					break;
				case Field::ftUnixTime:
					if (!dbd.write_time((*it)->get_ut(m_expfields[i])))
						return false;
					break;
				case Field::ftIdentifier:
					if (!dbd.write_id((*it)->get_ri(m_expfields[i])))
						return false;
					break;
				default:
					return false;
			}
		}
		
		if (!dbd.write_endrecord())
			return false;
	}
	
	/* Commit the new files */
	if (!dbh.commit())
		return false;
	
	if (!dbd.commit())
	{
		dbh.revert();
		return false;
	}
	
	if (!dbh.cleanup() || !dbd.cleanup())
		return false;
	
	/* Update the index */
	if (!me->index().exists(m_name) && (!me->index().add(m_name) ||
			!me->save_index()))
	{
		me->core().log(CriticalTwilight(), "Can't save index file "
				"(the database files for %s will be ignored on reload!)",
				m_name.c_str());
		return false;
	}
	
	/* Mark the collection as unchanged ;) */
	/*coll()->status_remove(Collection::csChanges);*/
	
	/* And a bit of log */
	if (me->core().debuglevel())
		me->core().log(DebugTwilight(), "Saved database %s", m_name.c_str());
	
	return true;
}

/*************************************************************************/

bool Database::save_all()
{
	twcontainer_type::iterator it, end;
	
	for (it = g_twlist.begin(), end = g_twlist.end(); it != end; ++it)
	{
		/* Ensure we can save the databases */
		if (!((*it)->perms() & dpExport) || !(*it)->writable())
			continue;
	
		/* Don't save the database if its collection has no changes. */
		/*if (!((*it)->coll()->status() & Collection::csChanges))
			continue;*/
	
		/* Then save them. */
		if (!(*it)->save())
			me->core().log(CriticalTwilight(), "Saving database %s failed",
					(*it)->name().c_str());
	}
	
	return true;
}

/*************************************************************************/

modules::Interface *Database::supports(modules::PSIID sid)
{
	if (sid == sDelayed)
		return this;
	else
		return databases::Database::supports(sid);;
}

/*************************************************************************/

bool Database::start_real()
{
	/* Set file names */
	m_dbhname = m_name + ".dbh";
	m_dbdname = m_name + ".dbd";
	
	/* Do initial import if requested. */
	if ((perms() & dpInitial) && !initial_import())
	{
		me->core().log(CriticalTwilight(),
				"Can't do initial import of database %s", m_name.c_str());
		return false;
	}
	
	if ((perms() & dpExport) && writable() && !coll()->update_new())
	{
		me->core().log(CriticalTwilight(),
				"Update of new fields failed for database %s", m_name.c_str());
		return false;
	}
	
	/* Save the data, so that the eventual version change is set in stone.
	 * Also, this will automatically clear all the data in the database
	 * if there was no initial loading. */
	if ((perms() & dpExport) && writable() && !save())
	{
		me->core().log(CriticalTwilight(), "Can't save database %s",
				m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool DBRemove::clear()
{
	databases::XMLIndex::const_iterator it, end;
		
	for (it = me->index().begin(), end = me->index().end(); it != end; ++it)
	{
		std::string dbd = misc::absolute_path(me->cd().directory,
			(*it) + ".dbd");
		std::string dbh = misc::absolute_path(me->cd().directory,
				(*it) + ".dbh");
		
		if (!::remove(dbd.c_str()) == -1 || !::remove(dbh.c_str()) == -1)
		{
			errorf("Could not remove database file: %s", strerror(errno));
			return false;
		}
		
		me->core().log(InfoTwilight(), "Removed database %s", 
				it->c_str());
	}
	
	me->index().clear();
	
	if (!me->save_index())
	{
		errorf("All databases removed, but index couldn't be saved.");
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool DBRemove::remove(std::string const &database)
{
	if (me->index().exists(database))
	{
		if (!me->index().remove(database) || !me->save_index())
		{
			errorf("Could not remove database from index");
			return false;
		}
	} else {
		errorf("Database doesn't exist");
		return false;
	}
	
	std::string dbd = misc::absolute_path(me->cd().directory,
			database + ".dbd");
	std::string dbh = misc::absolute_path(me->cd().directory,
			database + ".dbh");
	
	if (::remove(dbd.c_str()) == -1 || ::remove(dbh.c_str()) == -1)
	{
		errorf("Database removed from index, "
				"but could not remove database files: %s", strerror(errno));
		return false;
	}
	
	me->core().log(InfoTwilight(), "Removed database %s", 
			database.c_str());
	return true;
}

/*************************************************************************/

} /* namespace twilight */

/*************************************************************************/
