/* Database file.
 *
 * Twilight database module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef DBFILE_H
#define DBFILE_H

/*************************************************************************/

#include <string>

#include <databases/record.h>
#include <misc/string.h>

/*************************************************************************/

namespace databases { class Field; }

/*************************************************************************/
	
namespace twilight
{

/*************************************************************************/

/// Base class for the database file reader/writer.

class DBFile
{
	public:
		static uint16_t const dbfile_version = 1;
		static uint16_t const id_fieldid = 0xFFFE;
		static uint16_t const end_fieldid = 0xFFFF;
		static uint16_t const end_type = 0xFFFF;
	
		/// Constructor setting the filename.
		DBFile(std::string const &filename); 
		/// Destructor.
		~DBFile();
		
		/// Closes the database file.
		void close();
		
		/// Returns the filename.
		std::string const &filename() const { return m_filename; }
		/// Returns the absolute filename.
		std::string const &absfilename() const { return m_absfilename; }
	protected:
		/// The filename, as set by the user.
		std::string m_filename;
		/// The absolute filename.
		std::string m_absfilename;
		
		/// The file handle.
		FILE *m_handle;
		
		/// Logs a file error (with errno appended if needed).
		void error(char const *msg, char const *filename) const; 
	private:
		DBFile(DBFile const &);
		DBFile &operator=(DBFile const &);
};

/*************************************************************************/

/// Database file reader.

class DBReader : public DBFile
{
	public:
		/// Constructor.
		DBReader(std::string const &filename) : DBFile(filename), m_str(0),
				m_maxlen(0), m_fieldname(0) { reserve(4096); }
		
		/// Opens the file.
		bool open();
		
		/* Functions for the header file. */
		
		/// Reads (and checks) version.
		bool read_version();
		/// Reads database info.
		bool read_dbinfo(uint16_t *version);
		/// Reads field info.
		bool read_fieldinfo(uint16_t *fieldid, char const **name,
				uint16_t *type, uint32_t *size, uint16_t *attrs,
				char const **defval);
		/// Reads field info, fills fewer info.
		bool read_fieldinfo(uint16_t *fieldid, char const **name);
		
		/* Functions for the data file */
		
		/// Reads the count (at the beginning of the file.
		bool read_count(uint32_t *count);
		/// This needs to be called to know what the next value is.
		bool read_valueinfo(uint16_t *fieldid, uint16_t *type);
		
		/* Functions for reading values */
		
		/// This can only read one string at once!
		bool read_string(char const **str);
		template <class T> bool read_integer(T rn);
		template <class T> bool read_unsigned(T rn);
		bool read_time(time_t *t);
		bool read_id(databases::Record::identifier_type *id);
	protected:
		/// String for read_string().
		misc::cstring m_str;
		/// Maximum length reserved for the string.
		size_t m_maxlen;
		
		/// Additional string for read_fieldinfo().
		misc::cstring m_fieldname;
		
		/// Reserves space for the string.
		void reserve(size_t len);
};

/*************************************************************************/

/// Database file writer.

/** The database file writer adds a .new to the filename it writes to, and
  * has a special functions commit()and cleanup() to actually replace the
  * old file with the new one. */

class DBWriter : public DBFile
{
	public:
		/// Constructor.
		DBWriter(std::string const &filename) : DBFile(filename),
				m_newafn(m_absfilename + ".new"),
				m_oldafn(m_absfilename + ".old") { }
		
		/// Opens the file.
		bool open();
		
		/// Replaces the new file with the old one (expects an open file).
		bool commit();
		
		/// Cleanups the .old file.
		bool cleanup();
		/// Revert the old file to the main one.
		bool revert();
		
		/* Functions for the header file. */
		
		/// Writes the database file version.
		bool write_version();
		/// Writes database info.
		bool write_dbinfo(uint16_t version);
		/// Writes field info.
		bool write_fieldinfo(uint16_t fieldid, char const *name, uint16_t type,
				uint32_t size, uint16_t attrs, char const *defval);
		/// Writes field info (Field * version).
		bool write_fieldinfo(uint16_t fieldid, databases::Field const *field);
		/// This needs to be called after all fields have been written.
		bool write_endfields();
		
		/* Functions for the data file */
		
		/// Writes the count (at the beginning of the file.
		bool write_count(uint32_t count);
		/// This needs to be called before writing the value
		bool write_valueinfo(uint16_t fieldid, uint16_t type);
		/// This needs to be called when all values have been set for a record.
		bool write_endrecord();
		
		/* Functions for writing values */
		
		bool write_string(char const *str);
		bool write_integer(int32_t n);
		bool write_unsigned(uint32_t n);
		bool write_time(time_t t);
		bool write_id(databases::Record::identifier_type id);
	protected:
		std::string m_newafn;
		std::string m_oldafn;
};

/*************************************************************************/

template <class T>
bool DBReader::read_integer(T rn)
{
	uint32_t n = 0;
	
	for (int i = 3; i >= 0; --i)
	{
		int c;
		
		if ((c = fgetc(m_handle)) == EOF)
		{
			error("Can't read from %s (corrupted file?)", m_absfilename.c_str());
			return false;
		}
		
		n <<= 8;
		n |= (c & 0xFF);
	}
	
	*rn = n;
	return true;
}

/*************************************************************************/

template <class T>
bool DBReader::read_unsigned(T rn)
{
	uint32_t n = 0;
	
	for (int i = 3; i >= 0; --i)
	{
		int c;
		
		if ((c = fgetc(m_handle)) == EOF)
		{
			error("Can't read from %s (corrupted file?)",
					m_absfilename.c_str());
			return false;
		}
		
		n <<= 8;
		n |= (c & 0xFF);
	}
	
	*rn = n;
	return true;
}

/*************************************************************************/

} /* namespace twilight */

/*************************************************************************/

#endif /* DBFILE_H */
