/* MemoServ records management.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#include <system/headers.h>

#include <core/directives.h>
#include <core/epona.h>
#include <databases/collection.h>
#include <databases/database.h>
#include <databases/field.h>
#include <databases/record.h>
#include <mail/mail.h>

#include "epona.h"
#include "language.h"
#include "memoserv.h"
#include "nickserv.h"

#include "services.h"

using namespace std;

using databases::Collection;
using databases::Database;
using databases::Field;
using databases::Record;

/*************************************************************************/

Memos memos;

databases::Field *Memo::f_owner = 0;
databases::Field *Memo::f_number = 0;
databases::Field *Memo::f_flags = 0;
databases::Field *Memo::f_sender = 0;
databases::Field *Memo::f_target = 0;
databases::Field *Memo::f_created = 0;
databases::Field *Memo::f_text = 0;

/*************************************************************************/

Memo::Memo() : databases::Record(), m_owner(0), m_number(0),
		m_flags(mfUnread), m_sender(), m_target(), m_created(time(0)), m_text()
{
}

/*************************************************************************/

Memo::Memo(identifier_type id) : databases::Record(id), m_owner(0),
		m_number(0), m_flags(mfUnread), m_sender(), m_target(),
		m_created(time(0)), m_text()
{
}

/*************************************************************************/

bool Memo::check_text() const
{
	/* Don't allow \r character */
	if (strchr(m_text, '\r'))
		return false;
	
	if (!(flags() & mfMultiLine))
	{
		if (strchr(m_text, '\n'))
		{
			epona->log(ErrorMemoServList(), "Non-multiline memo #" RECORD_FORMAT
					" with newline character", id());
			return false;
		}
		
		if (strlen(m_text) > line_len)
		{
			epona->log(ErrorMemoServList(), "Memo #" RECORD_FORMAT
					" has a too long text", id());
			return false;
		}
		
		return true;
	}
	
	char const *beg = m_text, *end;
	
	while ((end = strchr(beg, '\n')))
	{
		if ((size_t)(end - beg) > line_len)
		{
			epona->log(ErrorMemoServList(), "Multiline memo #" RECORD_FORMAT
					" has a too long line", id());
			return false;
		}
		
		beg = end + 1;
	}
	
	return true;
}

/*************************************************************************/

time_t Memo::created(time_t created)
{
	if (db_status() & dbsUpdating)
	{
		/* Don't allow dates in the future */
		if (!created || created > time(0))
			created = time(0);
	}
	
	if (m_created != created) 
	{
		if (!memos.change_ut(this, f_created, created))
			return false;
		
		m_created = created;
	}
	
	return m_created;
}

/*************************************************************************/

bool Memo::end_update()
{
	if (!Record::end_update())
		return false;
	
	if (!check_text())
		return false;
	
	/* If number is 0 ... */
	if (!m_number)
	{	
		uint32_t number = 0;
		
		/* Compute a number. */
		if (!m_owner->memos.memos().empty()) 
			number = (m_owner->memos.memos().back()->number() + 1);
		else
			number = 1;
		
		if (!memos.change_ui(this, f_number, number))
			return false;
		
		m_number = number;
	}
	
	return true;
}

/*************************************************************************/

uint16_t Memo::flags(uint16_t flags)
{
	/* Assume (db_status() & dbsUpdating) == true */
	
	flags &= ~mfUnused;
	
	if (m_flags != flags) 
	{
		m_flags = flags;
	}
	
	return m_flags;
}

/*************************************************************************/

uint16_t Memo::flags_add(uint16_t flags)
{
	if ((m_flags & flags) != flags) 
	{
		uint16_t tflags = (m_flags | flags);
		
		if (!memos.change_ui(this, f_flags, tflags))
			return m_flags;
		
		m_flags = tflags;
	}
	
	return m_flags;
}

/*************************************************************************/

uint16_t Memo::flags_remove(uint16_t flags)
{
	if ((m_flags & flags) != 0) 
	{
		uint16_t tflags = (m_flags & (~flags));
		
		if (!memos.change_ui(this, f_flags, tflags))
			return m_flags;
		
		m_flags = tflags;
	}
	
	return m_flags;
}
    
/*************************************************************************/

Record::identifier_type Memo::get_ri(Field const *field) const
{
	if (field == f_owner)
		return (m_owner ? m_owner->id() : 0);
	else
		return Record::get_ri(field);
}

/*************************************************************************/

char const *Memo::get_s(Field const *field) const
{
	if (field == f_sender)
		return sender();
	else if (field == f_target)
		return target();
	else if (field == f_text)
		return text();
	else
		return Record::get_s(field);
}

/*************************************************************************/

uint32_t Memo::get_ui(Field const *field) const
{
	if (field == f_flags)
		return flags();
	else if (field == f_number)
		return number();
	else
		return Record::get_ui(field);
}

/*************************************************************************/

time_t Memo::get_ut(Field const *field) const
{
	if (field == f_created)
		return created();
	else
		return Record::get_ut(field);
}

/*************************************************************************/

size_t Memo::memory_usage() const
{
	size_t result = Record::memory_usage();
	
	result += sizeof(Memo);
	
	result += strlen(m_sender) + 1;
	if (m_target)
		result += strlen(m_target) + 1;
	result += strlen(m_text) + 1;
	
	return result;
}

/*************************************************************************/

uint32_t Memo::number(uint32_t number)
{
	if (db_status() & dbsUpdating)
	{
		/* Don't allow number changes. */
		if (m_number)
			return m_number;
	}
	
	return (m_number = number);
}

/*************************************************************************/

NickGroup *Memo::owner(NickGroup *owner)
{
	if (m_owner)
		return m_owner;
	
	return (m_owner = owner);
}

/*************************************************************************/

Record::identifier_type Memo::owner_id(identifier_type owner_id)
{
	/* Assume (db_status() & dbsUpdating == true). */
	
	/* Don't allow owner changes. */
	if (m_owner)
		return m_owner->id();
	
	NickGroup *owner;
	
	/* Don't allow no ownership. */
	if (!owner_id || !(owner = (NickGroup *)groups.find(owner_id)))
	{
		epona->log(ErrorNickServList(), "Ignoring memo #" RECORD_FORMAT
				" with bad owner #" RECORD_FORMAT, id(), owner_id);
		
		/* Update error. */
		db_status_add(dbsUpdateError);
		/* Don't need to change the value, the record will
		 * be deleted. */
		return owner_id;
	}
	
	m_owner = owner;
	
	return m_owner->id();
}

/*************************************************************************/

char const *Memo::sender(char const *sender)
{
	if (db_status() & dbsUpdating)
	{
		/* Don't allow NULL sender */
		if (!sender)
		{
			/* If there's already a sender, just keep the old. */
			if (m_sender)
				return m_sender;
			/* Else, it is an error */
			db_status_add(dbsUpdateError);
			/* Return the NULL sender... */
			return sender;
		}
	}
	
	if (m_sender.compare(sender, sender_len) != 0)
	{	
		misc::cstring tsender(sender, sender_len);
		
		if (!memos.change_s(this, f_sender, tsender))
			return m_sender;
		
		m_sender.swap(tsender);
	}
	
	return m_sender;
}

/*************************************************************************/

Record::identifier_type Memo::set_ri(Field const *field,
		identifier_type value)
{
	if (!(db_status() & dbsUpdating))
		return Record::set_ri(field, value);
	
	if (field == f_owner)
		return owner_id(value);
	else
		return Record::set_ri(field, value);
}

/*************************************************************************/

char const *Memo::set_s(Field const *field, char const *value)
{
	if (!(db_status() & dbsUpdating))
		return Record::set_s(field, value);
	
	if (field == f_sender)
		return sender(value);
	else if (field == f_target)
		return target(value);
	else if (field == f_text)
		return text(value);
	else
		return Record::set_s(field, value);
}

/*************************************************************************/

uint32_t Memo::set_ui(Field const *field, uint32_t value)
{
	if (!(db_status() & dbsUpdating))
		return Record::set_ui(field, value);
	
	if (field == f_flags)
		return flags(value);
	else if (field == f_number)
		return number(value);
	else
		return Record::set_ui(field, value);
}

/*************************************************************************/

time_t Memo::set_ut(Field const *field, time_t value)
{
	if (!(db_status() & dbsUpdating))
		return Record::set_ut(field, value);
	
	if (field == f_created)
		return created(value);
	else
		return Record::set_ut(field, value);
}

/*************************************************************************/

char const *Memo::target(char const *target)
{
	if (m_target.compare(target, target_len) != 0)
	{	
		misc::cstring ttarget(target, target_len);
		
		if (!memos.change_s(this, f_target, ttarget))
			return m_target;
		
		m_target.swap(ttarget);
	}
	
	return m_target;
}

/*************************************************************************/

char const *Memo::text(char const *text)
{
	if (db_status() & dbsUpdating)
	{
		/* Don't allow NULL text */
		if (!text)
		{
			/* If there's already a text, just keep the old. */
			if (m_text)
				return m_text;
			/* Else, it is an error */
			db_status_add(dbsUpdateError);
			/* Return the NULL text... */
			return text;
		}
	}
	
	if (m_text.compare(text, text_len) != 0)
	{	
		misc::cstring ttext(text, text_len);
		
		if (!memos.change_s(this, f_text, ttext))
			return m_text;
		
		m_text.swap(ttext);
	}
	
	return m_text;
}

/*************************************************************************/

MemoInfo::MemoInfo(NickGroup *ng) : m_ng(ng), m_memos()
{
	
}

/*************************************************************************/

int16_t MemoInfo::memomax() const
{
	return m_ng->maxmemos();
}

/*************************************************************************/

int16_t MemoInfo::memomax(int16_t memomax)
{
	return m_ng->maxmemos(memomax);
}

/*************************************************************************/

bool Memos::add(databases::Record *record, int *err)
{
	if (!Collection::add(record, err))
		return false;
	
	Memo *m = (Memo *)record;
	
	if (!(status() & csInitial))
	{
		NickAlias *na;
		
		if (m->owner()->flags() & NickGroup::gfMemoReceive) 
		{
			VectorNA::const_iterator it, end;
			
			for (it = m->owner()->aliases().begin(),
					end = m->owner()->aliases().end(); it != end; ++it)
			{
				na = (NickAlias *)(*it);
				if (na->u && nick_identified(na->u))
					notice_lang(s_MemoServ, na->u, MEMO_NEW_MEMO_ARRIVED,
							m->sender(), s_MemoServ, m->number());
			}
		}
		
		if ((m->owner()->flags() & NickGroup::gfMemoForward) &&
				(m->owner()->flags() & NickGroup::gfAuth))
			mail_memo(m);
	}
	
	return true;
}

/*************************************************************************/

bool Memos::add_list(Record *record)
{
	Memo *m = (Memo *)record;
	MemoInfo *mi = &m->owner()->memos;
	
	MemoInfo::list_type::iterator it, end;
	
	for (it = mi->memos().begin(), end = mi->memos().end(); it != end; ++it)
	{
		Memo *m2 = (*it);
		
		/* The number already exists! Set a better number and retry. */
		if (m->number() == m2->number())
		{
			m->number(mi->memos().back()->number() + 1);
			return add_list(record);
		}
		
		if (m->number() < m2->number())
			break;
	}
	
	it = mi->memos().insert(it, m);
	
	if (!Collection::add_list(record))
	{
		mi->memos().erase(it);
		return false;
	}
	
	return true;
}

/*************************************************************************/

Record *Memos::create(Record::identifier_type id) const
{
	if (id)
		return new Memo(id);
	else
		return new Memo;
}

/*************************************************************************/

void Memos::mail_memo(Memo const *m) const
{
	if (!epona->cd().memoserv.forward && !mail::Mail::configured())
		return;
	
	User *u = 0;
	
	if (!(status() & csImport))
		u = finduser(m->sender());
	
	char subject[BUFSIZE];
	
	if (m->target())
		snprintf(subject, sizeof(subject),
				getstring2(m->owner(), MEMO_SET_FORWARD_SUBJECT_TARGET),
				m->target(), (char const *)epona->cd().services.network_name);
	else
		snprintf(subject, sizeof(subject),
				getstring2(m->owner(), MEMO_SET_FORWARD_SUBJECT), 
				(char const *)epona->cd().services.network_name);
	
	mail::Mail mail;
	
	if (u && u->na)
		mail.from_name(u->na->ng()->name());
	else
		mail.from_name(m->sender());
	if (u && u->na && u->na->ng()->email())
		mail.from_address(u->na->ng()->email());
	mail.subject(subject);
	mail.to_name(m->owner()->name());
	mail.to_address(m->owner()->email());
	if (u)
		mail.origin(u->host);
	else
		mail.origin(m->sender());
	
	mail.text_add(m->text());
	
	uint32_t mflags = 0;
	
	if (u && is_oper(u))
		mflags |= mail::Mail::mfNoQuota;
	if (m->owner()->flags() & NickGroup::gfAuth)
		mflags |= mail::Mail::mfAuthenticated;
	
	uint32_t err;
	std::string errdesc;
	
	if (!mail.send(mflags, &err, &errdesc))
	{
		if (err & mail::Mail::meQuota)
			epona->log(InfoMemoServ(),
					"Could not forward memo to %s: %s",
					m->owner()->email(), errdesc.c_str());
		else
			epona->log(ErrorMemoServ(),
					"Could not forward memo to %s: %s",
					m->owner()->email(), errdesc.c_str());
	}
}

/*************************************************************************/

void Memos::remove_list(Record *record)
{
	Memo *rm = (Memo *)record;
	MemoInfo *mi = &rm->owner()->memos;
	
	MemoInfo::list_type::iterator it, end;
	
	for (it = mi->memos().begin(), end = mi->memos().end(); it != end; ++it)
	{
		Memo *m = (*it);
		
		if (m != rm)
			continue;
	
		mi->memos().erase(it);
		break;
	}
	
	Collection::remove_list(record);
}

/*************************************************************************/

bool Memos::remove_owned_db(databases::Collection const *coll,
				databases::Record const *record, int *err)
{
	if (coll != &groups)
		return true;
	
	NickGroup *ng = (NickGroup *)record;
	
	if (ng->memos.memos().empty())
		return true;
	
	return Collection::remove_owned_db(coll, record, err);
}

/*************************************************************************/

bool Memos::remove_owned_list(databases::Collection const *coll,
				databases::Record const *record, int *err)
{
	if (coll != &groups)
		return true;
	
	NickGroup *ng = (NickGroup *)record;
	
	while (!ng->memos.memos().empty())
	{
		Memo *m = (Memo *)ng->memos.memos().back();
		
		m->db_status_remove(Record::dbsInsert);
		
		if (!remove(m, err))
			return false;
	}
	
	return true;
}

/*************************************************************************/

void Memos::setup_database(databases::Database *db)
{
	db->coll(this);
	db->name("nickserv_memos");
	db->version(Database::vtExport, 1);
	db->version(Database::vtMinimum, 1);
	db->perms_add(Database::dpInitial |
			Database::dpImport | Database::dpExport);
	
	Memo::f_owner = db->field_create("owner", 1, Field::ftIdentifier,
			0, Field::faOwner, "0", 1);
	Memo::f_number = db->field_create("number", 1, Field::ftUnsigned,
			4, 0, "0", 2);
	Memo::f_flags = db->field_create("flags", 1, Field::ftUnsigned,
			2, 0, "1", 3);
	Memo::f_sender = db->field_create("sender", 1, Field::ftString,
			Memo::sender_len, 0, 0, 4);
	Memo::f_target = db->field_create("target", 1, Field::ftString,
			Memo::target_len, 0, 0, 5);
	Memo::f_created = db->field_create("created", 1, Field::ftUnixTime,
			0, 0, "0", 6);
	Memo::f_text = db->field_create("text", 1, Field::ftString,
			Memo::text_len, 0, 0, 7);
	
	if (!db->field_add(Memo::f_owner) || 
			!db->field_add(Memo::f_number) ||
			!db->field_add(Memo::f_flags) || 
			!db->field_add(Memo::f_sender) || 
			!db->field_add(Memo::f_target) || 
			!db->field_add(Memo::f_created) ||
			!db->field_add(Memo::f_text))
		epona->fatal("Can't add fields to database nickserv_memos");
	
	this->db(db);
}

/*************************************************************************/

void Memos::setup_owner()
{
	groups.set_owned(*this);
}

/*************************************************************************/
