/* Network bans.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * Based on the original code of Services by Andy Church.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
#include <system/headers.h>

#if HAVE_ARPA_INET_H
# include <arpa/inet.h>
#endif

#include <core/directives.h>
#include <core/epona.h>
#include <log/levels.h>
#include <misc/misc.h>
#include <scheduler/scheduler.h>

#include "epona.h"
#include "netbans.h"

using namespace std;

/*************************************************************************/

NetForce netforce;

/*************************************************************************/

namespace
{

/*************************************************************************/

LOGLEVELH(DebugNetBan, xlog::LogDebug, "Debug::NetBan", "debug: netban: ");
LOGLEVELH(InfoNetBan, xlog::LogInfo, "Info::NetBan", "netban: ");

/*************************************************************************/

class NickChanHash : public containers::Hash<NickChanNetBan *, char const *>
{
	public:
		virtual int compare(char const *const &key,
				NickChanNetBan *const &value) const 
				{ return stricmp(key, value->mask()); }	
		virtual uint32_t hash(char const *const &key, uint32_t max) const
				{ return (misc::crc16_str(key, misc::ctLowerCase) % max); }
		virtual bool match(char const *const &key,
				NickChanNetBan *const &value) const
				{ return !stricmp(key, value->mask()); }
};

/*************************************************************************/

class HostHash : public containers::Hash<UmaskNetBan *, char const *>
{
	public:
		virtual int compare(char const *const &key,
				UmaskNetBan *const &value) const 
				{ return stricmp(key, value->host()); }	
		virtual uint32_t hash(char const *const &key, uint32_t max) const
				{ return (misc::crc16_str(key, misc::ctLowerCase) % max); }
		virtual bool match(char const *const &key,
				UmaskNetBan *const &value) const
				{ return !stricmp(key, value->host()); }
};

/*************************************************************************/

class IPHash : public containers::Hash<IPNetBan *, uint32_t>
{
	public:
		virtual int compare(uint32_t const &key,
				IPNetBan *const &value) const 
				{ return ((key == value->ip().s_addr) ? 0 :
						((key < value->ip().s_addr) ? -1 : 1)); }	
		virtual uint32_t hash(uint32_t const &key, uint32_t max) const
				{ return ((key >> 19) & 0xFFF); }
		virtual bool match(uint32_t const &key,
				IPNetBan *const &value) const
				{ return (key == value->ip().s_addr); }
};

/*************************************************************************/

class RealHash : public containers::Hash<RealNetBan *, char const *>
{
	public:
		virtual int compare(char const *const &key,
				RealNetBan *const &value) const 
				{ return stricmp(key, value->mask()); }	
		virtual uint32_t hash(char const *const &key, uint32_t max) const
				{ return (misc::crc16_str(key, misc::ctLowerCase) % max); }
		virtual bool match(char const *const &key,
				RealNetBan *const &value) const
				{ return !stricmp(key, value->mask()); }
};

/*************************************************************************/

class LeaveTask : public scheduler::TaskTimeout
{
	public:
		/// Constructor.
		LeaveTask(std::string const &name) :
				TaskTimeout(epona->cd().services.inhabit, epona, name, "") { }

		/// Overrides Task::run().
		virtual int run()
				{ send_cmd(s_GlobalNoticer, "PART %s", m_name.c_str()); return rFree; }

		/// Overrides Task::description().
		virtual std::string description() const
				{ return std::string("Leave channel ") + name() +
						" after channel network ban"; }
};

/*************************************************************************/

NickChanHash nc_hash;
HostHash host_hash;
IPHash ip_hash;
RealHash real_hash;

/*************************************************************************/

} /* anonymous namespace */

/*************************************************************************/

ChanNameNetBanner::ChanNameNetBanner() : ChanNetBanner(),
		m_hlist(nc_hash), m_list()
{
	
}

/*************************************************************************/

ChanNameNetBanner &ChanNameNetBanner::operator=(ChanNameNetBanner const &right)
{
	if (this != &right)
	{
		ChanNetBanner::operator=(right);
		
		m_hlist = right.m_hlist;
		m_list = right.m_list;
	}
	
	return *this;
}

/*************************************************************************/

bool ChanNameNetBanner::add(NetBan *ban_)
{
	NickChanNetBan *ban = (NickChanNetBan *) ban_;
	
	if (!(ban->flags() & NetBan::bfComplete) ||
			!(ban->flags() & NickChanNetBan::bfChannel))
		return false;
	
	if (!(ban->flags() & NickChanNetBan::bfWild))
	{
		if (m_hlist.add_sorted(ban->mask(), ban) == m_hlist.end())
			return false;
		
		if (epona->debuglevel() >= 2)
			epona->log(DebugNetBan(),
					"Added channel name network ban %s (flags: %X)",
					ban->mask(), ban->flags());
		return true;
	}
	
	list_type::const_iterator it, end;
	
	for (it = m_list.begin(), end = m_list.end(); it != end; ++it)
	{
		/* Ensure the same ban is not already in the list. */
		if (!stricmp((*it)->mask(), ban->mask()))
			return false;
	}
	
	if (m_list.insert(m_list.end(), ban) == m_list.end())
		return false;
	
	if (epona->debuglevel() >= 2)
		epona->log(DebugNetBan(),
				"Added channel name network ban %s (flags: %X)",
				ban->mask(), ban->flags());
	
	return true;
}

/*************************************************************************/

bool ChanNameNetBanner::enforce(Channel *channel) const
{
	NetBan *nb = matches(channel);
	
	if (nb)
	{
		epona->log(InfoNetBan(), "Enforcing network ban for channel %s (matches "
				" channel name mask %s)", channel->name, nb->mask());
		
		nb->apply();
	}
	
	return false;
}

/*************************************************************************/

NetBan *ChanNameNetBanner::matches(Channel const *channel) const
{
	hlist_type::const_iterator hit, hend = m_hlist.end();
	
	if ((hit = m_hlist.find(channel->name)) != hend)
		return (*hit);
	
	list_type::const_iterator it, end;
	NickChanNetBan *nb = 0;
	
	for (it = m_list.begin(), end = m_list.end(); it != end; ++it)
	{
		nb = *it;
		
		if (misc::irc_match(nb->mask(), channel->name))
			return nb;
	}
	
	return 0;
}

/*************************************************************************/

bool ChanNameNetBanner::remove(NetBan *ban_)
{
	NickChanNetBan *ban = (NickChanNetBan *) ban_;
	
	if (!(ban->flags() & NetBan::bfComplete) ||
			!(ban->flags() & NickChanNetBan::bfChannel))
		return false;
	
	if (!(ban->flags() & NickChanNetBan::bfWild))
	{
		if (!m_hlist.remove_sorted(ban->mask()))
			return false;
		ban->unapply();
		
		if (epona->debuglevel() >= 2)
			epona->log(DebugNetBan(), "Removed channel name network ban %s",
					ban->mask());
		return true;
	}
	
	list_type::iterator it, end;
	
	for (it = m_list.begin(), end = m_list.end(); it != end; ++it)
	{
		if ((*it) == ban)
		{
			m_list.erase(it);
			ban->unapply();
			
			epona->log(DebugNetBan(), "Removed channel name network ban %s",
					ban->mask());
			return true;
		}
	}
	
	return false;
}

/*************************************************************************/

IPNetBan::IPNetBan(NetBanner *nbr) : NetBan(nbr), m_ip(), m_netmask(),
		m_rmask(0)
{
	m_ip.s_addr = m_netmask.s_addr = 0;
}

/*************************************************************************/

IPNetBan::IPNetBan(IPNetBan const &nb) : NetBan(nb), m_ip(), m_netmask(),
		m_rmask(nb.m_rmask)
{
	m_ip = nb.m_ip;
	m_netmask = nb.m_netmask;
}

/*************************************************************************/

IPNetBan &IPNetBan::operator=(IPNetBan const &right)
{
	if (this != &right)
	{
		NetBan::operator=(right);
		
		m_ip = right.m_ip;
		m_netmask = right.m_netmask;
		m_rmask = right.m_rmask;
	}
	
	return *this;
}

/*************************************************************************/

void IPNetBan::apply() const
{
	if (!(m_flags & bfComplete))
		return;

#if defined(IRC_BAHAMUT) || defined(IRC_UNREAL)	
	uint16_t crc = crc16_str(m_mask);
#endif
	
#if defined(IRC_BAHAMUT)
	send_cmd(NULL, "AKILL %s * %d %s %ld :%s [#%X]",
			m_rmask.str(), 86400*2,
			m_creator.str(), (long) time(0), /* Shouldn't be set to
			                                  * the creation time! */
			m_reason.str(), crc);
#elif defined(IRC_UNREAL)
	send_cmd(NULL, "TKL + Z * %s %s %ld %ld :%s [#%X]",
			m_rmask.str(), m_creator.str(), 
			(long) time(NULL)+86400*2, /* Avoids filling the SZLINE list of servers
		    	                        * too much */
			(long) m_creation_time, m_reason.str(), crc);
#endif
}

/*************************************************************************/

void IPNetBan::generate_rmask(char const *mask)
{
	if (m_flags & bfCIDR4)
	{
		m_rmask.reserve(19);
		snprintf(m_rmask, 19, "%s/%u", inet_ntoa(m_ip),
				misc::netmask_to_cidr4(ntohl(m_netmask.s_addr)));
	} else if (m_flags & bfWild) {
		m_rmask = mask;
	} else {
		m_rmask = inet_ntoa(m_ip);
	}
}

/*************************************************************************/

bool IPNetBan::mask(char const *mask)
{
	if (m_flags & bfComplete)
		return false;
	
	if (!mask || !*mask)
		return false;
	
	char const *end = mask;
	int digits = 0, wild = 0, wildq = 0, dots = 0, slash = 0, other = 0;
	
	while (*end)
	{
		switch (*end++)
		{
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				digits++;
				break;
			case '.':
				dots++;
				break;
			case '/':
				slash++;
				break;
			case '?':
				wildq++;
			case '*':
				wild++;
				break;
			default:
				other++;
		}
	}
	
	/* Some sanity checks. */
	if (other || (!wild && dots < 3) || dots > 3 || (slash > 1) ||
			(wild && slash))
		return false;
	
	/* Don't ban the whole world. */
	if (wild && digits <= 1)
		return false;

#ifndef HAS_CIDR4_AKILL
	if (slash)
		return false;
#endif
	
	bool update_mask = true;
	misc::cstring mymask(mask);
	
#ifdef HAS_CIDR4_AKILL
	/* Check for some easy wild-to-CIDR conversions */
	if (wild && dots && !wildq && isdigit(*mask))
	{
		dots = 0; 
		end = mask;
		
		while (*end && (isdigit(*end) || *end == '.'))
		{
			if (*end++ == '.')
				++dots;
		}
		
		if (end != mask && !isdigit(*(end-1)))
		{
			char const *end2 = end;
			
			while (*end && (*end == '*' || *end == '.'))
				end++;
			
			if (!*end && dots)
			{
				mymask.reserve(19);
				char *end3 = stpncpy(mymask, mask, (end2-mask));
				
				switch (dots)
				{
					case 1:
						strcpy(end3, "0.0.0/8");
						break;
					case 2:
						strcpy(end3, "0.0/16");
						break;
					case 3:
						strcpy(end3, "0/24");
						break;
				}
				
				wild = 0;
				slash = 1;
				
				update_mask = false;
			}
		}
	}
#endif

	if (!wild)
	{	
		if (slash)
		{
			char *sep = strchr(mymask, '/');
			*sep++ = 0;
			
			if (!*sep)
				return false;
			
			char *end2;
			
			uint8_t cidr = strtoul(sep, &end2, 10);
			if (*end || (cidr < 1) || (cidr > 32))
				return false;
			
			m_netmask.s_addr = htonl(misc::cidr4_to_netmask(cidr));
			if (cidr != 32)
				m_flags |= bfCIDR4;
		} else {
			m_netmask.s_addr = misc::cidr4_to_netmask(32);
		}
		
		if (!inet_aton(mymask, &m_ip))
			return false;
		
		m_ip.s_addr &= m_netmask.s_addr;
	} else {
		m_flags |= bfWild;
	}
	
	generate_rmask(mask);
	m_mask = (update_mask ? m_rmask : mask);
	
	m_flags |= bfComplete;
	return true;
}

/*************************************************************************/

size_t IPNetBan::memory_usage() const
{
	size_t result = NetBan::memory_usage() - sizeof(NetBan) +
			sizeof(IPNetBan);
	
	if (m_flags & bfComplete)
		result += strlen(m_rmask) + 1;
	
	return result;
}

/*************************************************************************/

void IPNetBan::unapply() const
{
	if (!(m_flags & bfComplete))
		return;
	
#if defined(IRC_BAHAMUT)
	send_cmd(NULL, "RAKILL %s *", m_rmask.str());
#elif defined(IRC_UNREAL)
	send_cmd(NULL, "TKL - Z * %s %s", m_rmask.str(), 
			epona->cd().services.sname.value().c_str());
#endif
}

/*************************************************************************/

bool IPNetBan::wider(NetBan const *ban) const
{
	return false;
}

/*************************************************************************/

IPNetBanner::IPNetBanner() : UserNetBanner(),
		m_hlist(ip_hash), m_c4list(), m_wlist()
{
	
}

/*************************************************************************/

IPNetBanner &IPNetBanner::operator=(IPNetBanner const &right)
{
	if (this != &right)
	{
		UserNetBanner::operator=(right);
		
		m_hlist = right.m_hlist;
		m_c4list = right.m_c4list;
		m_wlist = right.m_wlist;
	}
	
	return *this;
}

/*************************************************************************/

bool IPNetBanner::add(NetBan *ban_)
{
	IPNetBan *ban = (IPNetBan *) ban_;
	
	if (!(ban->flags() & NetBan::bfComplete))
		return false;
	
	if (!(ban->flags() & IPNetBan::bfWild) &&
			!(ban->flags() & IPNetBan::bfCIDR4))
	{
		if (m_hlist.add_sorted(ban->ip().s_addr, ban) == m_hlist.end())
			return false;
		
		if (epona->debuglevel() >= 2)
			epona->log(DebugNetBan(), "Added IP network ban %s "
					"(real: %s; flags: %X)", ban->mask(), ban->rmask(),
					ban->flags());
		return true;
	}
	
	list_type::const_iterator it, end;
	
	if (ban->flags() & IPNetBan::bfCIDR4)
	{
		for (it = m_c4list.begin(), end = m_c4list.end(); it != end; ++it)
		{
			/* Ensure the same ban is not already in the list. */
			if (((*it)->ip().s_addr == ban->ip().s_addr) && 
					((*it)->netmask().s_addr == ban->netmask().s_addr))
				return false;
		}
		
		if (m_c4list.insert(m_c4list.end(), ban) == m_c4list.end())
			return false;
	} 
	else if (ban->flags() & IPNetBan::bfWild)
	{
		for (it = m_wlist.begin(), end = m_wlist.end(); it != end; ++it)
		{
			/* Ensure the same ban is not already in the list. */
			if (!stricmp((*it)->mask(), ban->mask()))
				return false;
		}
		
		if (m_wlist.insert(m_wlist.end(), ban) == m_wlist.end())
			return false;
	} else {
		return false;
	}
	
	if (epona->debuglevel() >= 2)
		epona->log(DebugNetBan(), "Added IP network ban %s "
				"(real: %s; flags: %X)", ban->mask(), ban->rmask(),
				ban->flags());
	
	return true;
}

/*************************************************************************/

bool IPNetBanner::enforce(User *user) const
{
#ifdef HAS_NICKIP
	IPNetBan *nb = (IPNetBan *)matches(user);
	
	if (nb)
	{
		epona->log(InfoNetBan(), "Enforcing network ban for user %s (matches "
				" IP mask %s (real: %s))", user->nick, nb->mask(),
				nb->rmask());
		
		char nick[NICKMAX];
		misc::strncpyzt(nick, user->nick, sizeof(nick));
		
		nb->apply();
		
		if (!finduser(nick))
			return true;
	}
#endif
	return false;
}

/*************************************************************************/

NetBan *IPNetBanner::matches(User const *user) const
{
#ifdef HAS_NICKIP
	hlist_type::const_iterator hit, hend = m_hlist.end();
	
	if ((hit = m_hlist.find(user->ip.s_addr)) != hend)
		return (*hit);
	
	IPNetBan *nb = 0;
	list_type::const_iterator it, end;
	
	for (it = m_c4list.begin(), end = m_c4list.end(); it != end; ++it)
	{
		nb = (*it);
		
		if ((user->ip.s_addr & nb->netmask().s_addr) == nb->ip().s_addr)
			return nb;
	}
	
	for (it = m_wlist.begin(), end = m_wlist.end(); it != end; ++it)
	{
		nb = (*it);
		
		if (misc::irc_match(nb->rmask(), inet_ntoa(user->ip)))
			return nb;
	}
#endif
	return 0;
}

/*************************************************************************/

bool IPNetBanner::remove(NetBan *ban_)
{
	IPNetBan *ban = (IPNetBan *) ban_;
	
	if (!(ban->flags() & NetBan::bfComplete))
		return false;
	
	if (!(ban->flags() & IPNetBan::bfWild) &&
			!(ban->flags() & IPNetBan::bfCIDR4))
	{
		if (!m_hlist.remove_sorted(ban->ip().s_addr))
			return false;
		ban->unapply();
		
		if (epona->debuglevel() >= 2)
			epona->log(DebugNetBan(), "Removed IP network ban %s (real: %s)",
					ban->mask(), ban->rmask());
		return true;
	}
	
	if (ban->flags() & (IPNetBan::bfCIDR4 | IPNetBan::bfWild))
	{
		list_type &list = (ban->flags() & IPNetBan::bfCIDR4) ?
				m_c4list : m_wlist;
		list_type::iterator it, end;
		
		for (it = list.begin(), end = list.end(); it != end; ++it)
		{
			if ((*it) == ban)
			{
				list.erase(it);
				ban->unapply();
				
				epona->log(DebugNetBan(),
						"Removed IP network ban %s (real: %s)",
						ban->mask(), ban->rmask());
				return true;
			}
		}
	}
	
	return false;
}

/*************************************************************************/

NetBan &NetBan::operator=(NetBan const &right)
{
	if (this != &right)
	{
		m_banner = right.m_banner;
		
		m_mask = right.m_mask;
		m_reason = right.m_reason;
		m_creator = right.m_creator;
		m_creation_time = right.m_creation_time;
		m_flags = right.m_flags;
	}
	
	return *this;
}

/*************************************************************************/

bool NetBan::creation_time(time_t creation_time)
{
	m_creation_time = creation_time;
	return true;
}

/*************************************************************************/

bool NetBan::creator(char const *creator)
{
	if (!creator)
		return false;
	
	m_creator = creator;
	return true;
}

/*************************************************************************/

size_t NetBan::memory_usage() const
{
	size_t result = sizeof(NetBan);
	
	if (m_flags & bfComplete)
		result += strlen(m_mask) + 1;
	result += strlen(m_creator) + 1;
	result += strlen(m_reason) + 1;
	
	return result;
}

/*************************************************************************/

bool NetBan::reason(char const *reason)
{
	if (!reason)
		return false;
	
	m_reason = reason;
	return true;
}

/*************************************************************************/

bool NetBan::wider(NetBan const *ban) const
{
	if (!(m_flags & bfComplete) || !(ban->flags() & bfComplete))
		return false;
	
	return misc::irc_match(m_mask, ban->mask());
}

/*************************************************************************/

NetForce &NetForce::operator=(NetForce const &right)
{
	if (this != &right)
	{
		m_nicknb = right.m_nicknb;
		m_umasknb = right.m_umasknb;
#ifdef HAS_IP_AKILL
		m_ipnb = right.m_ipnb;
#endif
#ifdef HAS_REALNAME_AKILL
		m_realnb = right.m_realnb;
#endif
		m_chnamenb = right.m_chnamenb;
	}
	
	return *this;
}

/*************************************************************************/

bool NetForce::ban_enforce(Channel *channel)
{
	return m_chnamenb.enforce(channel);
}

/*************************************************************************/

bool NetForce::ban_enforce(User *user)
{
	/* Always refuse to ban opers */
	if (is_oper(user))
		return false;
	
	return (m_umasknb.enforce(user)
#ifdef HAS_IP_AKILL
			|| m_ipnb.enforce(user)
#endif
#ifdef HAS_REALNAME_AKILL
			|| m_realnb.enforce(user)
#endif
			|| m_nicknb.enforce(user)
			);
}

/*************************************************************************/

NetBanner *NetForce::ban_supports(modules::PSIID iid)
{
	if (iid == NickNetBanner::IID)
		return &m_nicknb;
	else if (iid == UmaskNetBanner::IID)
		return &m_umasknb;
#ifdef HAS_IP_AKILL
	else if (iid == IPNetBanner::IID)
		return &m_ipnb;
#endif
#ifdef HAS_REALNAME_AKILL
	else if (iid == RealNetBanner::IID)
		return &m_realnb;
#endif
	else if (iid == ChanNameNetBanner::IID)
		return &m_chnamenb;
	
	return 0;
}

/*************************************************************************/

void NickChanNetBan::apply() const
{
	if (!(m_flags & bfComplete))
		return;

	uint16_t crc = crc16_str(m_mask);	

	if (m_flags & bfChannel)
	{
#ifdef IRC_BAHAMUT
		send_cmd(NULL, "SQLINE %s :%s [#%X]", m_mask.str(), m_reason.str(), crc);
#endif
		if (m_flags & bfWild)
		{
			for (int i = 0; i < 1024; i++) 
			{
				for (Channel *c = chanlist[i], *next = 0; c; c = next) 
				{
					next = c->next;
					
					if ((m_flags & bfWild) ? !misc::irc_match(m_mask, c->name) :
							stricmp(m_mask, c->name))
						continue;
					
					clear_chan(c, crc);
				}
			}
		} else {
			Channel *c = findchan(m_mask);
			
			if (c)
				clear_chan(c, crc);
		}
	}
	else
	{
#ifdef IRC_UNREAL
		send_cmd(NULL, "TKL + Q * %s %s 0 %ld :%s [#%X]", m_mask.str(),
				m_creator.str(), m_creation_time, m_reason.str(), crc);
#else
		send_cmd(NULL, "SQLINE %s :%s [#%X]", m_mask.str(), m_reason.str(), crc);
#endif
	
		if (m_flags & bfWild)
		{
			for (int i = 0; i < 1024; i++) 
			{
				for (User *u = userlist[i], *next = 0; u; u = next) 
				{
					next = u->next;
					if (misc::irc_match(m_mask, u->nick))
						s_svsnick(s_GlobalNoticer, u);
				}
			}
		} else {
			User *u = finduser(m_mask);
			
			if (u)
				s_svsnick(s_GlobalNoticer, u);
		}
	}
}

/*************************************************************************/

void NickChanNetBan::clear_chan(Channel *c, uint16_t crc) const
{
	char const *av[3];
	
#ifndef IRC_BAHAMUT
	LeaveTask *t = new LeaveTask(c->name);
	
	if (epona->tasks().add(*t))
	{
		send_cmd(s_GlobalNoticer, "JOIN %s", c->name);
	
		av[0] = "+b";
		av[1] = "*!*@*";
		s_mode(s_GlobalNoticer, c->name, av[0], av[1]);
		chan_set_modes(s_GlobalNoticer, c, 2, av, 0);
	}
#endif
	
	for (Channel::c_userlist *cu = c->users, *cunext = 0;
			cu; cu = cunext) 
	{
		cunext = cu->next;
		
		if (is_oper(cu->user))
			continue;
			
		av[0] = c->name;
		av[1] = cu->user->nick;
		av[2] = m_reason.str();
		send_cmd(s_GlobalNoticer, "KICK %s %s :Channel is forbidden: %s [#%X]",
				av[0], av[1], av[2], crc);
		do_kick(s_GlobalNoticer, 3, av);
	}
}

/*************************************************************************/

bool NickChanNetBan::mask(char const *mask)
{
	if (m_flags & bfComplete)
		return false;
	
	/* Don't allow empty masks! */
	if (!mask || !*mask)
		return false;

	char const *end = mask;
	
	int wild = 0, other = 0;
	
	while (*end)
	{
		switch (*end++)
		{
			case '#':
				break;
			case '*':
			case '?':
				++wild;
				break;
			default:
				++other;
				break;
		}
	}
	
	/* Ensure we won't be banning the whole world -- make sure
	 * there are significant characters other than wildcards */
	if (wild && !other)
		return false;
	
	if (wild)
		m_flags |= bfWild;
	if (*mask == '#')
		m_flags |= bfChannel;
	m_mask = mask;
	
	m_flags |= bfComplete;
	return true;
}

/*************************************************************************/

void NickChanNetBan::unapply() const
{
	if (!(m_flags & bfComplete))
		return;
	
	if (m_flags & bfChannel)
#ifdef IRC_BAHAMUT
		send_cmd(NULL, "UNSQLINE 0 %s", m_mask.str());
#else
		;
#endif
	else
#if defined(IRC_BAHAMUT)
		send_cmd(NULL, "UNSQLINE 0 %s", m_mask.str());
#elif defined(IRC_UNREAL)
		send_cmd(NULL, "TKL - Q * %s %s", m_mask.str(),
				epona->cd().services.sname.value().c_str());
#else
		send_cmd(NULL, "UNSQLINE %s", m_mask.str());
#endif
}

/*************************************************************************/

bool NickChanNetBan::wider(NetBan const *ban) const
{
	if ((flags() & bfChannel) != (ban->flags() & bfChannel))
		return false;
	
	return NetBan::wider(ban);
}

/*************************************************************************/

NickNetBanner::NickNetBanner() : UserNetBanner(),
		m_hlist(nc_hash), m_list()
{
	
}

/*************************************************************************/

NickNetBanner &NickNetBanner::operator=(NickNetBanner const &right)
{
	if (this != &right)
	{
		UserNetBanner::operator=(right);
		
		m_hlist = right.m_hlist;
		m_list = right.m_list;
	}
	
	return *this;
}

/*************************************************************************/

bool NickNetBanner::add(NetBan *ban_)
{
	NickChanNetBan *ban = (NickChanNetBan *) ban_;
	
	if (!(ban->flags() & NetBan::bfComplete) ||
			(ban->flags() & NickChanNetBan::bfChannel))
		return false;
	
	if (!(ban->flags() & NickChanNetBan::bfWild))
	{
		if (m_hlist.add_sorted(ban->mask(), ban) == m_hlist.end())
			return false;
		
		if (epona->debuglevel() >= 2)
			epona->log(DebugNetBan(), "Added nickname network ban %s (flags: %X)",
					ban->mask(), ban->flags());
		return true;
	}
	
	list_type::const_iterator it, end;
	
	for (it = m_list.begin(), end = m_list.end(); it != end; ++it)
	{
		/* Ensure the same ban is not already in the list. */
		if (!stricmp((*it)->mask(), ban->mask()))
			return false;
	}
	
	if (m_list.insert(m_list.end(), ban) == m_list.end())
		return false;
	
	if (epona->debuglevel() >= 2)
		epona->log(DebugNetBan(), "Added nickname network ban %s (flags: %X)",
				ban->mask(), ban->flags());
	
	return true;
}

/*************************************************************************/

bool NickNetBanner::enforce(User *user) const
{
	NetBan *nb = matches(user);
	
	if (nb)
	{
		epona->log(InfoNetBan(), "Enforcing network ban for user %s (matches "
				" nickname mask %s)", user->nick, nb->mask());
		
		nb->apply();
	}
	
	return false;
}

/*************************************************************************/

NetBan *NickNetBanner::matches(User const *user) const
{
	hlist_type::const_iterator hit, hend = m_hlist.end();
	
	if ((hit = m_hlist.find(user->nick)) != hend)
		return (*hit);
		
	list_type::const_iterator it, end;
	NickChanNetBan *nb = 0;
	
	for (it = m_list.begin(), end = m_list.end(); it != end; ++it)
	{
		nb = *it;
		
		if (misc::irc_match(nb->mask(), user->nick))
			return nb;
	}
	
	return 0;
}

/*************************************************************************/

bool NickNetBanner::remove(NetBan *ban_)
{
	NickChanNetBan *ban = (NickChanNetBan *) ban_;
	
	if (!(ban->flags() & NetBan::bfComplete) ||
			(ban->flags() & NickChanNetBan::bfChannel))
		return false;
	
	if (!(ban->flags() & NickChanNetBan::bfWild))
	{
		if (!m_hlist.remove_sorted(ban->mask()))
			return false;
		ban->unapply();
		
		if (epona->debuglevel() >= 2)
			epona->log(DebugNetBan(), "Removed nickname network ban %s",
					ban->mask());
		return true;
	}
	
	list_type::iterator it, end;
	
	for (it = m_list.begin(), end = m_list.end(); it != end; ++it)
	{
		if ((*it) == ban)
		{
			m_list.erase(it);
			ban->unapply();
			
			epona->log(DebugNetBan(), "Removed nickname network ban %s",
					ban->mask());
			return true;
		}
	}
	
	return false;
}

/*************************************************************************/

void RealNetBan::apply() const
{
	if (!(m_flags & bfComplete))
		return;

	uint16_t crc = crc16_str(m_mask);

#if defined(IRC_BAHAMUT)
	send_cmd(NULL, "SGLINE %lu :%s:%s [#%X]", (long) strlen(m_mask.str()),
			m_mask.str(), m_reason.str(), crc);
#elif defined(IRC_UNREAL)
	misc::cstring rreason;
	
	size_t len = strlen(m_reason) + 9;
	snprintf(rreason.reserve(len), len, "%s [#%X]", m_reason.str(), crc);
	
	misc::cstring rreason2(rreason);
	
	char *end = rreason;
	
	while (*end)
	{
		if (*end == ' ')
			*end = '_';
		++end;
	}

	send_cmd(NULL, "SVSNLINE + %s :%s", rreason.str(), m_mask.str());
#else
	misc::cstring rreason;

        size_t len = strlen(m_reason) + 9;
        snprintf(rreason.reserve(len), len, "%s [#%X]", m_reason.str(), crc);
#endif

	for (int i = 0; i < 1024; i++) 
	{
		for (User *u = userlist[i], *next = 0; u; u = next) 
		{
			next = u->next;
			
			if ((m_flags & bfWild) ? misc::irc_match(m_mask, u->realname) :
					!stricmp(m_mask, u->realname))
#if defined(IRC_BAHAMUT)
				send_cmd(NULL, "SVSKILL %s :G-Lined: %s [#%X]", u->nick,
						 m_reason.str(), crc);
#elif defined(IRC_UNREAL)
				kill_user(s_GlobalNoticer, u->nick, rreason2);
#else
				kill_user(s_GlobalNoticer, u->nick, rreason);
#endif
		}
	}
}

/*************************************************************************/

bool RealNetBan::mask(char const *mask)
{
	if (m_flags & bfComplete)
		return false;
	
	/* Don't allow empty masks! */
	if (!mask || !*mask)
		return false;

	char const *end = mask;
	
	int wild = 0, other = 0;
	
	while (*end)
	{
		switch (*end++)
		{
			case '*':
			case '?':
				++wild;
				break;
			default:
				++other;
				break;
		}
	}
	
	/* Ensure we won't be banning the whole world -- make sure
	 * there are significant characters other than wildcards */
	if (wild && !other)
		return false;
	
	if (wild)
		m_flags |= bfWild;
	m_mask = mask;
	
	m_flags |= bfComplete;
	return true;
}

/*************************************************************************/

void RealNetBan::unapply() const
{
	if (!(m_flags & bfComplete))
		return;
	
#if defined(IRC_BAHAMUT)
		send_cmd(NULL, "UNSGLINE 0 :%s", m_mask.str());
#elif defined(IRC_UNREAL)
		send_cmd(NULL, "SVSNLINE - :%s", m_mask.str());
#endif
}

/*************************************************************************/

RealNetBanner::RealNetBanner() : UserNetBanner(),
		m_hlist(real_hash), m_list()
{
	
}

/*************************************************************************/

RealNetBanner &RealNetBanner::operator=(RealNetBanner const &right)
{
	if (this != &right)
	{
		UserNetBanner::operator=(right);
		
		m_hlist = right.m_hlist;
		m_list = right.m_list;
	}
	
	return *this;
}

/*************************************************************************/

bool RealNetBanner::add(NetBan *ban_)
{
	RealNetBan *ban = (RealNetBan *) ban_;
	
	if (!(ban->flags() & NetBan::bfComplete))
		return false;
	
	if (!(ban->flags() & RealNetBan::bfWild))
	{
		if (m_hlist.add_sorted(ban->mask(), ban) == m_hlist.end())
			return false;
		
		if (epona->debuglevel() >= 2)
			epona->log(DebugNetBan(), "Added realname network ban %s (flags: %X)",
					ban->mask(), ban->flags());
		return true;
	}
	
	list_type::const_iterator it, end;
	
	for (it = m_list.begin(), end = m_list.end(); it != end; ++it)
	{
		/* Ensure the same ban is not already in the list. */
		if (!stricmp((*it)->mask(), ban->mask()))
			return false;
	}
	
	if (m_list.insert(m_list.end(), ban) == m_list.end())
		return false;
	
	if (epona->debuglevel() >= 2)
		epona->log(DebugNetBan(), "Added realname network ban %s (flags: %X)",
				ban->mask(), ban->flags());
	
	return true;
}

/*************************************************************************/

bool RealNetBanner::enforce(User *user) const
{
	NetBan *nb = matches(user);
	
	if (nb)
	{
		epona->log(InfoNetBan(), "Enforcing network ban for user %s (matches "
				" realname mask %s)", user->nick, nb->mask());
		
		char nick[NICKMAX];
		misc::strncpyzt(nick, user->nick, sizeof(nick));
		
		nb->apply();
		
		if (!finduser(nick))
			return true;
	}
	
	return false;
}

/*************************************************************************/

NetBan *RealNetBanner::matches(User const *user) const
{
	hlist_type::const_iterator hit, hend = m_hlist.end();
	
	if ((hit = m_hlist.find(user->realname)) != hend)
		return (*hit);
		
	list_type::const_iterator it, end;
	RealNetBan *nb = 0;
	
	for (it = m_list.begin(), end = m_list.end(); it != end; ++it)
	{
		nb = *it;
		
		if (misc::irc_match(nb->mask(), user->realname))
			return nb;
	}
	
	return 0;
}

/*************************************************************************/

bool RealNetBanner::remove(NetBan *ban_)
{
	RealNetBan *ban = (RealNetBan *) ban_;
	
	if (!(ban->flags() & NetBan::bfComplete))
		return false;
	
	if (!(ban->flags() & RealNetBan::bfWild))
	{
		if (!m_hlist.remove_sorted(ban->mask()))
			return false;
		ban->unapply();
		
		if (epona->debuglevel() >= 2)
			epona->log(DebugNetBan(), "Removed realname network ban %s",
					ban->mask());
		return true;
	}
	
	list_type::iterator it, end;
	
	for (it = m_list.begin(), end = m_list.end(); it != end; ++it)
	{
		if ((*it) == ban)
		{
			m_list.erase(it);
			ban->unapply();
			
			epona->log(DebugNetBan(), "Removed realname network ban %s",
					ban->mask());
			return true;
		}
	}
	
	return false;
}

/*************************************************************************/

UmaskNetBan &UmaskNetBan::operator=(UmaskNetBan const &right)
{
	if (this != &right)
	{
		NetBan::operator=(right);
		
		m_user = right.m_user;
		m_host = right.m_host;
	}
	
	return *this;
}

/*************************************************************************/

void UmaskNetBan::apply() const
{
	if (!(m_flags & bfComplete))
		return;
	
	uint16_t crc = crc16_str(m_mask);
	
#if defined(IRC_BAHAMUT)
	send_cmd(NULL, "AKILL %s %s %d %s %ld :%s [#%X]",
			((m_flags & bfHost) ? m_host.str() : "*"),
			((m_flags & bfUser) ? m_user.str() : "*"), 86400*2,
			m_creator.str(), (long) time(0), /* Shouldn't be set to
			                                  * the creation time! */
			m_reason.str(), crc);
#elif defined(IRC_UNREAL)
	send_cmd(NULL, "TKL + G %s %s %s %ld %ld :%s [#%X]",
			((m_flags & bfUser) ? m_user.str() : "*"),
			((m_flags & bfHost) ? m_host.str() : "*"), m_creator.str(), 
			(long) time(NULL)+86400*2, /* Avoids filling the akill list of servers
		    	                        * too much */
			(long) m_creation_time, m_reason.str(), crc);
#elif defined(IRC_DREAMFORGE)
	send_cmd(NULL, "AKILL %s %s :%s [#%X]",
			((m_flags & bfHost) ? m_host.str() : "*"),
			((m_flags & bfUser) ? m_user.str() : "*"), m_reason.str(),
			crc);
#endif
}

/*************************************************************************/

char const *UmaskNetBan::host() const
{
	if (!(m_flags & bfComplete))
		return 0;
	
	return ((m_flags & bfHost) ? m_host.str() : "*");
}

/*************************************************************************/

bool UmaskNetBan::mask(char const *mask)
{
	if (m_flags & bfComplete)
		return false;
	
	if (!mask || !*mask)
		return false;
	
	misc::cstring mymask(mask);
	
	char const *user = mymask;
	char *host = strchr(mymask, '@');
	
	if (host) 
	{
		/* If there is a host, setup the separation with the user part. */
		*host = 0;
		++host;
	} else
	{
		/* ... Else, set the host to be the whole mask, and set user to "*" */
		host = mymask;
		user = "*";
	}
	
	/* Don't allow empty host or username! */
	if (!*user || !*host)
		return false;

	char const *end = user;
	
	int utildes = 0, uwildq = 0, uwilds = 0, uother = 0;
	
	while (*end)
	{
		switch (*end++)
		{
			case '~':
				++utildes;
				break;
			case '*':
				++uwilds;
				break;
			case '?':
				++uwildq;
				break;
			default:
				++uother;
				break;
		}
	}
	
	end = host;
	
	int hdots = 0, hwildq = 0, hwilds = 0, hother = 0;
	
	while (*end)
	{
		switch (*end++)
		{
			case '.':
				++hdots;
				break;
			case '*':
				++hwilds;
				break;
			case '?':
				++hwildq;
				break;
			default:
				++hother;
				break;
		}
	}
	
	/* Ensure we won't be banning the whole world -- make sure
	 * there are significant characters other than wildcards */
	if ((uwilds && !uother) && (hwilds && !hother))
		return false;
	
	if (utildes || uwildq || uother)
	{
		m_flags |= bfUser;
		if (uwildq || uwilds)
			m_flags |= bfUserWild;
		m_user = user;
	}
	
	if (hdots || hwildq || hother)
	{
		m_flags |= bfHost;
		if (hwildq || hwilds)
			m_flags |= bfHostWild;
		m_host = host;
	}
	
	size_t size = (m_user ? strlen(m_user) : 1) +
			(m_host ? strlen(m_host) : 1) + 2;
	snprintf(m_mask.reserve(size), size, "%s@%s",
			(m_user ? m_user.str() : "*"), (m_host ? m_host.str() : "*"));
	
	m_flags |= bfComplete;
	return true;
}

/*************************************************************************/

size_t UmaskNetBan::memory_usage() const
{
	size_t result = NetBan::memory_usage() - sizeof(NetBan) +
			sizeof(UmaskNetBan);
	
	if (m_flags & bfComplete)
	{
		if (m_flags & bfUser)
			result += strlen(m_user) + 1;
		if (m_flags & bfHost)
			result += strlen(m_host) + 1;
	}
	
	return result;
}

/*************************************************************************/

void UmaskNetBan::unapply() const
{
	if (!(m_flags & bfComplete))
		return;
	
#if defined(IRC_UNREAL)
	send_cmd(NULL, "TKL - G %s %s %s",
			((m_flags & bfUser) ? m_user.str() : "*"),
			((m_flags & bfHost) ? m_host.str() : "*"),
			epona->cd().services.sname.value().c_str());
#else
	send_cmd(NULL, "RAKILL %s %s", ((m_flags & bfHost) ? m_host.str() : "*"),
			((m_flags & bfUser) ? m_user.str() : "*"));
#endif
}

/*************************************************************************/

char const *UmaskNetBan::user() const
{
	if (!(m_flags & bfComplete))
		return 0;
	
	return ((m_flags & bfUser) ? m_user.str() : "*");
}

/*************************************************************************/

UmaskNetBanner::UmaskNetBanner() : UserNetBanner(),
		m_hlist(host_hash), m_list()
{
	
}

/*************************************************************************/

UmaskNetBanner &UmaskNetBanner::operator=(UmaskNetBanner const &right)
{
	if (this != &right)
	{
		UserNetBanner::operator=(right);
		
		m_hlist = right.m_hlist;
		m_list = right.m_list;
	}
	
	return *this;
}

/*************************************************************************/

bool UmaskNetBanner::add(NetBan *ban_)
{
	UmaskNetBan *ban = (UmaskNetBan *) ban_;
	
	if (!(ban->flags() & NetBan::bfComplete))
		return false;
	
	if ((ban->flags() & UmaskNetBan::bfHost) &&
			!(ban->flags() & UmaskNetBan::bfHostWild))
	{
		if (m_hlist.add_sorted(ban->host(), ban) == m_hlist.end())
			return false;
		
		if (epona->debuglevel() >= 2)
			epona->log(DebugNetBan(), "Added usermask network ban %s (flags: %X)",
					ban->mask(), ban->flags());
		return true;
	}
	
	list_type::const_iterator it, end;
	
	for (it = m_list.begin(), end = m_list.end(); it != end; ++it)
	{
		/* Ensure the same ban is not already in the list. */
		if (!stricmp( (*it)->mask(), ban->mask() ))
			return false;
	}
	
	if (m_list.insert(m_list.end(), ban) == m_list.end())
		return false;
	
	if (epona->debuglevel() >= 2)
		epona->log(DebugNetBan(), "Added usermask network ban %s (flags: %X)",
				ban->mask(), ban->flags());
	
	return true;
}

/*************************************************************************/

bool UmaskNetBanner::enforce(User *user) const
{
	NetBan *nb = matches(user);
	
	if (nb)
	{
		epona->log(InfoNetBan(), "Enforcing network ban for user %s (matches "
				" usermask %s)", user->nick, nb->mask());
		
		char nick[NICKMAX];
		misc::strncpyzt(nick, user->nick, sizeof(nick));
		
		nb->apply();
		
		if (!finduser(nick))
			return true;
	}
	
	return false;
}

/*************************************************************************/

bool UmaskNetBanner::hash_matches(User const *user, UmaskNetBan const *ban) const
{
	/* The host part matches at this point.
	 * Check for the user part. */
	if (ban->flags() & UmaskNetBan::bfUser)
	{
		if (ban->flags() & UmaskNetBan::bfUserWild)
		{
			if (misc::irc_match(ban->user(), user->username))
				return true;
		} else {
			if (!stricmp(ban->user(), user->username))
				return true;
		}
	} else {
		return true;
	}
	
	return false;
}

/*************************************************************************/

NetBan *UmaskNetBanner::matches(User const *user) const
{
	hlist_type::const_iterator hit, hend = m_hlist.end();
	
	if ((hit = m_hlist.find(user->host)) != hend)
	{
		if (hash_matches(user, (*hit)))
			return (*hit);
	}
	
#ifdef HAS_VHOST_AKILL
	if (user->vhost && (hit = m_hlist.find(user->vhost)) != hend)
	{
		if (hash_matches(user, (*hit)))
			return (*hit);
	}
#endif
	
	list_type::const_iterator it, end;
	UmaskNetBan *nb = 0;
	
	for (it = m_list.begin(),
			end = m_list.end();
			it != end;
			 ++it)
	{
		nb = *it;
		
		if (nb->flags() & UmaskNetBan::bfUser)
		{
			if (nb->flags() & UmaskNetBan::bfUserWild)
			{
				if (!misc::irc_match(nb->user(), user->username))
					continue;
			} else {
				if (stricmp(nb->user(), user->username))
					continue;
			}
		}
		
		if (nb->flags() & UmaskNetBan::bfHost)
		{
			if (nb->flags() & UmaskNetBan::bfHostWild)
			{
				if (!misc::irc_match(nb->host(), user->host)
#ifdef HAS_VHOST_AKILL
						&& (!user->vhost || !misc::irc_match(nb->host(), user->vhost))
#endif
						)
					continue;
			} else {
				if (stricmp(nb->host(), user->host)
#ifdef HAS_VHOST_AKILL
						&& (!user->vhost || !stricmp(nb->host(), user->vhost))
#endif
						)
					continue;
			}
		}
		
		return nb;
	}
	
	return 0;
}

/*************************************************************************/

bool UmaskNetBanner::remove(NetBan *ban_)
{
	UmaskNetBan *ban = (UmaskNetBan *) ban_;
	
	if (!(ban->flags() & NetBan::bfComplete))
		return false;
	
	if ((ban->flags() & UmaskNetBan::bfHost) &&
			!(ban->flags() & UmaskNetBan::bfHostWild))
	{
		if (!m_hlist.remove_sorted(ban->host()))
			return false;
		ban->unapply();
		
		if (epona->debuglevel() >= 2)
			epona->log(DebugNetBan(), "Removed usermask network ban %s",
					ban->mask());
		return true;
	}
	
	list_type::iterator it, end;
	
	for (it = m_list.begin(), end = m_list.end(); it != end; ++it)
	{
		if ((*it) == ban)
		{
			m_list.erase(it);
			ban->unapply();
			
			epona->log(DebugNetBan(), "Removed usermask network ban %s",
					ban->mask());
			return true;
		}
	}
	
	return false;
}

/*************************************************************************/
