/* Network bans.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * Based on the original code of Services by Andy Church.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef NETBANS_H
#define NETBANS_H

/*************************************************************************/

#include <list>

#include <containers/hashlist.h>
#include <misc/string.h>
#include <modules/interface.h>

#include "services.h"

/*************************************************************************/

class NetBanner;

/*************************************************************************/

/// Base class for network bans.

/** This class is the base of all network bans class, providing a standard
  * interface for callers.
  *
  * Setting the mask is the only thing required to do before you can add
  * the ban to a list. All others fields have default values.
  *
  * The apply(), unapply() and mask() functions need to be
  * overriden in derived classes as to support different ban types. */

class NetBan
{
	public:
		/// Constructor.
		/** Constructor.
		  *
		  * The reason, creator and creation time fields of the object
		  * are set to default values so that a ban is complete immediately
		  * after a mask is set successfully (even though you may change
		  * these fields later.)
		  *
		  * \param[in] nbr The network banner that created the ban. */
		NetBan(NetBanner *nbr) : m_banner(nbr), m_mask(0),
				m_reason("No reason"), m_creator("unknown"),
				m_creation_time(time(0)), m_flags(0) { }
		/// Copy constructor.
		NetBan(NetBan const &nb) : m_banner(nb.m_banner), m_mask(nb.m_mask),
				m_reason(nb.m_reason), m_creator(nb.m_creator),
				m_creation_time(nb.m_creation_time), m_flags(nb.m_flags) { }
		/// Assignment operator.
		virtual NetBan &operator=(NetBan const &right);
		/// Destructor.
		virtual ~NetBan() { }
		
		/// Applies the ban.
		/** Applies the ban. This will send relevant message(s) to
		  * the uplink IRC server so that all matching targets
		  * will be banned from the IRC network. */
		virtual void apply() const = 0;
		
		/// Unapplies the ban.
		/** Unapplies the ban. This will send relevant message(s) to
		  * the uplink IRC server so that targets matching this ban
		  * stop to be banned by this ban. */
		virtual void unapply() const = 0;
		
		/// Returns the network banner this ban was created by.
		NetBanner *banner() const { return m_banner; }
		
		/// Returns the network ban mask.
		/** Returns the network ban mask. The format of the ban
		  * mask is completely free-form, and may not be exactly
		  * the same as the one submitted to the mask() function.
		  *
		  * This function can be overriden in derived classes.
		  *
		  * \return The network ban mask, in human-readable
		  * form, if possible. */
		virtual char const *mask() const
				{ return ((m_flags & bfComplete) ? m_mask.str() : 0); }
		/// Sets the network ban mask.
		/** Sets the network ban mask. The format of the ban
		  * mask depends on the network ban type. You must not
		  * allow the mask to be changed again after it has been
		  * set successfully once.
		  *
		  * This function must be overriden in derived classes.
		  *
		  * \param[in] mask The network ban mask.
		  * \return true if the given mask is valid, false otherwise or
		  * if the mask was already set. */
		virtual bool mask(char const *mask) = 0;

		/// Returns the network ban reason.
		char const *reason() const { return m_reason; }
		/// Sets the network ban reason.
		/** Sets the network ban reason.
		  *
		  * \param[in] reason The ban reason.
		  * \return true if the given reason is valid, false
		  * otherwise. */
		virtual bool reason(char const *reason);
		
		/// Returns the network ban creator.
		char const *creator() const { return m_creator; }
		/// Sets the network ban creator.
		/** Sets the network ban creator.
		  *
		  * \param[in] creator The ban creator.
		  * \return true if the given creator is valid, false
		  * otherwise. */
		virtual bool creator(char const *creator);
		
		/// Returns the network ban creation time.
		time_t creation_time() const { return m_creation_time; }
		/// Sets the network ban creation time.
		/** Sets the network ban creation time.
		  *
		  * \param[in] creator The ban creation time.
		  * \return true if the given creation time is valid, false
		  * otherwise. */
		virtual bool creation_time(time_t creation_time);
		
		/// Possible ban flags.
		enum
		{
			/// The ban is complete.
			bfComplete = 0x0001
		};
		
		/// Returns ban flags.
		/** Returns ban flags, which is an or'ed set of derived classes'
		  * defined flags and the following ban flag that this class
		  * defines:
		  *
		  *   - bfComplete: The ban is complete (generally set after
		  *     the ban mask has been set by the mask(char const *)
		  *     function.)
		  *
		  * \return The ban flags. */
		uint16_t flags() const { return m_flags; }
		
		/// Returns whether the ban has wider range than another one.
		/** Returns whether the ban has wider range than another
		  * given one (that is, whether the given ban is useless
		  * because its mask is overlapped by this ban).
		  *
		  * The default implementation just calls
		  * misc::irc_match(mask(), ban->mask()), after having ensured
		  * both bans are complete.
		  *
		  * \param[in] ban The other ban.
		  * \return true if this ban has wider range, false
		  * otherwise. */
		virtual bool wider(NetBan const *ban) const;
		
		/// Returns the memory usage of the ban.
		/** Returns the memory usage of the ban.
		  *
		  * \return The memory usage of the ban. */
		virtual size_t memory_usage() const;
	protected:
		/// Net banner.
		NetBanner *m_banner;
		
		/// Ban mask.
		misc::cstring m_mask;
		/// Ban reason.
		misc::cstring m_reason;
		/// Ban creator.
		misc::cstring m_creator;
		/// Ban creation time.
		time_t m_creation_time;
		/// Ban flags.
		uint16_t m_flags;
};

/*************************************************************************/

/// Class for nickname and channel network bans.

/** This class handles nickname-based and channel-based network bans. */

class NickChanNetBan : public NetBan
{
	public:
		/// Constructor.
		NickChanNetBan(NetBanner *nbr) : NetBan(nbr) { }
		
		/// Implements NetBan::apply().
		virtual void apply() const;
		
		/// Implements NetBan::unapply().
		virtual void unapply() const;
		
		/// Reexpose NetBan::mask().
		virtual char const *mask() const { return NetBan::mask(); }
		/// Sets the network ban nickname or channel name.
		/** Sets the network ban nickname or channel name, with
		  * optional wildcard characters if needed.
		  *
		  * This is an implementation of NetBan::mask(char const *).
		  *
		  * \param[in] mask The ban mask.
		  * \return true if the given mask is valid, false otherwise. */
		virtual bool mask(char const *mask);
		
		/// Additional ban flags.
		enum
		{
			/// There are wildcards in the mask.
			bfWild = 0x8000,
			/// It's a channel.
			bfChannel = 0x4000,
			
			/// Flags set by mask() -- for internal use.
			bfMaskFlags = 0xC000
		};
		
		/// Overrides NetBan::wider().
		virtual bool wider(NetBan const *ban) const;
	private:
		void clear_chan(Channel *chan, uint16_t crc) const;
};

/*************************************************************************/

/// Class for usermask network bans.

/** This class handles usermask-based network bans (usermasks have the
  * syntax user@host). */

class UmaskNetBan : public NetBan
{
	public:
		/// Constructor.
		UmaskNetBan(NetBanner *nbr) : NetBan(nbr), m_user(0), m_host(0) { }
		/// Copy constructor.
		UmaskNetBan(UmaskNetBan const &nb) : NetBan(nb), m_user(nb.m_user),
				m_host(nb.m_host) { }
		/// Assignment operator.
		virtual UmaskNetBan &operator=(UmaskNetBan const &right);
		
		/// Implements NetBan::apply().
		virtual void apply() const;
		
		/// Implements NetBan::unapply().
		virtual void unapply() const;
		
		/// Reexpose NetBan::mask().
		virtual char const *mask() const { return NetBan::mask(); }
		/// Sets the network ban usermask.
		/** Sets the network ban usermask. It should be in the
		  * format user@host. However, the function will automatically
		  * prepend a mask which has a missing user part with "*@".
		  *
		  * This is an implementation of NetBan::mask(char const *).
		  *
		  * \param[in] mask The ban mask.
		  * \return true if the given mask is valid, false otherwise. */
		virtual bool mask(char const *mask);
		
		/// Additional ban flags.
		enum
		{
			/// There is a user part (i.e. non "*").
			bfUser =     0x8000,
			/// There is a host part (i.e. non "*").
			bfHost =     0x4000,
			/// There are wildcards in the user part.
			bfUserWild = 0x2000,
			/// There are wildcards in the host part.
			bfHostWild = 0x1000,
			
			/// Flags set by mask() -- for internal use.
			bfMaskFlags = 0xF000
		};
		
		/// Returns the user part of the mask, if any.
		char const *user() const;
		/// Returns the host part of the mask, if any.
		char const *host() const;
		
		/// Overrides NetBan::memory_usage().
		virtual size_t memory_usage() const;
	protected:
		/// User part of the mask.
		misc::cstring m_user;
		/// Host part of the mask.
		misc::cstring m_host;
};

/*************************************************************************/

/// Class for IP network bans.

/** This class handles IP-based network bans (with CIDR on supporting
  * networks). */

class IPNetBan : public NetBan
{
	public:
		/// Constructor.
		IPNetBan(NetBanner *nbr);
		/// Copy constructor.
		IPNetBan(IPNetBan const &nb);
		/// Assignment operator.
		virtual IPNetBan &operator=(IPNetBan const &right);
		
		/// Implements NetBan::apply().
		virtual void apply() const;
		
		/// Implements NetBan::unapply().
		virtual void unapply() const;
		
		/// Reexpose NetBan::mask().
		virtual char const *mask() const { return NetBan::mask(); }
		/// Sets the network ban IP mask.
		/** Sets the network ban IP mask.
		  *
		  * If this function is successful, the mask() function
		  * returns exactly the mask that was passed to this
		  * function, even though that may not be the mask that
		  * is used internally; that's because conversion may happen
		  * (such as converting some masks with wildcards to CIDR4
		  * notation) that may not be supported by all IRCds. To
		  * get the real mask, use the rmask() function.
		  *
		  * This is an implementation of NetBan::mask(char const *).
		  *
		  * \param[in] mask The ban mask.
		  * \return true if the given mask is valid, false otherwise. */
		virtual bool mask(char const *mask);
		
		/// Additional ban flags.
		enum
		{
			/// There are wildcards in the mask
			bfWild = 0x8000,
			/// The mask is in CIDR4 notation
			bfCIDR4 = 0x4000,
			/// Flags set by mask() -- for internal use.
			bfMaskFlags = 0xC000
		};
		
		/// Returns the IP address (valid if the bfWild flag is not set.)
		struct in_addr const &ip() const { return m_ip; }
		/// Returns the netmask (valid if the bfCIDR4 flag is set.)
		struct in_addr const &netmask() const { return m_netmask; }
		
		/// Returns the real mask (as sent to the IRC server.)
		char const *rmask() const { return m_rmask; }
		
		/// Overrides NetBan::memory_usage().
		virtual size_t memory_usage() const;
		
		/// Overrides NetBan::wider().
		virtual bool wider(NetBan const *ban) const;
	protected:
		/// IP address (valid if no bfWild).
		struct in_addr m_ip;
		/// Netmask (valid if bfCIDR4)
		struct in_addr m_netmask;
		
		/// Real mask.
		misc::cstring m_rmask;
		
		void generate_rmask(char const *mask);
};

/*************************************************************************/

/// Class for realname network bans.

/** This class handles realname-based network bans. */

class RealNetBan : public NetBan
{
	public:
		/// Constructor.
		RealNetBan(NetBanner *nbr) : NetBan(nbr) { }
		
		/// Implements NetBan::apply().
		virtual void apply() const;
		
		/// Implements NetBan::unapply().
		virtual void unapply() const;
		
		/// Reexpose NetBan::mask().
		virtual char const *mask() const { return NetBan::mask(); }
		/// Sets the network ban nickname.
		/** Sets the network ban nickname. It should be a nickname,
		  * with optional wildcard characters if needed.
		  *
		  * This is an implementation of NetBan::mask(char const *).
		  *
		  * \param[in] mask The ban mask.
		  * \return true if the given mask is valid, false otherwise. */
		virtual bool mask(char const *mask);
		
		/// Additional ban flags.
		enum
		{
			/// There are wildcards in the mask.
			bfWild = 0x8000,
			
			/// Flags set by mask() -- for internal use.
			bfMaskFlags = 0xC000
		};
};

/*************************************************************************/

/// Base class for a type-specific ban list.
/** This class is the base class for type-specific ban list. It it responsible
  * to add, remove and check for bans in its internal list. Details on how
  * this is done is left to the derived classes.
  *
  * The create(), add() and remove functions must be overriden in derived
  * classes. */

class NetBanner : public modules::Interface
{
	public:
		/// Class interface ID.
		static modules::PSIID const IID = 0x428fd399;
		
		/// Creates a ban.
		/** Creates a ban suitable for addition to this ban list.
		  *
		  * \return A newly-allocated ban, or 0 if no ban could be
		  * created. The ban should be freed using operator delete
		  * when it's no longer needed. */
		virtual NetBan *create() = 0;
		
		/// Adds a ban to the list.
		/** Adds a ban to the list. No copy of the ban is made,
		  * and so it must remain valid until it is removed from
		  * the list.
		  *
		  * \param[in] ban The ban to add.
		  * \return true if the ban has been added to the list,
		  * false if the ban is incomplete or a similar ban
		  * already exists. */
		virtual bool add(NetBan *ban) = 0;
		
		/// Removes a ban from the list.
		/** Removes a ban from the list. This function should also
		  * call the NetBan::unapply() function to tell other servers
		  * this ban is no longer set, if needed.
		  *
		  * \param[in] ban The ban to remove.
		  * \return true if the ban has been removed from the
		  * list, false otherwise. */
		virtual bool remove(NetBan *ban) = 0;
	private:
		
};

/*************************************************************************/

/// Extends NetBanner for user targets.

/** An extension of NetBanner for user targets.
  *
  * The matches() and enforce() functions must be overriden in derived
  * classes. */

class UserNetBanner : public NetBanner
{
	public:
		/// Class interface ID.
		static modules::PSIID const IID = 0x4292415d;
	
		/// Checks whether a user matches a ban in the list.
		/** Checks whether a user matches a ban in the list.
		  *
		  * This function doesn't enforce bans if the user
		  * matches. Use the enforce() function for that purpose. 
		  *
		  * \param[in] user The target user.
		  * \return A pointer to a ban matching the user, or
		  * 0 if none was found.
		  * \sa enforce() */
		virtual NetBan *matches(User const *user) const = 0;
		
		/// Bans a user matching a ban in the list.
		/** Bans a user matching a ban in the list.
		  *
		  * \param[in] user The target user.
		  * \return true if the user record is now invalid,
		  * false otherwise. */
		virtual bool enforce(User *user) const = 0;
	private:
};

/*************************************************************************/

/// Extends NetBanner for channel targets.

/** An extension of NetBanner for channel targets.
  *
  * The matches() and enforce() functions must be overriden in derived
  * classes. */

class ChanNetBanner : public NetBanner
{
	public:
		/// Class interface ID.
		static modules::PSIID const IID = 0x42924173;
	
		/// Checks whether a channel matches a ban in the list.
		/** Checks whether a channel matches a ban in the list.
		  *
		  * This function doesn't enforce bans if the channel
		  * matches. Use the enforce() function for that purpose. 
		  *
		  * \param[in] channel The target channel.
		  * \return A pointer to a ban matching the channel, or
		  * 0 if none was found.
		  * \sa enforce() */
		virtual NetBan *matches(Channel const *channel) const = 0;
		
		/// Bans a channel matching a ban in the list.
		/** Bans a channel matching a ban in the list.
		  *
		  * \param[in] channel The target channel.
		  * \return true if the channel record is now invalid,
		  * false otherwise. */
		virtual bool enforce(Channel *channel) const = 0;
	private:
};

/*************************************************************************/

/// A network banner targetting users through nicknames.

/** A network banner targetting users through nicknames.
  *
  * Overriden functions of this class with a NetBan * parameter expect
  * it to be castable to NickChanNetBan *. NetBan * pointers returned
  * by some overriden functions of this class can be casted to
  * a NickChanNetBan * pointer too.  */

class NickNetBanner : public UserNetBanner
{
	public:
		/// Hash list type.
		typedef containers::HashList<NickChanNetBan *,
				char const *, 1024> hlist_type;
		/// Non-hash list type.
		typedef std::list<NickChanNetBan *> list_type;
	
		/// Class interface ID.
		static modules::PSIID const IID = 0x43e665c5;
		
		/// Constructor.
		NickNetBanner();
		/// Copy constructor.
		NickNetBanner(NickNetBanner const &nb) :
				UserNetBanner(nb), m_hlist(nb.m_hlist), m_list(nb.m_list) { }
		/// Assignment operator.
		NickNetBanner &operator=(NickNetBanner const &right);
		
		/// Implements NetBanner::create().
		virtual NetBan *create() { return new NickChanNetBan(this); }
		
		/// Implements NetBanner::add(). 
		virtual bool add(NetBan *ban);
		
		/// Implements NetBanner::remove().
		virtual bool remove(NetBan *ban);
		
		/// Implements UserNetBanner::matches().
		virtual NetBan *matches(User const *user) const;
		
		/// Implements UserNetBanner::enforce().
		virtual bool enforce(User *user) const;
	protected:
		/// Hashed list of bans.
		hlist_type m_hlist;
		/// List of non-hashed bans.
		list_type m_list;
};

/*************************************************************************/

/// A network banner targetting users through usermasks.

/** A network banner targetting users through usermasks.
  *
  * Overriden functions of this class with a NetBan * parameter expect
  * it to be castable to UmaskNetBan *. NetBan * pointers returned
  * by some overriden functions of this class can be casted to
  * a UmaskNetBan * pointer too.  */

class UmaskNetBanner : public UserNetBanner
{
	public:
		/// Hash list type.
		typedef containers::HashList<UmaskNetBan *,
				char const *, 1024> hlist_type;
		/// Non-hash list type.
		typedef std::list<UmaskNetBan *> list_type;
	
		/// Class interface ID.
		static modules::PSIID const IID = 0x4292420f;
		
		/// Constructor.
		UmaskNetBanner();
		/// Copy constructor.
		UmaskNetBanner(UmaskNetBanner const &nb) :
				UserNetBanner(nb), m_hlist(nb.m_hlist), m_list(nb.m_list) { }
		/// Assignment operator.
		UmaskNetBanner &operator=(UmaskNetBanner const &right);
		
		/// Implements NetBanner::create().
		virtual NetBan *create() { return new UmaskNetBan(this); }
		
		/// Implements NetBanner::add(). 
		virtual bool add(NetBan *ban);
		
		/// Implements NetBanner::remove().
		virtual bool remove(NetBan *ban);
		
		/// Implements UserNetBanner::matches().
		virtual NetBan *matches(User const *user) const;
		
		/// Implements UserNetBanner::enforce().
		virtual bool enforce(User *user) const;
	protected:
		/// Hashed list of bans.
		hlist_type m_hlist;
		/// List of non-hashed bans.
		list_type m_list;
	private:
		bool hash_matches(User const *user, UmaskNetBan const *ban) const;
};

/*************************************************************************/

/// A network banner targetting users through IPs.

/** A network banner targetting users through IPs.
  *
  * Overriden functions of this class with a NetBan * parameter expect
  * it to be castable to IPNetBan *. NetBan * pointers returned
  * by some overriden functions of this class can be casted to
  * a IPNetBan * pointer too.  */

class IPNetBanner : public UserNetBanner
{
	public:
		/// Hash list type.
		typedef containers::HashList<IPNetBan *, uint32_t, 4096> hlist_type;
		/// Non-hash list type.
		typedef std::list<IPNetBan *> list_type;
	
		/// Class interface ID.
		static modules::PSIID const IID = 0x43ea643b;
		
		/// Constructor.
		IPNetBanner();
		/// Copy constructor.
		IPNetBanner(IPNetBanner const &nb) :
				UserNetBanner(nb), m_hlist(nb.m_hlist), m_c4list(nb.m_c4list),
				m_wlist(nb.m_wlist) { }
		/// Assignment operator.
		IPNetBanner &operator=(IPNetBanner const &right);
		
		/// Implements NetBanner::create().
		virtual NetBan *create() { return new IPNetBan(this); }
		
		/// Implements NetBanner::add(). 
		virtual bool add(NetBan *ban);
		
		/// Implements NetBanner::remove().
		virtual bool remove(NetBan *ban);
		
		/// Implements UserNetBanner::matches().
		virtual NetBan *matches(User const *user) const;
		
		/// Implements UserNetBanner::enforce().
		virtual bool enforce(User *user) const;
	protected:
		/// Hashed list of IP addresses.
		hlist_type m_hlist;
		/// List of CIDR4 IP addresses.
		list_type m_c4list;
		/// List of IP addresses with wildcards.
		list_type m_wlist;
};

/*************************************************************************/

/// A network banner targetting users through realnames.

/** A network banner targetting users through realnames.
  *
  * Overriden functions of this class with a NetBan * parameter expect
  * it to be castable to RealNetBan *. NetBan * pointers returned
  * by some overriden functions of this class can be casted to
  * a RealNetBan * pointer too.  */

class RealNetBanner : public UserNetBanner
{
	public:
		/// Hash list type.
		typedef containers::HashList<RealNetBan *,
				char const *, 1024> hlist_type;
		/// Non-hash list type.
		typedef std::list<RealNetBan *> list_type;
	
		/// Class interface ID.
		static modules::PSIID const IID = 0x43ebc7e4;
		
		/// Constructor.
		RealNetBanner();
		/// Copy constructor.
		RealNetBanner(RealNetBanner const &nb) :
				UserNetBanner(nb), m_hlist(nb.m_hlist), m_list(nb.m_list) { }
		/// Assignment operator.
		RealNetBanner &operator=(RealNetBanner const &right);
		
		/// Implements NetBanner::create().
		virtual NetBan *create() { return new RealNetBan(this); }
		
		/// Implements NetBanner::add(). 
		virtual bool add(NetBan *ban);
		
		/// Implements NetBanner::remove().
		virtual bool remove(NetBan *ban);
		
		/// Implements UserNetBanner::matches().
		virtual NetBan *matches(User const *user) const;
		
		/// Implements UserNetBanner::enforce().
		virtual bool enforce(User *user) const;
	protected:
		/// Hashed list of bans.
		hlist_type m_hlist;
		/// List of non-hashed bans.
		list_type m_list;
};

/*************************************************************************/

/// A network banner targetting channels through channel names.

/** A network banner targetting channels through channel names.
  *
  * Overriden functions of this class with a NetBan * parameter expect
  * it to be castable to NickChanNetBan *. NetBan * pointers returned
  * by some overriden functions of this class can be casted to
  * a NickChanNetBan * pointer too.  */

class ChanNameNetBanner : public ChanNetBanner
{
	public:
		/// Hash list type.
		typedef containers::HashList<NickChanNetBan *,
				char const *, 1024> hlist_type;
		/// Non-hash list type.
		typedef std::list<NickChanNetBan *> list_type;
	
		/// Class interface ID.
		static modules::PSIID const IID = 0x43e69eae;
		
		/// Constructor.
		ChanNameNetBanner();
		/// Copy constructor.
		ChanNameNetBanner(ChanNameNetBanner const &nb) :
				ChanNetBanner(nb), m_hlist(nb.m_hlist), m_list(nb.m_list) { }
		/// Assignment operator.
		ChanNameNetBanner &operator=(ChanNameNetBanner const &right);
		
		/// Implements NetBanner::create().
		virtual NetBan *create() { return new NickChanNetBan(this); }
		
		/// Implements NetBanner::add(). 
		virtual bool add(NetBan *ban);
		
		/// Implements NetBanner::remove().
		virtual bool remove(NetBan *ban);
		
		/// Implements UserNetBanner::matches().
		virtual NetBan *matches(Channel const *channel) const;
		
		/// Implements UserNetBanner::enforce().
		virtual bool enforce(Channel *channel) const;
	protected:
		/// Hashed list of bans.
		hlist_type m_hlist;
		/// List of non-hashed bans.
		list_type m_list;
};

/*************************************************************************/

/* XXX: The NetForce class is, in fact, an incomplete and early pseudo-server
 * class for the network modules. Thus, the name will need to be changed
 * and the class probably moved to another header file when the network
 * modules will be implemented. */

class NetForce
{
	public:
		/// Constructor.
		NetForce() : m_nicknb(), m_umasknb(),
#ifdef HAS_IP_AKILL
				m_ipnb(),
#endif
#ifdef HAS_REALNAME_AKILL
				m_realnb(),
#endif
				m_chnamenb() { }
		/// Copy constructor.
		NetForce(NetForce const &nf) : m_nicknb(nf.m_nicknb),
				m_umasknb(nf.m_umasknb),
#ifdef HAS_IP_AKILL
				m_ipnb(nf.m_ipnb),
#endif
#ifdef HAS_REALNAME_AKILL
				m_realnb(nf.m_realnb),
#endif
				m_chnamenb(nf.m_chnamenb) { }
		/// Assignment operator.
		NetForce &operator=(NetForce const &right);
		
		/// Returns a network banner interface (if available).
		NetBanner *ban_supports(modules::PSIID iid);
		/// Iterates through network banners as long as NetBanner::enforce() yields false.
		bool ban_enforce(User *user);
		/// Iterates through network banners as long as NetBanner::enforce() yields false.
		bool ban_enforce(Channel *channel);
		
		/* XXX: other modules should not get banners that way --
		 * should be made private. */
		NickNetBanner *nicknb() { return &m_nicknb; }
		UmaskNetBanner *umasknb() { return &m_umasknb; }
		ChanNameNetBanner *chnamenb() { return &m_chnamenb; }
	private:
		NickNetBanner m_nicknb;
		UmaskNetBanner m_umasknb;
#ifdef HAS_IP_AKILL
		IPNetBanner m_ipnb;
#endif
#ifdef HAS_REALNAME_AKILL
		RealNetBanner m_realnb;
#endif
		ChanNameNetBanner m_chnamenb;
};

/*************************************************************************/

extern NetForce netforce;

/*************************************************************************/

#endif /* NETBANS_H */
