/* Session Limiting functions.
 *
 * by Andrew Kempe (TheShadow)
 *     E-mail: <theshadow@shadowfire.org>
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#include <system/headers.h>

#include <core/directives.h>
#include <core/epona.h>
#include <databases/database.h>
#include <log/levels.h>
#include <misc/misc.h>
#include <modules/manager.h>
#include <modules/module.h>

#include "epona.h"

#include "operserv.h"

#include "services.h"
#include "pseudo.h"

using namespace std;
using namespace core;

using misc::duration_scan;

/*************************************************************************/

/* SESSION LIMITING
 *
 * The basic idea of session limiting is to prevent one host from having more 
 * than a specified number of sessions (client connections/clones) on the 
 * network at any one time. To do this we have a list of sessions and 
 * exceptions. Each session structure records information about a single host, 
 * including how many clients (sessions) that host has on the network. When a 
 * host reaches it's session limit, no more clients from that host will be 
 * allowed to connect.
 *
 * When a client connects to the network, we check to see if their host has 
 * reached the default session limit per host, and thus whether it is allowed 
 * any more. If it has reached the limit, we kill the connecting client; all 
 * the other clients are left alone. Otherwise we simply increment the counter 
 * within the session structure. When a client disconnects, we decrement the 
 * counter. When the counter reaches 0, we free the session.
 *
 * Exceptions allow one to specify custom session limits for a specific host 
 * or a range thereof. The first exception that the host matches is the one 
 * used.
 *
 * "Session Limiting" is likely to slow down services when there are frequent 
 * client connects and disconnects. The size of the exception list can also 
 * play a large role in this performance decrease. It is therefore recommened 
 * that you keep the number of exceptions to a minimum. A very simple hashing 
 * method is currently used to store the list of sessions. I'm sure there is 
 * room for improvement and optimisation of this, along with the storage of 
 * exceptions. Comments and suggestions are more than welcome!
 *
 * -TheShadow (02 April 1999)
 */

/*************************************************************************/

/* I'm sure there is a better way to hash the list of hosts for which we are
 * storing session information. This should be sufficient for the mean time.
 * -TheShadow */

#define HASH(host)      (((host)[0]&31)<<5 | ((host)[1]&31))

static Session *sessionlist[1024];
static int32_t nsessions = 0;

ExceptionList exceptions;

/*************************************************************************/

LOGLEVELH(ErrorSessions, xlog::LogError, "Error::Sessions", "Error: sessions: ");

/*************************************************************************/

static int exception_msgs[ListCmd::cmListCmdTotal];

class ExceptionCmd : public ListCmd
{
	public:
		ExceptionCmd() : 
				ListCmd("EXCEPTION", s_OperServ, exception_msgs, true) { }		
	protected:
		virtual container_type *container(User *u);
		
		virtual bool allow_dispatch(User *u, container_type &ctn,
				char const *sub);
		virtual int dispatch(User *u, container_type &ctn, char const *sub);
		virtual bool changeable(User *u, container_type &ctn,
				char const *sub) const;
		
		virtual int add(User *u, container_type &ctn, char const *sub);
		
		virtual bool del_record(User *u, container_type &ctn,
				char const *sub, databases::Record *rec);
		
		virtual bool list_record(User *u, container_type &ctn,
				char const *sub, uint16_t num, databases::Record const *rec,
				uint16_t count);
		
		virtual bool clear_records(User *u, container_type &ctn,
				char const *sub);
};

/*************************************************************************/

static Session *findsession(const char *host);

/*************************************************************************/
/****************************** Statistics *******************************/
/*************************************************************************/

void get_session_stats(long *nrec, long *memuse)
{
    Session *session;
    long mem;
    int i;

    mem = sizeof(Session) * nsessions;
    for (i = 0; i < 1024; i++) {
    	for (session = sessionlist[i]; session; session = session->next) {
	    mem += strlen(session->host)+1;
        }
    }

    *nrec = nsessions;
    *memuse = mem;
}

/*************************************************************************/
/************************* Session List Display **************************/
/*************************************************************************/

/* Syntax: SESSION LIST threshold
 *	Lists all sessions with atleast threshold clients.
 *	The threshold value must be greater than 1. This is to prevent 
 * 	accidental listing of the large number of single client sessions.
 *
 * Syntax: SESSION VIEW host
 *	Displays detailed session information about the supplied host.
 */

void do_session(User *u)
{
    Session *session;
    Exception *exception;
    char *cmd = strtok(NULL, " ");
    char *param1 = strtok(NULL, " ");
    int mincount;
    int i;

    if (!epona->cd().session.enable) {
	notice_lang(s_OperServ, u, OPER_SESSION_DISABLED);
	return;
    }

	if (!cmd)
	{
		syntax_error(s_OperServ, u, "SESSION", OPER_SESSION_SYNTAX);
	} else if (stricmp(cmd, "LIST") == 0) {
		if (!param1) {
		    syntax_error(s_OperServ, u, "SESSION", OPER_SESSION_LIST_SYNTAX);

	} else if ((mincount = atoi(param1)) <= 1) {
	    notice_lang(s_OperServ, u, OPER_SESSION_INVALID_THRESHOLD);

	} else {
	    notice_lang(s_OperServ, u, OPER_SESSION_LIST_HEADER, mincount);
	    notice_lang(s_OperServ, u, OPER_SESSION_LIST_COLHEAD);
    	    for (i = 0; i < 1024; i++) {
            	for (session = sessionlist[i]; session; session=session->next) 
		{
            	    if (session->count >= mincount)
                        notice_lang(s_OperServ, u, OPER_SESSION_LIST_FORMAT,
                    	            session->count, session->host);
                }
    	    }
	}
    } else if (stricmp(cmd, "VIEW") == 0) {
        if (!param1) {
	    syntax_error(s_OperServ, u, "SESSION", OPER_SESSION_VIEW_SYNTAX);

        } else {
	    session = findsession(param1);
	    if (!session) {
		notice_lang(s_OperServ, u, OPER_SESSION_NOT_FOUND, param1);
	    } else {
	        exception = exceptions.find_match(param1);

	        notice_lang(s_OperServ, u, OPER_SESSION_VIEW_FORMAT,
				param1, session->count, 
				exception ? exception->limit() :
				epona->cd().session.default_limit);
	    }
        }

    } else {
	syntax_error(s_OperServ, u, "SESSION", OPER_SESSION_SYNTAX);
    }
}

/*************************************************************************/
/********************* Internal Session Functions ************************/
/*************************************************************************/

static Session *findsession(const char *host)
{
    Session *session;
    int i;

    if (!host)
	return NULL;

    for (i = 0; i < 1024; i++) {
        for (session = sessionlist[i]; session; session = session->next) {
            if (stricmp(host, session->host) == 0) {
                return session;
            }
        }
    }

    return NULL;
}

/* Attempt to add a host to the session list. If the addition of the new host
 * causes the the session limit to be exceeded, kill the connecting user.
 * Returns 1 if the host was added or 0 if the user was killed.
 */

int add_session(const char *nick, const char *host)
{
    Session *session, **list;
    Exception *exception;
    int sessionlimit = 0;

    session = findsession(host);

    if (session) {
	exception = exceptions.find_match(host);
	sessionlimit = exception ? exception->limit() :
			epona->cd().session.default_limit;

	if (sessionlimit != 0 && session->count >= sessionlimit) 
	{
		conf::ConfStrCharA::container_type::const_iterator it,
				end = epona->cd().session.message->end();
		
		for (it = epona->cd().session.message->begin(); it != end; ++it)
			notice(s_OperServ, nick, (char const *) *it, host);

	    /* We don't use kill_user() because a user stucture has not yet
	     * been created. Simply kill the user. -TheShadow
	     */
#ifdef IRC_BAHAMUT
			send_cmd(NULL, "SVSKILL %s :Session limit exceeded", nick);
#else
            send_cmd(s_OperServ, "KILL %s :%s (Session limit exceeded)", nick, s_OperServ);
#endif
            session->hits++;
            if (epona->cd().session.max_kills &&
            		session->hits >= epona->cd().session.max_kills) 
            {
            	char akillmask[BUFSIZE];
            	snprintf(akillmask, sizeof(akillmask), "*@%s", host);
            	akill_cmd->add_record(0, akillmask, s_OperServ,
            			time(NULL) + epona->cd().session.expire_akills,
            			"Session limit exceeded");
            	wallops(s_OperServ,
            			"Added a temporary AKILL for \2%s\2 due to excessive "
            			"connections", akillmask);
            }
	    return 0;
	} else {
	    session->count++;
	    return 1;
	}
    }

    nsessions++;
    session = static_cast<Session *>(scalloc(sizeof(Session), 1));
    session->host = sstrdup(host);
    list = &sessionlist[HASH(session->host)];
    session->next = *list;
    if (*list)
        (*list)->prev = session;
    *list = session;
    session->count = 1;

    return 1;
}

void del_session(const char *host)
{
    Session *session;

    session = findsession(host);

    if (!session) {
	wallops(s_OperServ, 
		"WARNING: Tried to delete non-existant session: \2%s", host);
	epona->log(ErrorSessions(), "Tried to delete non-existant session: %s", host);
	return;
    }

    if (session->count > 1) {
	session->count--;
	return;
    }

    if (session->prev)
        session->prev->next = session->next;
    else
        sessionlist[HASH(session->host)] = session->next;
    if (session->next)
        session->next->prev = session->prev;

    free(session->host);
    free(session);

    nsessions--;
}


/*************************************************************************/
/********************** External Exception Functions *********************/
/*************************************************************************/

bool exceptions_load()
{
	if (!epona->modules().load(epona->cd().session.database_exceptions))
	{
		epona->log(ErrorSessions(), "Can't load database module %s: %s",
				epona->cd().session.database_exceptions.value().c_str(),
				epona->modules().error().c_str());
		return false;
	}
	
	exception_msgs[ListCmd::cmSyntax] = OPER_EXCEPTION_SYNTAX;
	exception_msgs[ListCmd::cmDisabled] = OPER_EXCEPTION_DISABLED;
	exception_msgs[ListCmd::cmEmpty] = OPER_EXCEPTION_EMPTY;
	exception_msgs[ListCmd::cmSyntaxDel] = OPER_EXCEPTION_DEL_SYNTAX;
	exception_msgs[ListCmd::cmNoSuchEntry] = OPER_EXCEPTION_NO_SUCH_ENTRY;
	exception_msgs[ListCmd::cmNoMatch] = OPER_EXCEPTION_NO_MATCH;
	exception_msgs[ListCmd::cmDeletedOne] = OPER_EXCEPTION_DELETED_ONE;
	exception_msgs[ListCmd::cmDeletedSeveral] = OPER_EXCEPTION_DELETED_SEVERAL;
	exception_msgs[ListCmd::cmNotFound] = OPER_EXCEPTION_NOT_FOUND;
	exception_msgs[ListCmd::cmDeleted] = OPER_EXCEPTION_DELETED;
	exception_msgs[ListCmd::cmSyntaxList] = -1; /* Not used for EXCEPTION */
	exception_msgs[ListCmd::cmListHeader] = OPER_EXCEPTION_LIST_HEADER;
	exception_msgs[ListCmd::cmListedOne] = OPER_EXCEPTION_LISTED_ONE;
	exception_msgs[ListCmd::cmListedSeveral] = OPER_EXCEPTION_LISTED_SEVERAL;
	exception_msgs[ListCmd::cmClearPartially] = OPER_EXCEPTION_CLEAR_PARTIALLY;
	exception_msgs[ListCmd::cmClear] = OPER_EXCEPTION_CLEAR;
	
	return true;
}

/*************************************************************************/

void exceptions_notify(modules::Module &sender, int msg, void *arg)
{
	switch (msg)
	{
		case modules::Module::mnLoaded:
			if (sender.name() == epona->cd().session.database_exceptions.value())
			{
				databases::Database *db = (databases::Database *)
						sender.factory(databases::Database::IID);
				if (!db)
				{
					epona->fatalf("Can't create database operserv_exceptions: "
							"module %s does not expose the Database interface",
							sender.name().c_str());	
				}
				
				/* Add a reference to the module */
				sender.references_add();
				/* Setup the database */
				exceptions.setup_database(db);
			}
		default:
			break;
	}
}

/*************************************************************************/

void exceptions_start()
{
	if (!exceptions.db() || !exceptions.db()->start())
		epona->fatal("Can't start database operserv_exceptions");
}

/*************************************************************************/
/************************ Exception Manipulation *************************/
/*************************************************************************/

int ExceptionCmd::add(User *u, container_type &ctn, char const *sub)
{
	if (exceptions.size() == ctn.max_size())
	{
		notice_lang(m_service, u, OPER_EXCEPTION_TOO_MANY);
		return crStop;
	}
	
	char *expiry = NULL, *mask = strtok(NULL, " ");
	
	if (mask && *mask == '+')
	{
		expiry = mask;
		mask = strtok(NULL, " ");
	}
	
	char *limitstr = strtok(NULL, " "), *reason = strtok(NULL, "");
	
	if (!mask || !limitstr || !reason)
	{
		syntax_error(m_service, u, m_name, OPER_EXCEPTION_ADD_SYNTAX);
		return crStop;
	}
	
	/* Computes the expiration time */
	time_t expires = epona->cd().session.expire_exceptions;
	
	if (expiry && !duration_scan(expiry, &expires))
	{
		notice_lang(m_service, u, BAD_EXPIRY_TIME);
		return crStop;
	}
	
	if (expires)
		expires += time(0);
	
	/* Validates the mask */
	
	if (strlen(mask) > Exception::mask_len)
	{
		notice_lang(m_service, u, OPER_EXCEPTION_HOSTMASK_TOO_LONG,
				Exception::mask_len);
		return crStop;
	}
	
	if (strchr(mask, '!') || strchr(mask, '@'))
	{
		notice_lang(m_service, u, OPER_EXCEPTION_INVALID_HOSTMASK);
		return crStop;
	}
	
	/* Computes the limit */
	
	char *lend;
	uint16_t limit = strtoul(limitstr, &lend, 10);
	
	if (*lend || (limit > epona->cd().session.max_limit))
	{
		notice_lang(m_service, u, OPER_EXCEPTION_INVALID_LIMIT,
				epona->cd().session.max_limit.value());
		return crStop;
	}
	
	/* Validates the reason */
	
	if (strlen(reason) > Exception::reason_len)
	{
		notice_lang(m_service, u, OPER_EXCEPTION_REASON_TOO_LONG,
				Exception::reason_len);
		return crStop;
	}
	
	/* All data is correct, proceed to the creation of the exception */
	
	Exception *ex = (Exception *)exceptions.create();
	
	ex->mask(mask);
	ex->limit(limit);
	ex->creator(u->nick);
	ex->reason(reason);
	ex->expires(expires);
	
	/* Try to add it to the list */
	
	int error;
	
	if (!exceptions.add(ex, &error))
	{
		if (error == databases::ArrayCollection::ceName)
			 notice_lang(s_OperServ, u, OPER_EXCEPTION_ALREADY_PRESENT,
			 		ex->mask());
		else
			notice_lang(m_service, u, m_msgs[cmDisabled]);
		
		delete ex;
	} else {
		notice_lang(m_service, u, OPER_EXCEPTION_ADDED, ex->mask(), ex->limit());
	}
	
	return crStop;
}

/*************************************************************************/

bool ExceptionCmd::allow_dispatch(User *u, container_type &ctn,
		char const *sub)
{
	return ListCmd::allow_dispatch(u, ctn, sub);
}

/*************************************************************************/

bool ExceptionCmd::changeable(User *u, container_type &ctn,
		char const *sub) const
{
	return exceptions.db()->writable();
}

/*************************************************************************/

bool ExceptionCmd::clear_records(User *u, container_type &ctn,
		char const *sub)
{
	if (!exceptions.clear())
		return true; /* Partial clear (on paper :P) */
	
	ctn.clear();
	return false;
}

/*************************************************************************/

ExceptionCmd::container_type *ExceptionCmd::container(User *u)
{
	if (!epona->cd().session.enable) 
	{
		notice_lang(s_OperServ, u, OPER_EXCEPTION_DISABLED);
		return 0;
	}
	
	return &exceptions.array();
}

/*************************************************************************/

bool ExceptionCmd::del_record(User *u, container_type &ctn, char const *sub,
		databases::Record *rec)
{
	return exceptions.remove(rec);
}

/*************************************************************************/

int ExceptionCmd::dispatch(User *u, container_type &ctn, char const *sub)
{
	if (!stricmp(sub, "VIEW"))
		return list(u, ctn, sub);
	else
		return ListCmd::dispatch(u, ctn, sub);
}

/*************************************************************************/

bool ExceptionCmd::list_record(User *u, container_type &ctn, char const *sub,
		uint16_t num, databases::Record const *rec, uint16_t count)
{
	Exception *ex = (Exception *)rec;
	
	if (!stricmp(sub, "LIST"))
	{
		if (count == 0)
			notice_lang(m_service, u, OPER_EXCEPTION_LIST_COLHEAD);
	
		notice_lang(m_service, u, OPER_EXCEPTION_LIST_FORMAT, num,
				ex->limit(), ex->mask());
	} else if (!stricmp(sub, "VIEW")) 
	{
		time_t t = ex->since();
		struct tm tm = *localtime(&t);
		
		char timebuf[32];
		strftime_lang(timebuf, sizeof(timebuf), u,
				STRFTIME_SHORT_DATE_FORMAT, &tm);
		
		char expirebuf[256];
		expire_left(u->na, expirebuf, sizeof(expirebuf), ex->expires());
		
		notice_lang(s_OperServ, u, OPER_EXCEPTION_VIEW_FORMAT, num, ex->mask(),
				ex->creator(), timebuf, expirebuf, ex->limit(), ex->reason());
	}
	
	return true;
}

/*************************************************************************/

void do_exception(User *u)
{
	static ExceptionCmd cmd;
	cmd.execute(u);
}

/*************************************************************************/
