/* Routines to maintain a list of online users.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#include <system/headers.h>

#if HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif
#if HAVE_ARPA_INET_H
# include <arpa/inet.h>
#endif

#include <core/directives.h>
#include <core/epona.h>
#include <log/levels.h>
#include <misc/misc.h>
#include <misc/string.h>

#include "blacklist.h"
#include "epona.h"
#include "netbans.h"
#include "nickserv.h"
#include "operserv.h"
#include "users.h"

#include "services.h"

using namespace std;
using namespace core;

#define HASH(nick)	(((nick)[0]&31)<<5 | ((nick)[1]&31))
User *userlist[1024];

uint32_t usercnt = 0, opcnt = 0;

/*************************************************************************/

LOGLEVELH(DebugUser, xlog::LogDebug, "Debug::User", "debug: user: "); 
LOGLEVELH(DebugUserList, DebugUser, "Debug::User::List",
		"debug: user list: ");

LOGLEVEL (InfoUserCount, InfoUser, "Info::User::Count");
LOGLEVELH(InfoUserLogging, InfoUser, "Info::User::Logging", "log-users: ");

/*************************************************************************/

static unsigned long umodes[128] = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,			
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,			
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,			
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,			
	0, UMODE_A, 0, 0, 0, 0, 0, 0, 
#ifdef IRC_SEQUANA
	UMODE_H,
#else
	0,
#endif 
#ifdef IRC_BAHAMUT
	UMODE_I,
#else
	0,
#endif
	0, 0, 0, 0, 0, 0,
	0, 0, 													
#ifdef IRC_BAHAMUT
	UMODE_R,												
#else
	0,
#endif
	0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, UMODE_a, 0, 0, 0, 0, 0, 
#ifdef IRC_DREAMFORGE
	UMODE_g,
#else
	0,
#endif 
	UMODE_h, UMODE_i, 0, 0, 0, 0, 0, UMODE_o,		
	0, 0, UMODE_r, 0, 0, 0, 0, UMODE_w, 
#if defined(IRC_ULTIMATE) || defined(IRC_UNREAL)
	UMODE_x,
#else
	0,
#endif 
	0, 0, 0, 0, 0, 0, 0
};

/*************************************************************************/
/*************************************************************************/

/* Allocate a new User structure, fill in basic values, link it to the
 * overall list, and return it.  Always successful.
 */

static User *new_user(const char *nick)
{
    User *user, **list;

    user = static_cast<User *>(scalloc(sizeof(User), 1));
    if (!nick)
	nick = "";
    strscpy(user->nick, nick, NICKMAX);
    list = &userlist[HASH(user->nick)];
    user->next = *list;
    if (*list)
	(*list)->prev = user;
    *list = user;
    user->na = aliases.find(nick);
    if (user->na)
	user->na->u = user;
    
    usercnt++;
    stats.make_record();
  
    if (stats_rec && usercnt > stats_rec->maxusers()) 
    {
    	stats_rec->new_max(usercnt, time(0));
		if (epona->cd().services.log_maxusers)
			epona->log(InfoUserCount(), "New maximum user count: %d", 
					stats_rec->maxusers());
    }
    return user;
}

/*************************************************************************/

/* Change the nickname of a user, and move pointers as necessary. */

static void change_user_nick(User *user, const char *nick)
{
    User **list;
    int is_same = (!stricmp(user->nick, nick) ? 1 : 0);

    if (user->prev)
	user->prev->next = user->next;
    else
	userlist[HASH(user->nick)] = user->next;
    if (user->next)
	user->next->prev = user->prev;
    user->nick[1] = 0;	/* paranoia for zero-length nicks */
    strscpy(user->nick, nick, NICKMAX);
    list = &userlist[HASH(user->nick)];
    user->next = *list;
    user->prev = NULL;
    if (*list)
	(*list)->prev = user;
    *list = user;
    
    /* Only if old and new nick aren't the same; no need to waste time */
    if (!is_same) {
    	if (user->na)
    	user->na->u = NULL;
    	user->na = aliases.find(nick);
    	if (user->na)
    	user->na->u = user;
    }
}

/*************************************************************************/

#ifdef HAS_VHOST

static void update_host(User *user)
{	
	if (user->na && (nick_identified(user) || 
			(!(user->na->ng()->flags() & NickGroup::gfSecure) && nick_recognized(user))))
		user->na->last_usermask(user->username, GetHost(user));
    
	if (epona->debuglevel())
		epona->log(DebugUser(), "%s changes its host to %s", user->nick, GetHost(user));
}

#endif

/*************************************************************************/

#ifdef HAS_CHGHOST

/* Change the (virtual) hostname of a user. */

void change_user_host(User *user, const char *host)
{
#ifdef HAS_VHOST
    if (user->vhost) 
    	free(user->vhost);
    user->vhost = sstrdup(host);

    if (epona->debuglevel())
    	epona->log(DebugUser(), "%s changes its vhost to %s", user->nick, host);
    
    update_host(user);
    if (netforce.umasknb())
		netforce.umasknb()->enforce(user);
#else
	if (user->host) 
    	free(user->host);
    user->host = sstrdup(host);

    if (epona->debuglevel())
    	epona->log(DebugUser(), "%s changes its host to %s", user->nick, host);
    
    if (user->na && (nick_identified(user) || 
			(!(user->na->ng()->flags() & NickGroup::gfSecure) && nick_recognized(user))))
		user->na->last_usermask(user->username, user->host);
	if (netforce.umasknb())
		netforce.umasknb()->enforce(user);
#endif
}

#endif

/*************************************************************************/

#if defined(IRC_ULTIMATE) || defined(IRC_UNREAL)

/* Change the realname of a user. */

void change_user_realname(User *user, const char *realname)
{
	if (user->realname) 
		free(user->realname);
	user->realname = sstrdup(realname);
    
	if (user->na && (nick_identified(user) || 
			(!(user->na->ng()->flags() & NickGroup::gfSecure) && nick_recognized(user))))
		user->na->last_realname(realname);
    
	if (epona->debuglevel())
		epona->log(DebugUser(), "%s changes its realname to %s", user->nick, realname);
}


/*************************************************************************/

/* Change the username of a user. */

void change_user_username(User *user, const char *username)
{
	if (user->username) 
		free(user->username);
	user->username = sstrdup(username);
    
	if (user->na && (nick_identified(user) || 
			(!(user->na->ng()->flags() & NickGroup::gfSecure) && nick_recognized(user))))
		user->na->last_usermask(user->username, GetHost(user));

	if (epona->debuglevel())
		epona->log(DebugUser(), "%s changes its username to %s", user->nick, username);
}

#endif

/*************************************************************************/

void set_umode(User *user, int ac, char const *const *av)
{	
	int add = 1;		/* 1 if adding modes, 0 if deleting */
	char const *modes = av[0];
		
	ac--;
	
	if (epona->debuglevel())
		epona->log(DebugUser(), "Changing mode for %s to %s", user->nick, modes);
    
    while (*modes) 
    {	
    	add ? (user->mode |= umodes[(int)*modes]) : (user->mode &= ~umodes[(int)*modes]);
    	
		switch (*modes++) {
	    	case '+': 
	    		add = 1; 
	    		break;
	    	case '-': 
	    		add = 0;
	    		break;
			case 'd':
				if (ac == 0) {
#if !defined(IRC_ULTIMATE) && !defined(IRC_UNREAL)
					epona->log(ErrorUser(), "umode +d with no parameter (?) for user %s",
							user->nick);
#endif
					break;
				}
				
				ac--; av++;
				user->svid = strtoul(*av, NULL, 0);
				break;
	    	case 'o':
				if (add) {
		    		opcnt++;
		    	
		    		if (epona->cd().operserv.wall & Epona::Directives::OperServ::Wall::waNewOper)
						wallops(s_OperServ, "\2%s\2 is now an IRC operator.", user->nick);
		    		display_news(user, News::ntOper);
				} else {
		    		opcnt--;
				}
				break;
			case 'r':
	    		if (add && !nick_identified(user)) {
		    		s_svsmode_user(s_NickServ, user->nick, user->timestamp, "-r", NULL);
		    		user->mode &= ~UMODE_r;
		    	}
				break;
#if defined(IRC_SEQUANA)
			case 'H':
#elif defined(IRC_ULTIMATE) || defined(IRC_UNREAL)
			case 'x':
#endif
#if defined(IRC_SEQUANA) || defined(IRC_ULTIMATE) || defined(IRC_UNREAL)
				update_host(user);
				break;
#endif
		}
    }
}

/*************************************************************************/

/* Remove and free a User structure. */

static void delete_user(User *user)
{
    User::u_chanlist *c, *c2;
    User::u_chaninfolist *ci, *ci2;

	if (epona->debuglevel())
		epona->log(DebugUser(), "Deleting user %s", user->nick);

	if (epona->cd().services.log_users) {
#ifdef HAS_VHOST
		epona->log(InfoUserLogging(), "%s (%s@%s => %s) (%s) left the network (%s).",
				user->nick, user->username, user->host, 
				(user->vhost ? user->vhost: "(none)"), user->realname, 
				(user->server ? user->server->name : "(unknown)"));
#else
		epona->log(InfoUserLogging(), "%s (%s@%s) (%s) left the network (%s).",
				user->nick, user->username, user->host, 
				user->realname, 
				(user->server ? user->server->name : "(unknown)"));
#endif
	}
	
    usercnt--;
    if (is_oper(user))
	opcnt--;
    
    free(user->username);
    free(user->host);
#ifdef HAS_VHOST
    if (user->vhost)
    	free(user->vhost);
#endif
#ifdef HAS_CHGHOST
	if (user->ivhost)
		free(user->ivhost);
#endif
    free(user->realname);
    
    c = user->chans;
    while (c) {
		c2 = c->next;
		chan_deluser(user, c->chan);
		free(c);
		c = c2;
	}
	
	/* This called only here now */
	cancel_user(user);
	if (user->na) 
    user->na->u = NULL;
    
    ci = user->founder_chans;
    while (ci) {
	ci2 = ci->next;
	free(ci);
	ci = ci2;
    }
    
    if (user->prev)
	user->prev->next = user->next;
    else
	userlist[HASH(user->nick)] = user->next;
    if (user->next)
	user->next->prev = user->prev;
    
    free(user);
}

/*************************************************************************/
/*************************************************************************/

/* Return statistics.  Pointers are assumed to be valid. */

void get_user_stats(long *nusers, long *memuse)
{
    long count = 0, mem = 0;
    int i;
    User *user;
    User::u_chanlist *uc;
    User::u_chaninfolist *uci;

    for (i = 0; i < 1024; i++) {
	for (user = userlist[i]; user; user = user->next) {
	    count++;
	    mem += sizeof(*user);
	    if (user->username)
		mem += strlen(user->username)+1;
	    if (user->host)
		mem += strlen(user->host)+1;
#ifdef HAS_VHOST
        if (user->vhost)
        mem += strlen(user->vhost)+1;
#endif
#ifdef HAS_CHGHOST
        if (user->ivhost)
        mem += strlen(user->ivhost)+1;
#endif
	    if (user->realname)
		mem += strlen(user->realname)+1;
	    for (uc = user->chans; uc; uc = uc->next)
		mem += sizeof(*uc);
	    for (uci = user->founder_chans; uci; uci = uci->next)
		mem += sizeof(*uci);
	}
    }
    *nusers = count;
    *memuse = mem;
}

/*************************************************************************/

/* Find a user by nick.  Return NULL if user could not be found. */

User *finduser(const char *nick)
{
    User *user;

    if (epona->debuglevel() >= 3)
	epona->log(DebugUserList(), "finduser(%p)", nick);
    user = userlist[HASH(nick)];
    while (user && stricmp(user->nick, nick) != 0)
	user = user->next;
    if (epona->debuglevel() >= 3)
	epona->log(DebugUserList(), "finduser(%s) -> %p", nick,
			static_cast<void *>(user));
    return user;
}

/*************************************************************************/

/* Iterate over all users in the user list.  Return NULL at end of list. */

static User *current;
static int next_index;

User *firstuser(void)
{
    next_index = 0;
    while (next_index < 1024 && current == NULL)
	current = userlist[next_index++];
    if (epona->debuglevel() >= 3)
	epona->log(DebugUserList(), "firstuser() returning %s",
			current ? current->nick : "NULL (end of list)");
    return current;
}

User *nextuser(void)
{
    if (current)
	current = current->next;
    if (!current && next_index < 1024) {
	while (next_index < 1024 && current == NULL)
	    current = userlist[next_index++];
    }
    if (epona->debuglevel() >= 3)
	epona->log(DebugUserList(), "nextuser() returning %s",
			current ? current->nick : "NULL (end of list)");
    return current;
}

/*************************************************************************/
/*************************************************************************/

/* Handle a server NICK command. */

User *do_nick(char const *source, char *nick, char *username, char *host,
		char *server, char *realname, time_t ts, uint32_t svid,
		char const *modes, ...)
{
#ifdef HAS_NICKIP
	struct in_addr addr;
#endif
	
    User *user;

    NickAlias *old_na;	/* Old nick rec */
    int nc_changed = 1;	/* Did nick core change? */
    int status = 0;     /* Status to apply */

    if (!*source) 
    {
#ifdef HAS_VHOST
		char *vhost = NULL;
#endif
#if defined(HAS_NICKIP) || defined(HAS_NICKVHOST)
		va_list args;
		va_start(args, modes);
#endif

#ifdef HAS_NICKIP
# if defined(IRC_BAHAMUT)
		addr.s_addr = va_arg(args, uint32_t);
		addr.s_addr = htonl(addr.s_addr);
# elif defined(IRC_UNREAL)
		char *b64ip = va_arg(args, char *);
		if (*b64ip != '*' && (strlen(b64ip) == 8))
		{
			if (misc::base64_decode(b64ip, 8, (char *)&addr.s_addr) != 4)
				addr.s_addr = 0; /* Invalid IP address */
		} else {
			addr.s_addr = 0;
		}
# endif
#endif

#ifdef HAS_NICKVHOST
		vhost = va_arg(args, char *);
# ifdef IRC_UNREAL
 		if (!strcmp(vhost, "*"))
 			vhost = NULL;
# endif
#endif

#if defined(HAS_NICKIP) || defined(HAS_NICKVHOST)
		va_end(args);
#endif
    	
		/* This is a new user; create a User structure for it. */
		if (epona->debuglevel())
	    	epona->log(DebugUser(), "new user: %s", nick);

		if (epona->cd().services.log_users) 
		{
#ifdef HAS_NICKVHOST
# ifdef HAS_NICKIP
			epona->log(InfoUserLogging(), "%s (%s@%s => %s) (%s) [%s] connected to the network (%s).",
					nick, username, host, vhost, realname, inet_ntoa(addr), server);
# else
			epona->log(InfoUserLogging(), "%s (%s@%s => %s) (%s) connected to the network (%s).",
					nick, username, host, vhost, realname, server);
# endif
#else
# ifdef HAS_NICKIP
			epona->log(InfoUserLogging(), "%s (%s@%s) (%s) [%s] connected to the network (%s).",
					nick, username, host, realname, inet_ntoa(addr), server);
# else
			epona->log(InfoUserLogging(), "%s (%s@%s) (%s) connected to the network (%s).",
					nick, username, host, realname, server);
# endif
#endif
		}
	    
#ifndef STREAMLINED
        /* Now check for session limits */
        if (epona->cd().session.enable && !add_session(nick, host))
            return NULL;
#endif

		/* Allocate User structure and fill it in. */
		user = new_user(nick);
		user->username = sstrdup(username);
		user->host = sstrdup(host);
		user->server = server_find(server);
		user->realname = sstrdup(realname);
		user->timestamp = ts;
		user->svid = svid;
		user->my_signon = time(NULL);
#ifdef HAS_VHOST
		user->vhost = vhost ? sstrdup(vhost) : sstrdup(host);
#endif
#ifdef HAS_NICKIP
		user->ip.s_addr = addr.s_addr;
#endif
		
		/* Return NULL only if the user record is invalid after
		 * that point. Else the user is *STILL* connected to the
		 * network, and we must consider it as being a normal
		 * client, even though the function may have issued an
		 * AKILL to ban it from the network. */
		if (netforce.ban_enforce(user))
			return NULL;

		/* Send news only if the user was not connected to
		 * the network. */
		if (user->svid == 0) 
		{
			display_news(user, News::ntLogon);
			display_news(user, News::ntRandom);
		}
		
		if (static_cast<time_t>(user->svid) == user->timestamp && user->na) {
			/* Timestamp and svid match, and nick is registered; automatically
			 * identify the nick */
			user->na->status_add(NickAlias::asRecognized | NickAlias::asIdentified);
	    	nc_changed = 0;
    	} else if (svid != 1) {
    		/* Reset the svid because it doesn't match */
    		user->svid = 1;
			s_svsmode_user(s_NickServ, user->nick, user->timestamp, "+d", "1");
		} else {
			user->svid = 1;
		}
		
		if (modes)
			set_umode(user, 1, &modes);
    } else {
		/* An old user changing nicks. */
		user = finduser(source);
		if (!user) {
	    	epona->log(ErrorUser(), "NICK from nonexistent nick %s", source);
	    	return NULL;
		}
		
		if (epona->debuglevel())
	    	epona->log(DebugUser(), "%s changes nick to %s", source, nick);

		if (epona->cd().services.log_users) {
#ifdef HAS_VHOST
		epona->log(InfoUserLogging(), "%s (%s@%s => %s) (%s) changed his nick to %s (%s).",
				user->nick, user->username, user->host, 
				(user->vhost ? user->vhost: "(none)"), user->realname, 
				nick, (user->server ? user->server->name : "(unknown)"));
#else
		epona->log(InfoUserLogging(), "%s (%s@%s) (%s) changed his nick to %s (%s).",
				user->nick, user->username, user->host, 
				user->realname, nick, (user->server ? user->server->name : "(unknown)"));
#endif
		}

		user->timestamp = ts;

#if defined(IRC_BAHAMUT) || defined(IRC_UNREAL)
		user->mode &= ~UMODE_r;
#endif

		if (stricmp(nick, user->nick) == 0) {
			/* No need to redo things */
			change_user_nick(user, nick);
	    	nc_changed = 0;
	    } else {
	    	/* Update this only if nicks aren't the same */
	    	user->my_signon = time(NULL);
	    	
			old_na = user->na;
			if (old_na) {
				if (nick_recognized(user))
					user->na->last_seen(time(0));
				status = old_na->status() & NickAlias::asTransGroup;
				cancel_user(user);
			}
	    
			change_user_nick(user, nick);
			
			if ((old_na ? old_na->ng() : NULL) == (user->na ? user->na->ng() : NULL))
				nc_changed = 0;
	
			if (!nc_changed && (user->na)) {
				user->na->status_add(status);
			} else {
				/* Can't call user->na->deidentified() since we're
				 * using the new alias (if any), not the old. */
				change_user_mode(s_NickServ, user,
						((user->mode & UMODE_r) ? "-r+d" : "+d"), "1");
#ifdef HAS_CHGHOST
				custom_host_unset(user);
#endif
			}
    	}
    	
    	if (netforce.nicknb()->enforce(user))
    		return NULL;
    
    }	/* if (!*source) */

    if (nc_changed || !nick_recognized(user)) {
		validate_user(user);
	} else {
		if (nick_identified(user)) 
		{
			user->na->post_identified();
    		
			epona->log(InfoNickServRights(),
					"%s!%s@%s automatically identified for nick %s",
					user->nick, user->username, GetHost(user), user->nick);
		}
	}

	/* Bahamut sets -r on every nick changes, so we must test
	 * it even if nc_changed == 0
	 *
	 * ... Uh? Is this _really_ necessary? */
/*#ifdef IRC_BAHAMUT
	if (nick_identified(user))
	{
		if ((time_t)user->svid != user->timestamp) 
		{
			char tsbuf[16];
			snprintf(tsbuf, sizeof(tsbuf), "%ld", (long) user->timestamp);
			change_user_mode(s_NickServ, user,
					(!(user->mode & UMODE_r) ? "+rd" : "+d"), tsbuf);
	    } else {
	    	if (!(user->mode & UMODE_r))
				change_user_mode(s_NickServ, user, "+r", NULL);
	    }
	}
#endif*/

	return user;
}

/*************************************************************************/

/* Handle a MODE command for a user.
 *	av[0] = nick to change mode for
 *	av[1] = modes
 */

void do_umode(const char *source, int ac, char **av)
{
    User *user;

    if (stricmp(source, av[0]) != 0) {
		epona->log(ErrorUser(), "MODE %s %s from different nick %s!", av[0], av[1], source);
		wallops(NULL, "%s attempted to change mode %s for %s", source, av[1], av[0]);
		return;
    }
    
    user = finduser(source);
    if (!user) {
		epona->log(ErrorUser(), "MODE %s for nonexistent nick %s: %s", av[1],
				source, merge_args(ac, av));
		return;
    }
    
    set_umode(user, ac-1, &av[1]);
}

/*************************************************************************/

/* Handle a QUIT command.
 *	av[0] = reason
 */

void do_quit(const char *source, int ac, char **av)
{
    User *user;
    
    user = finduser(source);
    if (!user) {
	epona->log(ErrorUser(), "QUIT from nonexistent user %s: %s", source,
							merge_args(ac, av));
	return;
    }
    if (epona->debuglevel())
	epona->log(DebugUser(), "%s quits", source);
    if (nick_recognized(user)) 
    	user->na->post_signoff(*av[0] ? av[0] : 0);
    
#ifndef STREAMLINED
# if defined(HAS_CHGHOST) && !defined(HAS_VHOST)
	if (epona->cd().session.enable)
	    del_session(user->ivhost ? user->ivhost : user->host);
# else
    if (epona->cd().session.enable)
	    del_session(user->host);
# endif
#endif
    delete_user(user);
}

/*************************************************************************/

/* Handle a KILL command.
 *	av[0] = nick being killed
 *	av[1] = reason
 */

void do_kill(const char *source, int ac, char const *const *av)
{
    User *user;

    user = finduser(av[0]);
    if (!user)
	return;
    if (epona->debuglevel())
	epona->log(DebugUser(), "%s killed", av[0]);
    if (nick_recognized(user))
   		user->na->post_signoff(*av[1] ? av[1] : 0);
   	
#ifndef STREAMLINED
# if defined(HAS_CHGHOST) && !defined(HAS_VHOST)
	if (epona->cd().session.enable)
	    del_session(user->ivhost ? user->ivhost : user->host);
# else
    if (epona->cd().session.enable)
	    del_session(user->host);
# endif
#endif
    delete_user(user);
}

/*************************************************************************/
/*************************************************************************/

/* Is the given nick an oper? */

int is_oper(User *user)
{
    return (user->mode & UMODE_o);
}

/*************************************************************************/
/*************************************************************************/

/* Does the user's usermask match the given mask (either nick!user@host or
 * just user@host)?
 */

int match_usermask(const char *mask, User *user, bool restrict)
{
	misc::cstring mask2(mask);
	char *nick, *username, *host;
	
	if (strchr(mask2, '!')) {
		nick = strtok(mask2, "!");
		username = strtok(NULL, "@");
	} else {
		nick = NULL;
		username = strtok(mask2, "@");
	}
	host = strtok(NULL, "");
	if (!username || !host)
		return 0;
	
	if (nick && !match_wild_nocase(nick, user->nick)) 
		return 0;
	
	if (!match_wild_nocase(username, user->username))
		return 0;
	
	if (restrict)
		return match_wild_nocase(host, GetHost(user));
	else
		return (match_wild_nocase(host, user->host)
#ifdef HAS_VHOST
				|| (user->vhost && match_wild_nocase(host, user->vhost))
#endif
				);
	
	return 1;
}

/*************************************************************************/

/* Split a usermask up into its constitutent parts.  Returned strings are
 * malloc()'d, and should be free()'d when done with.  Returns "*" for
 * missing parts.
 */

void split_usermask(const char *mask, char **nick, char **user, char **host)
{
    misc::cstring mask2(mask);

	char const *tnick, *tuser, *thost;

    tnick = strtok(mask2, "!");
    tuser = strtok(NULL, "@");
    thost = strtok(NULL, "");
    /* Handle special case: mask == user@host */
    if (tnick && !tuser && strchr(tnick, '@')) {
	tnick = NULL;
	tuser = strtok(mask2, "@");
	thost = strtok(NULL, "");
    }
    if (!tnick)
	tnick = "*";
    if (!tuser)
	tuser = "*";
    if (!thost)
	thost = "*";
	
    *nick = misc::nstrdup(tnick);
    *user = misc::nstrdup(tuser);
    *host = misc::nstrdup(thost);
}

/*************************************************************************/

/* Given a user, return a mask that will most likely match any address the
 * user will have from that location.  For IP addresses, wildcards the
 * appropriate subnet mask (e.g. 35.1.1.1 -> 35.*; 128.2.1.1 -> 128.2.*);
 * for named addresses, wildcards the leftmost part of the name unless the
 * name only contains two parts.  If the username begins with a ~, delete
 * it.  The returned character string is malloc'd and should be free'd
 * when done with.
 */

char *create_mask(User *u, bool vhost)
{
    char *mask, *s, *end;
    char const *host;
    int ulen = strlen(u->username);

	host = (vhost ? GetHost(u) : u->host);

    /* Get us a buffer the size of the username plus hostname.  The result
     * will never be longer than this (and will often be shorter), thus we
     * can use strcpy() and sprintf() safely.
     */
    end = mask = static_cast<char *>(smalloc(ulen + strlen(host) + 3));
    end += sprintf(end, "%s%s@",
    	(ulen < (*(u->username) == '~' ? USERMAX+1 : USERMAX) ? "*" : ""),
    	(*(u->username) == '~' ? u->username+1 : u->username));
    
    if (strspn(host, "0123456789.") == strlen(host)
		&& (s = strchr(host, '.'))
		&& (s = strchr(s+1, '.'))
		&& (s = strchr(s+1, '.'))
		&& (   !strchr(s+1, '.'))) 
    {    /* IP addr */
		s = sstrdup(host);
		*strrchr(s, '.') = 0;
		
		sprintf(end, "%s.*", s);
		free(s);
    } else {
		if ((s = strchr(host, '.')) && strchr(s+1, '.')) {
	    		s = sstrdup(strchr(host, '.')-1);
			*s = '*';
			strcpy(end, s);
			free(s);
		} else {
			strcpy(end, host);
		}
    }
    return mask;
}

/*************************************************************************/
